/*
 * This File: languages.h
 *
 * This file contains all the English and German menu item names and various 
 * other strings necessary for both an English and German interface, and is 
 * generalizable to new languages.  (Feedback on this aspect is greatly 
 * appreciated, since I only speak bad American and a tiny bit of German.)
 *
 * Notice that the 4 German characters that do not appear in the English
 * alphabet have octal values of ss=337, ue=374, ae=344, and oe=366.  At
 * least those are their values in "variable" font, and some others I
 * checked.  However, those characters do not seem to be defined in "fixed"
 * and certain others.  The selection of an appropriate font is critical
 * for the German interface to work smoothly.
 *
 * FUTURE: \Ue
 *
 * Special thanks to Sven Dack and Klaus Reichenberger who did the original 
 * German translation for the menus.
 */

/*
 * The following are some general terms that appear throughout the interface
 */
static char *english_ok = "OK";
static char *german_ok = "OK";

static char *english_cancel = "Cancel";
static char *german_cancel = "Abbrechen";

static char *english_english = "English";
static char *german_english = "Englisch";

static char *english_german = "German";
static char *german_german = "Deutsch";

static char *english_help = "Help";
static char *german_help = "Hilfe";

static char *english_status1 =
	  "STATUS:  1 scrawl window currently open.  -->  ";
static char *german_status1 = 
          "STATUS:  1 scrawl Fenster ist im Moment offen.  -->  ";
static char *english_status =
	  "STATUS:  %d scrawl windows currently open.  -->  ";
static char *german_status = 
	  "STATUS:  %d scrawl Fenster sind im Moment offen.  -->  ";

/*
 * window manager string
 */
static char *english_name = 
	       "WSCRAWL 2.01 by Brian Wilson: a product of eXcess engineering";

static char *german_name = 
	      "WSCRAWL 2.01 von Brian Wilson: a product of eXcess engineering";

/*
 * The following are the pull down menu text in both English and German
 */
static char *english_menu_bar_labels[] = {"File", "Edit", "Options",
			"Colors", "Fonts", "VideoConference", "****"};

static char *german_menu_bar_labels[] = {"Datei", "Bearbeiten", "Optionen",
			"Farbe", "Zeichensatz", "VideoConference", "****"};

static char *english_menu0[] = {"Add Display...", "Join Existing Session...",
			"Attract Attention...",
			"----", "Save Selection...", 
			"Save Entire Whiteboard...", "Read In a File...",
			"----", "Quit                      <ctnl>q", "****"};

static char *german_menu0[] = {"Neuen Teilnehmer anmelden...", 
			"Als Teilnehmer eintreten...",
			"Teilnehmer Signalisierung...",
			"----", "Bildausschnitt speichern...", 
			"Gesamte Arbeitsfl\344che speichern...", 
			"Datei hereinladen...", "----", 
			"Programm verlassen                  <ctnl>q", 
			"****"};

static char *english_menu1[] = {"Cut              <cntl>x", 
			 "Copy             <cntl>c",
                         "Paste            <cntl>v", 
                         "Paste X-Selection", "Clear", 
			 "Fatbits Selection", "----", 
			 "Select Visible Area", "Select All", "----", 
			 "Clear Whiteboard", "****"};

static char *german_menu1[] = {"Ausschneiden                 <cntl>x", 
			 "Kopieren                        <cntl>c",
                         "Einf\374gen                        <cntl>v", 
                         "Einf\374gen X-Selection", 
			 "L\366schen ", 
			 "Fatbits Selection", "----", 
			 "Sichtbaren Bereich ausw\344hlen", 
			 "Alles ausw\344hlen", "----", 
			 "Arbeitsfl\344che l\366schen", "****"};

static char *english_menu2[] = {"Language...", "Scrollbar Settings...", 
			 "Pointer Options...", "----",
			 "Telepointer Label...",
			 "Super Selection...", "Brush Shapes...",
			 "Line Settings...", "Eraser Options...",
			 "Square Settings...", "Oval Settings...",
			 "****"};

static char *german_menu2[] = {"Sprache...", "Scrollbar-Einstellungen...", 
                         "Mauszeiger Optionen...", "----", 
			 "Zeigername...",
			 "Bildschirmausschnitt...", "Pinseleinstellungen...",
			 "Linieneinstellungen...", 
			 "Radiergummieinstellungen...",
			 "Rechteck Einstellungen...", 
			 "Ellipsen Einstellungen...",
			 "****"};

static char *english_menu3[] = {"white",  "magenta", "red", "blue", "cyan",
                         "LightSteelBlue", "navy blue", "green", "coral",
                         "light grey", "orange", "plum", "yellow", "black",
                         "----", "Custom...", "****"};

static char *german_menu3[] = {"Wei\337",  "Magenta", "Rot", "Blau", "Cyan", 
			 "Stahlblau (hell)", "Dunkelblau", "Gr\374n", 
			 "Korralle", "Hellgrau", "Orange", "Lila", "Gelb", 
			 "Schwarz", "----", "Eigene...", "****"};

static char *english_menu4[] = {"fixed", "variable", "timR12", "helvO12", 
			 "courR12", "courR14", "timBI18", "courR24", 
			 "timR24", "timBI24", "ncenR24", "vbee-36", "vr-40", 
			 "vgl-40", "vsgn-57", "****"};

static char *german_menu4[] = {"fixed", "variable", "timR12", "helvO12", 
			 "courR12", "courR14", "timBI18", "courR24", 
			 "timR24", "timBI24", "ncenR24", "vbee-36", "vr-40", 
			 "vgl-40", "vsgn-57", "****"};

/*
 * The following is the text for some generic dialog box prompts 
 */
static char *english_add_display_prompt = 
		  "Enter the name of the display to add:";

static char *german_add_display_prompt = 
		  "Bitte geben Sie die Adresse\ndes neuen Teilnehmers ein:";

static char *english_join_session_prompt = 
	          "Enter the name of the machine\nyou wish to be added to:";

static char *german_join_session_prompt = 
	          "Bitte geben Sie die Adresse\neines Teilnehmers ein:";

static char *english_language_prompt = 
	          "Select the language you wish\nthe interface to be in:";

static char *german_language_prompt = 
   "Bitte w\344hlen Sie die Sprache, in\nder die Men\374s erscheinen sollen:";

static char *english_line_width_prompt = 
		  "Enter the line width for straight lines:";

static char *german_line_width_prompt = 
		  "Bitte Linienbreite eingeben fuer Normale Linien:";

static char *english_square_width_prompt = 
		  "Enter the line width for outline squares:";

static char *german_square_width_prompt = 
		  "Bitte Linienbreite eingeben fuer Rechteckumrandung:";

static char *english_oval_width_prompt = 
		  "Enter the line width for outline ovals:";

static char *german_oval_width_prompt = 
		  "Bitte Linienbreite eingeben fuer Ellipsenumrandung:";


/*
 * dirty_add_display_dialog prompts and buttons
 */
static char *english_dirty_add_display_prompt = 
"The shared drawing surface has been drawn on.\n\
Copying this drawing to the new display can\n\
be very time consuming.  Please select one of\n\
the following options:";

static char *german_dirty_add_display_prompt = 
"Auf die gemeinsame Arbeitsfl\344che ist schon gezeichnet worden.\n\
Dem neuen Teilnehmer die gesamte Zeichnung zu kopieren, kann sehr\n\
zeitaufwendig werden.  W\344hlen Sie bitte eine der folgenden\n\
M\366glichkeiten:";

static char *english_dirty_add_display_but1 = 
                  "Copy Entire Whiteboard (VERY time consuming)";

static char *german_dirty_add_display_but1 = 
                  "Gesamte Arbeitsfl\344che kopieren (Sehr zeitaufwendig)";

static char *english_dirty_add_display_but2 = 
                  "Copy Visible Region Only (less time consuming)";

static char *german_dirty_add_display_but2 = 
                  "Nur den sichbaren Bereich kopieren (weniger zeitaufwendig)";

static char *english_dirty_add_display_but3 = 
                  "Clear All Whiteboards (very fast)";

static char *german_dirty_add_display_but3 = 
                  "\334berall die Arbeitsfl\344eche l\366schen (sehr schnell)";

/*
 * scrollbar_dialog prompt and buttons
 */
static char *english_scrollbar_prompt = 
			    "Select The Scrollbar Behavior You Wish:";

static char *german_scrollbar_prompt = 
		    "W\344hlen sie das gew\374nschte Verhalten der Scrollbar:";

static char *english_scrollbar_but1 = "Scrollbars All Locked Together";

static char *german_scrollbar_but1 = "Scrollbars werden gemeinsam benutzt";

static char *english_scrollbar_but2 = "Scrollbars Operate Independently";

static char *german_scrollbar_but2 = 
		   "Scrollbars arbeiten unabh\344ngig voneinander";


/*
 * save_file_format_dialog prompt and buttons
 */
static char *english_save_file_format_prompt = 
			    "The file to be saved is in which Format?";

static char *german_save_file_format_prompt = 
		    "In welchem Format soll das Bild gespeichert werden?";

static char *english_save_file_format_but1 = 
			    "Bitmap Format (X-Windows bitmap)";

static char *german_save_file_format_but1 = 
			    "Bitmap Format (X-Windows bitmap)";

static char *english_save_file_format_but2 = 
			    "Image Format (xwd format)";

static char *german_save_file_format_but2 = 
			    "Image Format (xwd format)";

static char *english_save_file_format_but3 = 
			    "Greyscale Postscript";

static char *german_save_file_format_but3 = 
			    "Greyscale Postscript";

static char *english_save_file_format_but4 = "PPM Format (Portable Pixmap)";

static char *german_save_file_format_but4 = "PPM Format (Portable Pixmap)";


/*
 * read_file_format_dialog prompt and buttons
 */
static char *english_read_file_format_prompt = 
			    "The file to be read is in which Format?";

static char *german_read_file_format_prompt = 
                "Welches Format hat das Bild, das Sie hereinladen m\366chten?";

static char *english_read_file_format_but1 = 
			    "Bitmap Format (X-Windows bitmap)";

static char *german_read_file_format_but1 = 
			    "Bitmap Format (X-Windows bitmap)";

static char *english_read_file_format_but2 = 
			    "Image Format (xwd format)";

static char *german_read_file_format_but2 = 
			    "Image Format (xwd format)";

static char *english_read_file_format_but3 = "Text File";

static char *german_read_file_format_but3 =  "Text File";

static char *english_read_file_format_but4 = "PPM Format (Portable Pixmap)";

static char *german_read_file_format_but4 = "PPM Format (Portable Pixmap)";


/*
 * color_picker prompt and buttons
 */
static char *english_color_picker_prompt = 
			    "Dial in the color you would like to use:";

static char *german_color_picker_prompt = 
			    "Stellen Sie die gew\374nschte Farbe ein:";

static char *english_color_picker_set_col_but = "Set Color";

static char *german_color_picker_set_col_but = "Farbe \374bernehmen";

static char *english_color_picker_close_but = "Close";

static char *german_color_picker_close_but = "Schlie\337en";

static char *english_color_picker_grab_color_but = "Grab Color";

static char *german_color_picker_grab_color_but = "Farbe aufnehmen";


/*
 * pointer options dialog prompt and buttons
 */
static char *english_mouse_dialog_prompt = 
			    "Select any combination:";

static char *german_mouse_dialog_prompt = 
			    "Jede Kombination moeglich:";

static char *english_mouse_dialog_toggle1 = 
			    "Show Pointer Coordinates";

static char *german_mouse_dialog_toggle1 = 
			    "Zeige Mauskoordinaten";

static char *english_mouse_dialog_toggle2 = 
			    "Always Show TelePointer";

static char *german_mouse_dialog_toggle2 = 
			    "TelePointer immer anzeigen";

