/*
 *			      WSCRAWL 2.0
 * This file:  wscrawl.c
 *
 * Author: Brian Wilson
 *  email: brianw@apple.com, or brianw@cs.stanford.edu
 * Last Modified: 10/24/92
 */

#include "wscrawl.h"
#include "pause_curs.h"
#include "languages.h"
#include "icon.h"

int num_disps;
int DOUBLE_CLICK_THRESHOLD[MAX_NUM_DISPS];		/* in milliseconds */
XtAppContext appCon;

struct disp_info_struct disp_info[MAX_NUM_DISPS];
extern Widget create_application();
extern XImage *read_image_from_disk();
int vsbar_movementCB();
int hsbar_movementCB();
XtTimerCallbackProc spraypaint_spotCB();
XtInputCallbackProc handle_knock_knockCB();
XtEventHandler handle_wscrawl_window_moveCB();
XtEventHandler icon_exposeCB();
float my_rand();
int wm_kill_window();

struct global_values_struct global_values;
int draw_area_height, draw_area_width;	      /* shared among all displays */
int sync_scrollbars_bool;	      


main(argc,argv)
int argc;
char **argv;
{
    int i, n, knock;
    char location[256];
    Arg args[MAX_ARGS];
    Widget main_window;
    int disp_num;
    XEvent *the_event;
    unsigned int copy_of_argc;
    char *copy_of_argv[100];

    /*
     * initialize all displays as not used
     */
    for (i=0; i < MAX_NUM_DISPS; i++)
	disp_info[i].in_session_bool = FALSE;
    disp_info[0].in_session_bool = TRUE;
    num_disps = 1;

    /* 
     * copy argv and argc for future use since XtOpenDisplay alters them
     */
    copy_of_argc = argc;
    for (i=0; i < argc; i++)
    {
        copy_of_argv[i] = (char *) malloc(strlen(argv[i]) + 1);
        strcpy(copy_of_argv[i], argv[i]);
    }

    /*
     * Initialize toolkit and open the display.
     */
    XtToolkitInitialize();
    appCon = XtCreateApplicationContext();

    if ((disp_info[0].disp = XtOpenDisplay(appCon, NULL, "wscrawl", "WScrawl",
	                                   NULL, 0, &argc, argv)) == NULL)
    {
	XtWarning("wscrawl: cannot open display, exiting...");
	exit(0);
    }

    set_default_resources(0, 0, NULL, copy_of_argc, copy_of_argv);

    disp_info[0].NUM_EFF_MENUS = NUM_MENUS;
       
    read_one_time_only_global_resources(disp_info[0].disp, &global_values);
    global_values.screen_dirty_bit = FALSE;
     
    /* XSynchronize(disp_info[0].disp, True);  */
    parse_command_line(argv, argc, &knock, location, &global_values);
    init_communication(0, knock, location);

    if (disp_info[0].my_socket != 0)
        XtAppAddInput(appCon, disp_info[0].my_socket, (XtPointer)XtInputReadMask, 
		  (XtInputCallbackProc)handle_knock_knockCB, 0);
    /*
     * Create application shell.
     */
    n = 0;
    XtSetArg(args[n], XmNdeleteResponse, XmDO_NOTHING); n++;
    disp_info[0].app_shell = XtAppCreateShell("wscrawl", "WScrawl",
	          applicationShellWidgetClass, disp_info[0].disp, args, n);
    XtAddEventHandler(disp_info[0].app_shell, StructureNotifyMask, False,
		       (XtEventHandler)handle_wscrawl_window_moveCB, 0);

    n = 0;
    XtSetArg(args[n], XmNmappedWhenManaged, False); n++;
    XtSetValues(disp_info[0].app_shell, args, n);

    /*
     * Create and realize main application window.
     */
    main_window = create_application(disp_info[0].app_shell, 0);
    disp_info[0].draw_win_id = NULL;
    XtRealizeWidget(disp_info[0].app_shell);

    /*
     * Initialize other stuff, like Xlib vars...
     */
    init_misc(0);
    adjust_graphics_contexts(0);

    if (global_values.edit_file_bool == TRUE)
        read_xwd_file_into_drawing_area(0, global_values.edit_filename);

    /*
     * Get and displatch events.
     */
    XtMapWidget(disp_info[0].app_shell);
    XtAppMainLoop(appCon); 
}


/*
 * init_misc - this function initializes various variables for the display
 *             passed in.  The display has just been added.
 */
init_misc(disp_num)
int disp_num;
{
    XWindowAttributes window_attr;
    XWindowChanges win_change;
    XSetWindowAttributes set_attr;
    XColor scrn_def_ret, exact_def_ret;
    XGCValues gcvalues;
    XFontStruct *the_font_struct;
    XSizeHints size_hints;
    XWMHints wmhints;
    GC default_gc;
    XColor fg1, bg1;
    Pixel fg, bg;
    Pixmap source, mask;
    Pixmap icon_pix, icon_bit;
    Widget icon_widget, icon_draw_area;
    int i, j, left, top, n;
    Dimension width, height;
    Arg args[MAX_ARGS];
    char tmp_str[60], tmp_str2[60];
    char *disp_name;
    char *ptr;
    unsigned long dummy=0;
    unsigned long pixel_ret;
    XColor the_color;
    Atom xa_WM_DELETE_WINDOW;
    Atom xa_WM_PROTOCOLS;

    disp_info[disp_num].draw_win_id = XtWindow(disp_info[disp_num].draw_area);

    disp_name = DisplayString(disp_info[disp_num].disp);
    strncpy(disp_info[disp_num].disp_name, disp_name, 45);
    /* 
     * decipher a coherent user understandable name for this display, instead
     * of stuff like ":0" or "local:0"
     */
    if ((strncmp(disp_info[disp_num].disp_name, "local", 5) == 0) ||
        (strncmp(disp_info[disp_num].disp_name, "unix", 4) == 0) ||
        (strncmp(disp_info[disp_num].disp_name, ":", 1) == 0))
    {
        /*
         * this is a wscrawl window on the invoker's home display, so
         * change it to be meaningful.
         */
        gethostname(tmp_str, 30);
        for (ptr=disp_info[disp_num].disp_name; *ptr != ':'; ptr++)
	    ;
        sprintf(tmp_str2, "%s%s", tmp_str, ptr);
        strncpy(disp_info[disp_num].disp_name, tmp_str2, 40);
    }
    strncpy(disp_info[disp_num].rubber_pointer.pointer_str, 
	    disp_info[disp_num].disp_name, 45);
    
    DOUBLE_CLICK_THRESHOLD[disp_num] = 600;
    /* should be = XtGetMultiClickTime(disp), but doesn't work */

    create_selected_regions(disp_num);	/* create the selected regions */

    for (i=0; i < disp_info[disp_num].NUM_EFF_MENUS; i++)
    {
        XtRealizeWidget(disp_info[disp_num].menu[i].menu_shell);
        XtRealizeWidget(disp_info[disp_num].menu[i].menu_bar_button);
	
	win_change.border_width = 0;
	XConfigureWindow(disp_info[disp_num].disp,
		XtWindow(disp_info[disp_num].menu[i].menu_shell),
		CWBorderWidth, &win_change);
    }
    fill_out_menu_data_structures(disp_num);

    set_attr.cursor = XCreateFontCursor(disp_info[disp_num].disp,XC_draped_box);
    set_attr.background_pixel = 
	  WhitePixelOfScreen(DefaultScreenOfDisplay(disp_info[disp_num].disp));
    disp_info[disp_num].eraser_win_id = XCreateWindow(disp_info[disp_num].disp,
	    disp_info[disp_num].draw_win_id, 0, 0, 20, 20, 2, 0, 
	    CopyFromParent, CopyFromParent, CWCursor | CWBackPixel, &set_attr);

    strncpy(disp_info[disp_num].cur_color_str, "black", 40);
    strncpy(disp_info[disp_num].cur_font_str, "fixed", 40);

    disp_info[disp_num].pointer_state = NOT_PRESSED;
    disp_info[disp_num].input_mode = PAINTING;
    disp_info[disp_num].pointer_snap_to_grid = FALSE;
    disp_info[disp_num].paintbrush.paint_mode = NORMAL;
    disp_info[disp_num].paintbrush.which_paintb = 2;
    disp_info[disp_num].paintbrush.brush_width = 9;
    disp_info[disp_num].eraser_win_width = 40;
    disp_info[disp_num].eraser_win_height = 40;
    disp_info[disp_num].line_dialog.line_width = 0;
    disp_info[disp_num].square_dialog.line_width = 0;
    disp_info[disp_num].oval_dialog.line_width = 0;

    n = 0;
    XtSetArg(args[n], XmNwidth, &width); n++;
    XtSetArg(args[n], XmNheight, &height); n++;
    XtSetArg(args[n], XmNbackground, &bg); n++;
    XtGetValues(disp_info[disp_num].draw_area, args, n);
    disp_info[disp_num].draw_win_width = (int) width;
    disp_info[disp_num].draw_win_height = (int) height;

    gcvalues.foreground = (unsigned long) bg;
    disp_info[disp_num].backing_store_gc = XCreateGC(disp_info[disp_num].disp, 
		         disp_info[disp_num].draw_win_id, GCForeground, 
			 &gcvalues);
    if (XAllocColorCells(disp_info[disp_num].disp,
	DefaultColormap(disp_info[disp_num].disp, DefaultScreen(
	disp_info[disp_num].disp)), False, &dummy, 0, &pixel_ret, 1) != 0)
    {
        disp_info[disp_num].color_picker_dialog.color_cell_allocated = TRUE;
        gcvalues.foreground = pixel_ret;
        disp_info[disp_num].color_picker_dialog.the_color_cell = pixel_ret;
	the_color.flags = DoRed | DoGreen | DoBlue;
	the_color.red = 0;
	the_color.green = 0;
	the_color.blue = 0;
	the_color.pixel =disp_info[disp_num].color_picker_dialog.the_color_cell;
        XStoreColor(disp_info[disp_num].disp, DefaultColormap(
		disp_info[disp_num].disp, DefaultScreen(
		disp_info[disp_num].disp)), &the_color);
    }
    else
    {
        disp_info[disp_num].color_picker_dialog.color_cell_allocated = FALSE;
        gcvalues.foreground = WhitePixelOfScreen(
                             DefaultScreenOfDisplay(disp_info[disp_num].disp));
    }
    disp_info[disp_num].color_picker_dialog.the_gc = XCreateGC(
			 disp_info[disp_num].disp, 
		         disp_info[disp_num].draw_win_id, GCForeground, 
			 &gcvalues);

    gcvalues.line_width = 5;
    gcvalues.function = GXinvert;
    gcvalues.background = BlackPixelOfScreen(
                             DefaultScreenOfDisplay(disp_info[disp_num].disp));
    gcvalues.foreground = WhitePixelOfScreen(
                             DefaultScreenOfDisplay(disp_info[disp_num].disp));
    disp_info[disp_num].type_info.cursor_gc =XCreateGC(disp_info[disp_num].disp,
	       disp_info[disp_num].draw_win_id, 
               GCFunction | GCBackground | GCForeground | GCLineWidth,
	       &gcvalues);
    the_font_struct = XLoadQueryFont(disp_info[disp_num].disp, "fixed");
    gcvalues.font = the_font_struct->fid;
    gcvalues.subwindow_mode = IncludeInferiors;
    disp_info[disp_num].rubber_band_gc = XCreateGC(disp_info[disp_num].disp, 
         disp_info[disp_num].draw_win_id, 
	 GCFunction | GCBackground | GCForeground | GCFont | GCSubwindowMode,
	 &gcvalues);
    disp_info[disp_num].super_rubber_band_gc = 
      XCreateGC(disp_info[disp_num].disp, 
         XDefaultRootWindow(disp_info[disp_num].disp),
	 GCFunction | GCBackground | GCForeground | GCFont | GCSubwindowMode,
	 &gcvalues);
    gcvalues.font = the_font_struct->fid;
    gcvalues.line_width = 2;
    gcvalues.function = GXcopy;
    gcvalues.foreground = BlackPixelOfScreen(
                             DefaultScreenOfDisplay(disp_info[disp_num].disp));
    disp_info[disp_num].menu_gc = XCreateGC(disp_info[disp_num].disp, 
         disp_info[disp_num].draw_win_id, 
	 GCFunction | GCForeground | GCFont | GCLineWidth,
	 &gcvalues);
    /* 
     * create the backing store and clear it to the background color
     */
    disp_info[disp_num].backing_store = XCreatePixmap(disp_info[disp_num].disp,
	                  disp_info[disp_num].draw_win_id, 
			  (unsigned int) width, (unsigned int) height, 
	                  DefaultDepth(disp_info[disp_num].disp,
	                  DefaultScreen(disp_info[disp_num].disp)));
    XFillRectangle(disp_info[disp_num].disp, disp_info[disp_num].backing_store,
		   disp_info[disp_num].backing_store_gc, 0, 0, width, height);

    /*
     * create the "Please Wait" cursor
     */
    fg1.flags = DoRed | DoGreen | DoBlue;
    bg1.flags = DoRed | DoGreen | DoBlue;
    fg1.red = fg1.blue = fg1.green = 0;
    bg1.red = bg1.blue = bg1.green = 0;

    source = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, pause_curs_bits,
                      pause_curs_width, pause_curs_height, 1, 0, 1);
    mask = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, pause_curs_bits,
                      pause_curs_width, pause_curs_height, 1, 0, 1);

    disp_info[disp_num].pause_cursor = XCreatePixmapCursor(
                   disp_info[disp_num].disp, source, mask, &fg1, &bg1,
                   pause_curs_x_hot, pause_curs_y_hot);
    
    /*
     * init the telepointer (rubber pointer)
     */
    disp_info[disp_num].rubber_pointer.height =
                                   15 + the_font_struct->max_bounds.ascent +
                                   the_font_struct->max_bounds.descent;
    disp_info[disp_num].rubber_pointer.is_mapped_bool = FALSE;
    disp_info[disp_num].rubber_pointer.width = 15 + XTextWidth(the_font_struct,
                  disp_info[disp_num].rubber_pointer.pointer_str,
                  strlen(disp_info[disp_num].rubber_pointer.pointer_str));

    XMoveResizeWindow(disp_info[disp_num].disp, 
		  XtWindow(disp_info[disp_num].app_shell), 5 + (disp_num * 150),
		  5 + (disp_num * 150), 800, 550);

    put_bitmaps_into_tool_palette(disp_num);
    put_bitmaps_into_brush_dialog(disp_num);
    put_bitmaps_into_fatbits_dialog(disp_num);
    XDefineCursor(disp_info[disp_num].disp, disp_info[disp_num].draw_win_id,
	          disp_info[disp_num].paintbrush.paintb_curs[2]);

    n = 0;
    XtSetArg(args[n], XmNverticalScrollBar, &(disp_info[disp_num].vsbar)); n++;
    XtSetArg(args[n], XmNhorizontalScrollBar,&(disp_info[disp_num].hsbar)); n++;
    XtGetValues(disp_info[disp_num].swindow, args, n);

    n = 0;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    XtSetValues(disp_info[disp_num].vsbar, args, n);
    n = 0;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    XtSetValues(disp_info[disp_num].hsbar, args, n);

    XtAddCallback(disp_info[disp_num].vsbar, XmNdecrementCallback,
                  (XtCallbackProc)vsbar_movementCB, (XtPointer)(disp_num << 8 | DECR));
    XtAddCallback(disp_info[disp_num].vsbar, XmNincrementCallback,
                  (XtCallbackProc)vsbar_movementCB, (XtPointer)(disp_num << 8 | INCR));
    XtAddCallback(disp_info[disp_num].vsbar, XmNdragCallback,
                  (XtCallbackProc)vsbar_movementCB, (XtPointer)(disp_num << 8 | GENERIC_MOVE));
    XtAddCallback(disp_info[disp_num].vsbar, XmNpageDecrementCallback,
                  (XtCallbackProc)vsbar_movementCB, (XtPointer)(disp_num << 8 | GENERIC_MOVE));
    XtAddCallback(disp_info[disp_num].vsbar, XmNpageIncrementCallback,
                  (XtCallbackProc)vsbar_movementCB, (XtPointer)(disp_num << 8 | GENERIC_MOVE));

    XtAddCallback(disp_info[disp_num].hsbar, XmNdecrementCallback,
                  (XtCallbackProc)hsbar_movementCB, (XtPointer)(disp_num << 8 | DECR));
    XtAddCallback(disp_info[disp_num].hsbar, XmNincrementCallback,
                  (XtCallbackProc)hsbar_movementCB, (XtPointer)(disp_num << 8 | INCR));
    XtAddCallback(disp_info[disp_num].hsbar, XmNdragCallback,
                  (XtCallbackProc)hsbar_movementCB, (XtPointer)(disp_num << 8 | GENERIC_MOVE));
    XtAddCallback(disp_info[disp_num].hsbar, XmNpageDecrementCallback,
                  (XtCallbackProc)hsbar_movementCB, (XtPointer)(disp_num << 8 | GENERIC_MOVE));
    XtAddCallback(disp_info[disp_num].hsbar, XmNpageIncrementCallback,
                  (XtCallbackProc)hsbar_movementCB, (XtPointer)(disp_num << 8 | GENERIC_MOVE));

    read_special_resources(disp_num);

    make_selection_dependant_menus_stippled(disp_num);

    n = 0;
    XtSetArg(args[n], XmNsensitive, False); n++;
    XtSetValues(disp_info[disp_num].menu[1].menu_item[2], args, n); /* Paste */
    disp_info[disp_num].menu[1].selectable[2] = FALSE;

    /*
     * create the icon for any window manager that will listen
     */
    if (XAllocNamedColor(disp_info[disp_num].disp, XDefaultColormapOfScreen(
                         XDefaultScreenOfDisplay(disp_info[disp_num].disp)),
                         disp_info[disp_num].icon_fg_str, &scrn_def_ret, 
			 &exact_def_ret))
    {
        fg = scrn_def_ret.pixel;  /*great!  it worked!*/
    }
    else
    {
        fg=BlackPixelOfScreen(DefaultScreenOfDisplay(disp_info[disp_num].disp));
    }

    if (XAllocNamedColor(disp_info[disp_num].disp, XDefaultColormapOfScreen(
                         XDefaultScreenOfDisplay(disp_info[disp_num].disp)),
                         disp_info[disp_num].icon_bg_str, &scrn_def_ret, 
			 &exact_def_ret))
    {
        bg = scrn_def_ret.pixel;  /*great!  it worked!*/
    }
    else
    {
        bg=WhitePixelOfScreen(DefaultScreenOfDisplay(disp_info[disp_num].disp));
    }

    icon_bit = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, icon_bits,
                      icon_width, icon_height, 1, 0, 1);

    icon_pix = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, icon_bits,
                      icon_width, icon_height, fg, bg, 
		      DefaultDepth(disp_info[disp_num].disp,
		      DefaultScreen(disp_info[disp_num].disp)));
    disp_info[disp_num].icon_pix = icon_pix;

    size_hints.flags = USPosition | PPosition;
    size_hints.x = 5 + (disp_num * 80);
    size_hints.y = 5 + (disp_num * 80);
    XSetStandardProperties(disp_info[disp_num].disp, 
			   XtWindow(disp_info[disp_num].app_shell),
			   english_name, "wscrawl", icon_bit, NULL, 0, 
			   &size_hints);
    /*
     * create the icon window, and some other WM hints
     */
    set_attr.background_pixel = 
	  BlackPixelOfScreen(DefaultScreenOfDisplay(disp_info[disp_num].disp));
    n = 0;
    XtSetArg(args[n], XmNwidth, 80); n++;
    XtSetArg(args[n], XmNheight, 67); n++;
    XtSetArg(args[n], XmNmappedWhenManaged, False); n++;
    icon_widget = XtCreatePopupShell("icon_shell", overrideShellWidgetClass,
				    disp_info[disp_num].main_window, args, n);
    XtManageChild(icon_widget);
    XtAddEventHandler(icon_widget, ExposureMask, False,
							 (XtEventHandler)icon_exposeCB, (XtPointer)disp_num);

    wmhints.input = True;
    wmhints.icon_window = XtWindow(icon_widget);
    wmhints.flags = InputHint | IconWindowHint;
    XSetWMHints(disp_info[disp_num].disp, 
		XtWindow(disp_info[disp_num].app_shell), &wmhints);
}


/*
 * add_display - this function adds the display with name passed in.  Name
 *               will be something like "foo_machine:0.0".
 */
add_display(disp_name, source_disp_num, dirty_screen_option)
char *disp_name;
int source_disp_num, dirty_screen_option;
{
    int i, n, disp_num;
    Arg args[MAX_ARGS];
    Widget main_window;
    XWindowAttributes window_attr;
    int argc = 0;
    char **argv=(char **)NULL;
    char tmp_str[200];

    set_paused_cursors();

    /*
     * find an open record in the global array
     */
    for (i=0; i < MAX_NUM_DISPS; i++)
    {
	if (disp_info[i].in_session_bool == FALSE)
	    break;
    }

    if (i == MAX_NUM_DISPS)
    {
	printf("ERROR: no space left in internal data structures.\n");
        unset_paused_cursors();
	return;
    }
    else 
    {
	if (i == num_disps)
            num_disps++;	/* one more display in the world */
        disp_num = i;
        disp_info[disp_num].in_session_bool = TRUE;
    }

    strcpy(disp_info[disp_num].disp_name, disp_name);

    /*
     * Open the display.
     */
    if ((disp_info[disp_num].disp = XtOpenDisplay(appCon, disp_name, "wscrawl", 
	 "WScrawl", NULL, 0, &argc, argv)) == NULL)
    {
	sprintf(tmp_str, "Cannot Open display named:\n\n%s\n\n%s", disp_name,
	    "That display might not exist,\nor it might need to xhost you.");
	popup_info_dialog(source_disp_num, tmp_str);
        disp_info[disp_num].in_session_bool = FALSE;
        num_disps--;	/* one less display in the world */
        unset_paused_cursors();
	return;
    }
    
    argc = 0;
    set_default_resources(source_disp_num, disp_num, disp_name, argc, argv);

    disp_info[disp_num].NUM_EFF_MENUS = NUM_MENUS;
    /*
     * Create application shell.
     */
    n = 0;
    XtSetArg(args[n], XmNdeleteResponse, XmDO_NOTHING); n++;
    disp_info[disp_num].app_shell = XtAppCreateShell("wscrawl", "WScrawl",
		                 applicationShellWidgetClass, 
				 disp_info[disp_num].disp, args, n);
    XtAddEventHandler(disp_info[disp_num].app_shell, StructureNotifyMask,False,
		       (XtEventHandler)handle_wscrawl_window_moveCB, (XtPointer)disp_num);
    n = 0;
    XtSetArg(args[n], XmNmappedWhenManaged, False); n++;
    XtSetValues(disp_info[disp_num].app_shell, args, n);

    /*
     * Create and realize main application window.
     */
    main_window = create_application(disp_info[disp_num].app_shell, disp_num);
    disp_info[disp_num].draw_win_id = NULL;
    XtRealizeWidget(disp_info[disp_num].app_shell);

    /*
     * Initialize other stuff...
     */
    init_misc(disp_num);
    adjust_graphics_contexts(disp_num);

    if (dirty_screen_option != -1)
    {
	if (dirty_screen_option == CLEAR_ALL)
	{
	    process_menu_selection(disp_num, EDIT_MENU, CLEAR_WHITEBOARD);
	}
	else
	{
	    copy_image_to_new_display(source_disp_num, disp_num,
				      dirty_screen_option);
	}
    }

    copy_clipboards_to_new_display(disp_num);
    copy_selections_to_new_display(disp_num);

    /*
     * Map it.
     */
    XtMapWidget(disp_info[disp_num].app_shell);
    unset_paused_cursors();

    for (i=0; i < num_disps; i++)
        if (disp_info[i].in_session_bool && (i != disp_num))
	    draw_status_win_message(i);

    /*
     * Now turn off the screen saver and bing the bell once for good measure
     */
    XForceScreenSaver(disp_info[disp_num].disp, ScreenSaverReset);
    XBell(disp_info[disp_num].disp, 100);
}


/* 
 * copy_image_to_new_display - this function copies the image from the 
 *                             source screen to the destination screen
 *                             according to the option passed in.
 */
copy_image_to_new_display(source_disp_num, disp_num, copy_option)
int source_disp_num, disp_num, copy_option;
{
    XImage *the_image;
    Dimension d_width, d_height;
    int x, y, width, height, depth;
    char tmp_name[50], tmp_str[100];
    int n;
    Arg args[MAX_ARGS];
    Widget clip_window;
    Window dummy;

    sprintf(tmp_name, "/tmp/ws.%d.%d", source_disp_num, getpid());

    if (copy_option == COPY_ALL)
    {
	x = 0;
	y = 0;
	width = global_values.draw_area_width;
	height = global_values.draw_area_height;
    }
    else if (copy_option == COPY_VIS)
    {
        n = 0;
        XtSetArg(args[n], XmNclipWindow, &clip_window); n++;
        XtGetValues(disp_info[source_disp_num].swindow, args, n);
    
        n = 0;
        XtSetArg(args[n], XmNwidth, &d_width); n++;
        XtSetArg(args[n], XmNheight, &d_height); n++;
        XtGetValues(clip_window, args, n);
	width = (int) d_width;
	height = (int) d_height;
    
        XTranslateCoordinates(disp_info[source_disp_num].disp,
               XtWindow(clip_window), disp_info[source_disp_num].draw_win_id, 
               0, 0, &x, &y, &dummy);
	if (x < 0)	
	    x = 0;
	if (y < 0)
	    y = 0;
    }
    else
    {
	printf("ERROR: unknown copy option in copy_image_to_new_display.\n");
    }

    save_image_on_disk(disp_info[source_disp_num].disp,
		disp_info[source_disp_num].draw_win_id,
		disp_info[source_disp_num].backing_store,
		x, y, width, height, tmp_name);

    if ((the_image = read_image_from_disk(disp_info[disp_num].disp,
                           disp_info[disp_num].draw_win_id,
                           tmp_name, &width, &height, &depth)) == NULL)
    {
         popup_error_dialog(source_disp_num,
                           "ERROR in copy_image_to_new_display.");
         unlink(tmp_name);
         return;
    }

    /*
     * it is a valid xwd file, place it
     */
    if (DefaultDepth(disp_info[disp_num].disp,
                     DefaultScreen(disp_info[disp_num].disp)) != depth)
    {
         sprintf(tmp_str,
                   "Image not displayed on display %s.\nInconsistent Depths.",
                    disp_info[disp_num].disp_name);
         popup_info_dialog(source_disp_num, tmp_str);
         XDestroyImage(the_image);       /*free the image memory*/
    }
    else
    {
         XPutImage(disp_info[disp_num].disp,
                   disp_info[disp_num].backing_store,
                   disp_info[disp_num].win_gc[disp_num], the_image,
                   0, 0, x, y, width, height);
         XDestroyImage(the_image);
    }

    unlink(tmp_name);
}


/*
 * icon_exposeCB - this function is called when the icon window gets exposed
 */
XtEventHandler icon_exposeCB(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int disp_num;

    disp_num = ((int) client_data);

    XCopyArea(disp_info[disp_num].disp, disp_info[disp_num].icon_pix,
	      XtWindow(w), disp_info[disp_num].win_gc[disp_num], 
	      0, 0, icon_width, icon_height, 0, 0);
}


/*
 * read_xwd_file_into_drawing_area - this function reads the indicated xwd file 
 *                               into the upper left corner of the drawing
 *                               area for this display only.
 */
read_xwd_file_into_drawing_area(disp_num, file_name)
int disp_num;
char *file_name;
{
    XImage *the_image;
    int width, height, depth;
    char tmp_str[100];
    FILE *fp;

    if ((fp = fopen(file_name, "r")) == NULL)
	return;   /* file does not exist yet */
    else
	fclose(fp);

    if ((the_image = read_image_from_disk(disp_info[disp_num].disp,
                           disp_info[disp_num].draw_win_id,
                           file_name, &width, &height, &depth)) == NULL)
    {
         popup_error_dialog(disp_num,
                           "ERROR in read_xwd_file_into_drawing_area.");
         return;
    }

    /*
     * it is a valid xwd file, place it
     */
    if (DefaultDepth(disp_info[disp_num].disp,
                     DefaultScreen(disp_info[disp_num].disp)) != depth)
    {
         sprintf(tmp_str,
                   "Image not displayed on display %s.\nInconsistent Depths.",
                    disp_info[disp_num].disp_name);
         popup_info_dialog(disp_num, tmp_str);
         XDestroyImage(the_image);       /*free the image memory*/
    }
    else
    {
         XPutImage(disp_info[disp_num].disp,
                   disp_info[disp_num].backing_store,
                   disp_info[disp_num].win_gc[disp_num], the_image,
                   0, 0, 0, 0, width, height);
         XDestroyImage(the_image);
    }
}


/*
 * save_xwd_from_drawing_area - this function saves the picture in the drawing
 *                              area of this display into the filename.  This
 *                              function first "clips" the picture to only
 *                              the area that has paint on it, plus a small
 *                              border.
 */
save_xwd_from_drawing_area(disp_num, file_name)
int disp_num;
char *file_name;
{
    XImage *the_image;
    int x, y, width, height;
    int top, left, bottom, right;
    int n;
    Arg args[MAX_ARGS];
    Widget clip_window;
    unsigned long the_pixel;
    unsigned long bg_pixel;
    Pixel bg;

    x = 0;
    y = 0;
    width = global_values.draw_area_width;
    height = global_values.draw_area_height;
    
    if ((the_image = XGetImage(disp_info[disp_num].disp,
                disp_info[disp_num].backing_store,
                0, 0, width, height, (~0), ZPixmap)) == NULL)
    {
        printf("ERROR: could not get the_image save_xwd_from_drawing_area.\n");
    }
    
    n = 0;
    XtSetArg(args[n], XmNbackground, &bg); n++;
    XtGetValues(disp_info[disp_num].draw_area, args, n);
    bg_pixel = (unsigned long) bg;
    
    /*
     * now find the clipping box that has any and all painting in it
     */
    top = height + 1;
    left = width + 1;
    bottom = -1;
    right = -1;

    for (y=0; y < height; y++)
    {
        for (x=0; x < width; x++)
        {
            the_pixel = XGetPixel(the_image, x, y);

            if (the_pixel != bg_pixel)
            {
		if (x < left)
		    left = x;
		if (x > right)
		    right = x;
		if (y < top)
		    top = y;
		if (y > bottom)
		    bottom = y;
            }
        }
    }

    if ((top == height + 1) || (left == width + 1) || (bottom == -1) || 
	(right == -1))
    {
	/*
	 * the screen was blank, so save a very small dummy picture
	 */
	top = 0;
	left = 0;
	bottom = 3;
	right = 3;
    }

    save_image_on_disk(disp_info[disp_num].disp,
		disp_info[disp_num].draw_win_id,
		disp_info[disp_num].backing_store,
		left, top, right - left, bottom - top, file_name);
}


/*
 * wm_kill_window - this function is called when someone pulls down the
 *                  "Close" mwm window manager box.
 */
wm_kill_window(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int disp_num;

    disp_num = ((int) client_data);

    printf("wn_kill_window, disp_num=%d.\n", disp_num);
    process_menu_selection(disp_num, FILE_MENU, QUIT);
    printf("BACK OUT wn_kill_window, disp_num=%d.\n", disp_num);
}


/*
 * handle_wscrawl_window_moveCB - this function is called when the wscrawl
 *                                window has been moved (or configured at all).
 *                                This function then moves the selected 
 *                                region if appropriate.
 */
XtEventHandler handle_wscrawl_window_moveCB(w, client_data, our_event, cbool)
Widget w;
caddr_t client_data;
XEvent *our_event;
Boolean *cbool;
{
   int disp_num, i, x, y, root_x, root_y;
   Window dummy;
   XConfigureEvent *the_event;

   disp_num = ((int) client_data);

   if (our_event->type != ConfigureNotify)
       return;          /* only interested in XMoveWindow */

   the_event = (XConfigureEvent *) our_event;

   for (i=0; i < num_disps; i++)
   {
        if (disp_info[i].in_session_bool)
	{
            if (disp_info[i].selected_region.region_is_selected == TRUE)
	    {
		if (i == disp_num)    /* my region is different than others */
		{
		    x = disp_info[i].selected_region.cur_x;
		    y = disp_info[i].selected_region.cur_y;

                    XTranslateCoordinates(disp_info[disp_num].disp, 
                          disp_info[disp_num].draw_win_id,
                          RootWindow(disp_info[disp_num].disp,
                          DefaultScreen(disp_info[disp_num].disp)),
                          x, y, &root_x, &root_y, &dummy);
		}
		else
		{
		    root_x = disp_info[i].selected_region.cur_x;
		    root_y = disp_info[i].selected_region.cur_y;
		}
		
                XtMoveWidget(
	              disp_info[i].selected_region.selection_shell[disp_num], 
		      root_x, root_y);
		XRaiseWindow(disp_info[disp_num].disp, XtWindow(
		    disp_info[i].selected_region.selection_shell[disp_num]));
	    }
	}
   }
}


/*
 * handle_wscrawl_window_enterCB - this function is a hack to get keybord
 *                                 events to the correct place.  See the
 *                                 comment on "handle_enter_or_leave_draw_area"
 *                                 below.
 */
XtEventHandler handle_wscrawl_window_enterCB(w, client_data, our_event, cbool)
Widget w;
caddr_t client_data;
XEvent *our_event;
Boolean *cbool;
{
   int disp_num;

   disp_num = ((int) client_data);

   if (our_event->type == EnterNotify)
   {
       /*
       XtSetKeyboardFocus(disp_info[disp_num].app_shell, 
	                   disp_info[disp_num].draw_area);
       printf("Setting the input focus, disp_num=%d.\n", disp_num);
       XSetInputFocus(disp_info[disp_num].disp, disp_info[disp_num].draw_win_id,
                      RevertToPointerRoot, CurrentTime);
       */
   }
   else if (our_event->type == LeaveNotify)
   {
       ;
   }
   else
   {
       printf("ERROR: in handle_wscrawl_window_enter, unknown event type.\n");
   }
}


/*
 * handle_enter_or_leave_draw_area - this function WAS a hack to get keybord
 *                               events to appear 100 percent of the time.
 *                               Now I've found that setting
 *                               wscrawl*keyboardFocusPolicy: pointer
 *                               seems to fix my woes.
 *                               I do not know why this is necessary.
 *                               This function is also for other stuff now.
 */
XtEventHandler handle_enter_or_leave_draw_area(w, client_data, our_event, cbool)
Widget w;
caddr_t client_data;
XEvent *our_event;
Boolean *cbool;
{
    int disp_num, result;
    int i, oldtop, oldleft;

    disp_num = ((int) client_data);

    if (our_event->type == EnterNotify)
    {
	/*
        result = XGrabKeyboard(disp_info[disp_num].disp, 
	             disp_info[disp_num].draw_win_id,
		     False, GrabModeAsync, GrabModeAsync, CurrentTime);
        if (result == GrabSuccess)
	   printf("Grabbed keyboard successfully.\n");
	else
	   printf("ERROR: tried to grab keyboard, but got %d error.\n", result);
        */

        if (disp_info[disp_num].show_telepointer_always == TRUE)
            disp_info[disp_num].rubber_pointer.is_mapped_bool = TRUE;
    }
    else if (our_event->type == LeaveNotify)
    {
	/*
        XUngrabKeyboard(disp_info[disp_num].disp, CurrentTime); 
	 */
        if (disp_info[disp_num].show_telepointer_always == TRUE)
            unmap_pointer(disp_num);
    }
    else
    {
        printf("ERROR: handle_enter_or_leave_draw_area,unknown event type.\n");
    }
}


/*
 * draw_area_inputCB - this function handles all draw area input events
 *                     such as ButonPress or KeyPress.
 */
draw_area_inputCB(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int disp_num;
    XmDrawingAreaCallbackStruct *cb_struct;
    XEvent *the_event;

    cb_struct = (XmDrawingAreaCallbackStruct *) call_data;
    the_event = (XEvent *) cb_struct->event;
    disp_num = (int) client_data;

    switch(the_event->type)
    {
        case ButtonPress:
	    handle_ButtonPress_event(the_event, disp_num);
	    break;
        case ButtonRelease:
	    handle_ButtonRelease_event(the_event, disp_num);
	    break;
        case KeyPress:
	    handle_KeyPress_event(the_event, disp_num);
	    break;
        case KeyRelease:
	    break;
        case Expose:
	    printf("ERROR: Expose event in display: %d.\n", disp_num);
	    break;
	default:
	    printf("Unknown event type in draw_area_inputCB().\n");
	    break;
    }
}


/*
 * set_x_y_width_height - this function is used in several places.  It sorts
 *                        the coordinates to give the proper "upper-left"
 * 			  corner, and width and height.
 */
set_x_y_width_height(disp_num, x, y, width, height)
int disp_num;
int *x, *y, *width, *height;
{
    *x = (disp_info[disp_num].orig_rubberband_point.x <
         disp_info[disp_num].last_rubberband_point.x) ?
         disp_info[disp_num].orig_rubberband_point.x :
         disp_info[disp_num].last_rubberband_point.x;

    *y = (disp_info[disp_num].orig_rubberband_point.y <
         disp_info[disp_num].last_rubberband_point.y) ?
         disp_info[disp_num].orig_rubberband_point.y :
         disp_info[disp_num].last_rubberband_point.y;

    *width = abs(disp_info[disp_num].last_rubberband_point.x -
                disp_info[disp_num].orig_rubberband_point.x);

    *height = abs(disp_info[disp_num].last_rubberband_point.y -
                 disp_info[disp_num].orig_rubberband_point.y);
}


/*
 * my_isspace - this function returns TRUE if the character passed in is
 *              a newline, or space, or tab, and FALSE if not.
 */
my_isspace(c)
char c;
{
    switch ((int) c)
    {
        case 9:                 /* ^i tab */
            return(TRUE);
            break;
        case 10:                /* ^j linefeed */
            return(TRUE);
            break;
        case 13:                /* ^m carriage return */
            return(TRUE);
            break;
        case 32:                /* space */
            return(TRUE);
            break;
        default:
            return(FALSE);
            break;
    }
}


/*
 * get_dimensions_of_text_file - this function finds and returns the
 *              dimensions (in in screen pixels) of the text file that
 *              is passed in as an argument.
 */
get_dimensions_of_text_file(disp_num, text_file_ptr, width, height)
int disp_num;
FILE *text_file_ptr;
int *width, *height;
{
    int c, length, current_width, current_height, max_width_found;
    char current_line[512];

    for (max_width_found=0, current_height=0; (c=getc(text_file_ptr)) != EOF;)
    {
        ungetc(c, text_file_ptr);
        fgets(current_line, 510, text_file_ptr);       /*get the line*/
        for (length=0; current_line[length] != '\n'; length++)
            ;

        current_height += disp_info[disp_num].type_info.char_height;

        if ((current_width = XTextWidth(
              disp_info[disp_num].type_info.the_font_struct,
              current_line, length)) > max_width_found)
        {
            max_width_found = current_width;
        }
    }

    current_height += disp_info[disp_num].type_info.char_height;

    *width = max_width_found;
    *height = current_height;
}


/*
 * set_paused_cursors - this function sets all cursors on all displays to be
 *            "Pause" cursors.  This function is called when some action is
 *            done that totally sucks all cycles for a minute or two, like
 *            reading an image in.
 *
 */
set_paused_cursors()
{
    int i, j;

    for (i=0; i < num_disps; i++)
    {
        if (disp_info[i].in_session_bool) /*if window is alive*/
        {
            XDefineCursor(disp_info[i].disp, disp_info[i].draw_win_id,
                         disp_info[i].pause_cursor);

	    for (j=0; j < disp_info[i].NUM_EFF_MENUS; j++)
	    {
                XDefineCursor(disp_info[i].disp, 
	                     XtWindow(disp_info[i].menu[j].menu_bar_button),
                             disp_info[i].pause_cursor);
	    }

	    for (j=0; j < 12; j++)
	    {
                XDefineCursor(disp_info[i].disp, XtWindow(disp_info[i].tool[j]),
                             disp_info[i].pause_cursor);
	    }

	    /*
	     * regions also !!!
	     */
            XFlush(disp_info[i].disp);
        }
    }
}


/*
 * unset_paused_cursors - this function sets all cursors on all displays to be
 *            the cursors they were before the "Pause" occured.
 */
unset_paused_cursors()
{
    int i, j, the_cursor, which;

    for (i=0; i < num_disps; i++)
    {
        if (disp_info[i].in_session_bool) /*if window is alive*/
        {
            switch (disp_info[i].input_mode)
            {
                case POINTING:
                    the_cursor = XC_target;
                    break;
                case SELECTING:
                    the_cursor = XC_crosshair;
                    break;
                case PENCILING:
                    the_cursor = XC_pencil;
                    break;
                case TYPING:
                    the_cursor = XC_xterm;
                    break;
                case PAINTING:
	            which = disp_info[i].paintbrush.which_paintb;
	            XDefineCursor(disp_info[i].disp,
			    disp_info[i].draw_win_id,
			    disp_info[i].paintbrush.paintb_curs[which]);
                    break;
                case SPRAYPAINTING:
                    the_cursor = XC_spraycan;
                    break;
                case ERASING:
                    the_cursor = XC_draped_box;
                    break;
                case DRAWING_LINES:
                case OUTLINING_SQUARES:
                case FILLING_SQUARES:
                case OUTLINING_OVALS:
                case FILLING_OVALS:
                    the_cursor = XC_crosshair;
                    break;
                default:
                    break;
            }

	    if (disp_info[i].input_mode != PAINTING)
	    {
                XDefineCursor(disp_info[i].disp, disp_info[i].draw_win_id,
                     XCreateFontCursor(disp_info[i].disp, the_cursor));
	    }

	    for (j=0; j < disp_info[i].NUM_EFF_MENUS; j++)
	    {
                XDefineCursor(disp_info[i].disp, 
                      XtWindow(disp_info[i].menu[j].menu_bar_button),
                      XCreateFontCursor(disp_info[i].disp, XC_left_ptr));
	    }

	    for (j=0; j < 12; j++)
	    {
                XDefineCursor(disp_info[i].disp, XtWindow(disp_info[i].tool[j]),
                         XCreateFontCursor(disp_info[i].disp, XC_left_ptr));
	    }

            XFlush(disp_info[i].disp);
        }
    }
}


/*
 * load_moron_fonts - this function loads fonts for real morons and owners
 *              of Sun workstations.  The thing is, both morons and
 *              owners of Sun workstations have systems that are so screwed
 *              up, they can't find R4 fonts when you try to allocate them
 *              using the font name.  Therefore, this is a really, really ugly
 *              hack to allow their systems to find ONLY MY default fonts
 *              using xlfd strings.  This will work with NO OTHER fonts.
 *              This function returns TRUE if it allocated a font, and FALSE
 *              if not.  Hopefully this function will keep 90% of the Sun
 *              users happy, and thinking they got their moneys worth when
 *              buying Sun, when actually they made a drastic mistake.
 */
load_moron_fonts(disp_num, i, the_font_struct, font_file_name)
int disp_num, i;
XFontStruct **the_font_struct;
char *font_file_name;
{
    char xlfd_string[512];

    if (strcmp(font_file_name, "fixed") == 0)
        strcpy(xlfd_string, "fixed");
    else if (strcmp(font_file_name, "variable") == 0)
        strcpy(xlfd_string, "-*-helvetica-bold-r-normal-*-*-120-*-*-*-*-*-*");
    else if (strcmp(font_file_name, "timR12") == 0)
        strcpy(xlfd_string,
               "-adobe-times-medium-r-normal--12-120-75-75-p-64-iso8859-1");
    else if (strcmp(font_file_name, "helvO12") == 0)
        strcpy(xlfd_string,
               "-adobe-helvetica-medium-o-normal--12-120-75-75-p-67-iso8859-1");
    else if (strcmp(font_file_name, "courR12") == 0)
        strcpy(xlfd_string,
               "-adobe-courier-medium-r-normal--12-120-75-75-m-70-iso8859-1");
    else if (strcmp(font_file_name, "courR14") == 0)
        strcpy(xlfd_string,
               "-adobe-courier-medium-r-normal--14-140-75-75-m-90-iso8859-1");
    else if (strcmp(font_file_name, "timBI18") == 0)
        strcpy(xlfd_string,
               "-adobe-times-bold-i-normal--18-180-75-75-p-98-iso8859-1");
    else if (strcmp(font_file_name, "courR24") == 0)
        strcpy(xlfd_string,
               "-adobe-courier-medium-r-normal--24-240-75-75-m-150-iso8859-1");
    else if (strcmp(font_file_name, "timR24") == 0)
        strcpy(xlfd_string,
               "-adobe-times-medium-r-normal--24-240-75-75-p-124-iso8859-1");
    else if (strcmp(font_file_name, "timBI24") == 0)
        strcpy(xlfd_string,
               "-adobe-times-bold-i-normal--24-240-75-75-p-128-iso8859-1");
    else if (strcmp(font_file_name, "ncenR24") == 0)
        strcpy(xlfd_string,
 "-adobe-new century schoolbook-medium-r-normal--24-240-75-75-p-137-iso8859-1");
    else if (strcmp(font_file_name, "vbee-36") == 0)
        strcpy(xlfd_string, "vbee-36");
    else if (strcmp(font_file_name, "vr-40") == 0)
        strcpy(xlfd_string, "vr-40");
    else if (strcmp(font_file_name, "vgl-40") == 0)
        strcpy(xlfd_string, "vgl-40");
    else if (strcmp(font_file_name, "vgl-40") == 0)
        strcpy(xlfd_string, "vsgn-57");
    else
        return(FALSE);  /*not one of these fonts*/

    if ((*the_font_struct = XLoadQueryFont(disp_info[i].disp,
          xlfd_string)) != NULL)
    {
        return(TRUE);
    }
    else
    {
        return(FALSE);
    }
}


/*
 * my_rand - NOTE: THIS MAY REQUIRE A PORT: due to the non-standard "rand"
 *           function, this routine is not entirely portable.  The reason
 *           this function is here is to allow you to hack on it to get it
 *           to return a value between 0 and 1.  This only affects the
 *           airbrush.
 */
float my_rand()
{
    float ret_val;

    ret_val = (float) ((double)rand()/(double)MAX_RAND);

    if (ret_val > 1.0)
        ret_val = (ret_val - ((float) (1 * ((int) ret_val))));

    return(ret_val);
}


/*
 * my_sleep - this function sleeps using a busy wait.  It sleeps any float
 *            number of seconds.
 */
my_sleep(duration)
float duration;
{
    static struct timezone dummy;
    static struct timeval start_time, cur_time;
    unsigned long lapsed_seconds, lapsed_milli_seconds, duration_milli_seconds;

    duration_milli_seconds = (unsigned long) (duration * 1000.0);
    gettimeofday(&start_time, &dummy);

    lapsed_milli_seconds = 0;

    while (lapsed_milli_seconds < duration_milli_seconds)
    {
        gettimeofday(&cur_time, &dummy);

        lapsed_seconds = cur_time.tv_sec - start_time.tv_sec;

        if (lapsed_seconds == 0)
        {
            lapsed_milli_seconds = ((cur_time.tv_usec - start_time.tv_usec)  
				    / 1000);
        }
        else
        {
            lapsed_milli_seconds = (((1000000 - start_time.tv_usec) / 1000) +
                                    ((lapsed_seconds - 1) * 1000) + 
		                    (cur_time.tv_usec / 1000));
        }
    }
}


/*
 * handle_knock_knockCB - this function is called when there is input pending
 *                        on "my_socket", meaning someone is knocking on our
 *                        door to be let in.
 */
XtInputCallbackProc handle_knock_knockCB(client_data, source, id)
int *client_data;
int *source;
XtInputId *id;
{
    int disp_num, s, n, cli_addr_len;
    struct sockaddr cli_addr;
    struct knock_knock_msg_struct knock_msg;
    char tmp_str[200];

    disp_num = (int) client_data;

    cli_addr_len = sizeof(cli_addr);
    s = accept(disp_info[disp_num].my_socket, (struct sockaddr *) &cli_addr, 
	       &cli_addr_len);

    if (s < 0)
        printf("ERROR: in init_communication.\n");
    else
    {
        n = my_readn(s, (char *) &knock_msg, sizeof(knock_msg));
	sprintf(tmp_str, 
	     "Just got a Knock-Knock.  Attempting to\nadd display: %s.\n",
	     knock_msg.newhostname);
        popup_info_dialog(disp_num, tmp_str);
	add_display(knock_msg.newhostname, disp_num, COPY_ALL);
    }
}


/*
 * knock_knock - this function "knocks on the door" of the wscrawl running
 *               at the machine name "dest_machine".  If everything is groovey,
 *               then this wscrawl exits, and the dest_machine adds this person
 *               to its wscrawl session.
 */
knock_knock(disp_num, dest_machine)
int disp_num;
char *dest_machine;
{
    int i, n;
    int dest_socket;
    struct knock_knock_msg_struct the_mesg;
    char *disp_str, *ptr, tmp_str[50], tmp_str2[50];

    /*
     * correct for silly users who don't understand the difference between
     * a machine or a display, then open the socket.
     */
    for (i=0; dest_machine[i] != '\0' && dest_machine[i] != ':'; i++)
	;
    if (dest_machine[i] == ':')
	dest_machine[i] = '\0';

    dest_socket = my_open_socket_to_send_data(dest_machine, SERV_TCP_PORT);

    disp_str = XDisplayString(disp_info[disp_num].disp);
    strncpy(disp_info[disp_num].disp_name, disp_str, 40);

    if ((strncmp(disp_info[disp_num].disp_name, "local", 5) == 0) ||
        (strncmp(disp_info[disp_num].disp_name, "unix", 4) == 0) ||
        (strncmp(disp_info[disp_num].disp_name, ":", 1) == 0))
    {
        /*
         * this is a wscrawl window on the invoker's home display, so
         * change it to be meaningful.
         */
        gethostname(tmp_str, 30);
        for (ptr=disp_info[disp_num].disp_name; *ptr != ':'; ptr++)
	    ;
        sprintf(tmp_str2, "%s%s", tmp_str, ptr);
        strncpy(disp_info[disp_num].disp_name, tmp_str2, 40);
    }
    disp_str = disp_info[disp_num].disp_name;

    the_mesg.type = KNOCK_KNOCK_MESG;
    strncpy(the_mesg.newhostname, disp_str, MAX_HOST_NAME_LENGTH);

    n = sizeof(the_mesg);
    if (my_writen(dest_socket, (char *) &the_mesg, n) != n)
        printf("ERROR: in knock_knock.\n");

    close(dest_socket);
    return(TRUE);
}


/*
 * my_writen - this function writes "n" bytes down a socket before returning.
 */
my_writen(fd, ptr, nbytes)
register int fd;
register char *ptr;
register int nbytes;
{
    int nleft, nwritten;

    for (nleft = nbytes; nleft > 0; )
    {
        nwritten = write(fd, ptr, nleft);
        if (nwritten <= 0)
            return(nwritten);

        nleft -= nwritten;
        ptr += nwritten;
    }

    return(nbytes - nleft);
}


/*
 * my_readn - this function reads "n" bytes from a socket before returning.
 */
my_readn(fd, ptr, nbytes)
register int fd;
register char *ptr;
register int nbytes;
{
    int nleft, nread;

    for (nleft = nbytes; nleft > 0; )
    {
        nread = read(fd, ptr, nleft);
        if (nread <= 0)
	    return(nread);

	nleft -= nread;
	ptr += nread;
    }

    return(nbytes - nleft);
}
