/*
 * This File: wsdialog1.c
 *
 * This file contains half of the dialog creation and handling functions for 
 * wscrawl.
 */

#include "wscrawl.h"
#include "languages.h"

#include "brush0.h"	/* brushes in the paintbrush shapes dialog */
#include "brush1.h"
#include "brush2.h"
#include "brush3.h"
#include "brush4.h"
#include "brush5.h"
#include "brush6.h"
#include "brush7.h"
#include "brush8.h"
#include "brush9.h"
#include "brush10.h"
#include "brush11.h"
#include "brush12.h"
#include "brush13.h"
#include "brush14.h"
#include "brush15.h"
#include "brush16.h"
#include "brush17.h"
#include "brush18.h"
#include "brush19.h"

extern struct disp_info_struct disp_info[MAX_NUM_DISPS];
extern int num_disps;
extern XtAppContext appCon;
static XmStringCharSet charset = (XmStringCharSet) XmSTRING_DEFAULT_CHARSET;
XtTimerCallbackProc wait_freeze_super_selectCB();
extern struct global_values_struct global_values;
extern int handle_read_file_selection_dialog();
extern char *parse_and_fix_dialog_text();

struct brush_bitmap_struct
{
    int width;
    int height;
    int x_hot;
    int y_hot;
    char *bits;
};

struct brush_bitmap_struct brush_bitmaps[20];


/*
 * handle_save_file_selection_dialog - this function is the callback for the 
 *                        savefile selection dialog.
 */
handle_save_file_selection_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int n, i;
    Arg args[MAX_ARGS];
    int disp_num, which_button;
    FILE *fp;
    XmFileSelectionBoxCallbackStruct *fcb;
    char *fname, *filename;
    char tmp_str[256];

    disp_num = ((int) client_data) >> 8;
    which_button = ((int) client_data) & 0x00FF;

    switch(which_button)
    {
	case CANCEL:
	    XtUnmanageChild(
		disp_info[disp_num].save_file_dialog.file_selection_dialog);
	    jiggle_dialog(
		disp_info[disp_num].save_file_dialog.file_selection_dialog);
	    break;
	case OK:
	    XtUnmanageChild(
		disp_info[disp_num].save_file_dialog.file_selection_dialog);
	    jiggle_dialog(
		disp_info[disp_num].save_file_dialog.file_selection_dialog);
	    fcb = (XmFileSelectionBoxCallbackStruct *) call_data;
	    XmStringGetLtoR(fcb->value, charset, &fname);
	    filename = parse_and_fix_dialog_text(fname);
	    free(fname);
	    if (filename == NULL)
	    {
                popup_info_dialog(disp_num, "Could not parse filename.");
		return;
	    }
	    else
	    {
                if ((fp = fopen(filename, "w" )) == NULL)
		{
		    sprintf(tmp_str, 
			   "Could not open file called:\n%s\nfor reading.", 
			   filename);
		    popup_info_dialog(disp_num, tmp_str);
		    return;
                }
		else
		{
		    fclose(fp);
		    switch(disp_info[disp_num].save_file_dialog.format_chosen)
		    {
	                case BITMAP_FORMAT:
		            save_region_into_bitmap_file(disp_num, filename);
			    break;
	                case XWD_FORMAT:
		            save_region_into_xwd_file(disp_num, filename);
			    break;
	                case GREYSCALE_POSTSCRIPT:
			    save_region_into_ps_file(disp_num, filename);
			    break;
	                case PPM_FORMAT:
			    save_region_into_ppm_file(disp_num, filename);
			    break;
			default:
			    popup_error_dialog(disp_num, "unknown format");
			    break;
		    }
		}
	    }
	    break;
        default:
	    popup_error_dialog(disp_num, 
		 "ERROR: bad button in handle_save_file_selection_dialog.");
            break;
    }
}


/*
 * handle_save_file_dialog - this function is the callback for the savefile
 *                        dialog.
 */
handle_save_file_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int n, i;
    Arg args[MAX_ARGS];
    int disp_num, which_button;

    disp_num = ((int) client_data) >> 8;
    which_button = ((int) client_data) & 0x00FF;

    switch(which_button)
    {
	case BITMAP_FORMAT:
	case XWD_FORMAT:
	case GREYSCALE_POSTSCRIPT:
	case PPM_FORMAT:
	    disp_info[disp_num].save_file_dialog.format_chosen = which_button;
	    XtUnmanageChild(disp_info[disp_num].save_file_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].save_file_dialog.the_dialog);
	    XtManageChild(
		  disp_info[disp_num].save_file_dialog.file_selection_dialog);
	    break;
	case FILE_IO_CANCEL:
	    XtUnmanageChild(disp_info[disp_num].save_file_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].save_file_dialog.the_dialog);
	    break;
	default:
	    printf("ERROR: unknown event in handle_save_file_dialog.\n");
	    break;
    }
}


/*
 * create_save_file_dialog - this function creates the popup dialog that
 *                           prompts for file types to be save in.
 *
 */
Widget create_save_file_dialog(disp_num)
int disp_num;
{
    Widget the_dialog;
    Widget the_separator;
    Widget the_frame, the_form;
    Arg args[MAX_ARGS];
    int n;
    XmString mesg_string = NULL;

    n = 0;
    XtSetArg(args[n], XmNautoUnmanage, False); n++;
    XtSetArg(args[n], XmNtraversalOn, True); n++;
    the_dialog = XmCreateFormDialog(disp_info[disp_num].app_shell, 
			   "Save_File_Dialog", args, n);
    n = 0;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
    the_frame = XmCreateFrame(the_dialog, "dialog_frame", args, n);
    XtManageChild(the_frame);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNwidth, 400); n++;
    the_form = XmCreateForm(the_frame, "dialog_form", args, n);
    XtManageChild(the_form);

    mesg_string = XmStringCreateLtoR(
		    "Save the current selection in what format?",
                    charset);
    n = 0;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNlabelString, mesg_string); n++;
    disp_info[disp_num].save_file_dialog.label = 
		   XmCreateLabelGadget(the_form, "prompt", args, n);
    XtManageChild(disp_info[disp_num].save_file_dialog.label);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].save_file_dialog.label); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].save_file_dialog.bitmap_button =
                XmCreatePushButtonGadget(the_form, 
		      "Bitmap Format (X-Windows bitmap)", args, n);
    XtManageChild(disp_info[disp_num].save_file_dialog.bitmap_button);
    XtAddCallback(disp_info[disp_num].save_file_dialog.bitmap_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_save_file_dialog,
        (XtPointer)(disp_num << 8 | BITMAP_FORMAT));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].save_file_dialog.bitmap_button); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].save_file_dialog.image_button =
                XmCreatePushButtonGadget(the_form, 
		      "Image Format (xwd format)", args, n);
    XtManageChild(disp_info[disp_num].save_file_dialog.image_button);
    XtAddCallback(disp_info[disp_num].save_file_dialog.image_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_save_file_dialog,
        (XtPointer)(disp_num << 8 | XWD_FORMAT));
    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].save_file_dialog.image_button); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].save_file_dialog.greyscale_ps_button =
                XmCreatePushButtonGadget(the_form, 
		      "Greyscale Postscript", args, n);
    XtManageChild(disp_info[disp_num].save_file_dialog.greyscale_ps_button);
    XtAddCallback(disp_info[disp_num].save_file_dialog.greyscale_ps_button, 
	     XmNactivateCallback, (XtCallbackProc)handle_save_file_dialog,
        (XtPointer)(disp_num << 8 | GREYSCALE_POSTSCRIPT));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].save_file_dialog.greyscale_ps_button); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].save_file_dialog.ppm_button =
                XmCreatePushButtonGadget(the_form, 
		      "PPM Format (Portable Pixmap)", args, n);
    XtManageChild(disp_info[disp_num].save_file_dialog.ppm_button);
    XtAddCallback(disp_info[disp_num].save_file_dialog.ppm_button, 
	       XmNactivateCallback,
          (XtCallbackProc)handle_save_file_dialog, (XtPointer)(disp_num << 8 | PPM_FORMAT));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 0); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 0); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget,
             disp_info[disp_num].save_file_dialog.ppm_button); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    the_separator = XmCreateSeparatorGadget(the_form, "separator", args, n);
    XtManageChild(the_separator);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNleftPosition, 40); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].save_file_dialog.ppm_button); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNrightPosition, 60); n++;
    XtSetArg(args[n], XmNbottomOffset, 5); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].save_file_dialog.cancel_button =
                XmCreatePushButtonGadget(the_form, "Cancel", args, n);
    XtManageChild(disp_info[disp_num].save_file_dialog.cancel_button);
    XtAddCallback(disp_info[disp_num].save_file_dialog.cancel_button, 
		  XmNactivateCallback,
	     (XtCallbackProc)handle_save_file_dialog, (XtPointer)(disp_num << 8 | FILE_IO_CANCEL));

    if (mesg_string)
	XtFree((char*)mesg_string);
    
    set_dialog_wm_name(the_dialog, "Read In File Type");

    n = 0;
    XtSetArg(args[n], XmNtraversalOn, True); n++;
    disp_info[disp_num].save_file_dialog.file_selection_dialog = 
       XmCreateFileSelectionDialog(disp_info[disp_num].app_shell, 
				   "File_to_save_the_selection_as", 
				   args, n);
    set_dialog_wm_name(
	     disp_info[disp_num].save_file_dialog.file_selection_dialog,
	     "Save File As");
    XtAddCallback(disp_info[disp_num].save_file_dialog.file_selection_dialog,
		  XmNokCallback, (XtCallbackProc)handle_save_file_selection_dialog, 
		  (XtPointer)(disp_num << 8 | OK));
    XtAddCallback(disp_info[disp_num].save_file_dialog.file_selection_dialog,
		  XmNcancelCallback, (XtCallbackProc)handle_save_file_selection_dialog, 
		  (XtPointer)(disp_num << 8 | CANCEL));
    return(the_dialog);
}


/*
 * handle_scrollbar_dialog - this function is the callback for the scrollbar
 *                        dialog.
 */
handle_scrollbar_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int n, i;
    Arg args[MAX_ARGS];
    int disp_num, which_button;

    disp_num = ((int) client_data) >> 8;
    which_button = ((int) client_data) & 0x00FF;

    switch(which_button)
    {
	case SCROLL_SYNC:
	    global_values.sync_scrollbars_bool = TRUE;
	    sync_drawing_areas_with(disp_num);
	    XtUnmanageChild(disp_info[disp_num].scrollbar_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].scrollbar_dialog.the_dialog);
            merge_resources(disp_info[disp_num].disp, 
			    "wscrawl.syncScrollbars: True\n");
	    break;
	case SCROLL_A_SYNC:
	    global_values.sync_scrollbars_bool = FALSE;
	    XtUnmanageChild(disp_info[disp_num].scrollbar_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].scrollbar_dialog.the_dialog);
            merge_resources(disp_info[disp_num].disp, 
			    "wscrawl.syncScrollbars: False\n");
	    break;
	case SCROLL_CANCEL:
	    XtUnmanageChild(disp_info[disp_num].scrollbar_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].scrollbar_dialog.the_dialog);
	    break;
	default:
	    printf("ERROR: unknown event in handle_scrollbar_dialog.\n");
	    break;
    }
}


/*
 * create_scrollbar_dialog - this function creates the popup dialog that
 *                           prompts for scrollbar options.
 *
 */
Widget create_scrollbar_dialog(disp_num)
int disp_num;
{
    Widget the_dialog;
    Widget the_separator;
    Widget the_frame, the_form;
    Arg args[MAX_ARGS];
    int n;
    XmString mesg_string = NULL;

    n = 0;
    XtSetArg(args[n], XmNautoUnmanage, False); n++;
    XtSetArg(args[n], XmNtraversalOn, True); n++;
    the_dialog = XmCreateFormDialog(disp_info[disp_num].app_shell, 
			   "Scrollbar_Settings_Dialog", args, n);
    n = 0;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    the_frame = XmCreateFrame(the_dialog, "dialog_frame", args, n);
    XtManageChild(the_frame);

    n = 0;
    the_form = XmCreateForm(the_frame, "dialog_form", args, n);
    XtManageChild(the_form);

    mesg_string = XmStringCreateLtoR(
		    "Select The Scrollbar Behavior You Wish:", charset);
    n = 0;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNlabelString, mesg_string); n++;
    disp_info[disp_num].scrollbar_dialog.label = 
		   XmCreateLabelGadget(the_form, "prompt", args, n);
    XtManageChild(disp_info[disp_num].scrollbar_dialog.label);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].scrollbar_dialog.label); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].scrollbar_dialog.scroll_sync_button =
                XmCreatePushButtonGadget(the_form, 
		      "Scrollbars All Locked Together", args, n);
    XtManageChild(disp_info[disp_num].scrollbar_dialog.scroll_sync_button);
    XtAddCallback(disp_info[disp_num].scrollbar_dialog.scroll_sync_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_scrollbar_dialog,
        (XtPointer)(disp_num << 8 | SCROLL_SYNC));
    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].scrollbar_dialog.scroll_sync_button); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].scrollbar_dialog.scroll_a_sync_button =
                XmCreatePushButtonGadget(the_form, 
		      "Scrollbars Operate Independently", args, n);
    XtManageChild(disp_info[disp_num].scrollbar_dialog.scroll_a_sync_button);
    XtAddCallback(disp_info[disp_num].scrollbar_dialog.scroll_a_sync_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_scrollbar_dialog,
        (XtPointer)(disp_num << 8 | SCROLL_A_SYNC));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 0); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 0); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget,
             disp_info[disp_num].scrollbar_dialog.scroll_a_sync_button); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    the_separator = XmCreateSeparatorGadget(the_form, "separator", args, n);
    XtManageChild(the_separator);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNleftPosition, 40); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].scrollbar_dialog.scroll_a_sync_button); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNrightPosition, 60); n++;
    XtSetArg(args[n], XmNbottomOffset, 5); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].scrollbar_dialog.cancel_button =
                XmCreatePushButtonGadget(the_form, "Cancel", args, n);
    XtManageChild(disp_info[disp_num].scrollbar_dialog.cancel_button);
    XtAddCallback(disp_info[disp_num].scrollbar_dialog.cancel_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_scrollbar_dialog,
        (XtPointer)(disp_num << 8 | SCROLL_CANCEL));

    if (mesg_string)
	XtFree((char*)mesg_string);
    
    set_dialog_wm_name(the_dialog, "Scrollbar Settings");
    return(the_dialog);
}


/*
 * handle_dirty_add_display_dialog - this function is the callback for the 
 *                        obvious dialog.
 */
handle_dirty_add_display_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int n, i;
    Arg args[MAX_ARGS];
    int disp_num, which_button;

    disp_num = ((int) client_data) >> 8;
    which_button = ((int) client_data) & 0x00FF;

    switch(which_button)
    {
	case COPY_ALL:
	case COPY_VIS:
	case CLEAR_ALL:
	    XtUnmanageChild(
		     disp_info[disp_num].dirty_add_display_dialog.the_dialog);
	    jiggle_dialog(
		     disp_info[disp_num].dirty_add_display_dialog.the_dialog);
	    add_display(
	       disp_info[disp_num].dirty_add_display_dialog.new_dispname_str, 
	       disp_num, which_button);
	    break;
	case ADD_CANCEL:
	    XtUnmanageChild(
		     disp_info[disp_num].dirty_add_display_dialog.the_dialog);
	    jiggle_dialog(
		     disp_info[disp_num].dirty_add_display_dialog.the_dialog);
	    break;
	default:
	    printf("ERROR: unknown event handle_dirty_add_display_dialog.\n");
	    break;
    }
}


/*
 * create_dirty_add_display_dialog - this function creates the popup dialog 
 *                         that prompts for dirty copy options when adding
 *                         a display.
 *
 */
Widget create_dirty_add_display_dialog(disp_num)
int disp_num;
{
    Widget the_dialog;
    Widget the_separator;
    Widget the_frame, the_form;
    Arg args[MAX_ARGS];
    int n;
    char tmp_str[1024];
    XmString mesg_string = NULL;

    n = 0;
    XtSetArg(args[n], XmNautoUnmanage, False); n++;
    XtSetArg(args[n], XmNtraversalOn, True); n++;
    the_dialog = XmCreateFormDialog(disp_info[disp_num].app_shell, 
			   "Add_Display_Options_Dialog", args, n);

    n = 0;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
    the_frame = XmCreateFrame(the_dialog, "dialog_frame", args, n);
    XtManageChild(the_frame);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNwidth, 400); n++;
    the_form = XmCreateForm(the_frame, "dialog_form", args, n);
    XtManageChild(the_form);

    sprintf(tmp_str, "%s\n%s\n%s\n%s",
	    "The shared drawing surface has been drawn on.",
	    "Copying this drawing to the new display can",
	    "be very time consuming.  Please select one of",
	    "the following options:");

    mesg_string = XmStringCreateLtoR(tmp_str, charset);

    n = 0;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
    XtSetArg(args[n], XmNlabelString, mesg_string); n++;
    disp_info[disp_num].dirty_add_display_dialog.label = 
		   XmCreateLabelGadget(the_form, "prompt", args, n);
    XtManageChild(disp_info[disp_num].dirty_add_display_dialog.label);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].dirty_add_display_dialog.label); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].dirty_add_display_dialog.copy_all_button =
       XmCreatePushButtonGadget(the_form, 
       "Copy Entire Whiteboard (VERY time consuming)", args, n);
    XtManageChild(disp_info[disp_num].dirty_add_display_dialog.copy_all_button);
    XtAddCallback(disp_info[disp_num].dirty_add_display_dialog.copy_all_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_dirty_add_display_dialog,
        (XtPointer)(disp_num << 8 | COPY_ALL));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].dirty_add_display_dialog.copy_all_button); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].dirty_add_display_dialog.copy_vis_button =
       XmCreatePushButtonGadget(the_form, 
       "Copy Visible Region Only (less time consuming)", args, n);
    XtManageChild(disp_info[disp_num].dirty_add_display_dialog.copy_vis_button);
    XtAddCallback(disp_info[disp_num].dirty_add_display_dialog.copy_vis_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_dirty_add_display_dialog,
        (XtPointer)(disp_num << 8 | COPY_VIS));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].dirty_add_display_dialog.copy_vis_button); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].dirty_add_display_dialog.clear_all_button =
       XmCreatePushButtonGadget(the_form, 
              "Clear All Whiteboards (very fast)", args, n);
   XtManageChild(disp_info[disp_num].dirty_add_display_dialog.clear_all_button);
    XtAddCallback(disp_info[disp_num].dirty_add_display_dialog.clear_all_button,
		  XmNactivateCallback, (XtCallbackProc)handle_dirty_add_display_dialog,
        (XtPointer)(disp_num << 8 | CLEAR_ALL));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 0); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 0); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget,
           disp_info[disp_num].dirty_add_display_dialog.clear_all_button); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    the_separator = XmCreateSeparatorGadget(the_form, "separator", args, n);
    XtManageChild(the_separator);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNleftPosition, 40); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, 
         disp_info[disp_num].dirty_add_display_dialog.clear_all_button); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNrightPosition, 60); n++;
    XtSetArg(args[n], XmNbottomOffset, 5); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].dirty_add_display_dialog.cancel_button =
                XmCreatePushButtonGadget(the_form, "Cancel", args, n);
    XtManageChild(disp_info[disp_num].dirty_add_display_dialog.cancel_button);
    XtAddCallback(disp_info[disp_num].dirty_add_display_dialog.cancel_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_dirty_add_display_dialog,
        (XtPointer)(disp_num << 8 | ADD_CANCEL));

    if (mesg_string)
	XtFree((char*)mesg_string);
    
    set_dialog_wm_name(the_dialog, "Add Display Options");
    return(the_dialog);
}


/*
 * handle_read_file_dialog - this function is the callback for the readfile
 *                        dialog.
 */
handle_read_file_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int n, i;
    Arg args[MAX_ARGS];
    int disp_num, which_button;

    disp_num = ((int) client_data) >> 8;
    which_button = ((int) client_data) & 0x00FF;

    switch(which_button)
    {
	case BITMAP_FORMAT:
	case XWD_FORMAT:
	case TEXT_FILE_FORMAT:
	case GREYSCALE_POSTSCRIPT:
	case PPM_FORMAT:
	    disp_info[disp_num].read_file_dialog.format_chosen = which_button;
	    XtUnmanageChild(disp_info[disp_num].read_file_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].read_file_dialog.the_dialog);
	    XtManageChild(
		  disp_info[disp_num].read_file_dialog.file_selection_dialog);
	    break;
	case FILE_IO_CANCEL:
	    XtUnmanageChild(disp_info[disp_num].read_file_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].read_file_dialog.the_dialog);
	    break;
	default:
	    printf("ERROR: unknown event in handle_read_file_dialog.\n");
	    break;
    }
}


/*
 * create_read_file_dialog - this function creates the popup dialog that
 *                           prompts for file types to be read in.
 *
 */
Widget create_read_file_dialog(disp_num)
int disp_num;
{
    Widget the_dialog;
    Widget the_separator;
    Widget the_frame, the_form;
    Arg args[MAX_ARGS];
    int n;
    XmString mesg_string = NULL;

    n = 0;
    XtSetArg(args[n], XmNautoUnmanage, False); n++;
    XtSetArg(args[n], XmNtraversalOn, True); n++;
    the_dialog = XmCreateFormDialog(disp_info[disp_num].app_shell, 
			   "Read_File_Dialog", args, n);

    n = 0;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
    the_frame = XmCreateFrame(the_dialog, "dialog_frame", args, n);
    XtManageChild(the_frame);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNwidth, 400); n++;
    the_form = XmCreateForm(the_frame, "dialog_form", args, n);
    XtManageChild(the_form);

    mesg_string = XmStringCreateLtoR("The file to be read is in which Format?",
       charset);

    n = 0;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNlabelString, mesg_string); n++;
    disp_info[disp_num].read_file_dialog.label = 
		   XmCreateLabelGadget(the_form, "prompt", args, n);
    XtManageChild(disp_info[disp_num].read_file_dialog.label);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].read_file_dialog.label); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].read_file_dialog.bitmap_button =
                XmCreatePushButtonGadget(the_form, 
		      "Bitmap Format (X-Windows bitmap)", args, n);
    XtManageChild(disp_info[disp_num].read_file_dialog.bitmap_button);
    XtAddCallback(disp_info[disp_num].read_file_dialog.bitmap_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_read_file_dialog,
        (XtPointer)(disp_num << 8 | BITMAP_FORMAT));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].read_file_dialog.bitmap_button); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].read_file_dialog.image_button =
                XmCreatePushButtonGadget(the_form, 
		      "Image Format (xwd format)", args, n);
    XtManageChild(disp_info[disp_num].read_file_dialog.image_button);
    XtAddCallback(disp_info[disp_num].read_file_dialog.image_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_read_file_dialog,
        (XtPointer)(disp_num << 8 | XWD_FORMAT));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].read_file_dialog.image_button); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].read_file_dialog.text_file_button =
                XmCreatePushButtonGadget(the_form, 
		      "Text File", args, n);
    XtManageChild(disp_info[disp_num].read_file_dialog.text_file_button);
    XtAddCallback(disp_info[disp_num].read_file_dialog.text_file_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_read_file_dialog,
        (XtPointer)(disp_num << 8 | TEXT_FILE_FORMAT));

    /*
    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].read_file_dialog.text_file_button); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].read_file_dialog.monochrome_ps_button =
                XmCreatePushButtonGadget(the_form, 
		      "Greyscale Postscript", args, n);
    XtManageChild(disp_info[disp_num].read_file_dialog.monochrome_ps_button);
    XtAddCallback(disp_info[disp_num].read_file_dialog.monochrome_ps_button, 
	  XmNactivateCallback,
          handle_read_file_dialog, disp_num << 8 | GREYSCALE_POSTSCRIPT);
    */

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].read_file_dialog.text_file_button); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].read_file_dialog.ppm_button =
                XmCreatePushButtonGadget(the_form, 
		      "PPM Format (Portable Pixmap)", args, n);
    XtManageChild(disp_info[disp_num].read_file_dialog.ppm_button);
    XtAddCallback(disp_info[disp_num].read_file_dialog.ppm_button, 
	     XmNactivateCallback, (XtCallbackProc)handle_read_file_dialog,
        (XtPointer)(disp_num << 8 | PPM_FORMAT));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 0); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 0); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget,
             disp_info[disp_num].read_file_dialog.ppm_button); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    the_separator = XmCreateSeparatorGadget(the_form, "separator", args, n);
    XtManageChild(the_separator);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNleftPosition, 40); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].read_file_dialog.ppm_button); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNrightPosition, 60); n++;
    XtSetArg(args[n], XmNbottomOffset, 5); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].read_file_dialog.cancel_button =
                XmCreatePushButtonGadget(the_form, "Cancel", args, n);
    XtManageChild(disp_info[disp_num].read_file_dialog.cancel_button);
    XtAddCallback(disp_info[disp_num].read_file_dialog.cancel_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_read_file_dialog,
        (XtPointer)(disp_num << 8 | FILE_IO_CANCEL));

    if (mesg_string)
	XtFree((char*)mesg_string);
    
    set_dialog_wm_name(the_dialog, "Read In File Type");

    n = 0;
    XtSetArg(args[n], XmNtraversalOn, True); n++;
    disp_info[disp_num].read_file_dialog.file_selection_dialog = 
       XmCreateFileSelectionDialog(disp_info[disp_num].app_shell, 
				   "Name_The_File_To_Read_In", args, n);
    set_dialog_wm_name(
	     disp_info[disp_num].read_file_dialog.file_selection_dialog,
	     "Read In File");
    XtAddCallback(disp_info[disp_num].read_file_dialog.file_selection_dialog,
		  XmNokCallback, (XtCallbackProc)handle_read_file_selection_dialog, 
		  (XtPointer)(disp_num << 8 | OK));
    XtAddCallback(disp_info[disp_num].read_file_dialog.file_selection_dialog,
		  XmNcancelCallback, (XtCallbackProc)handle_read_file_selection_dialog, 
		  (XtPointer)(disp_num << 8 | CANCEL));
    return(the_dialog);
}


/*
 * handle_brush_dialog - this function is the callback for the brush
 *                        dialog.
 */
handle_brush_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int n, i, j, index, x, y, width, height;
    Arg args[MAX_ARGS];
    Widget clip_window;
    int disp_num, which_brush;
    char *the_text;
    Pixel fg, bg;
    Pixmap source, mask;
    XGCValues gcvalues;
    Window dummy;

    disp_num = ((int) client_data) >> 8;
    which_brush = ((int) client_data) & 0x00FF;

    if (which_brush == BRUSH_OK)
    {
        XtUnmanageChild(disp_info[disp_num].brush_dialog.the_dialog);
	jiggle_dialog(disp_info[disp_num].brush_dialog.the_dialog);
	return;
    }

    switch(which_brush)
    {
	case 0:
	    disp_info[disp_num].paintbrush.paint_mode = NORMAL;
	    disp_info[disp_num].paintbrush.brush_width = 3;
	    break;
	case 1:
	    disp_info[disp_num].paintbrush.paint_mode = NORMAL;
	    disp_info[disp_num].paintbrush.brush_width = 5;
	    break;
	case 2:
	    disp_info[disp_num].paintbrush.paint_mode = NORMAL;
	    disp_info[disp_num].paintbrush.brush_width = 8;
	    break;
	case 3:
	    disp_info[disp_num].paintbrush.paint_mode = NORMAL;
	    disp_info[disp_num].paintbrush.brush_width = 18;
	    break;
	case 4:
	    disp_info[disp_num].paintbrush.paint_mode = NORMAL;
	    disp_info[disp_num].paintbrush.brush_width = 32;
	    break;
	case 5:
	    disp_info[disp_num].paintbrush.paint_mode = BITMAP_BRUSH;
	    disp_info[disp_num].paintbrush.num_points = 7;
	    for (i = -3; i < 4; i++)
	    {
	        disp_info[disp_num].paintbrush.x[i + 3] = i;
	        disp_info[disp_num].paintbrush.y[i + 3] = (0 - i);
	    }
	    break;
	case 6:
	    disp_info[disp_num].paintbrush.paint_mode = BITMAP_BRUSH;
	    disp_info[disp_num].paintbrush.num_points = 17;
	    for (i = -8; i < 9; i++)
	    {
	        disp_info[disp_num].paintbrush.x[i + 8] = i;
	        disp_info[disp_num].paintbrush.y[i + 8] = (0 - i);
	    }
	    break;
	case 7:
	    disp_info[disp_num].paintbrush.paint_mode = BITMAP_BRUSH;
	    disp_info[disp_num].paintbrush.num_points = 5;
	    for (i = 0; i < 5; i++)
	    {
	        disp_info[disp_num].paintbrush.x[i] = i*3 - 6;
	        disp_info[disp_num].paintbrush.y[i] = (0 - (i*3 - 6));
	    }
	    break;
	case 8:
	    disp_info[disp_num].paintbrush.paint_mode = BITMAP_BRUSH;
	    disp_info[disp_num].paintbrush.num_points = 7;
	    for (i = 0; i < 7; i++)
	    {
	        disp_info[disp_num].paintbrush.x[i] = i*3 - 9;
	        disp_info[disp_num].paintbrush.y[i] = (0 - (i*3 - 9));
	    }
	    break;
	case 9:
	    disp_info[disp_num].paintbrush.paint_mode = BITMAP_BRUSH;
	    disp_info[disp_num].paintbrush.num_points = 22;
	    for (i = 0; i < 11; i++)
	    {
	        disp_info[disp_num].paintbrush.x[i] = i - 5;
	        disp_info[disp_num].paintbrush.y[i] = (0 - (i - 5));
	    }
	    for (i = 0; i < 11; i++)
	    {
	        disp_info[disp_num].paintbrush.x[i+11] = i - 5;
	        disp_info[disp_num].paintbrush.y[i+11] = i - 5;
	    }
	    break;
	case 10:
	    disp_info[disp_num].paintbrush.paint_mode = BITMAP_BRUSH;
	    disp_info[disp_num].paintbrush.num_points = 14;
	    for (i = -3; i < 4; i++)
	    {
	        disp_info[disp_num].paintbrush.x[i + 3] = i;
	        disp_info[disp_num].paintbrush.y[i + 3] = (0 - i);
	        disp_info[disp_num].paintbrush.x[i + 10] = i+1;
	        disp_info[disp_num].paintbrush.y[i + 10] = (0 - i);
	    }
	    break;
	case 11:
	    disp_info[disp_num].paintbrush.paint_mode = BITMAP_BRUSH;
	    disp_info[disp_num].paintbrush.num_points = 34;
	    for (i = -8; i < 9; i++)
	    {
	        disp_info[disp_num].paintbrush.x[i + 8] = i;
	        disp_info[disp_num].paintbrush.y[i + 8] = (0 - i);
	        disp_info[disp_num].paintbrush.x[i + 25] = i + 1;
	        disp_info[disp_num].paintbrush.y[i + 25] = (0 - i);
	    }
	    break;
	case 12:
	    disp_info[disp_num].paintbrush.paint_mode = BITMAP_BRUSH;
	    disp_info[disp_num].paintbrush.num_points = 14;
	    for (i = -3; i < 4; i++)
	    {
	        disp_info[disp_num].paintbrush.x[i + 3] = i;
	        disp_info[disp_num].paintbrush.y[i + 3] = i;
	        disp_info[disp_num].paintbrush.x[i + 10] = i;
	        disp_info[disp_num].paintbrush.y[i + 10] = i+1;
	    }
	    break;
	case 13:
	    disp_info[disp_num].paintbrush.paint_mode = BITMAP_BRUSH;
	    disp_info[disp_num].paintbrush.num_points = 34;
	    for (i = -8; i < 9; i++)
	    {
	        disp_info[disp_num].paintbrush.x[i + 8] = i;
	        disp_info[disp_num].paintbrush.y[i + 8] = i;
	        disp_info[disp_num].paintbrush.x[i + 25] = i;
	        disp_info[disp_num].paintbrush.y[i + 25] = i + 1;
	    }
	    break;
	case 14:
	    disp_info[disp_num].paintbrush.paint_mode = BITMAP_BRUSH;
	    disp_info[disp_num].paintbrush.num_points = 0;
	    for (j=0; j < 7; j++)
	    {
	        for (i=0; i < 7; i++)
	        {
	            disp_info[disp_num].paintbrush.x[j*7 + i] = i-3;
	            disp_info[disp_num].paintbrush.y[j*7 + i] = j-3;
	            (disp_info[disp_num].paintbrush.num_points)++;
	        }
	    }
	    break;
	case 15:
	    disp_info[disp_num].paintbrush.paint_mode = BITMAP_BRUSH;
	    disp_info[disp_num].paintbrush.num_points = 8;
	    for (i = 0; i < 8; i++)
	    {
	        disp_info[disp_num].paintbrush.x[i] = 0;
	        disp_info[disp_num].paintbrush.y[i] = i-3;
	    }
	    break;
	case 16:
	    disp_info[disp_num].paintbrush.paint_mode = BITMAP_BRUSH;
	    disp_info[disp_num].paintbrush.num_points = 19;
	    for (i = 0; i < 19; i++)
	    {
	        disp_info[disp_num].paintbrush.x[i] = 0;
	        disp_info[disp_num].paintbrush.y[i] = i-9;
	    }
	    break;
	case 17:
	    disp_info[disp_num].paintbrush.paint_mode = BITMAP_BRUSH;
	    disp_info[disp_num].paintbrush.num_points = 10;
	    for (i = 0; i < 10; i++)
	    {
	        disp_info[disp_num].paintbrush.x[i] = i - 4;
	        disp_info[disp_num].paintbrush.y[i] = 0;
	    }
	    break;
	case 18:
	    disp_info[disp_num].paintbrush.paint_mode = BITMAP_BRUSH;
	    disp_info[disp_num].paintbrush.num_points = 22;
	    for (i = 0; i < 22; i++)
	    {
	        disp_info[disp_num].paintbrush.x[i] = i - 10;
	        disp_info[disp_num].paintbrush.y[i] = 0;
	    }
	    break;
	case 19:
	    disp_info[disp_num].paintbrush.paint_mode = BITMAP_BRUSH;
	    disp_info[disp_num].paintbrush.num_points = 25;
	    index=0;
	    for (j=0; j < 4; j++)
	    {
	        for (i=(-j); i <= j; i++)
	        {
	            disp_info[disp_num].paintbrush.x[index] = i;
	            disp_info[disp_num].paintbrush.y[index] = j-3;
		    index++;
	        }
	    }

	    for (j=2; j >= 0; j--)
	    {
	        for (i=(-j); i <= j; i++)
	        {
	            disp_info[disp_num].paintbrush.x[index] = i;
	            disp_info[disp_num].paintbrush.y[index] = 3-j;
		    index++;
	        }
	    }
	    break;
	default:
	    printf("ERROR: unknown event in handle_brush_dialog.\n");
	    break;
    }

    disp_info[disp_num].input_mode = PAINTING;
    disp_info[disp_num].paintbrush.which_paintb = which_brush;

    XDefineCursor(disp_info[disp_num].disp, disp_info[disp_num].draw_win_id,
		  disp_info[disp_num].paintbrush.paintb_curs[which_brush]);

    if (disp_info[disp_num].paintbrush.paint_mode == NORMAL)
    {
        for (i=0; i < num_disps; i++)
        {
            if (disp_info[i].in_session_bool)
            {
                gcvalues.line_width =disp_info[disp_num].paintbrush.brush_width;
                XChangeGC(disp_info[i].disp, disp_info[disp_num].win_gc[i],
                              GCLineWidth, &gcvalues);
            }
        }
    }
    else if (disp_info[disp_num].paintbrush.paint_mode == BITMAP_BRUSH)
    {
        for (i=0; i < num_disps; i++)
        {
            if (disp_info[i].in_session_bool)
            {
                gcvalues.line_width = 0;
                XChangeGC(disp_info[i].disp, disp_info[disp_num].win_gc[i],
                              GCLineWidth, &gcvalues);
            }
        }
    }
    else
    {
	printf("ERROR in handle_brush_dialog, unknown paint_mode.\n");
    }
}


/*
 * put_bitmaps_into_brush_dialog - this function puts the brush patterns
 * 				   into the brush dialog.
 */
put_bitmaps_into_brush_dialog(disp_num)
int disp_num;
{
    Pixmap button_pixmap;
    Pixel fg, bg;
    XColor fg1, bg1;
    int n, i;
    Arg args[MAX_ARGS];
    Pixmap source, mask;

    brush_bitmaps[0].width = brush0_width;
    brush_bitmaps[0].height = brush0_height;
    brush_bitmaps[0].bits = brush0_bits;
    brush_bitmaps[0].x_hot = brush0_x_hot;
    brush_bitmaps[0].y_hot = brush0_y_hot;

    brush_bitmaps[1].width = brush1_width;
    brush_bitmaps[1].height = brush1_height;
    brush_bitmaps[1].bits = brush1_bits;
    brush_bitmaps[1].x_hot = brush1_x_hot;
    brush_bitmaps[1].y_hot = brush1_y_hot;

    brush_bitmaps[2].width = brush2_width;
    brush_bitmaps[2].height = brush2_height;
    brush_bitmaps[2].bits = brush2_bits;
    brush_bitmaps[2].x_hot = brush2_x_hot;
    brush_bitmaps[2].y_hot = brush2_y_hot;

    brush_bitmaps[2].width = brush2_width;
    brush_bitmaps[2].height = brush2_height;
    brush_bitmaps[2].bits = brush2_bits;
    brush_bitmaps[2].x_hot = brush2_x_hot;
    brush_bitmaps[2].y_hot = brush2_y_hot;

    brush_bitmaps[3].width = brush3_width;
    brush_bitmaps[3].height = brush3_height;
    brush_bitmaps[3].bits = brush3_bits;
    brush_bitmaps[3].x_hot = brush3_x_hot;
    brush_bitmaps[3].y_hot = brush3_y_hot;

    brush_bitmaps[4].width = brush4_width;
    brush_bitmaps[4].height = brush4_height;
    brush_bitmaps[4].bits = brush4_bits;
    brush_bitmaps[4].x_hot = brush4_x_hot;
    brush_bitmaps[4].y_hot = brush4_y_hot;

    brush_bitmaps[5].width = brush5_width;
    brush_bitmaps[5].height = brush5_height;
    brush_bitmaps[5].bits = brush5_bits;
    brush_bitmaps[5].x_hot = brush5_x_hot;
    brush_bitmaps[5].y_hot = brush5_y_hot;

    brush_bitmaps[6].width = brush6_width;
    brush_bitmaps[6].height = brush6_height;
    brush_bitmaps[6].bits = brush6_bits;
    brush_bitmaps[6].x_hot = brush6_x_hot;
    brush_bitmaps[6].y_hot = brush6_y_hot;

    brush_bitmaps[7].width = brush7_width;
    brush_bitmaps[7].height = brush7_height;
    brush_bitmaps[7].bits = brush7_bits;
    brush_bitmaps[7].x_hot = brush7_x_hot;
    brush_bitmaps[7].y_hot = brush7_y_hot;

    brush_bitmaps[8].width = brush8_width;
    brush_bitmaps[8].height = brush8_height;
    brush_bitmaps[8].bits = brush8_bits;
    brush_bitmaps[8].x_hot = brush8_x_hot;
    brush_bitmaps[8].y_hot = brush8_y_hot;

    brush_bitmaps[9].width = brush9_width;
    brush_bitmaps[9].height = brush9_height;
    brush_bitmaps[9].bits = brush9_bits;
    brush_bitmaps[9].x_hot = brush9_x_hot;
    brush_bitmaps[9].y_hot = brush9_y_hot;

    brush_bitmaps[10].width = brush10_width;
    brush_bitmaps[10].height = brush10_height;
    brush_bitmaps[10].bits = brush10_bits;
    brush_bitmaps[10].x_hot = brush10_x_hot;
    brush_bitmaps[10].y_hot = brush10_y_hot;

    brush_bitmaps[11].width = brush11_width;
    brush_bitmaps[11].height = brush11_height;
    brush_bitmaps[11].bits = brush11_bits;
    brush_bitmaps[11].x_hot = brush11_x_hot;
    brush_bitmaps[11].y_hot = brush11_y_hot;

    brush_bitmaps[12].width = brush12_width;
    brush_bitmaps[12].height = brush12_height;
    brush_bitmaps[12].bits = brush12_bits;
    brush_bitmaps[12].x_hot = brush12_x_hot;
    brush_bitmaps[12].y_hot = brush12_y_hot;

    brush_bitmaps[13].width = brush13_width;
    brush_bitmaps[13].height = brush13_height;
    brush_bitmaps[13].bits = brush13_bits;
    brush_bitmaps[13].x_hot = brush13_x_hot;
    brush_bitmaps[13].y_hot = brush13_y_hot;

    brush_bitmaps[14].width = brush14_width;
    brush_bitmaps[14].height = brush14_height;
    brush_bitmaps[14].bits = brush14_bits;
    brush_bitmaps[14].x_hot = brush14_x_hot;
    brush_bitmaps[14].y_hot = brush14_y_hot;

    brush_bitmaps[15].width = brush15_width;
    brush_bitmaps[15].height = brush15_height;
    brush_bitmaps[15].bits = brush15_bits;
    brush_bitmaps[15].x_hot = brush15_x_hot;
    brush_bitmaps[15].y_hot = brush15_y_hot;

    brush_bitmaps[16].width = brush16_width;
    brush_bitmaps[16].height = brush16_height;
    brush_bitmaps[16].bits = brush16_bits;
    brush_bitmaps[16].x_hot = brush16_x_hot;
    brush_bitmaps[16].y_hot = brush16_y_hot;

    brush_bitmaps[17].width = brush17_width;
    brush_bitmaps[17].height = brush17_height;
    brush_bitmaps[17].bits = brush17_bits;
    brush_bitmaps[17].x_hot = brush17_x_hot;
    brush_bitmaps[17].y_hot = brush17_y_hot;

    brush_bitmaps[18].width = brush18_width;
    brush_bitmaps[18].height = brush18_height;
    brush_bitmaps[18].bits = brush18_bits;
    brush_bitmaps[18].x_hot = brush18_x_hot;
    brush_bitmaps[18].y_hot = brush18_y_hot;

    brush_bitmaps[19].width = brush19_width;
    brush_bitmaps[19].height = brush19_height;
    brush_bitmaps[19].bits = brush19_bits;
    brush_bitmaps[19].x_hot = brush19_x_hot;
    brush_bitmaps[19].y_hot = brush19_y_hot;

    /*
     * define the first paintbrush cursor
     */
    fg1.flags = DoRed | DoGreen | DoBlue;
    bg1.flags = DoRed | DoGreen | DoBlue;
    fg1.red = fg1.blue = fg1.green = 0;
    bg1.red = bg1.blue = bg1.green = 0;

    fg = WhitePixelOfScreen(DefaultScreenOfDisplay(disp_info[disp_num].disp));
    bg = BlackPixelOfScreen(DefaultScreenOfDisplay(disp_info[disp_num].disp));

    for (i=0; i < 20; i++)
    {
        source = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, brush_bitmaps[i].bits,
                      brush_bitmaps[i].width, brush_bitmaps[i].height, 1, 0, 1);
        mask = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, brush_bitmaps[i].bits,
                      brush_bitmaps[i].width, brush_bitmaps[i].height, 1, 0, 1);

        disp_info[disp_num].paintbrush.paintb_curs[i] = XCreatePixmapCursor(
		   disp_info[disp_num].disp, source, mask, &fg1, &bg1,
		   brush_bitmaps[i].x_hot, brush_bitmaps[i].y_hot);
    }

    n = 0;
    XtSetArg(args[n], XmNforeground, &fg); n++;
    XtSetArg(args[n], XmNbackground, &bg); n++;
    XtGetValues(disp_info[disp_num].brush_dialog.brush_shape[0], args, n);

    for (i=0; i < 20; i++)
    {
        button_pixmap = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, brush_bitmaps[i].bits,
                      brush_bitmaps[i].width, brush_bitmaps[i].height, fg, bg,
                      DefaultDepth(disp_info[disp_num].disp,
                      DefaultScreen(disp_info[disp_num].disp)));
        n = 0;
        XtSetArg(args[n], XmNlabelType, XmPIXMAP); n++;
        XtSetArg(args[n], XmNlabelPixmap, button_pixmap); n++;
        XtSetArg(args[n], XmNfillOnArm, False); n++;
        XtSetArg(args[n], XmNwidth, 48); n++;
        XtSetArg(args[n], XmNheight, 48); n++;
        XtSetValues(disp_info[disp_num].brush_dialog.brush_shape[i], args, n);
    }
}


/*
 * create_brush_dialog - this function creates the popup dialog that
 *                        controls the brush shapes.
 *
 */
Widget create_brush_dialog(disp_num)
int disp_num;
{
    Widget the_dialog;
    Widget the_separator;
    Widget the_frame, the_form;
    Arg args[MAX_ARGS];
    int i, j, n, the_offset = 3;
    XmString mesg_string = NULL;

    n = 0;
    XtSetArg(args[n], XmNautoUnmanage, False); n++;
    XtSetArg(args[n], XmNtraversalOn, True); n++;
    the_dialog = XmCreateFormDialog(disp_info[disp_num].app_shell, 
			   "Brush_Shapes_Dialog", args, n);

    n = 0;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
    the_frame = XmCreateFrame(the_dialog, "dialog_frame", args, n);
    XtManageChild(the_frame);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNwidth, 263); n++;
    the_form = XmCreateForm(the_frame, "dialog_form", args, n);
    XtManageChild(the_form);
  
    /*
     * first row of brush shapes
     */
    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNwidth, 50); n++;
    XtSetArg(args[n], XmNheight, 50); n++;
    XtSetArg(args[n], XmNleftOffset, the_offset); n++;
    XtSetArg(args[n], XmNrightOffset, the_offset); n++;
    XtSetArg(args[n], XmNtopOffset, the_offset); n++;
    XtSetArg(args[n], XmNbottomOffset, the_offset); n++;
    disp_info[disp_num].brush_dialog.brush_shape[0] = 
		    XmCreatePushButton(the_form, "shape", args, n);
    XtManageChild(disp_info[disp_num].brush_dialog.brush_shape[0]);
    XtAddCallback(disp_info[disp_num].brush_dialog.brush_shape[0], 
		  XmNactivateCallback, (XtCallbackProc)handle_brush_dialog,
        (XtPointer)(disp_num << 8 | 0));

    for (i=1; i < 4; i++)
    {
        n = 0;
        XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
        XtSetArg(args[n], XmNleftWidget, 
		 disp_info[disp_num].brush_dialog.brush_shape[i-1]); n++;
        XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
        XtSetArg(args[n], XmNwidth, 50); n++;
        XtSetArg(args[n], XmNheight, 50); n++;
        XtSetArg(args[n], XmNleftOffset, the_offset); n++;
        XtSetArg(args[n], XmNrightOffset, the_offset); n++;
        XtSetArg(args[n], XmNtopOffset, the_offset); n++;
        XtSetArg(args[n], XmNbottomOffset, the_offset); n++;
        disp_info[disp_num].brush_dialog.brush_shape[i] = 
		    XmCreatePushButton(the_form, "shape", args, n);
        XtManageChild(disp_info[disp_num].brush_dialog.brush_shape[i]);
        XtAddCallback(disp_info[disp_num].brush_dialog.brush_shape[i], 
		      XmNactivateCallback, (XtCallbackProc)handle_brush_dialog,
            (XtPointer)(disp_num << 8 | i));
    }
	
    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftWidget, 
		 disp_info[disp_num].brush_dialog.brush_shape[i-1]); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNwidth, 50); n++;
    XtSetArg(args[n], XmNheight, 50); n++;
    XtSetArg(args[n], XmNleftOffset, the_offset); n++;
    XtSetArg(args[n], XmNrightOffset, the_offset); n++;
    XtSetArg(args[n], XmNtopOffset, the_offset); n++;
    XtSetArg(args[n], XmNbottomOffset, the_offset); n++;
    disp_info[disp_num].brush_dialog.brush_shape[i] = 
		    XmCreatePushButton(the_form, "shape", args, n);
    XtManageChild(disp_info[disp_num].brush_dialog.brush_shape[i]);
    XtAddCallback(disp_info[disp_num].brush_dialog.brush_shape[i], 
		  XmNactivateCallback, (XtCallbackProc)handle_brush_dialog,
        (XtPointer)(disp_num << 8 | i));
    /*
     * rest of brush shapes
     */
    for (j=1; j <= 3; j++)
    {
        n = 0;
        XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
        XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
        XtSetArg(args[n], XmNtopWidget, 
	 disp_info[disp_num].brush_dialog.brush_shape[(j-1)*5]); n++;
        XtSetArg(args[n], XmNwidth, 50); n++;
        XtSetArg(args[n], XmNheight, 50); n++;
        XtSetArg(args[n], XmNleftOffset, the_offset); n++;
        XtSetArg(args[n], XmNrightOffset, the_offset); n++;
        XtSetArg(args[n], XmNtopOffset, the_offset); n++;
        XtSetArg(args[n], XmNbottomOffset, the_offset); n++;
        disp_info[disp_num].brush_dialog.brush_shape[j*5] = 
		    XmCreatePushButton(the_form, "shape", args, n);
        XtManageChild(disp_info[disp_num].brush_dialog.brush_shape[j*5]);
        XtAddCallback(disp_info[disp_num].brush_dialog.brush_shape[j*5], 
		      XmNactivateCallback, (XtCallbackProc)handle_brush_dialog,
            (XtPointer)(disp_num << 8 | j*5));

        for (i=1; i < 4; i++)
        {
            n = 0;
            XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
            XtSetArg(args[n], XmNleftWidget, 
	     disp_info[disp_num].brush_dialog.brush_shape[j*5 + i-1]); n++;
            XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
            XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].brush_dialog.brush_shape[(j-1)*5+i]); n++;
            XtSetArg(args[n], XmNwidth, 50); n++;
            XtSetArg(args[n], XmNheight, 50); n++;
            XtSetArg(args[n], XmNleftOffset, the_offset); n++;
            XtSetArg(args[n], XmNrightOffset, the_offset); n++;
            XtSetArg(args[n], XmNtopOffset, the_offset); n++;
            XtSetArg(args[n], XmNbottomOffset, the_offset); n++;
            disp_info[disp_num].brush_dialog.brush_shape[j*5 + i] = 
		        XmCreatePushButton(the_form, "shape", args, n);
            XtManageChild(
		   disp_info[disp_num].brush_dialog.brush_shape[j*5+i]);
            XtAddCallback(
		   disp_info[disp_num].brush_dialog.brush_shape[j*5+i], 
		      XmNactivateCallback, (XtCallbackProc)handle_brush_dialog,
            (XtPointer)(disp_num << 8 | j*5 + i));
        }
    	
        n = 0;
        XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
        XtSetArg(args[n], XmNleftWidget, 
          disp_info[disp_num].brush_dialog.brush_shape[j*5+i-1]); n++;
        XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
        XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
        XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].brush_dialog.brush_shape[(j-1)*5+i]); n++;
        XtSetArg(args[n], XmNwidth, 50); n++;
        XtSetArg(args[n], XmNheight, 50); n++;
        XtSetArg(args[n], XmNleftOffset, the_offset); n++;
        XtSetArg(args[n], XmNrightOffset, the_offset); n++;
        XtSetArg(args[n], XmNtopOffset, the_offset); n++;
        XtSetArg(args[n], XmNbottomOffset, the_offset); n++;
        disp_info[disp_num].brush_dialog.brush_shape[j*5+i] = 
		        XmCreatePushButton(the_form, "shape", args, n);
        XtManageChild(disp_info[disp_num].brush_dialog.brush_shape[j*5+i]);
        XtAddCallback(disp_info[disp_num].brush_dialog.brush_shape[j*5+i], 
		      XmNactivateCallback, (XtCallbackProc)handle_brush_dialog,
            (XtPointer)(disp_num << 8 | j*5 + i));
    }
    
    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 0); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 0); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, 
	disp_info[disp_num].brush_dialog.brush_shape[15]); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    the_separator = XmCreateSeparatorGadget(the_form, "separator", args, n);
    XtManageChild(the_separator);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNleftPosition, 40); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, 
	   disp_info[disp_num].brush_dialog.brush_shape[15]); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNrightPosition, 60); n++;
    XtSetArg(args[n], XmNbottomOffset, 5); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].brush_dialog.ok_button = 
                XmCreatePushButtonGadget(the_form, "OK", args, n);
    XtManageChild(disp_info[disp_num].brush_dialog.ok_button);
    XtAddCallback(disp_info[disp_num].brush_dialog.ok_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_brush_dialog,
        (XtPointer)(disp_num << 8 | BRUSH_OK));

    if (mesg_string)
	XtFree((char*)mesg_string);
    
    set_dialog_wm_name(the_dialog, "Brush Shapes");
    XtRealizeWidget(the_dialog);
    return(the_dialog);
}


/*
 * handle_select_dialog - this function is the callback for the select
 *                        dialog.
 */
handle_select_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int n, i, x, y, width, height;
    Arg args[MAX_ARGS];
    Widget clip_window;
    int disp_num, which_button;
    char *the_text;
    Window dummy;

    disp_num = ((int) client_data) >> 8;
    which_button = ((int) client_data) & 0x00FF;
    disp_info[disp_num].select_dialog.pointer_down = FALSE;

    switch(which_button)
    {
	case SUPER_SEL:
	    XtUnmanageChild(disp_info[disp_num].select_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].select_dialog.the_dialog);
            if (XGrabPointer(disp_info[disp_num].disp, 
		     disp_info[disp_num].draw_win_id, False, 
		     PointerMotionMask | ButtonPressMask | ButtonReleaseMask,
                     GrabModeAsync, GrabModeAsync, None, 
                     XCreateFontCursor(disp_info[disp_num].disp, XC_crosshair), 
		     CurrentTime)	
                != GrabSuccess)
            {
                popup_info_dialog(disp_num,
	            "Couldn't grab the pointer.\nAborting Super Select.");
            }
            disp_info[disp_num].input_mode = SUPER_SELECTING;
	    break;
	case FREEZE_SUPER_SEL:
	    XtUnmanageChild(disp_info[disp_num].select_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].select_dialog.the_dialog);
	    XGrabServer(disp_info[disp_num].disp);
            if (XGrabPointer(disp_info[disp_num].disp, 
		     disp_info[disp_num].draw_win_id, False, 
		     PointerMotionMask | ButtonPressMask | ButtonReleaseMask,
                     GrabModeAsync, GrabModeAsync, None, 
                     XCreateFontCursor(disp_info[disp_num].disp, XC_crosshair), 
		     CurrentTime)	
                != GrabSuccess)
            {
                popup_info_dialog(disp_num,
	            "Couldn't grab the pointer.\nAborting Super Select.");
            }
            disp_info[disp_num].input_mode = SUPER_SELECTING;
	    break;
	case WAIT_FREEZE_SUPER_SEL:
	    XtUnmanageChild(disp_info[disp_num].select_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].select_dialog.the_dialog);
            XtAppAddTimeOut(appCon, 10000, (XtTimerCallbackProc)wait_freeze_super_selectCB, 
			    (int *) disp_num);
	    break;
	case SELECTION_CANCEL:
	    XtUnmanageChild(disp_info[disp_num].select_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].select_dialog.the_dialog);
	    break;
	default:
	    printf("ERROR: unknown event in handle_select_dialog.\n");
	    break;
    }
}


/*
 * jiggle_dialog - this function "jiggles" the dialog passed in.  For some
 *                 totally unknown reason this makes the dialog "come up"
 *                 10 times faster next time it is Managed.
 */
jiggle_dialog(the_dialog)
Widget the_dialog;
{
    Arg args[MAX_ARGS];
    int n, width;

    /*
    n = 0;
    XtSetArg(args[n], XmNwidth, &width); n++;
    XtGetValues(the_dialog, args, n);

    n = 0;
    XtSetArg(args[n], XmNwidth, width + 1); n++;
    XtSetValues(the_dialog, args, n);
    */
}


/*
 * wait_freeze_super_selectCB - this function is the function that is called
 * 				after a few seconds to grab the server, etc.
 * 				Then a super select occurs.
 */
XtTimerCallbackProc wait_freeze_super_selectCB(client_data, id)
int *client_data;
XtIntervalId *id;
{
    int disp_num;
    Window rr, cr;
    unsigned int mskr;
    int rxr, ryr, win_x, win_y;

    disp_num = (int) client_data;

    XQueryPointer(disp_info[disp_num].disp,
	          disp_info[disp_num].draw_win_id, &rr, &cr, &rxr,
		  &ryr, &win_x, &win_y, &mskr);
    if (mskr & Button1Mask)
       disp_info[disp_num].select_dialog.pointer_down = TRUE;
    else
       disp_info[disp_num].select_dialog.pointer_down = FALSE;

    XGrabServer(disp_info[disp_num].disp);
    if (XGrabPointer(disp_info[disp_num].disp, 
		     disp_info[disp_num].draw_win_id, False, 
		     PointerMotionMask | ButtonPressMask | ButtonReleaseMask,
                     GrabModeAsync, GrabModeAsync, None, 
                     XCreateFontCursor(disp_info[disp_num].disp, XC_crosshair), 
		     CurrentTime)	
        != GrabSuccess)
    {
        popup_info_dialog(disp_num, "Couldn't grab the pointer.");
    }

    disp_info[disp_num].input_mode = SUPER_SELECTING;
}


/*
 * create_select_dialog - this function creates the popup dialog that
 *                        controls the super-selection capabilities.
 *
 */
Widget create_select_dialog(disp_num)
int disp_num;
{
    Widget the_dialog;
    Widget the_separator;
    Widget the_frame, the_form;
    Arg args[MAX_ARGS];
    int n;

    n = 0;
    XtSetArg(args[n], XmNautoUnmanage, False); n++;
    the_dialog = XmCreateFormDialog(disp_info[disp_num].app_shell, 
			   "Selection_Dialog", args, n);

    n = 0;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    the_frame = XmCreateFrame(the_dialog, "dialog_frame", args, n);
    XtManageChild(the_frame);

    n = 0;
    the_form = XmCreateForm(the_frame, "dialog_form", args, n);
    XtManageChild(the_form);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].select_dialog.super_sel_button =
                XmCreatePushButtonGadget(the_form, 
	           "Super Select (select region external to wscrawl window)", 
		   args, n);
    XtManageChild(disp_info[disp_num].select_dialog.super_sel_button);
    XtAddCallback(disp_info[disp_num].select_dialog.super_sel_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_select_dialog,
        (XtPointer)(disp_num << 8 | SUPER_SEL));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
    XtSetArg(args[n], XmNrightWidget, 
	     disp_info[disp_num].select_dialog.super_sel_button); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].select_dialog.super_sel_button); n++;
    XtSetArg(args[n], XmNbottomOffset, 5); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].select_dialog.freeze_sel_button =
                XmCreatePushButtonGadget(the_form, 
		    "Freeze Moniter --> Super Select", args, n);
    XtManageChild(disp_info[disp_num].select_dialog.freeze_sel_button);
    XtAddCallback(disp_info[disp_num].select_dialog.freeze_sel_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_select_dialog,
        (XtPointer)(disp_num << 8 | FREEZE_SUPER_SEL));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
    XtSetArg(args[n], XmNrightWidget, 
	     disp_info[disp_num].select_dialog.freeze_sel_button); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].select_dialog.freeze_sel_button); n++;
    XtSetArg(args[n], XmNbottomOffset, 5); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].select_dialog.wait_freeze_sel_button =
                XmCreatePushButtonGadget(the_form, 
		    "Wait 10 secs --> Freeze Moniter --> Super Select", 
		    args, n);
    XtManageChild(disp_info[disp_num].select_dialog.wait_freeze_sel_button);
    XtAddCallback(disp_info[disp_num].select_dialog.wait_freeze_sel_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_select_dialog,
        (XtPointer)(disp_num << 8 | WAIT_FREEZE_SUPER_SEL));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 0); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 0); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget,
             disp_info[disp_num].select_dialog.wait_freeze_sel_button); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    the_separator = XmCreateSeparatorGadget(the_form, "separator", args, n);
    XtManageChild(the_separator);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNleftPosition, 40); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].select_dialog.wait_freeze_sel_button); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNrightPosition, 60); n++;
    XtSetArg(args[n], XmNbottomOffset, 5); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].select_dialog.cancel_button =
                XmCreatePushButtonGadget(the_form, "Cancel", args, n);
    XtManageChild(disp_info[disp_num].select_dialog.cancel_button);
    XtAddCallback(disp_info[disp_num].select_dialog.cancel_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_select_dialog,
        (XtPointer)(disp_num << 8 | SELECTION_CANCEL));

    set_dialog_wm_name(the_dialog, "Selection");
    return(the_dialog);
}


/*
 * handle_info_dialog - this function handles the callbacks from the info 
 *	                dialog
 */
handle_info_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int disp_num;

    disp_num = ((int) client_data);
    jiggle_dialog(disp_info[disp_num].info_dialog);
}


/*
 * popup_info_dialog - this function fills in the information into the info 
 *                     dialog, then pops it up on the correct display.
 */
popup_info_dialog(disp_num, mesg)
int disp_num;
char *mesg;
{
    int n;
    Arg args[MAX_ARGS];
    XmString mesg_string = NULL;

    mesg_string = XmStringCreateLtoR(mesg, charset);

    n = 0;
    XtSetArg(args[n], XmNmessageString, mesg_string); n++;
    XtSetValues(disp_info[disp_num].info_dialog, args, n);
    XtManageChild(disp_info[disp_num].info_dialog);
}


/*
 * create_info_dialog - this function creates the info dialog, used for
 *                      all sorts of stuff all over the place.
 */
Widget create_info_dialog(disp_num)
int disp_num;
{
    Widget info_dialog;
    Widget button;
    Widget info_label_text;
    Arg args[MAX_ARGS];
    int n;
    XmString mesg_string = NULL;
    XmString ok_mesg_string = NULL;

    mesg_string = XmStringCreateLtoR("Default Message", charset);
    ok_mesg_string = XmStringCreateLtoR("OK", charset);

    n = 0;
    XtSetArg(args[n], XmNmessageString, mesg_string); n++;
    XtSetArg(args[n], XmNokLabelString, ok_mesg_string); n++;
    info_dialog = XmCreateInformationDialog(disp_info[disp_num].app_shell,
			  "info_dialog", args, n);
    XtAddCallback(info_dialog, XmNokCallback,
                  (XtCallbackProc)handle_info_dialog, (XtPointer)disp_num);

    button = XmMessageBoxGetChild(info_dialog, XmDIALOG_HELP_BUTTON);
    XtUnmanageChild(button);
    button = XmMessageBoxGetChild(info_dialog, XmDIALOG_CANCEL_BUTTON);
    XtUnmanageChild(button);

    set_dialog_wm_name(info_dialog, "Information");

    if (mesg_string)
	XtFree((char*)mesg_string);
    if (ok_mesg_string)
	XtFree((char*)ok_mesg_string);

    return(info_dialog);
}

/*
 * handle_error_dialog - this function handles the callbacks from the error 
 *	                dialog
 */
handle_error_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int disp_num;

    disp_num = ((int) client_data);
    jiggle_dialog(disp_info[disp_num].error_dialog);
}


/*
 * popup_error_dialog - this function fills in the information into the info 
 *                     dialog, then pops it up on the correct display.
 */
popup_error_dialog(disp_num, mesg)
int disp_num;
char *mesg;
{
    int n;
    Arg args[MAX_ARGS];
    XmString mesg_string = NULL;

    mesg_string = XmStringCreateLtoR(mesg, charset);

    n = 0;
    XtSetArg(args[n], XmNmessageString, mesg_string); n++;
    XtSetValues(disp_info[disp_num].error_dialog, args, n);
    XtManageChild(disp_info[disp_num].error_dialog);
}


/*
 * create_error_dialog - this function creates the error dialog, used for
 *                      all sorts of stuff all over the place.
 */
Widget create_error_dialog(disp_num)
int disp_num;
{
    Widget error_dialog;
    Widget button;
    Widget error_label_text;
    Arg args[MAX_ARGS];
    int n;
    XmString mesg_string = NULL;
    XmString ok_mesg_string = NULL;

    mesg_string = XmStringCreateLtoR("Default Message", charset);
    ok_mesg_string = XmStringCreateLtoR("OK", charset);

    n = 0;
    XtSetArg(args[n], XmNmessageString, mesg_string); n++;
    XtSetArg(args[n], XmNokLabelString, ok_mesg_string); n++;
    error_dialog = XmCreateErrorDialog(disp_info[disp_num].app_shell,
			  "error_dialog", args, n);
    XtAddCallback(error_dialog, XmNokCallback,
        (XtCallbackProc)handle_error_dialog, (XtPointer)disp_num);

    button = XmMessageBoxGetChild(error_dialog, XmDIALOG_HELP_BUTTON);
    XtUnmanageChild(button);
    button = XmMessageBoxGetChild(error_dialog, XmDIALOG_CANCEL_BUTTON);
    XtUnmanageChild(button);

    set_dialog_wm_name(error_dialog, "ERROR");

    if (mesg_string)
	XtFree((char*)mesg_string);
    if (ok_mesg_string)
	XtFree((char*)ok_mesg_string);

    return(error_dialog);
}

