/*
 * This File: wsdialog2.c
 *
 * This file contains half of the dialog creation and handling functions for 
 * wscrawl.
 */

#include "wscrawl.h"
#include "languages.h"

extern struct disp_info_struct disp_info[MAX_NUM_DISPS];
extern int num_disps;
extern XtAppContext appCon;
static XmStringCharSet charset = (XmStringCharSet) XmSTRING_DEFAULT_CHARSET;
XtTimerCallbackProc wait_freeze_super_selectCB();
extern struct global_values_struct global_values;


/*
 * set_dialog_wm_name - this function sets the indicated dialog's name to
 *                      the indicated string.
 */
set_dialog_wm_name(the_dialog, the_name)
Widget the_dialog;
char *the_name;
{
    int n;
    Arg args[MAX_ARGS];

    n = 0;
    XtSetArg(args[n], XmNtitle, the_name); n++;
    XtSetValues(XtParent(the_dialog), args, n);
}


/*
 * set_dialog_width - this function sets the indicated dialog's width to
 *                    the indicated size.
 */
set_dialog_width(the_dialog, width)
Widget the_dialog;
int width;
{
    int n;
    Arg args[MAX_ARGS];

    n = 0;
    XtSetArg(args[n], XmNwidth, width); n++;
    XtSetValues(XtParent(the_dialog), args, n);
}


/*
 * handle_add_display_dialog - this function is the callback for the display
 *                             adding dialog.
 */
handle_add_display_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    XmSelectionBoxCallbackStruct *sbcb;		/* data from callback */
    char *dispname_str;
    int i, disp_num;

    sbcb = (XmSelectionBoxCallbackStruct *) call_data;
    disp_num = (int) client_data;

    switch(sbcb->reason)
    {
	case XmCR_OK:
	    XmStringGetLtoR(sbcb->value, charset, &dispname_str);
            if (global_values.screen_dirty_bit == TRUE)
	    {
		strncpy(
		  disp_info[disp_num].dirty_add_display_dialog.new_dispname_str,
		  dispname_str, 90);
	  	XtManageChild(
		      disp_info[disp_num].dirty_add_display_dialog.the_dialog);
	    }
	    else
	    {
	        add_display(dispname_str, disp_num, -1);
	    }
	    jiggle_dialog(disp_info[disp_num].add_display_dialog);
	    break;
	case XmCR_CANCEL:
	    jiggle_dialog(disp_info[disp_num].add_display_dialog);
	    break;
	default:
	    printf("ERROR: unknown event in handle_add_display_dialog.\n");
	    break;
    }
}


/*
 * create_add_display_dialog - this function creates the popup dialog that
 *                             prompts for a display name.
 */
Widget create_add_display_dialog(disp_num)
int disp_num;
{
    Widget the_dialog;
    Widget button;
    Widget edit_field;
    Widget label_text;
    Arg args[MAX_ARGS];
    int n;
    XmString mesg_string = NULL;
    XmString button_string = NULL;

    mesg_string = XmStringCreateLtoR("Enter the name of the display to add: ",
				     charset);
    n = 0;
    XtSetArg(args[n], XmNselectionLabelString, mesg_string); n++;
    XtSetArg(args[n], XmNtraversalOn, True); n++;
    the_dialog = XmCreatePromptDialog(disp_info[disp_num].app_shell, 
			   "Add_Display_Dialog", args, n);
    XtAddCallback(the_dialog,XmNokCallback,
                  (XtCallbackProc)handle_add_display_dialog, (XtPointer)disp_num);
    XtAddCallback(the_dialog, XmNcancelCallback,
                  (XtCallbackProc)handle_add_display_dialog, (XtPointer)disp_num);
    button = XmSelectionBoxGetChild(the_dialog, XmDIALOG_HELP_BUTTON);
    XtUnmanageChild(button);

    if (mesg_string)
	XtFree((char*)mesg_string);
    if (button_string)
	XtFree((char*)button_string);
    
    set_dialog_wm_name(the_dialog, "Add Display");
    return(the_dialog);
}


/*
 * handle_line_dialog - this function is the callback for the line settings
 *                      dialog.
 */
handle_line_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    XmSelectionBoxCallbackStruct *sbcb;		/* data from callback */
    char *tmp_str;
    int i, disp_num, width;
    XGCValues gcvalues;

    sbcb = (XmSelectionBoxCallbackStruct *) call_data;
    disp_num = (int) client_data;

    switch(sbcb->reason)
    {
	case XmCR_OK:
	    XmStringGetLtoR(sbcb->value, charset, &tmp_str);
	    width = -1;
	    sscanf(tmp_str, "%d", &width);
	    if (width != -1)
		disp_info[disp_num].line_dialog.line_width = width;

	    if (disp_info[disp_num].input_mode == DRAWING_LINES)
	    {
                for (i=0; i < num_disps; i++)
                {
                    if (disp_info[i].in_session_bool)
                    {
                        gcvalues.line_width = width;
                        XChangeGC(disp_info[i].disp, 
				  disp_info[disp_num].win_gc[i],
                                  GCLineWidth, &gcvalues);
                    }
                }
		gcvalues.line_width = width;
	        XChangeGC(disp_info[disp_num].disp,
			  disp_info[disp_num].rubber_band_gc, 
			  GCLineWidth, &gcvalues);
	    }
	    jiggle_dialog(disp_info[disp_num].line_dialog.the_dialog);
	    break;
	case XmCR_CANCEL:
	    jiggle_dialog(disp_info[disp_num].line_dialog.the_dialog);
	    break;
	default:
	    printf("ERROR: unknown event in handle_line_dialog.\n");
	    break;
    }
}


/*
 * create_line_dialog - this function creates the popup dialog that
 *                      prompts for line settings.
 */
Widget create_line_dialog(disp_num)
int disp_num;
{
    Widget the_dialog;
    Widget button;
    Widget edit_field;
    Widget label_text;
    Arg args[MAX_ARGS];
    int n;
    XmString mesg_string = NULL;
    XmString button_string = NULL;

    mesg_string = XmStringCreateLtoR(
		    "Enter the line width\nfor straight lines:", charset);
    n = 0;
    XtSetArg(args[n], XmNselectionLabelString, mesg_string); n++;
    XtSetArg(args[n], XmNtraversalOn, True); n++;
    the_dialog = XmCreatePromptDialog(disp_info[disp_num].app_shell, 
			   "Line_Dialog", args, n);
    XtAddCallback(the_dialog, XmNokCallback, (XtCallbackProc)handle_line_dialog, (XtPointer)disp_num);
    XtAddCallback(the_dialog, XmNcancelCallback, (XtCallbackProc)handle_line_dialog, (XtPointer)disp_num);
    button = XmSelectionBoxGetChild(the_dialog, XmDIALOG_HELP_BUTTON);
    XtUnmanageChild(button);

    if (mesg_string)
	XtFree((char*)mesg_string);
    if (button_string)
	XtFree((char*)button_string);
    
    set_dialog_wm_name(the_dialog, "Line Settings");
    return(the_dialog);
}


/*
 * handle_square_dialog - this function is the callback for the line settings
 *                      dialog.
 */
handle_square_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    XmSelectionBoxCallbackStruct *sbcb;		/* data from callback */
    char *tmp_str;
    int i, disp_num, width;
    XGCValues gcvalues;

    sbcb = (XmSelectionBoxCallbackStruct *) call_data;
    disp_num = (int) client_data;

    switch(sbcb->reason)
    {
	case XmCR_OK:
	    XmStringGetLtoR(sbcb->value, charset, &tmp_str);
	    width = -1;
	    sscanf(tmp_str, "%d", &width);
	    if (width != -1)
		disp_info[disp_num].square_dialog.line_width = width;

	    if (disp_info[disp_num].input_mode == OUTLINING_SQUARES)
	    {
                for (i=0; i < num_disps; i++)
                {
                    if (disp_info[i].in_session_bool)
                    {
                        gcvalues.line_width = width;
                        XChangeGC(disp_info[i].disp, 
				  disp_info[disp_num].win_gc[i],
                                  GCLineWidth, &gcvalues);
                    }
                }
		gcvalues.line_width = width;
	        XChangeGC(disp_info[disp_num].disp,
			  disp_info[disp_num].rubber_band_gc, 
			  GCLineWidth, &gcvalues);
	    }
	    jiggle_dialog(disp_info[disp_num].square_dialog.the_dialog);
	    break;
	case XmCR_CANCEL:
	    jiggle_dialog(disp_info[disp_num].square_dialog.the_dialog);
	    break;
	default:
	    printf("ERROR: unknown event in handle_square_dialog.\n");
	    break;
    }
}


/*
 * create_square_dialog - this function creates the popup dialog that
 *                      prompts for line settings.
 */
Widget create_square_dialog(disp_num)
int disp_num;
{
    Widget the_dialog;
    Widget button;
    Widget edit_field;
    Widget label_text;
    Arg args[MAX_ARGS];
    int n;
    XmString mesg_string = NULL;
    XmString button_string = NULL;

    mesg_string = XmStringCreateLtoR(
		    "Enter the line width\nfor outlined squares:", charset);
    n = 0;
    XtSetArg(args[n], XmNselectionLabelString, mesg_string); n++;
    XtSetArg(args[n], XmNtraversalOn, True); n++;
    the_dialog = XmCreatePromptDialog(disp_info[disp_num].app_shell, 
			   "Square_Dialog", args, n);
    XtAddCallback(the_dialog, XmNokCallback, (XtCallbackProc)handle_square_dialog, (XtPointer)disp_num);
    XtAddCallback(the_dialog, XmNcancelCallback, (XtCallbackProc)handle_square_dialog,(XtPointer)disp_num);
    button = XmSelectionBoxGetChild(the_dialog, XmDIALOG_HELP_BUTTON);
    XtUnmanageChild(button);

    if (mesg_string)
	XtFree((char*)mesg_string);
    if (button_string)
	XtFree((char*)button_string);
    
    set_dialog_wm_name(the_dialog, "Square Settings");
    return(the_dialog);
}


/*
 * handle_oval_dialog - this function is the callback for the line settings
 *                      dialog.
 */
handle_oval_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    XmSelectionBoxCallbackStruct *sbcb;		/* data from callback */
    char *tmp_str;
    int i, disp_num, width;
    XGCValues gcvalues;

    sbcb = (XmSelectionBoxCallbackStruct *) call_data;
    disp_num = (int) client_data;

    switch(sbcb->reason)
    {
	case XmCR_OK:
	    XmStringGetLtoR(sbcb->value, charset, &tmp_str);
	    width = -1;
	    sscanf(tmp_str, "%d", &width);
	    if (width != -1)
		disp_info[disp_num].oval_dialog.line_width = width;

	    if (disp_info[disp_num].input_mode == OUTLINING_OVALS)
	    {
                for (i=0; i < num_disps; i++)
                {
                    if (disp_info[i].in_session_bool)
                    {
                        gcvalues.line_width = width;
                        XChangeGC(disp_info[i].disp, 
				  disp_info[disp_num].win_gc[i],
                                  GCLineWidth, &gcvalues);
                    }
                }
		gcvalues.line_width = width;
	        XChangeGC(disp_info[disp_num].disp,
			  disp_info[disp_num].rubber_band_gc, 
			  GCLineWidth, &gcvalues);
	    }
	    jiggle_dialog(disp_info[disp_num].oval_dialog.the_dialog);
	    break;
	case XmCR_CANCEL:
	    jiggle_dialog(disp_info[disp_num].oval_dialog.the_dialog);
	    break;
	default:
	    printf("ERROR: unknown event in handle_oval_dialog.\n");
	    break;
    }
}


/*
 * create_oval_dialog - this function creates the popup dialog that
 *                      prompts for line settings.
 */
Widget create_oval_dialog(disp_num)
int disp_num;
{
    Widget the_dialog;
    Widget button;
    Widget edit_field;
    Widget label_text;
    Arg args[MAX_ARGS];
    int n;
    XmString mesg_string = NULL;
    XmString button_string = NULL;

    mesg_string = XmStringCreateLtoR(
		    "Enter the line width\nfor outlined ovals:", charset);
    n = 0;
    XtSetArg(args[n], XmNselectionLabelString, mesg_string); n++;
    XtSetArg(args[n], XmNtraversalOn, True); n++;
    the_dialog = XmCreatePromptDialog(disp_info[disp_num].app_shell, 
			   "Oval_Dialog", args, n);
    XtAddCallback(the_dialog, XmNokCallback, (XtCallbackProc)handle_oval_dialog, (XtPointer)disp_num);
    XtAddCallback(the_dialog, XmNcancelCallback, (XtCallbackProc)handle_oval_dialog, (XtPointer)disp_num);
    button = XmSelectionBoxGetChild(the_dialog, XmDIALOG_HELP_BUTTON);
    XtUnmanageChild(button);

    if (mesg_string)
	XtFree((char*)mesg_string);
    if (button_string)
	XtFree((char*)button_string);
    
    set_dialog_wm_name(the_dialog, "Oval Settings");
    return(the_dialog);
}


/*
 * handle_change_pointer_dialog - this function is the callback for the display
 *                             changing rubber pointer dialog.
 */
handle_change_pointer_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    XmSelectionBoxCallbackStruct *sbcb;		/* data from callback */
    XFontStruct *the_font_struct;
    int disp_num, i;
    char *name_str;
    char buf[256];

    sbcb = (XmSelectionBoxCallbackStruct *) call_data;
    disp_num = (int) client_data;

    switch(sbcb->reason)
    {
	case XmCR_OK:
	    XmStringGetLtoR(sbcb->value, charset, &name_str);
	    strncpy(disp_info[disp_num].rubber_pointer.pointer_str,name_str,48);
            the_font_struct = XLoadQueryFont(disp_info[disp_num].disp, "fixed");
            disp_info[disp_num].rubber_pointer.width = 15 +
                                     XTextWidth(the_font_struct, name_str, 
				     strlen(name_str));
            disp_info[disp_num].rubber_pointer.height =
                                   15 + the_font_struct->max_bounds.ascent
				   /* + the_font_struct->max_bounds.descent */ ;
	    for (i=0; i < num_disps; i++)
	        if (disp_info[i].in_session_bool)  
	            draw_status_win_message(i);
	    jiggle_dialog(disp_info[disp_num].change_pointer_dialog);
	    sprintf(buf, "wscrawl.telePointerName: %s\n", name_str);
            merge_resources(disp_info[disp_num].disp, buf);
	    break;
	case XmCR_CANCEL:
	    jiggle_dialog(disp_info[disp_num].change_pointer_dialog);
	    break;
	default:
	    printf("ERROR: unknown event in handle_change_pointer_dialog.\n");
	    break;
    }
}


/*
 * create_change_pointer_dialog - this function creates the popup dialog that
 *                             prompts for a display name.
 */
Widget create_change_pointer_dialog(disp_num)
int disp_num;
{
    Widget the_dialog;
    Widget button;
    Widget edit_field;
    Widget label_text;
    Arg args[MAX_ARGS];
    int n;
    XmString mesg_string = NULL;
    XmString button_string = NULL;

    mesg_string = XmStringCreateLtoR(
     "Enter your name.  This will be used\nas the label on your pointer: ",
			     charset);
    n = 0;
    XtSetArg(args[n], XmNselectionLabelString, mesg_string); n++;
    XtSetArg(args[n], XmNtraversalOn, True); n++;
    the_dialog = XmCreatePromptDialog(disp_info[disp_num].app_shell, 
			   "Change_Pointer_Dialog", args, n);
    XtAddCallback(the_dialog, XmNokCallback, (XtCallbackProc)handle_change_pointer_dialog,
                  (XtPointer)disp_num);
    XtAddCallback(the_dialog, XmNcancelCallback, (XtCallbackProc)handle_change_pointer_dialog, 
		            (XtPointer)disp_num);
    button = XmSelectionBoxGetChild(the_dialog, XmDIALOG_HELP_BUTTON);
    XtUnmanageChild(button);

    if (mesg_string)
	XtFree((char*)mesg_string);
    if (button_string)
	XtFree((char*)button_string);
    
    set_dialog_wm_name(the_dialog, "Change Pointer");
    return(the_dialog);
}


/*
 * handle_join_session_dialog - this function is the callback for the session
 *                              joining dialog.
 */
handle_join_session_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int disp_num;
    XmSelectionBoxCallbackStruct *sbcb;		/* data from callback */
    char *location_str;
    char tmp_str[200];

    sbcb = (XmSelectionBoxCallbackStruct *) call_data;
    disp_num = (int) client_data;

    switch(sbcb->reason)
    {
	case XmCR_OK:
	    XmStringGetLtoR(sbcb->value, charset, &location_str);
	    if (knock_knock(disp_num, location_str) == TRUE)
	        exit(0);
	    else
	    {
		sprintf(tmp_str, "Unable to knock-knock %s.\n\n%s",location_str,
		       "Probably there is no wscrawl running there.");
		popup_info_dialog(disp_num, tmp_str);
	    }
	    jiggle_dialog(disp_info[disp_num].join_session_dialog);
	    break;
	case XmCR_CANCEL:
	    jiggle_dialog(disp_info[disp_num].join_session_dialog);
	    break;
	default:
	    printf("ERROR: unknown event in handle_join_session_dialog.\n");
	    break;
    }
}


/*
 * create_join_session_dialog - this function creates the popup dialog that
 *                             prompts for a few things.
 */
Widget create_join_session_dialog(disp_num)
int disp_num;
{
    Widget the_dialog;
    Widget button;
    Widget edit_field;
    Widget label_text;
    Arg args[MAX_ARGS];
    int n;
    XmString mesg_string = NULL;
    XmString button_string = NULL;

    mesg_string = XmStringCreateLtoR(
               "Enter the name of the machine\nyou wish to be added to: ", 
	       charset);

    n = 0;
    XtSetArg(args[n], XmNselectionLabelString, mesg_string); n++;
    XtSetArg(args[n], XmNtraversalOn, True); n++;
    the_dialog = XmCreatePromptDialog(disp_info[disp_num].app_shell, 
			   "Join_Session_Dialog", args, n);
    XtAddCallback(the_dialog, XmNokCallback, (XtCallbackProc)handle_join_session_dialog, 
		  (XtPointer)disp_num);
    XtAddCallback(the_dialog, XmNcancelCallback, (XtCallbackProc)handle_join_session_dialog, 
		  (XtPointer)disp_num);
    button = XmSelectionBoxGetChild(the_dialog, XmDIALOG_HELP_BUTTON);
    XtUnmanageChild(button);

    if (mesg_string)
	XtFree((char*)mesg_string);
    if (button_string)
	XtFree((char*)button_string);
    
    set_dialog_wm_name(the_dialog, "Join Session");
    return(the_dialog);
}


/*
 * handle_mouse_dialog - this function is the callback for the mouse options
 *                          dialog.
 */
handle_mouse_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int n, disp_num, which_button;
    int oldtop, oldleft, i;
    Arg args[MAX_ARGS];
    XmToggleButtonCallbackStruct *tbcb;

    disp_num = ((int) client_data) >> 8;
    which_button = ((int) client_data) & 0x00FF;
    tbcb = (XmToggleButtonCallbackStruct *) call_data;

    switch(which_button)
    {
	case MOUSE_COORDINATES:
	    if (tbcb->set == True)
	    {
		disp_info[disp_num].show_pointer_coords = FALSE;
		XtUnmanageChild(disp_info[disp_num].pointer_x_coords_label);
		XtUnmanageChild(disp_info[disp_num].pointer_y_coords_label);
                merge_resources(disp_info[disp_num].disp, 
			    "wscrawl.showPointerCoordinates: False\n");
	    }
	    else
	    {
		disp_info[disp_num].show_pointer_coords = TRUE;
		XtManageChild(disp_info[disp_num].pointer_x_coords_label);
		XtManageChild(disp_info[disp_num].pointer_y_coords_label);
                merge_resources(disp_info[disp_num].disp, 
			    "wscrawl.showPointerCoordinates: True\n");
	    }
	    break;
	case MOUSE_TELEPOINTER:
	    if (tbcb->set == True)
	    {
		unmap_pointer(disp_num);
		disp_info[disp_num].show_telepointer_always = FALSE;
                merge_resources(disp_info[disp_num].disp, 
			    "wscrawl.continuousTelePointer: False\n");
	    }
	    else
            {
                disp_info[disp_num].last_telepointer_point.x = 0;
                disp_info[disp_num].last_telepointer_point.y = 0;
                disp_info[disp_num].rubber_pointer.is_mapped_bool = TRUE;
		disp_info[disp_num].show_telepointer_always = TRUE;
                merge_resources(disp_info[disp_num].disp, 
			    "wscrawl.continuousTelePointer: True\n");
	    }
	    break;
	case MOUSE_GRID:
	    printf("Mouse grid.\n");
	    break;
	case MOUSE_CLOSE:
	    XtUnmanageChild(disp_info[disp_num].mouse_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].mouse_dialog.the_dialog);
	    break;
	default:
	    printf("ERROR: unknown event in handle_mouse_dialog.\n");
	    break;
    }
}


/*
 * create_mouse_dialog - this function creates the popup dialog that
 *                       prompts for mouse options.
 */
Widget create_mouse_dialog(disp_num)
int disp_num;
{
    Widget the_dialog;
    Widget the_separator;
    Widget the_frame, the_form;
    Arg args[MAX_ARGS];
    int n;
    XmString mesg_string = NULL;

    n = 0;
    XtSetArg(args[n], XmNautoUnmanage, False); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    the_dialog = XmCreateFormDialog(disp_info[disp_num].app_shell, 
			   "Mouse_Dialog", args, n);

    n = 0;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    the_frame = XmCreateFrame(the_dialog, "dialog_frame", args, n);
    XtManageChild(the_frame);

    n = 0;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    the_form = XmCreateForm(the_frame, "dialog_form", args, n);
    XtManageChild(the_form);

    mesg_string = XmStringCreateLtoR("Select any combination: ", charset);
    n = 0;
    XtSetArg(args[n], XmNx, 15); n++;
    XtSetArg(args[n], XmNy, 15); n++;
    XtSetArg(args[n], XmNlabelString, mesg_string); n++;
    disp_info[disp_num].mouse_dialog.label = 
		   XmCreateLabelGadget(the_form, "prompt", args, n);
    XtManageChild(disp_info[disp_num].mouse_dialog.label);
    
    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 20); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 35); n++;
    XtSetArg(args[n], XmNindicatorType, XmN_OF_MANY); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    disp_info[disp_num].mouse_dialog.coordinates_toggle =
      XmCreateToggleButtonGadget(the_form, "Show Pointer Coordinates", args, n);
    XtManageChild(disp_info[disp_num].mouse_dialog.coordinates_toggle);
    XtAddCallback(disp_info[disp_num].mouse_dialog.coordinates_toggle, 
		  XmNarmCallback, (XtCallbackProc)handle_mouse_dialog,
        (XtPointer)(disp_num << 8 | MOUSE_COORDINATES));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 20); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, 
		  disp_info[disp_num].mouse_dialog.coordinates_toggle); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    XtSetArg(args[n], XmNindicatorType, XmN_OF_MANY); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    disp_info[disp_num].mouse_dialog.telepointer_toggle =
      XmCreateToggleButtonGadget(the_form, "Always Show TelePointer", args, n);
    XtManageChild(disp_info[disp_num].mouse_dialog.telepointer_toggle);
    XtAddCallback(disp_info[disp_num].mouse_dialog.telepointer_toggle, 
		  XmNarmCallback, (XtCallbackProc)handle_mouse_dialog,
        (XtPointer)(disp_num << 8 | MOUSE_TELEPOINTER));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 0); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 0); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget,
             disp_info[disp_num].mouse_dialog.telepointer_toggle); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    the_separator = XmCreateSeparatorGadget(the_form, "separator", args, n);
    XtManageChild(the_separator);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNleftPosition, 40); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNrightPosition, 60); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    XtSetArg(args[n], XmNtopWidget, the_separator); n++;
    XtSetArg(args[n], XmNbottomOffset, 5); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].mouse_dialog.close_button =
                XmCreatePushButtonGadget(the_form, "Close", args, n);
    XtManageChild(disp_info[disp_num].mouse_dialog.close_button);
    XtAddCallback(disp_info[disp_num].mouse_dialog.close_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_mouse_dialog,
        (XtPointer)(disp_num << 8 | MOUSE_CLOSE));
    if (mesg_string)
	XtFree((char*)mesg_string);
    
    set_dialog_wm_name(the_dialog, "Pointer");
    return(the_dialog);
}


/*
 * handle_language_dialog - this function is the callback for the language
 *                          dialog.
 */
handle_language_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int disp_num, which_language;
    int n;
    Arg args[MAX_ARGS];
    XmString cancel_label_str = NULL;

    disp_num = ((int) client_data) >> 8;
    which_language = ((int) client_data) & 0x00FF;

    switch(which_language)
    {
	case ENGLISH:
	    set_language_of_interface(disp_num, ENGLISH);
	    disp_info[disp_num].language_dialog.current_language = ENGLISH;
	    break;
	case GERMAN:
	    set_language_of_interface(disp_num, GERMAN);
	    disp_info[disp_num].language_dialog.current_language = GERMAN;
	    break;
	case LANGUAGE_OK:
	    XtUnmanageChild(disp_info[disp_num].language_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].language_dialog.the_dialog);
            disp_info[disp_num].current_language = 
			disp_info[disp_num].language_dialog.current_language;
	    if (disp_info[disp_num].current_language == GERMAN)
                cancel_label_str = XmStringCreateLtoR(german_cancel, charset);
	    else
                cancel_label_str = XmStringCreateLtoR(english_cancel, charset);
            n = 0;
            XtSetArg(args[n], XmNlabelString, cancel_label_str); n++;
            XtSetValues(disp_info[disp_num].language_dialog.cancel_button, 
                        args, n);
            if (cancel_label_str)
                XtFree((char*)cancel_label_str);
	    break;
	case LANGUAGE_CANCEL:
	    set_language_of_interface(disp_num,
                          disp_info[disp_num].current_language);
	    XtUnmanageChild(disp_info[disp_num].language_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].language_dialog.the_dialog);
	    break;
	default:
	    printf("ERROR: unknown event in handle_language_dialog.\n");
	    break;
    }
}


/*
 * create_language_dialog - this function creates the popup dialog that
 *                          prompts for a language change.
 */
Widget create_language_dialog(disp_num)
int disp_num;
{
    Widget the_dialog;
    Widget the_separator;
    Widget the_frame, the_form;
    Arg args[MAX_ARGS];
    int n;
    XmString mesg_string = NULL;

    n = 0;
    XtSetArg(args[n], XmNautoUnmanage, False); n++;
    XtSetArg(args[n], XmNtraversalOn, True); n++;
    the_dialog = XmCreateFormDialog(disp_info[disp_num].app_shell, 
			   "Language_Dialog", args, n);

    n = 0;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    the_frame = XmCreateFrame(the_dialog, "dialog_frame", args, n);
    XtManageChild(the_frame);

    n = 0;
    the_form = XmCreateForm(the_frame, "dialog_form", args, n);
    XtManageChild(the_form);

    mesg_string = XmStringCreateLtoR(
	           "Select the language you wish\nthe interface to be in: ", 
                   charset);
    n = 0;
    XtSetArg(args[n], XmNx, 15); n++;
    XtSetArg(args[n], XmNy, 15); n++;
    XtSetArg(args[n], XmNrightOffset, 15); n++;
    XtSetArg(args[n], XmNlabelString, mesg_string); n++;
    disp_info[disp_num].language_dialog.label = 
		   XmCreateLabelGadget(the_form, "prompt", args, n);
    XtManageChild(disp_info[disp_num].language_dialog.label);
    
    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 60); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].language_dialog.german_button =
                XmCreatePushButtonGadget(the_form, "German", args, n);
    XtManageChild(disp_info[disp_num].language_dialog.german_button);
    XtAddCallback(disp_info[disp_num].language_dialog.german_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_language_dialog,
        (XtPointer)(disp_num << 8 | GERMAN));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
    XtSetArg(args[n], XmNrightWidget, 
	     disp_info[disp_num].language_dialog.german_button); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].language_dialog.german_button); n++;
    XtSetArg(args[n], XmNbottomOffset, 5); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].language_dialog.english_button =
                XmCreatePushButtonGadget(the_form, "English", args, n);
    XtManageChild(disp_info[disp_num].language_dialog.english_button);
    XtAddCallback(disp_info[disp_num].language_dialog.english_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_language_dialog,
        (XtPointer)(disp_num << 8 | ENGLISH));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 0); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 0); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget,
             disp_info[disp_num].language_dialog.english_button); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    the_separator = XmCreateSeparatorGadget(the_form, "separator", args, n);
    XtManageChild(the_separator);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].language_dialog.english_button); n++;
    XtSetArg(args[n], XmNbottomOffset, 5); n++;
    XtSetArg(args[n], XmNrightOffset, 5); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    XtSetArg(args[n], XmNwidth, 60); n++;
    disp_info[disp_num].language_dialog.ok_button =
                XmCreatePushButtonGadget(the_form, "OK", args, n);
    XtManageChild(disp_info[disp_num].language_dialog.ok_button);
    XtAddCallback(disp_info[disp_num].language_dialog.ok_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_language_dialog,
        (XtPointer)(disp_num << 8 | LANGUAGE_OK));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftWidget, 
	     disp_info[disp_num].language_dialog.ok_button); n++;
    XtSetArg(args[n], XmNleftOffset, 10); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].language_dialog.english_button); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
    XtSetArg(args[n], XmNrightWidget, 
	     disp_info[disp_num].language_dialog.english_button); n++;
    XtSetArg(args[n], XmNbottomOffset, 5); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].language_dialog.cancel_button =
                XmCreatePushButtonGadget(the_form, "Cancel", args, n);
    XtManageChild(disp_info[disp_num].language_dialog.cancel_button);
    XtAddCallback(disp_info[disp_num].language_dialog.cancel_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_language_dialog,
        (XtPointer)(disp_num << 8 | LANGUAGE_CANCEL));

    if (mesg_string)
	XtFree((char*)mesg_string);
    
    set_dialog_wm_name(the_dialog, "Language");
    return(the_dialog);
}


/*
 * handle_attract_attention_dialog - this function is the callback for the 
 *                          attract attention dialog.
 */
handle_attract_attention_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int disp_num, which_language;
    int n, i, j, k;
    Arg args[MAX_ARGS];
    XmString cancel_label_str = NULL;
    unsigned long valuemask = CWOverrideRedirect | CWBackPixel;
    Window flash_window[MAX_NUM_DISPS];
    static XSetWindowAttributes blank_win;

    disp_num = ((int) client_data) >> 8;
    which_language = ((int) client_data) & 0x00FF;

    switch(which_language)
    {
	case BELL:
            for (k=0; k < 3; k++)
	    {
                for (i=0; i < num_disps; i++)
                    if (disp_info[i].in_session_bool)
		    {
                        XBell(disp_info[i].disp, 100);
			XFlush(disp_info[i].disp);
		    }
		my_sleep(0.3);
	    }
	    break;
	case FLASH:
	    blank_win.override_redirect = True;
	    blank_win.background_pixel = WhitePixelOfScreen(
			    DefaultScreenOfDisplay(disp_info[disp_num].disp));

            for (i=0; i < num_disps; i++)
                if (disp_info[i].in_session_bool)
	        {
	             flash_window[i] = XCreateWindow(disp_info[i].disp,
				XDefaultRootWindow(disp_info[i].disp), 
				0, 0, 1400, 1400, 0, 0, CopyFromParent, 
				CopyFromParent, valuemask, &blank_win);
		}
            for (k=0; k < 3; k++)
	    {
                for (i=0; i < num_disps; i++)
                    if (disp_info[i].in_session_bool)
		    {
			XForceScreenSaver(disp_info[i].disp, ScreenSaverReset);
                        XMapWindow(disp_info[i].disp, flash_window[i]);
			XFlush(disp_info[i].disp);
		    }
		my_sleep(1.0);
                for (i=0; i < num_disps; i++)
                    if (disp_info[i].in_session_bool)
		    {
                        XUnmapWindow(disp_info[i].disp, flash_window[i]);
			XFlush(disp_info[i].disp);
		    }
	    }
            for (i=0; i < num_disps; i++)
                if (disp_info[i].in_session_bool)
		    XDestroyWindow(disp_info[i].disp, flash_window[i]);
	    break;
	case ATTENTION_CLOSE:
	    XtUnmanageChild(
		 disp_info[disp_num].attract_attention_dialog.the_dialog);
	    jiggle_dialog(
		 disp_info[disp_num].attract_attention_dialog.the_dialog);
	    break;
	default:
	    printf("ERROR: unknown event handle_attract_attention_dialog.\n");
	    break;
    }
}


/*
 * create_attract_attention_dialog - this function creates the popup dialog that
 *                          prompts for a attention grabbing actions
 */
Widget create_attract_attention_dialog(disp_num)
int disp_num;
{
    Widget the_dialog;
    Widget the_separator;
    Widget the_frame, the_form;
    Arg args[MAX_ARGS];
    int n;
    XmString mesg_string = NULL;

    n = 0;
    XtSetArg(args[n], XmNautoUnmanage, False); n++;
    XtSetArg(args[n], XmNtraversalOn, True); n++;
    the_dialog = XmCreateFormDialog(disp_info[disp_num].app_shell, 
			   "Attract_Attention_Dialog", args, n);

    n = 0;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
    the_frame = XmCreateFrame(the_dialog, "dialog_frame", args, n);
    XtManageChild(the_frame);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNwidth, 230); n++;
    the_form = XmCreateForm(the_frame, "dialog_form", args, n);
    XtManageChild(the_form);

    mesg_string = XmStringCreateLtoR(
	           "Push buttons to attract the\nremote user's attention: ", 
                   charset);
    n = 0;
    XtSetArg(args[n], XmNx, 15); n++;
    XtSetArg(args[n], XmNy, 15); n++;
    XtSetArg(args[n], XmNlabelString, mesg_string); n++;
    disp_info[disp_num].attract_attention_dialog.label = 
		   XmCreateLabelGadget(the_form, "prompt", args, n);
    XtManageChild(disp_info[disp_num].attract_attention_dialog.label);
    
    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 20); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 20); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 60); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].attract_attention_dialog.ring_bell_button =
              XmCreatePushButtonGadget(the_form, "Ring Bell", args, n);
    XtManageChild(
	      disp_info[disp_num].attract_attention_dialog.ring_bell_button);
    XtAddCallback(disp_info[disp_num].attract_attention_dialog.ring_bell_button,
		  XmNactivateCallback, (XtCallbackProc)handle_attract_attention_dialog,
        (XtPointer)(disp_num << 8 | BELL));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 20); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
    XtSetArg(args[n], XmNrightWidget, 
           disp_info[disp_num].attract_attention_dialog.ring_bell_button); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNtopWidget, 
           disp_info[disp_num].attract_attention_dialog.ring_bell_button); n++;
    XtSetArg(args[n], XmNbottomOffset, 5); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].attract_attention_dialog.flash_screen_button =
                XmCreatePushButtonGadget(the_form, "Flash Screen", args, n);
    XtManageChild(
	   disp_info[disp_num].attract_attention_dialog.flash_screen_button);
    XtAddCallback(
	   disp_info[disp_num].attract_attention_dialog.flash_screen_button, 
	   XmNactivateCallback, (XtCallbackProc)handle_attract_attention_dialog,
      (XtPointer)(disp_num << 8 | FLASH));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 0); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 0); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget,
        disp_info[disp_num].attract_attention_dialog.flash_screen_button); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    the_separator = XmCreateSeparatorGadget(the_form, "separator", args, n);
    XtManageChild(the_separator);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNleftPosition, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNrightPosition, 70); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNtopWidget, 
         disp_info[disp_num].attract_attention_dialog.flash_screen_button); n++;
    XtSetArg(args[n], XmNbottomOffset, 5); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].attract_attention_dialog.close_button =
                XmCreatePushButtonGadget(the_form, "Close", args, n);
    XtManageChild(disp_info[disp_num].attract_attention_dialog.close_button);
    XtAddCallback(disp_info[disp_num].attract_attention_dialog.close_button, 
	     XmNactivateCallback, (XtCallbackProc)handle_attract_attention_dialog, 
	     (XtPointer)(disp_num << 8 | ATTENTION_CLOSE));
    if (mesg_string)
	 XtFree((char*)mesg_string);
    
    set_dialog_wm_name(the_dialog, "Attract Attention");
    return(the_dialog);
}



/*
 * example_area_exposeCB - this function gets called when an expose event
 *                     occurs in the example area.
 */
example_area_exposeCB(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int disp_num;
    XmDrawingAreaCallbackStruct *cb_struct;
    XExposeEvent *the_event;

    cb_struct = (XmDrawingAreaCallbackStruct *) call_data;
    the_event = (XExposeEvent *) cb_struct->event;
    disp_num = (int) client_data;

    if (the_event->type != Expose)
        printf("ERROR: in example_area_exposeCB.\n");

    XFillRectangle(disp_info[disp_num].disp, XtWindow(
	            disp_info[disp_num].color_picker_dialog.example_color),
		    disp_info[disp_num].color_picker_dialog.the_gc, 
		    0, 0, 50, 50);
}


/*
 * handle_color_picker_dialog - this function is the callback for the color
 *                              picker dialog.
 */
handle_color_picker_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int i, disp_num, which_widget;
    int n, r_val, g_val, b_val;
    Arg args[MAX_ARGS];
    XColor the_color;
    XGCValues gcvalues;

    disp_num = ((int) client_data) >> 8;
    which_widget = ((int) client_data) & 0x00FF;

    n = 0;
    XtSetArg(args[n], XmNvalue, &r_val); n++;
    XtGetValues(disp_info[disp_num].color_picker_dialog.rslider,args,n);
    n = 0;
    XtSetArg(args[n], XmNvalue, &g_val); n++;
    XtGetValues(disp_info[disp_num].color_picker_dialog.gslider,args,n);
    n = 0;
    XtSetArg(args[n], XmNvalue, &b_val); n++;
    XtGetValues(disp_info[disp_num].color_picker_dialog.bslider,args,n);

    the_color.flags = DoRed | DoGreen | DoBlue;
    the_color.red = r_val * 256;
    the_color.green = g_val * 256;
    the_color.blue = b_val * 256;

    switch(which_widget)
    {
	case RED_SLIDER:
	case GREEN_SLIDER:
	case BLUE_SLIDER:
	    if (disp_info[disp_num].color_picker_dialog.color_cell_allocated)
	    {
		the_color.pixel = 
		       disp_info[disp_num].color_picker_dialog.the_color_cell;
		XStoreColor(disp_info[disp_num].disp, DefaultColormap(
                        disp_info[disp_num].disp, DefaultScreen(
                        disp_info[disp_num].disp)), 
			&the_color);
                XSetForeground(disp_info[disp_num].disp, 
		  disp_info[disp_num].color_picker_dialog.the_gc,
		  the_color.pixel);
		/*
	        XFillRectangle(disp_info[disp_num].disp, XtWindow(
	            disp_info[disp_num].color_picker_dialog.example_color),
		    disp_info[disp_num].color_picker_dialog.the_gc, 
		    0, 0, 50, 50);
	         */
            }
	    break;
	case ALLOCATE_COLOR:
	    for (i=0; i < num_disps; i++)
            {
	        if (disp_info[i].in_session_bool)
	        {
                    the_color.flags = DoRed | DoGreen | DoBlue;
                    the_color.red = r_val * 256;
                    the_color.green = g_val * 256;
                    the_color.blue = b_val * 256;

		    if (!XAllocColor(disp_info[i].disp, DefaultColormap(
                        disp_info[i].disp, DefaultScreen(disp_info[i].disp)), 
			&the_color))
		    {
			continue;   /* no colors left */
		    }
		    else
		    {
			if (disp_num == i)
			    disp_info[disp_num].current_pixel = the_color.pixel;
                        gcvalues.foreground = the_color.pixel;
		        XChangeGC(disp_info[i].disp, 
			      disp_info[disp_num].win_gc[i],
			      GCForeground, &gcvalues);
		    }
	        }
            }
	    break;
	case GRAB_COLOR:
            if (XGrabPointer(disp_info[disp_num].disp, 
		     disp_info[disp_num].draw_win_id, False, 
		     ButtonPressMask | ButtonReleaseMask,
                     GrabModeAsync, GrabModeAsync, None, 
                     XCreateFontCursor(disp_info[disp_num].disp, XC_crosshair), 
		     CurrentTime)	
                != GrabSuccess)
            {
                popup_info_dialog(disp_num,
	            "Couldn't grab the pointer.\nAborting Color Grab.");
            }
            disp_info[disp_num].input_mode = GRABBING_COLOR;
	    break;
	case COLOR_CANCEL:
	    XtUnmanageChild(disp_info[disp_num].color_picker_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].color_picker_dialog.the_dialog);
	    break;
	default:
	    printf("ERROR: unknown event in handle_color_picker_dialog.\n");
	    break;
    }
}


/*
 * grab_color - this function grabs the color at the indicated spot on the
 *              display and sets the current pen color to that color.
 */
grab_color(disp_num, x, y)
int disp_num, x, y;
{
    XImage *the_image;
    XColor the_color;
    Arg args[MAX_ARGS];
    int i, n, r_val, g_val, b_val;
    XGCValues gcvalues;

    if ((the_image = XGetImage(disp_info[disp_num].disp, 
			       XDefaultRootWindow(disp_info[disp_num].disp),
			       x, y, 1, 1, (~0), ZPixmap)) == NULL) 
    {
        popup_error_dialog(disp_num, "ERROR: XGetImage failed.\n");
	return;
    }

    the_color.pixel = XGetPixel(the_image, 0, 0);
    the_color.flags = DoRed | DoGreen | DoBlue;
    XQueryColor(disp_info[disp_num].disp, 
		XDefaultColormapOfScreen(XDefaultScreenOfDisplay(
		disp_info[disp_num].disp)), &the_color);
    /*
     * show it in the example color widget in the color picker dialog
     */
    if (disp_info[disp_num].color_picker_dialog.color_cell_allocated)
    {
        the_color.flags = DoRed | DoGreen | DoBlue;
	the_color.pixel =disp_info[disp_num].color_picker_dialog.the_color_cell;
	XStoreColor(disp_info[disp_num].disp, DefaultColormap(
                    disp_info[disp_num].disp, DefaultScreen(
                    disp_info[disp_num].disp)), &the_color);
        XSetForeground(disp_info[disp_num].disp, 
	            disp_info[disp_num].color_picker_dialog.the_gc,
		    the_color.pixel);
    }

    /*
     * set the pen color on all displays
     */
    for (i=0; i < num_disps; i++)
    {
        if (disp_info[i].in_session_bool)
        {
            the_color.flags = DoRed | DoGreen | DoBlue;

	    if (!XAllocColor(disp_info[i].disp, DefaultColormap(
                   disp_info[i].disp, DefaultScreen(disp_info[i].disp)), 
		   &the_color))
	    {
		continue;   /* no colors left */
	    }
	    else
	    {
		if (disp_num == i)
		    disp_info[disp_num].current_pixel = the_color.pixel;
                gcvalues.foreground = the_color.pixel;
	        XChangeGC(disp_info[i].disp, 
		          disp_info[disp_num].win_gc[i],
			  GCForeground, &gcvalues);
	    }
        }
    }

    /*
     * set the RGB sliders approximately to where this color is
     */
    r_val = ((int) the_color.red) / 256;
    g_val = ((int) the_color.green) / 256;
    b_val = ((int) the_color.blue) / 256;

    n = 0;
    XtSetArg(args[n], XmNvalue, r_val); n++;
    XtSetValues(disp_info[disp_num].color_picker_dialog.rslider, args, n);

    n = 0;
    XtSetArg(args[n], XmNvalue, g_val); n++;
    XtSetValues(disp_info[disp_num].color_picker_dialog.gslider, args, n);
    n = 0;
    XtSetArg(args[n], XmNvalue, b_val); n++;
    XtSetValues(disp_info[disp_num].color_picker_dialog.bslider, args, n);
}


/*
 * create_color_picker_dialog - this function creates the popup dialog that
 *                          allows the user to pick a color.
 */
Widget create_color_picker_dialog(disp_num)
int disp_num;
{
    Widget the_dialog;
    Widget the_separator;
    Widget the_frame, the_form;
    Widget example_frame;
    Arg args[MAX_ARGS];
    int n;
    XmString mesg_string = NULL;
    XmString tsr = NULL;
    XmString tsg = NULL;
    XmString tsb = NULL;

    n = 0;
    XtSetArg(args[n], XmNautoUnmanage, False); n++;
    XtSetArg(args[n], XmNtraversalOn, True); n++;
    the_dialog = XmCreateFormDialog(disp_info[disp_num].app_shell, 
			   "Color_Picker_Dialog", args, n);

    n = 0;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    the_frame = XmCreateFrame(the_dialog, "dialog_frame", args, n);
    XtManageChild(the_frame);

    n = 0;
    the_form = XmCreateForm(the_frame, "dialog_form", args, n);
    XtManageChild(the_form);

    mesg_string = XmStringCreateLtoR(
       "Dial in the color you would like to use:  ", charset);
    n = 0;
    XtSetArg(args[n], XmNx, 15); n++;
    XtSetArg(args[n], XmNy, 15); n++;
    XtSetArg(args[n], XmNrightOffset, 15); n++;
    XtSetArg(args[n], XmNlabelString, mesg_string); n++;
    disp_info[disp_num].color_picker_dialog.label = 
		   XmCreateLabelGadget(the_form, "prompt", args, n);
    XtManageChild(disp_info[disp_num].color_picker_dialog.label);
    
    tsr = XmStringCreateLtoR("R", charset);
    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 0); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 45); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomOffset, 55); n++;
    XtSetArg(args[n], XmNminimum, 0); n++;
    XtSetArg(args[n], XmNmaximum, 255); n++;
    XtSetArg(args[n], XmNheight, 200); n++;
    XtSetArg(args[n], XmNshowValue, True); n++;
    XtSetArg(args[n], XmNtitleString, tsr); n++;
    disp_info[disp_num].color_picker_dialog.rslider =
                XmCreateScale(the_form, "slider", args, n);
    XtManageChild(disp_info[disp_num].color_picker_dialog.rslider);
    XtAddCallback(disp_info[disp_num].color_picker_dialog.rslider,
		  XmNdragCallback,
	          (XtCallbackProc)handle_color_picker_dialog, (XtPointer)(disp_num << 8 | RED_SLIDER));
    XtAddCallback(disp_info[disp_num].color_picker_dialog.rslider,
		  XmNvalueChangedCallback,
	          (XtCallbackProc)handle_color_picker_dialog, (XtPointer)(disp_num << 8 | RED_SLIDER));

    tsg = XmStringCreateLtoR("G", charset);
    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftOffset, 0); n++;
    XtSetArg(args[n], XmNleftWidget, 
	 disp_info[disp_num].color_picker_dialog.rslider); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 45); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomOffset, 55); n++;
    XtSetArg(args[n], XmNminimum, 0); n++;
    XtSetArg(args[n], XmNmaximum, 255); n++;
    XtSetArg(args[n], XmNheight, 200); n++;  /* no effect ?! */
    XtSetArg(args[n], XmNtitleString, tsg); n++;
    XtSetArg(args[n], XmNshowValue, True); n++;
    disp_info[disp_num].color_picker_dialog.gslider =
                XmCreateScale(the_form, "slider", args, n);
    XtManageChild(disp_info[disp_num].color_picker_dialog.gslider);
    XtAddCallback(disp_info[disp_num].color_picker_dialog.gslider,
		  XmNdragCallback, (XtCallbackProc)handle_color_picker_dialog,
        (XtPointer)(disp_num << 8 | GREEN_SLIDER));
    XtAddCallback(disp_info[disp_num].color_picker_dialog.gslider,
		  XmNvalueChangedCallback, (XtCallbackProc)handle_color_picker_dialog,
        (XtPointer)(disp_num << 8 | GREEN_SLIDER));

    tsb = XmStringCreateLtoR("B", charset);
    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftOffset, 0); n++;
    XtSetArg(args[n], XmNleftWidget, 
	 disp_info[disp_num].color_picker_dialog.gslider); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 45); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomOffset, 55); n++;
    XtSetArg(args[n], XmNminimum, 0); n++;
    XtSetArg(args[n], XmNmaximum, 255); n++;
    XtSetArg(args[n], XmNheight, 200); n++;
    XtSetArg(args[n], XmNshowValue, True); n++;
    XtSetArg(args[n], XmNtitleString, tsb); n++;
    disp_info[disp_num].color_picker_dialog.bslider =
                XmCreateScale(the_form, "slider", args, n);
    XtManageChild(disp_info[disp_num].color_picker_dialog.bslider);
    XtAddCallback(disp_info[disp_num].color_picker_dialog.bslider,
		  XmNdragCallback, (XtCallbackProc)handle_color_picker_dialog,
        (XtPointer)(disp_num << 8 | BLUE_SLIDER));
    XtAddCallback(disp_info[disp_num].color_picker_dialog.bslider,
		  XmNvalueChangedCallback, (XtCallbackProc)handle_color_picker_dialog,
        (XtPointer)(disp_num << 8 | BLUE_SLIDER));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftOffset, 10); n++;
    XtSetArg(args[n], XmNleftWidget, 
	 disp_info[disp_num].color_picker_dialog.bslider); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, 
	 disp_info[disp_num].color_picker_dialog.bslider); n++;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_IN); n++;
    example_frame = XmCreateFrame(the_form, "example_frame", args, n);
    XtManageChild(example_frame);

    n = 0;
    XtSetArg(args[n], XmNwidth, 50); n++;
    XtSetArg(args[n], XmNheight, 50); n++;
    disp_info[disp_num].color_picker_dialog.example_color =
                XmCreateDrawingArea(example_frame, "example_color", args, n);
    XtManageChild(disp_info[disp_num].color_picker_dialog.example_color);
    XtAddCallback(disp_info[disp_num].color_picker_dialog.example_color, 
		  XmNexposeCallback, (XtCallbackProc)example_area_exposeCB, (XtPointer)disp_num);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftWidget, 
	 disp_info[disp_num].color_picker_dialog.bslider); n++;
    XtSetArg(args[n], XmNleftOffset, 10); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 120); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].color_picker_dialog.grab_button =
                XmCreatePushButtonGadget(the_form, "Grab Color", args, n);
    XtManageChild(disp_info[disp_num].color_picker_dialog.grab_button);
    XtAddCallback(disp_info[disp_num].color_picker_dialog.grab_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_color_picker_dialog,
        (XtPointer)(disp_num << 8 | GRAB_COLOR));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 0); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 0); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget,
             disp_info[disp_num].color_picker_dialog.rslider); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    the_separator = XmCreateSeparatorGadget(the_form, "separator", args, n);
    XtManageChild(the_separator);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].color_picker_dialog.rslider); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNrightPosition, 45); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].color_picker_dialog.alloc_button =
                XmCreatePushButtonGadget(the_form, "Set Color", args, n);
    XtManageChild(disp_info[disp_num].color_picker_dialog.alloc_button);
    XtAddCallback(disp_info[disp_num].color_picker_dialog.alloc_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_color_picker_dialog,
        (XtPointer)(disp_num << 8 | ALLOCATE_COLOR));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftWidget, 
	disp_info[disp_num].color_picker_dialog.alloc_button); n++;
    XtSetArg(args[n], XmNleftOffset, 20); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].color_picker_dialog.rslider); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].color_picker_dialog.cancel_button =
                XmCreatePushButtonGadget(the_form, "Close", args, n);
    XtManageChild(disp_info[disp_num].color_picker_dialog.cancel_button);
    XtAddCallback(disp_info[disp_num].color_picker_dialog.cancel_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_color_picker_dialog,
        (XtPointer)(disp_num << 8 | COLOR_CANCEL));

    if (mesg_string)
	XtFree((char*)mesg_string);
    if (tsr)
	XtFree((char*)tsr);
    if (tsg)
	XtFree((char*)tsg);
    if (tsb)
	XtFree((char*)tsb);
    
    set_dialog_wm_name(the_dialog, "Color Picker");
    return(the_dialog);
}


/*
 * handle_eraser_dialog - this function is the callback for the eraser
 *                        dialog.
 */
handle_eraser_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int n, i, x, y, width, height;
    Dimension d_width, d_height;
    Arg args[MAX_ARGS];
    Widget clip_window;
    int disp_num, which_button;
    char *the_text;
    Window dummy;

    disp_num = ((int) client_data) >> 8;
    which_button = ((int) client_data) & 0x00FF;

    switch(which_button)
    {
	case ERASE_VISIBLE:
	    XtUnmanageChild(disp_info[disp_num].eraser_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].eraser_dialog.the_dialog);
   	    n = 0;
	    XtSetArg(args[n], XmNclipWindow, &clip_window); n++;
	    XtGetValues(disp_info[disp_num].swindow, args, n);
   	    n = 0;
	    XtSetArg(args[n], XmNwidth, &d_width); n++;
	    XtSetArg(args[n], XmNheight, &d_height); n++;
	    XtGetValues(clip_window, args, n);
	    width = (int) d_width;
	    height = (int) d_height;
            
            XTranslateCoordinates(disp_info[disp_num].disp, 
			  XtWindow(clip_window),
			  disp_info[disp_num].draw_win_id, 0,0, &x, &y, &dummy);

	    for (i=0; i < num_disps; i++)
	    {
	        if (disp_info[i].in_session_bool)  
		{
                    XClearArea(disp_info[i].disp, disp_info[i].draw_win_id,
		       x, y, width, height, False);
                    XFillRectangle(disp_info[i].disp,disp_info[i].backing_store,
                       disp_info[i].backing_store_gc, x, y, width, height);
                    XFlush(disp_info[i].disp);
		}
	    }
	    break;
	case ERASE_ALL_WHITEBOARD:
	    XtUnmanageChild(disp_info[disp_num].eraser_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].eraser_dialog.the_dialog);
	    process_menu_selection(disp_num, EDIT_MENU, CLEAR_WHITEBOARD);
	    break;
	case ERASER_OK:
	    XtUnmanageChild(disp_info[disp_num].eraser_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].eraser_dialog.the_dialog);
	    width = -1;
	    the_text = XmTextGetString( 
		    disp_info[disp_num].eraser_dialog.textfield);
	    sscanf(the_text, "%d", &width);
	    if (width > 2)
	    {
	        disp_info[disp_num].eraser_win_width = width;
	        disp_info[disp_num].eraser_win_height = width;
	    }
	    break;
	case ERASER_CANCEL:
	    XtUnmanageChild(disp_info[disp_num].eraser_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].eraser_dialog.the_dialog);
	    break;
	default:
	    printf("ERROR: unknown event in handle_eraser_dialog.\n");
	    break;
    }
}


/*
 * create_eraser_dialog - this function creates the popup dialog that
 *                        controls the eraser such as eraser size, and fast
 *
 */
Widget create_eraser_dialog(disp_num)
int disp_num;
{
    Widget the_dialog;
    Widget the_separator;
    Widget the_frame, the_form;
    Arg args[MAX_ARGS];
    int n;
    XmString mesg_string = NULL;

    n = 0;
    XtSetArg(args[n], XmNautoUnmanage, False); n++;
    XtSetArg(args[n], XmNtraversalOn, True); n++;
    the_dialog = XmCreateFormDialog(disp_info[disp_num].app_shell, 
			   "Eraser_Dialog", args, n);

    n = 0;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    the_frame = XmCreateFrame(the_dialog, "dialog_frame", args, n);
    XtManageChild(the_frame);

    n = 0;
    the_form = XmCreateForm(the_frame, "dialog_form", args, n);
    XtManageChild(the_form);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].eraser_dialog.erase_all_button =
                XmCreatePushButtonGadget(the_form, 
		    "Erase Entire Whiteboard", args, n);
    XtManageChild(disp_info[disp_num].eraser_dialog.erase_all_button);
    XtAddCallback(disp_info[disp_num].eraser_dialog.erase_all_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_eraser_dialog,
        (XtPointer)(disp_num << 8 | ERASE_ALL_WHITEBOARD));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
    XtSetArg(args[n], XmNrightWidget, 
	     disp_info[disp_num].eraser_dialog.erase_all_button); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].eraser_dialog.erase_all_button); n++;
    XtSetArg(args[n], XmNbottomOffset, 5); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].eraser_dialog.erase_vis_button =
                XmCreatePushButtonGadget(the_form, 
		      "Erase Visible Region of Whiteboard", args, n);
    XtManageChild(disp_info[disp_num].eraser_dialog.erase_vis_button);
    XtAddCallback(disp_info[disp_num].eraser_dialog.erase_vis_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_eraser_dialog,
        (XtPointer)(disp_num << 8 | ERASE_VISIBLE));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 0); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 0); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget,
             disp_info[disp_num].eraser_dialog.erase_vis_button); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    the_separator = XmCreateSeparatorGadget(the_form, "separator", args, n);
    XtManageChild(the_separator);

    mesg_string = XmStringCreateLtoR(
       "If you wish to change the eraser size,\nenter the new size in pixels:",
       charset);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget,
             disp_info[disp_num].eraser_dialog.erase_vis_button); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 15); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNx, 30); n++;
    XtSetArg(args[n], XmNlabelString, mesg_string); n++;
    disp_info[disp_num].eraser_dialog.label = 
		   XmCreateLabelGadget(the_form, "prompt", args, n);
    XtManageChild(disp_info[disp_num].eraser_dialog.label);
    
    n = 0;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].eraser_dialog.erase_vis_button); n++;
    XtSetArg(args[n], XmNtopOffset, 60); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    disp_info[disp_num].eraser_dialog.textfield = 
		   XmCreateText(the_form, "textfield", args, n);
    XtManageChild(disp_info[disp_num].eraser_dialog.textfield);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 0); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 0); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget,
             disp_info[disp_num].eraser_dialog.textfield); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    the_separator = XmCreateSeparatorGadget(the_form, "separator", args, n);
    XtManageChild(the_separator);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 30); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].eraser_dialog.textfield); n++;
    XtSetArg(args[n], XmNbottomOffset, 5); n++;
    XtSetArg(args[n], XmNrightOffset, 5); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    XtSetArg(args[n], XmNwidth, 60); n++;
    XtSetArg(args[n], XmNshowAsDefault, 1); n++;
    disp_info[disp_num].eraser_dialog.ok_button =
                XmCreatePushButtonGadget(the_form, "OK", args, n);
    XtManageChild(disp_info[disp_num].eraser_dialog.ok_button);
    XtAddCallback(disp_info[disp_num].eraser_dialog.ok_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_eraser_dialog,
        (XtPointer)(disp_num << 8 | ERASER_OK));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftWidget, 
	     disp_info[disp_num].eraser_dialog.ok_button); n++;
    XtSetArg(args[n], XmNleftOffset, 10); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, 
	     disp_info[disp_num].eraser_dialog.ok_button); n++;
    XtSetArg(args[n], XmNtopOffset, 5); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
    XtSetArg(args[n], XmNbottomWidget, 
	     disp_info[disp_num].eraser_dialog.ok_button); n++;
    XtSetArg(args[n], XmNbottomOffset, 5); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 30); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].eraser_dialog.cancel_button =
                XmCreatePushButtonGadget(the_form, "Cancel", args, n);
    XtManageChild(disp_info[disp_num].eraser_dialog.cancel_button);
    XtAddCallback(disp_info[disp_num].eraser_dialog.cancel_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_eraser_dialog,
        (XtPointer)(disp_num << 8 | ERASER_CANCEL));

    if (mesg_string)
	XtFree((char*)mesg_string);
    
    set_dialog_wm_name(the_dialog, "Eraser");
    return(the_dialog);
}

/*
 * parse_and_fix_dialog_text - this function takes a file name with tildas
 *                             and such and returns a file name that has
 *                             the tildas expanded, etc.
 */
char *parse_and_fix_dialog_text(fname)
char *fname;
{
    char *filename;

    filename = (char *) malloc(512);

    strncpy(filename, fname, 500);
    return(filename);
}


/*
 * handle_read_file_selection_dialog - this function is the callback for the 
 *                        readfile selection dialog.
 */
handle_read_file_selection_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int n, i;
    Arg args[MAX_ARGS];
    int disp_num, which_button;
    FILE *fp;
    XmFileSelectionBoxCallbackStruct *fcb;
    char *fname, *filename;
    char tmp_str[256];

    disp_num = ((int) client_data) >> 8;
    which_button = ((int) client_data) & 0x00FF;

    switch(which_button)
    {
	case CANCEL:
	    XtUnmanageChild(
		disp_info[disp_num].read_file_dialog.file_selection_dialog);
	    jiggle_dialog(
		disp_info[disp_num].read_file_dialog.file_selection_dialog);
	    break;
	case OK:
	    XtUnmanageChild(
		disp_info[disp_num].read_file_dialog.file_selection_dialog);
	    jiggle_dialog(
		disp_info[disp_num].read_file_dialog.file_selection_dialog);
	    fcb = (XmFileSelectionBoxCallbackStruct *) call_data;
	    XmStringGetLtoR(fcb->value, charset, &fname);
	    filename = parse_and_fix_dialog_text(fname);
	    free(fname);
	    if (filename == NULL)
	    {
                popup_info_dialog(disp_num, "Could not parse filename.");
		return;
	    }
	    else
	    {
                if ((fp = fopen(filename, "r" )) == NULL)
		{
		    sprintf(tmp_str, "Could not find file called:\n%s", 
			   filename);
		    popup_info_dialog(disp_num, tmp_str);
		    return;
                }
		else
		{
		    fclose(fp);
		    boink_tool(disp_num, SELECTION_BUTTON);
		    switch(disp_info[disp_num].read_file_dialog.format_chosen)
		    {
	                case BITMAP_FORMAT:
		            read_bitmap_file_into_region(disp_num, filename);
			    break;
	                case XWD_FORMAT:
		            read_xwd_file_into_region(disp_num, filename);
			    break;
	                case TEXT_FILE_FORMAT:
			    read_text_file_into_region(disp_num, filename);
			    break;
	                case GREYSCALE_POSTSCRIPT:
			    break;
	                case PPM_FORMAT:
		            read_ppm_file_into_region(disp_num, filename);
			    break;
			default:
			    popup_error_dialog(disp_num, "unknown format");
			    break;
		    }
		}
	    }
	    break;
        default:
	    popup_error_dialog(disp_num, 
		 "ERROR: bad button in handle_read_file_selection_dialog.");
            break;
    }
}
