/*
 * This file:  wsevents.c
 *
 * This file contains most of the event processing for wscrawl.
 */

#include "wscrawl.h"
#include "pause_curs.h"
#include "languages.h"

extern int num_disps;
extern XtAppContext appCon;

extern struct disp_info_struct disp_info[MAX_NUM_DISPS];
int vsbar_movementCB();
int hsbar_movementCB();
XtTimerCallbackProc spraypaint_spotCB();
XtInputCallbackProc handle_knock_knockCB();
XtEventHandler handle_wscrawl_window_moveCB();
float my_rand();

extern struct global_values_struct global_values;
extern int draw_area_height, draw_area_width;	/* shared among all displays */
extern int sync_scrollbars_bool;	      


/*
 * pointer_moveCB - this function is called when the pointer moves in the
 *                  drawing window.
 */
XtEventHandler pointer_moveCB(w, client_data, our_event, cbool)
Widget w;
caddr_t client_data;
XEvent *our_event;
Boolean *cbool;
{
    int i, disp_num, x, y, root_x, root_y;
    XMotionEvent *the_event;
    Window dummy;
    Widget dummy_w = (Widget)NULL;
    Boolean dummy_cbool;
    XmString labelstr = NULL;
    int n;
    Arg args[MAX_ARGS];
    char tmp_str[256];

    if (our_event->type != MotionNotify)
	printf("ERROR: wrong event type in pointer_moveCB.\n");

    the_event = (XMotionEvent *) our_event;
    disp_num = (int) client_data;

    if (disp_info[disp_num].show_pointer_coords == TRUE)
    {
        sprintf(tmp_str, "x=%d", the_event->x);
	labelstr = XmStringCreateLtoR(tmp_str, "fixed");
        n = 0;
        XtSetArg(args[n], XmNlabelString, labelstr); n++;
        XtSetValues(disp_info[disp_num].pointer_x_coords_label, args, n);
	XtFree((char*)labelstr);
        sprintf(tmp_str, "y=%d", the_event->y);
	labelstr = XmStringCreateLtoR(tmp_str, "fixed");
        n = 0;
        XtSetArg(args[n], XmNlabelString, labelstr); n++;
        XtSetValues(disp_info[disp_num].pointer_y_coords_label, args, n);
	XtFree((char*)labelstr);
    }

    if (disp_info[disp_num].pointer_state != PRESSED) 
    {
        if (disp_info[disp_num].show_telepointer_always == TRUE)
            handle_pointer_move_while_pointing(disp_num, the_event);
        return;
    }

    switch(disp_info[disp_num].input_mode)
    {
	case POINTING:
            handle_pointer_move_while_pointing(disp_num, the_event);
            break;
	case SELECTING:
            draw_rectangle(disp_num, disp_info[disp_num].rubber_band_gc);
            if (disp_info[disp_num].show_telepointer_always == TRUE)
                handle_pointer_move_while_pointing(disp_num, the_event);
	    disp_info[disp_num].last_rubberband_point.x = the_event->x;
	    disp_info[disp_num].last_rubberband_point.y = the_event->y;
            draw_rectangle(disp_num, disp_info[disp_num].rubber_band_gc);
	    XFlush(disp_info[disp_num].disp);
            if (disp_info[disp_num].show_telepointer_always == TRUE)
                map_pointer(disp_num);
            break;
	case SUPER_SELECTING:
	    if (disp_info[disp_num].select_dialog.pointer_down == TRUE)
	    {
		return;		/* not interested yet */
	    }
            draw_super_rectangle(disp_num, 
				 disp_info[disp_num].super_rubber_band_gc);
	    /*
	    XtTranslateCoords(disp_info[disp_num].draw_area, 
		          the_event->x, the_event->y, &root_x, &root_y);
	    */
	    myXtTranslateCoords(disp_num, disp_info[disp_num].draw_area, 
		          the_event->x, the_event->y, &root_x, &root_y);
	    disp_info[disp_num].last_rubberband_point.x = root_x;
	    disp_info[disp_num].last_rubberband_point.y = root_y;
            draw_super_rectangle(disp_num, 
				 disp_info[disp_num].super_rubber_band_gc);
	    XFlush(disp_info[disp_num].disp);
	    break;
	case PAINTING:
	    handle_pointer_move_while_painting(disp_num, the_event);
            if (disp_info[disp_num].show_telepointer_always == TRUE)
                handle_pointer_move_while_pointing(disp_num, the_event);
	    break;
	case PENCILING:
	    handle_pointer_move_while_penciling(disp_num, the_event);
            if (disp_info[disp_num].show_telepointer_always == TRUE)
                handle_pointer_move_while_pointing(disp_num, the_event);
	    break;
	case TYPING:
            if (disp_info[disp_num].show_telepointer_always == TRUE)
                handle_pointer_move_while_pointing(disp_num, the_event);
	    break;
	case ERASING:
	    handle_pointer_move_while_erasing(disp_num, the_event);
            if (disp_info[disp_num].show_telepointer_always == TRUE)
                handle_pointer_move_while_pointing(disp_num, the_event);
	    break;
	case SPRAYPAINTING:
            if (disp_info[disp_num].show_telepointer_always == TRUE)
                handle_pointer_move_while_pointing(disp_num, the_event);
	    break;
	case DRAWING_LINES:
            XDrawLine(disp_info[disp_num].disp, disp_info[disp_num].draw_win_id,
	            disp_info[disp_num].rubber_band_gc,
	            disp_info[disp_num].orig_rubberband_point.x,
	            disp_info[disp_num].orig_rubberband_point.y,
	            disp_info[disp_num].last_rubberband_point.x,
	            disp_info[disp_num].last_rubberband_point.y);
            if (disp_info[disp_num].show_telepointer_always == TRUE)
                handle_pointer_move_while_pointing(disp_num, the_event);
	    disp_info[disp_num].last_rubberband_point.x = the_event->x;
	    disp_info[disp_num].last_rubberband_point.y = the_event->y;
            XDrawLine(disp_info[disp_num].disp, disp_info[disp_num].draw_win_id,
	            disp_info[disp_num].rubber_band_gc,
	            disp_info[disp_num].orig_rubberband_point.x,
	            disp_info[disp_num].orig_rubberband_point.y,
	            disp_info[disp_num].last_rubberband_point.x,
	            disp_info[disp_num].last_rubberband_point.y);
	    break;
	case OUTLINING_SQUARES:
	case FILLING_SQUARES:
            draw_rectangle(disp_num, disp_info[disp_num].rubber_band_gc);
            if (disp_info[disp_num].show_telepointer_always == TRUE)
                handle_pointer_move_while_pointing(disp_num, the_event);
	    disp_info[disp_num].last_rubberband_point.x = the_event->x;
	    disp_info[disp_num].last_rubberband_point.y = the_event->y;
            draw_rectangle(disp_num, disp_info[disp_num].rubber_band_gc);
	    XFlush(disp_info[disp_num].disp);
	    break;
	case OUTLINING_OVALS:
	case FILLING_OVALS:
            draw_oval(disp_num, disp_info[disp_num].rubber_band_gc);
            if (disp_info[disp_num].show_telepointer_always == TRUE)
                handle_pointer_move_while_pointing(disp_num, the_event);
	    disp_info[disp_num].last_rubberband_point.x = the_event->x;
	    disp_info[disp_num].last_rubberband_point.y = the_event->y;
            draw_oval(disp_num, disp_info[disp_num].rubber_band_gc);
	    XFlush(disp_info[disp_num].disp);
	    break;
	case DRAGGING_SELECTION:
	    selection_pointer_moveCB(dummy_w, disp_num << 8 | disp_num, 
			   (XEvent *) the_event, &dummy_cbool);
	    break;
	case GRABBING_COLOR:
	    break;
        default:
	    printf("ERROR: unknown input mode in pointer_moveCB.\n");
	    break;
    }
}


/*
 * handle_pointer_move_while_penciling - this function is called when the
 *                   pointer has been dragged during penciling.  It must
 *                   draw the new paint segment on all the displays.
 */
handle_pointer_move_while_penciling(disp_num, the_event)
int disp_num;
XMotionEvent *the_event;
{
    int i, j;

    for (i=0; i < num_disps; i++)
    {
        if (disp_info[i].in_session_bool)
        {
            XDrawLine(disp_info[i].disp, disp_info[i].draw_win_id, 
    	           disp_info[disp_num].win_gc[i], 
    	           disp_info[disp_num].last_scrawl_point.x, 
	           disp_info[disp_num].last_scrawl_point.y,
    	           the_event->x, the_event->y);
            XDrawLine(disp_info[i].disp, disp_info[i].backing_store, 
    	           disp_info[disp_num].win_gc[i], 
    	           disp_info[disp_num].last_scrawl_point.x, 
	           disp_info[disp_num].last_scrawl_point.y,
    	           the_event->x, the_event->y);
            XFlush(disp_info[i].disp);
        }
    }

    disp_info[disp_num].last_scrawl_point.x = the_event->x;
    disp_info[disp_num].last_scrawl_point.y = the_event->y;
}


/*
 * handle_pointer_move_while_painting - this function is called when the
 *                   pointer has been dragged during painting.  It must
 *                   draw the new paint segment on all the displays.
 */
handle_pointer_move_while_painting(disp_num, the_event)
int disp_num;
XMotionEvent *the_event;
{
    int i, j;

    if (disp_info[disp_num].paintbrush.paint_mode == NORMAL)
    {
        for (i=0; i < num_disps; i++)
        {
            if (disp_info[i].in_session_bool)
            {
                XDrawLine(disp_info[i].disp, disp_info[i].draw_win_id, 
    	           disp_info[disp_num].win_gc[i], 
    	           disp_info[disp_num].last_scrawl_point.x, 
	           disp_info[disp_num].last_scrawl_point.y,
    	           the_event->x, the_event->y);
                XDrawLine(disp_info[i].disp, disp_info[i].backing_store, 
    	           disp_info[disp_num].win_gc[i], 
    	           disp_info[disp_num].last_scrawl_point.x, 
	           disp_info[disp_num].last_scrawl_point.y,
    	           the_event->x, the_event->y);
	        XFlush(disp_info[i].disp);
            }
        }
    }
    else if (disp_info[disp_num].paintbrush.paint_mode == BITMAP_BRUSH)
    {
        for (i=0; i < num_disps; i++)
        {
            if (disp_info[i].in_session_bool)
            {
		for (j=0; j < disp_info[disp_num].paintbrush.num_points; j++)
		{
                    XDrawLine(disp_info[i].disp, disp_info[i].draw_win_id, 
    	               disp_info[disp_num].win_gc[i], 
    	               disp_info[disp_num].last_scrawl_point.x +
		       disp_info[disp_num].paintbrush.x[j], 
	               disp_info[disp_num].last_scrawl_point.y +
		       disp_info[disp_num].paintbrush.y[j], 
    	               the_event->x + disp_info[disp_num].paintbrush.x[j], 
		       the_event->y + disp_info[disp_num].paintbrush.y[j]);
                    XDrawLine(disp_info[i].disp, disp_info[i].backing_store, 
    	               disp_info[disp_num].win_gc[i], 
    	               disp_info[disp_num].last_scrawl_point.x +
		       disp_info[disp_num].paintbrush.x[j], 
	               disp_info[disp_num].last_scrawl_point.y +
		       disp_info[disp_num].paintbrush.y[j], 
    	               the_event->x + disp_info[disp_num].paintbrush.x[j], 
		       the_event->y + disp_info[disp_num].paintbrush.y[j]);
		}

	        XFlush(disp_info[i].disp);
            }
        }
    }
    else
    {
        printf("ERROR: in handle_pointer_move_while_painting, paint_mode\n");
    }

    disp_info[disp_num].last_scrawl_point.x = the_event->x;
    disp_info[disp_num].last_scrawl_point.y = the_event->y;
}


/*
 * handle_pointer_move_while_pointing - this function is called when the
 *    		    pointer has been dragged during pointing.  It first
 *                  turns off the old pointer (by coping from the backing
 *                  store onto the windows), then then it paints
 *                  the new pointer, and then redraws all old pointers,
 *                  because they might have been obscured.
 */
handle_pointer_move_while_pointing(disp_num, the_event)
int disp_num;
XMotionEvent *the_event;
{
    int i, j, left, top;

    unmap_pointer(disp_num);

    disp_info[disp_num].last_telepointer_point.x = the_event->x;
    disp_info[disp_num].last_telepointer_point.y = the_event->y;

    for (i=0; i < num_disps; i++)
    {
        if (disp_info[i].in_session_bool)
	{
            draw_rubber_pointer(disp_num, i);
	    XFlush(disp_info[disp_num].disp);
	}
    }
}


/*
 * unmap_pointer - this function erases this display's rubber pointer from 
 *                 all the displays
 */
unmap_pointer(disp_num)
int disp_num;
{
    int i, oldleft, oldtop;

    disp_info[disp_num].rubber_pointer.is_mapped_bool = FALSE;
    oldleft = disp_info[disp_num].last_telepointer_point.x;
    oldtop = disp_info[disp_num].last_telepointer_point.y -
                                  disp_info[disp_num].rubber_pointer.height;
    /*
     * Replace chunk where old cursor was
     */
    for (i=0; i < num_disps; i++)
    {
        if (disp_info[i].in_session_bool)
	{
            XCopyArea(disp_info[i].disp, disp_info[i].backing_store,
		disp_info[i].draw_win_id, disp_info[disp_num].win_gc[i],
		oldleft, oldtop, 
		disp_info[disp_num].rubber_pointer.width,
		disp_info[disp_num].rubber_pointer.height, oldleft, oldtop);
	}
    }
}


/*
 * move_and_unmap_pointer - this function erases this display's rubber 
 *                pointer from all the displays, and moves its internal vars.
 */
move_and_unmap_pointer(disp_num, the_event)
int disp_num;
XMotionEvent *the_event;
{
    int i, oldleft, oldtop;

    disp_info[disp_num].rubber_pointer.is_mapped_bool = FALSE;
    oldleft = disp_info[disp_num].last_telepointer_point.x;
    oldtop = disp_info[disp_num].last_telepointer_point.y -
                                  disp_info[disp_num].rubber_pointer.height;
    /*
     * Replace chunk where old cursor was
     */
    for (i=0; i < num_disps; i++)
    {
        if (disp_info[i].in_session_bool)
	{
            XCopyArea(disp_info[i].disp, disp_info[i].backing_store,
		disp_info[i].draw_win_id, disp_info[disp_num].win_gc[i],
		oldleft, oldtop, 
		disp_info[disp_num].rubber_pointer.width,
		disp_info[disp_num].rubber_pointer.height, oldleft, oldtop);
	}
    }

    disp_info[disp_num].last_telepointer_point.x = the_event->x;
    disp_info[disp_num].last_telepointer_point.y = the_event->y;
}


/*
 * map_pointer - this function draws the rubber pointer of the display 
 *               indicated
 */
map_pointer(disp_num)
{
    int i;

    for (i=0; i < num_disps; i++)
    {
	if (disp_info[i].in_session_bool)
	{
	    draw_rubber_pointer(disp_num, i);
            XFlush(disp_info[i].disp);
	}
    }
}


/*
 * map_pointers - this function draws all the rubber pointers that should
 *                be drawn on the displays.
 */
map_pointers()
{
    int i, j;

    for (i=0; i < num_disps; i++)
    {
	if (disp_info[i].in_session_bool)
	{
            if (disp_info[i].pointer_mapped)
	    {    
                for (j=0; j < num_disps; j++)
                {
	            if (disp_info[j].in_session_bool)
	            {
	                draw_rubber_pointer(i, j);
	            }
                }
	    }    
	}
    }

    for (i=0; i < num_disps; i++)
        XFlush(disp_info[i].disp);
}


/*
 * handle_pointer_move_while_erasing - this function is called when the
 *                   pointer has been dragged during erasing.  It must
 *                   erase the new spot on all the displays.
 */
handle_pointer_move_while_erasing(disp_num, the_event)
int disp_num;
XMotionEvent *the_event;
{
    int i;

    XMoveWindow(disp_info[disp_num].disp,
                disp_info[disp_num].eraser_win_id,
                the_event->x - disp_info[disp_num].eraser_win_width/2,
                the_event->y - disp_info[disp_num].eraser_win_height/2);

    for (i=0; i < num_disps; i++)
    {
        if (disp_info[i].in_session_bool)
        {
            XClearArea(disp_info[i].disp, disp_info[i].draw_win_id,
                       the_event->x - disp_info[disp_num].eraser_win_width/2,
                       the_event->y - disp_info[disp_num].eraser_win_height/2,
                       disp_info[disp_num].eraser_win_width,
                       disp_info[disp_num].eraser_win_height, False);
            XFillRectangle(disp_info[i].disp, disp_info[i].backing_store,
                       disp_info[i].backing_store_gc, 
                       the_event->x - disp_info[disp_num].eraser_win_width/2,
                       the_event->y - disp_info[disp_num].eraser_win_height/2,
                       disp_info[disp_num].eraser_win_width,
                       disp_info[disp_num].eraser_win_height);
	    XFlush(disp_info[i].disp);
        }
    }

    map_pointers();
}


/*
 * draw_area_exposeCB - this function gets called when an expose even
 *                     occurs in the drawing area.
 */
draw_area_exposeCB(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int disp_num;
    XmDrawingAreaCallbackStruct *cb_struct;
    XExposeEvent *the_event;

    cb_struct = (XmDrawingAreaCallbackStruct *) call_data;
    the_event = (XExposeEvent *) cb_struct->event;
    disp_num = (int) client_data;
    
    if (the_event->type != Expose)
	printf("ERROR: in draw_area_exposeCB.\n");

    XCopyArea(disp_info[disp_num].disp, disp_info[disp_num].backing_store,
              disp_info[disp_num].draw_win_id, 
	      disp_info[disp_num].win_gc[disp_num], the_event->x, the_event->y,
	      the_event->width, the_event->height, the_event->x, the_event->y);
}


/*
 * status_area_exposeCB - this function gets called when an expose event
 *                     occurs in the status area.
 */
status_area_exposeCB(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int disp_num;
    XmDrawingAreaCallbackStruct *cb_struct;
    XExposeEvent *the_event;

    cb_struct = (XmDrawingAreaCallbackStruct *) call_data;
    the_event = (XExposeEvent *) cb_struct->event;
    disp_num = (int) client_data;
    
    if (the_event->type != Expose)
	printf("ERROR: in status_area_exposeCB.\n");

    draw_status_win_message(disp_num);
}


/*
 * draw_status_win_message - this function spits (in camel like fashion) out
 *                      the status window message to the display indicated.
 */
draw_status_win_message(disp_num)
{
    int i, num_disps2;
    char mesg[510], *cur_loc_ptr;

    for (i=0, num_disps2=0; i < num_disps; i++)
        if (disp_info[i].in_session_bool)  /*only if this window is alive*/
            num_disps2++;

    if (num_disps2 == 1)
        sprintf(mesg, disp_info[disp_num].status1_line_format);
    else
        sprintf(mesg, disp_info[disp_num].status_line_format, num_disps2);

    cur_loc_ptr = mesg + strlen(mesg);

    for (i=0; i < num_disps; i++)
    {
        if (disp_info[i].in_session_bool)  /*only if this window is alive*/
        {
            strncpy(cur_loc_ptr, disp_info[i].rubber_pointer.pointer_str, 35);
            cur_loc_ptr += strlen(disp_info[i].rubber_pointer.pointer_str);
            cur_loc_ptr[0] = ',';
            cur_loc_ptr[1] = ' ';          /*space delineates*/
            cur_loc_ptr += 2;
        }
    }
    cur_loc_ptr -= 2;
    cur_loc_ptr[0] = '\0';     /*end the string*/

    XClearWindow(disp_info[disp_num].disp, 
		 XtWindow(disp_info[disp_num].status_area));
    XDrawString(disp_info[disp_num].disp, 
		XtWindow(disp_info[disp_num].status_area),
                disp_info[disp_num].menu_gc, 10, 14, mesg, strlen(mesg));
    XFlush(disp_info[disp_num].disp);
}


/*
 * handle_non_typing_KeyPress_event - this function handles a KeyPress event 
 *                         in some window, but when this person is NOT typing,
 *                         so we handle things like menu accelerators, etc.
 */
handle_non_typing_KeyPress_event(the_char, disp_num)
char the_char;
int disp_num;
{
    int menu_num, i, x, y, width, height;

    switch ((int) the_char)
    {
        case 0:                     /* non-interesting like a "shift" */
            return(FALSE);
        case 3:                     /* "^c" for "Copy" */
	    if (disp_info[disp_num].menu[EDIT_MENU].selectable[COPY])
	        process_menu_selection(disp_num, EDIT_MENU, COPY);
            return(TRUE);
        case 17:                    /* "^q" for "Quit" */
	    if (disp_info[disp_num].menu[FILE_MENU].selectable[QUIT])
	        process_menu_selection(disp_num, FILE_MENU, QUIT);
            return(TRUE);
        case 22:                    /* "^v" for "Paste" */
	    if (disp_info[disp_num].menu[EDIT_MENU].selectable[PASTE])
	        process_menu_selection(disp_num, EDIT_MENU, PASTE);
            return(TRUE);
        case 24:                    /* "^x" for "Cut" */
	    if (disp_info[disp_num].menu[EDIT_MENU].selectable[CUT])
	        process_menu_selection(disp_num, EDIT_MENU, CUT);
            return(TRUE);
        default:                    /* something we are not interested in */
            break;
    }

    if ((the_char == 8) ||     	    /* 8 == Backspace */
        (the_char == 127))     	    /* 127 == Del or Rubout */
    {
        if (disp_info[disp_num].selected_region.region_is_selected == TRUE)
        {   
	    clear_region(disp_num);
	    return(TRUE);
        }
    }

    return(FALSE);		/* might be a typed character */
}


/*
 * handle_KeyPress_event - this function handles a KeyPress event in the
 *                         drawing window, most likely it is for typing.
 */
handle_KeyPress_event(our_event, disp_num)
XEvent *our_event;
int disp_num;
{
    XKeyPressedEvent *the_event;
    int menu_num, i;
    KeySym ksr;
    char the_char[256];

    the_event = (XKeyPressedEvent *) our_event;

    the_char[0] = 0;
    XLookupString((XKeyEvent *) (the_event), the_char, 10, &ksr, NULL);

    if ((handle_non_typing_KeyPress_event(the_char[0], disp_num) == TRUE) ||
        (disp_info[disp_num].input_mode != TYPING))
    {
	return;		/* not interested in this KeyPress event */
    }
    else
    {
        draw_char_on_screen(disp_num, the_char);
    }
}


/*
 * draw_char_on_screen - this function takes the character and draws it
 * 			 on the screen, adds it to the history, etc.
 */
draw_char_on_screen(disp_num, the_char)
int disp_num;
char *the_char;
{
    int i;

    if (the_char[0] == 0)
    {
        return;                        	/* it was a shift or some such */
    }
    else if ((the_char[0] == 13) ||    	/* 13 == Carriage Return */
             (the_char[0] == 10))    	/* 10 == Line Feed */
    {
	erase_type_cursor(disp_num);	/* erase cursor */
        push_letter_to_history(disp_num, the_char[0]);
        disp_info[disp_num].type_info.cur_pos.x = 
			      disp_info[disp_num].type_info.orig_pos.x;
        disp_info[disp_num].type_info.cur_pos.y +=
			      disp_info[disp_num].type_info.char_height;

	draw_type_cursor(disp_num);	/* redraw cursor */
    }
    else if ((the_char[0] == 8) ||     	/* 8 == Backspace */
	     (the_char[0] == 127))     	/* 127 == Del or Rubout */
    {
	erase_type_cursor(disp_num);	/* erase cursor */
        backspace_letter_from_history(disp_num);
	draw_type_cursor(disp_num);	/* redraw cursor */
    }
    else 				/* it is a valid letter */
    {
	erase_type_cursor(disp_num);	/* erase cursor */
        push_letter_to_history(disp_num, the_char[0]);

        for (i=0; i < num_disps; i++)
        {
            if (disp_info[i].in_session_bool)
            {
                XDrawString(disp_info[i].disp, disp_info[i].draw_win_id,
                            disp_info[disp_num].win_gc[i],
                            disp_info[disp_num].type_info.cur_pos.x,
                            disp_info[disp_num].type_info.cur_pos.y, 
			    the_char, 1);
                XDrawString(disp_info[i].disp, disp_info[i].backing_store,
                            disp_info[disp_num].win_gc[i],
                            disp_info[disp_num].type_info.cur_pos.x,
                            disp_info[disp_num].type_info.cur_pos.y, 
			    the_char, 1);
                XFlush(disp_info[i].disp);
            }
        }

        disp_info[disp_num].type_info.cur_pos.x +=
                   XTextWidth(disp_info[disp_num].type_info.the_font_struct,
                   the_char, 1);
	draw_type_cursor(disp_num);	/* redraw cursor */
    }

    global_values.screen_dirty_bit = TRUE;
}


/*
 * push_letter_to_history - this function takes a pointer to the top of
 *              the history stack and adds the character passed in to that
 *              stack.  This is to save the characters for future backspacing.
 */
push_letter_to_history(disp_num, the_char)
int disp_num;
char the_char;
{
    struct char_node *temp;

    temp = (struct char_node *) malloc(sizeof(struct char_node));
    temp->the_char = the_char;
    temp->next = disp_info[disp_num].type_info.type_history;

    if ((the_char == 13) ||  /* 13 == carriage return */
        (the_char == 10))    /* 10 == Line Feed */
    {
        temp->x = disp_info[disp_num].type_info.cur_pos.x;
        temp->y = disp_info[disp_num].type_info.cur_pos.y;
    }

    disp_info[disp_num].type_info.type_history = temp;
}
	

/*
 * backspace_letter_from_history - this function returns the most recently
 *                 typed character from the history stack, and sets the top of
 *                 the stack pointer appropriately.
 */
backspace_letter_from_history(disp_num)
int disp_num;
{
    int dr, far, fdr, i;
    struct char_node *char_node;
    XCharStruct the_fnt_strct;

    if (disp_info[disp_num].type_info.type_history == NULL)
    {
        return(0);		/* we are at the beginning already */
    }
    else
    {
        char_node = disp_info[disp_num].type_info.type_history;

        if ((char_node->the_char == 13) ||    /* it was a carriage return */
            (char_node->the_char == 10))      /* it was a line feed */
        {
            disp_info[disp_num].type_info.cur_pos.x = char_node->x;
            disp_info[disp_num].type_info.cur_pos.y = char_node->y;
        }
        else                               /*it was a normal letter*/
        {
            disp_info[disp_num].type_info.cur_pos.x -= 
	      XTextWidth(disp_info[disp_num].type_info.the_font_struct,
			 &(char_node->the_char), 1);

            XTextExtents(disp_info[disp_num].type_info.the_font_struct,
                   &(char_node->the_char), 1, &dr, &far, &fdr, &the_fnt_strct);

	    if (char_node->the_char != ' ')		/* no reason to */
	    {
                for (i=0; i < num_disps; i++)
                {
		    if (disp_info[i].in_session_bool == TRUE)
		    {
                        XClearArea(disp_info[i].disp, disp_info[i].draw_win_id,
                          disp_info[disp_num].type_info.cur_pos.x,
                          disp_info[disp_num].type_info.cur_pos.y - 
                          the_fnt_strct.ascent,
                          the_fnt_strct.lbearing + the_fnt_strct.rbearing,
                          the_fnt_strct.ascent + the_fnt_strct.descent, False);
                        XFillRectangle(disp_info[i].disp, 
                          disp_info[i].backing_store,
                          disp_info[i].backing_store_gc,
                          disp_info[disp_num].type_info.cur_pos.x,
                          disp_info[disp_num].type_info.cur_pos.y - 
                          the_fnt_strct.ascent,
                          the_fnt_strct.lbearing + the_fnt_strct.rbearing,
                          the_fnt_strct.ascent + the_fnt_strct.descent);
		    }
                }
	    }
        }
        /* now pop the top char node from the history stack */
        disp_info[disp_num].type_info.type_history =
                           disp_info[disp_num].type_info.type_history->next;
        free(char_node);
    }
}


/*
 * handle_ButtonPress_event - this function handles button press events in the
 *                            drawing window.
 */
handle_ButtonPress_event(our_event, disp_num)
XEvent *our_event;
int disp_num;
{
    XButtonPressedEvent *the_event;
    int left, top, menu_num, i, root_x, root_y;
    int x, y, width;
    Window dummy;

    the_event = (XButtonPressedEvent *) our_event;
    if (the_event->button != Button1)
    {
        if (disp_info[disp_num].input_mode != TYPING)
	    return;	/* not interested in this event */
    }
    disp_info[disp_num].pointer_state = PRESSED;

    if (disp_info[disp_num].selected_region.region_is_selected == TRUE)
	place_selection_permanently(disp_num);
    if (disp_info[disp_num].fatbits_dialog.selected_region.region_is_selected)
	place_fatbits_selection_permanently(disp_num);

    switch(disp_info[disp_num].input_mode)
    {
	case POINTING:
            disp_info[disp_num].last_telepointer_point.x = the_event->x;
            disp_info[disp_num].last_telepointer_point.y = the_event->y;
            disp_info[disp_num].rubber_pointer.is_mapped_bool = TRUE;
            /*
             * save new cursor area for the input cursor, then draw the
	     * pointer.
             */
            left = the_event->x;
            top = the_event->y - disp_info[disp_num].rubber_pointer.height;
            for (i=0; i < num_disps; i++)
            {
                if (disp_info[i].in_session_bool)   /*only if window is alive*/
                {
		    /*
                    XCopyArea(disp_info[i].disp, disp_info[i].backing_store,
                      disp_info[i].draw_win_id, disp_info[disp_num].win_gc[i],
		      left, top, disp_info[disp_num].rubber_pointer.width,
                      disp_info[disp_num].rubber_pointer.height, left, top);
		    */
                    draw_rubber_pointer(disp_num, i);
                }
            }
            break;
	case SELECTING:
	    disp_info[disp_num].orig_rubberband_point.x = the_event->x;
	    disp_info[disp_num].orig_rubberband_point.y = the_event->y;
	    disp_info[disp_num].last_rubberband_point.x = the_event->x;
	    disp_info[disp_num].last_rubberband_point.y = the_event->y;
            draw_rectangle(disp_num, disp_info[disp_num].rubber_band_gc);
	    break;
	case SUPER_SELECTING:
	    global_values.screen_dirty_bit = TRUE;
	    /*
	    XtTranslateCoords(disp_info[disp_num].draw_area, 
		          the_event->x, the_event->y, &root_x, &root_y);
	    */
	    myXtTranslateCoords(disp_num, disp_info[disp_num].draw_area, 
		          the_event->x, the_event->y, &root_x, &root_y);
	    disp_info[disp_num].last_rubberband_point.x = root_x;
	    disp_info[disp_num].last_rubberband_point.y = root_y;
	    disp_info[disp_num].orig_rubberband_point.x = root_x;
	    disp_info[disp_num].orig_rubberband_point.y = root_y;
            draw_super_rectangle(disp_num, 
				 disp_info[disp_num].super_rubber_band_gc);
	    break;
	case PAINTING:
	    global_values.screen_dirty_bit = TRUE;
            disp_info[disp_num].last_scrawl_point.x = the_event->x;
            disp_info[disp_num].last_scrawl_point.y = the_event->y;
            if (disp_info[disp_num].paintbrush.paint_mode != NORMAL)
	    {
	        handle_pointer_move_while_painting(disp_num, the_event);
	    }
	    else
            {
		x = the_event->x - disp_info[disp_num].paintbrush.brush_width/2;
		y = the_event->y - disp_info[disp_num].paintbrush.brush_width/2;
		width = disp_info[disp_num].paintbrush.brush_width;

                for (i=0; i < num_disps; i++)
                {
                    if (disp_info[i].in_session_bool)
                    {
                        XFillArc(disp_info[i].disp, disp_info[i].draw_win_id, 
    	                         disp_info[disp_num].win_gc[i], 
		                 x, y, width, width, 0, 23040);
                        XFillArc(disp_info[i].disp,disp_info[i].backing_store,
    	                         disp_info[disp_num].win_gc[i], 
		                 x, y, width, width, 0, 23040);
	                XFlush(disp_info[i].disp);
                    }
                }
            }
            break;
	case PENCILING:
	    global_values.screen_dirty_bit = TRUE;
            disp_info[disp_num].last_scrawl_point.x = the_event->x;
            disp_info[disp_num].last_scrawl_point.y = the_event->y;
	    handle_pointer_move_while_penciling(disp_num, the_event);
            break;
	case TYPING:
            if (the_event->button == Button2)  /* short cut Paste X-Selection */
	    {
		paste_x_selection_region(disp_num);
		return;
	    }
	    global_values.screen_dirty_bit = TRUE;
	    erase_type_cursor(disp_num);    /* erase old cursor */
            disp_info[disp_num].type_info.first_typed_letter_bool = TRUE;
            disp_info[disp_num].type_info.orig_pos.x = the_event->x;
            disp_info[disp_num].type_info.orig_pos.y = the_event->y;
            disp_info[disp_num].type_info.cur_pos.x = the_event->x;
            disp_info[disp_num].type_info.cur_pos.y = the_event->y;
            disp_info[disp_num].type_info.type_history = NULL;
	    /*
	     * draw the cursor 
	     */
	    draw_type_cursor(disp_num);
            break;
	case ERASING:
            if ((disp_info[disp_num].eraser_win_width - 4 > 0) &&
                (disp_info[disp_num].eraser_win_height - 4 > 0))
            {
                XMoveResizeWindow(disp_info[disp_num].disp,
                         disp_info[disp_num].eraser_win_id,
                         the_event->x - disp_info[disp_num].eraser_win_width/2,
                         the_event->y - disp_info[disp_num].eraser_win_height/2,
                         disp_info[disp_num].eraser_win_width - 4,
                         disp_info[disp_num].eraser_win_height - 4);
            }
            else
            {
                XMoveResizeWindow(disp_info[disp_num].disp,
                         disp_info[disp_num].eraser_win_id,
                         the_event->x - disp_info[disp_num].eraser_win_width/2,
                         the_event->y - disp_info[disp_num].eraser_win_height/2,
                         1, 1);
            }
            XMapWindow(disp_info[disp_num].disp,
                       disp_info[disp_num].eraser_win_id);
	    XFlush(disp_info[disp_num].disp);

	    /*
	     * erase first spot
	     */
            for (i=0; i < num_disps; i++)
            {
                if (disp_info[i].in_session_bool)
                {
                    XClearArea(disp_info[i].disp, disp_info[i].draw_win_id,
                       the_event->x - disp_info[disp_num].eraser_win_width/2,
                       the_event->y - disp_info[disp_num].eraser_win_height/2,
                       disp_info[disp_num].eraser_win_width,
                       disp_info[disp_num].eraser_win_height, False);
                    XFillRectangle(disp_info[i].disp,disp_info[i].backing_store,
                       disp_info[i].backing_store_gc, 
                       the_event->x - disp_info[disp_num].eraser_win_width/2,
                       the_event->y - disp_info[disp_num].eraser_win_height/2,
                       disp_info[disp_num].eraser_win_width,
                       disp_info[disp_num].eraser_win_height);
	            XFlush(disp_info[i].disp);
                }
            }
            break;
	case SPRAYPAINTING:
	    disp_info[disp_num].spraypaint_timer_id = XtAppAddTimeOut(appCon, 
			        1, (XtTimerCallbackProc)spraypaint_spotCB, (int *) disp_num);
            break;
	case DRAWING_LINES:
	    global_values.screen_dirty_bit = TRUE;
	    disp_info[disp_num].orig_rubberband_point.x = the_event->x;
	    disp_info[disp_num].orig_rubberband_point.y = the_event->y;
	    disp_info[disp_num].last_rubberband_point.x = the_event->x;
	    disp_info[disp_num].last_rubberband_point.y = the_event->y;
            XDrawLine(disp_info[disp_num].disp, disp_info[disp_num].draw_win_id,
	            disp_info[disp_num].rubber_band_gc,
	            disp_info[disp_num].orig_rubberband_point.x,
	            disp_info[disp_num].orig_rubberband_point.y,
	            disp_info[disp_num].last_rubberband_point.x,
	            disp_info[disp_num].last_rubberband_point.y);
            break;
	case OUTLINING_SQUARES:
	case FILLING_SQUARES:
	    global_values.screen_dirty_bit = TRUE;
	    disp_info[disp_num].orig_rubberband_point.x = the_event->x;
	    disp_info[disp_num].orig_rubberband_point.y = the_event->y;
	    disp_info[disp_num].last_rubberband_point.x = the_event->x;
	    disp_info[disp_num].last_rubberband_point.y = the_event->y;
            draw_rectangle(disp_num, disp_info[disp_num].rubber_band_gc);
	    break;
	case OUTLINING_OVALS:
	case FILLING_OVALS:
	    global_values.screen_dirty_bit = TRUE;
	    disp_info[disp_num].orig_rubberband_point.x = the_event->x;
	    disp_info[disp_num].orig_rubberband_point.y = the_event->y;
	    disp_info[disp_num].last_rubberband_point.x = the_event->x;
	    disp_info[disp_num].last_rubberband_point.y = the_event->y;
            draw_oval(disp_num, disp_info[disp_num].rubber_band_gc);
	    break;
	case DRAGGING_SELECTION:
	    printf("ERROR: drag_selection in handle_ButtonPress_event.\n");
	    break;
	case GRABBING_COLOR:
	    global_values.screen_dirty_bit = TRUE;
	    myXtTranslateCoords(disp_num, disp_info[disp_num].draw_area, 
		          the_event->x, the_event->y, &root_x, &root_y);
	    grab_color(disp_num, root_x, root_y);
	    break;
        default:
	    printf("ERROR: unknown input_mode in handle_ButtonPress_event.\n");
	    break;
    }
}


/*
 * draw_oval - this function draws the current oval with the GC
 *             passed in.  The main idea of this function is to reverse
 *             the coordinates if necessary for the XDrawArc call.
 */
draw_oval(disp_num, the_gc)
int disp_num;
GC the_gc;
{
    int x, y, width, height;

    x = (disp_info[disp_num].orig_rubberband_point.x < 
	 disp_info[disp_num].last_rubberband_point.x) ?
	 disp_info[disp_num].orig_rubberband_point.x :
	 disp_info[disp_num].last_rubberband_point.x;

    y = (disp_info[disp_num].orig_rubberband_point.y < 
	 disp_info[disp_num].last_rubberband_point.y) ?
	 disp_info[disp_num].orig_rubberband_point.y :
	 disp_info[disp_num].last_rubberband_point.y;

    width = abs(disp_info[disp_num].last_rubberband_point.x -
		disp_info[disp_num].orig_rubberband_point.x);

    height = abs(disp_info[disp_num].last_rubberband_point.y -
		 disp_info[disp_num].orig_rubberband_point.y);

    XDrawArc(disp_info[disp_num].disp, disp_info[disp_num].draw_win_id,
	     the_gc, x, y, width, height, 0, 23040);
}


/*
 * draw_rectangle - this function draws the current rectangle with the GC
 *                  passed in.  The main idea of this function is to reverse
 *                  the coordinates if necessary for the XDrawRectangle call.
 */
draw_rectangle(disp_num, the_gc)
int disp_num;
GC the_gc;
{
    int x, y, width, height;

    x = (disp_info[disp_num].orig_rubberband_point.x < 
	 disp_info[disp_num].last_rubberband_point.x) ?
	 disp_info[disp_num].orig_rubberband_point.x :
	 disp_info[disp_num].last_rubberband_point.x;

    y = (disp_info[disp_num].orig_rubberband_point.y < 
	 disp_info[disp_num].last_rubberband_point.y) ?
	 disp_info[disp_num].orig_rubberband_point.y :
	 disp_info[disp_num].last_rubberband_point.y;

    width = abs(disp_info[disp_num].last_rubberband_point.x -
		disp_info[disp_num].orig_rubberband_point.x);

    height = abs(disp_info[disp_num].last_rubberband_point.y -
		 disp_info[disp_num].orig_rubberband_point.y);

    XDrawRectangle(disp_info[disp_num].disp, 
		    disp_info[disp_num].draw_win_id,
	            the_gc, x, y, width, height);
}


/*
 * draw_super_rectangle - this function draws the current rectangle with the GC
 *                  passed in ON THE ROOT WINDOW.  The main idea of this 
 *                  function is to reverse the coordinates if necessary for 
 *                  the XDrawRectangle call.
 */
draw_super_rectangle(disp_num, the_gc)
int disp_num;
GC the_gc;
{
    int x, y, width, height;

    set_x_y_width_height(disp_num, &x, &y, &width, &height);

    XDrawRectangle(disp_info[disp_num].disp, 
		    XDefaultRootWindow(disp_info[disp_num].disp),
	            the_gc, x, y, width, height);
}


/*
 * draw_type_cursor - this function draws the type cursor on the indicated
 *                    display.  Everything needs to be set up correctly.
 */
draw_type_cursor(disp_num)
int disp_num;
{
    XDrawLine(disp_info[disp_num].disp, disp_info[disp_num].draw_win_id,
	      disp_info[disp_num].type_info.cursor_gc,
	      disp_info[disp_num].type_info.cur_pos.x,
	      disp_info[disp_num].type_info.cur_pos.y,
	      disp_info[disp_num].type_info.cur_pos.x +
	      disp_info[disp_num].type_info.prompt_width,
	      disp_info[disp_num].type_info.cur_pos.y);
}


/*
 * erase_type_cursor - this function erases the type cursor on the indicated
 *                    display.  Everything needs to be set up correctly.
 */
erase_type_cursor(disp_num)
int disp_num;
{
    XCopyArea(disp_info[disp_num].disp, disp_info[disp_num].backing_store,
	      disp_info[disp_num].draw_win_id, 
	      disp_info[disp_num].win_gc[disp_num],
	      disp_info[disp_num].type_info.cur_pos.x,
	      disp_info[disp_num].type_info.cur_pos.y - 2,
	      disp_info[disp_num].type_info.prompt_width, 5,
	      disp_info[disp_num].type_info.cur_pos.x,
	      disp_info[disp_num].type_info.cur_pos.y - 2);
}


/*
 * draw_rubber_pointer - this function draws the rubber pointer of the input
 *                 disp on the target disp at the indicated location.  The
 *                 indicated location is the upper left corner of the cursor
 *                 area, NOT the hotspot.  This simulated cursor consists of
 *                 the machine name that is drawing, and a small arrow pointing
 *                 to the lower left corner.
 */
draw_rubber_pointer(input_disp, target_disp)
int input_disp, target_disp;
{
    int x, y;
    char *disp_name;

    disp_name = disp_info[input_disp].rubber_pointer.pointer_str;

    x = disp_info[input_disp].last_telepointer_point.x;
    y = disp_info[input_disp].last_telepointer_point.y -
          disp_info[input_disp].rubber_pointer.height;

    /* Clear area under disp name */
    XClearArea(disp_info[target_disp].disp,
	       disp_info[target_disp].draw_win_id, x + 15 - 1 - 1, y,
	       disp_info[input_disp].rubber_pointer.width - 15 + 1 + 1,
	       disp_info[input_disp].rubber_pointer.height - 15 + 3, False);
    XDrawString(disp_info[target_disp].disp, 
		disp_info[target_disp].draw_win_id,
                disp_info[target_disp].menu_gc,
                x + 15 - 1, y + disp_info[input_disp].rubber_pointer.height-15,
                disp_name, strlen(disp_name));
    XDrawLine(disp_info[target_disp].disp, disp_info[target_disp].draw_win_id,
              disp_info[target_disp].menu_gc,
              x+2, y - 2 + disp_info[input_disp].rubber_pointer.height,
              x+12, y - 12 + disp_info[input_disp].rubber_pointer.height);
    XDrawLine(disp_info[target_disp].disp, disp_info[target_disp].draw_win_id,
              disp_info[target_disp].menu_gc,
              x+3, y - 2 + disp_info[input_disp].rubber_pointer.height,
              x+3, y - 9 + disp_info[input_disp].rubber_pointer.height);
    XDrawLine(disp_info[target_disp].disp, disp_info[target_disp].draw_win_id,
              disp_info[target_disp].menu_gc,
              x+3, y - 2 + disp_info[input_disp].rubber_pointer.height,
              x+9, y - 2 + disp_info[input_disp].rubber_pointer.height);
}


/*
 * spraypaint_spot - this function is a time-out callback that is added during
 *                   times that the spraypaint gun is being sprayed.  It
 *                   sprays one pattern at the current pointer location.
 */
XtTimerCallbackProc spraypaint_spotCB(client_data, id)
int *client_data;
XtIntervalId *id;
{
    int disp_num;
    Window rr, cr;
    unsigned int mskr;
    int i, rxr, ryr, win_x, win_y;
    int rx, ry, R2, RADIUS, NUM_DOTS;
    static XPoint spray_points[502];

    disp_num = (int) client_data;

    /* 
     * add new spray timeout for next spray
     */ 
    disp_info[disp_num].spraypaint_timer_id = XtAppAddTimeOut(appCon, 
			        20, (XtTimerCallbackProc)spraypaint_spotCB, (int *) disp_num);
    XQueryPointer(disp_info[disp_num].disp, disp_info[disp_num].draw_win_id,
                      &rr, &cr, &rxr, &ryr, &win_x, &win_y, &mskr);

    RADIUS = 20;            /* used to be RADIUS = penwidth/2 */
    NUM_DOTS = 20;          /* old command line argument */

    R2 = RADIUS * RADIUS;   /*radius squared*/

    for (i=0; i < NUM_DOTS; i++)
    {
        /*
         * -RADIUS < rx < RADIUS
         */
        rx = RADIUS * (2 * my_rand() - 1);
        ry = RADIUS * (2 * my_rand() - 1);

        /*
         * if actual radius is outside of given radius,
         * throw it out
         */
        if ((rx*rx + ry*ry) > R2)
        {
            i--;
            continue;
        }
        spray_points[i].x = win_x + rx;
        spray_points[i].y = win_y + ry;
    }

    for (i=0; i < num_disps; i++)
    {
        if (disp_info[i].in_session_bool) /*only if this window is alive*/
        {
            XDrawPoints(disp_info[i].disp, disp_info[i].draw_win_id,
                        disp_info[disp_num].win_gc[i],
                        spray_points, NUM_DOTS, CoordModeOrigin);
            XDrawPoints(disp_info[i].disp, disp_info[i].backing_store,
                        disp_info[disp_num].win_gc[i],
                        spray_points, NUM_DOTS, CoordModeOrigin);
            XFlush(disp_info[i].disp);
        }
    }
}


/*
 * handle_ButtonRelease_event - this function handles button release events.
 */
handle_ButtonRelease_event(our_event, disp_num)
XEvent *our_event;
int disp_num;
{
    XButtonPressedEvent *the_event;
    int menu_num, i;
    int x, y, root_x, root_y, width, height, oldleft, oldtop;
    Window dummy;

    the_event = (XButtonPressedEvent *) our_event;
    if (the_event->button != Button1)
    {
        if (disp_info[disp_num].input_mode != TYPING)
	    return;	/* not interested in this event */
    }
    disp_info[disp_num].pointer_state = NOT_PRESSED;

    switch(disp_info[disp_num].input_mode)
    {
	case POINTING:
            disp_info[disp_num].rubber_pointer.is_mapped_bool = FALSE;
            oldleft = disp_info[disp_num].last_telepointer_point.x;
            oldtop = disp_info[disp_num].last_telepointer_point.y -
                     disp_info[disp_num].rubber_pointer.height;
            for (i=0; i < num_disps; i++)
	    {
                if (disp_info[i].in_session_bool)
                {
                    XCopyArea(disp_info[i].disp, disp_info[i].backing_store,
                      disp_info[i].draw_win_id, disp_info[disp_num].win_gc[i],
		      oldleft, oldtop, disp_info[disp_num].rubber_pointer.width,
                      disp_info[disp_num].rubber_pointer.height, 
		      oldleft, oldtop);
                }
            }
	    break;
	case SELECTING:
            draw_rectangle(disp_num, disp_info[disp_num].rubber_band_gc);
	    disp_info[disp_num].last_rubberband_point.x = the_event->x;
	    disp_info[disp_num].last_rubberband_point.y = the_event->y;
	    define_selection(disp_num);
            break;
	case SUPER_SELECTING:
	    if (disp_info[disp_num].select_dialog.pointer_down == TRUE)
	    {
	        disp_info[disp_num].select_dialog.pointer_down = FALSE;
		return;		/* just letting up the pointer */
	    }
            draw_super_rectangle(disp_num, 
				 disp_info[disp_num].super_rubber_band_gc);
	    /*
	    XtTranslateCoords(disp_info[disp_num].draw_area, 
		              the_event->x, the_event->y, &root_x, &root_y);
	    */
	    myXtTranslateCoords(disp_num, disp_info[disp_num].draw_area, 
		              the_event->x, the_event->y, &root_x, &root_y);
	    disp_info[disp_num].last_rubberband_point.x = root_x;
	    disp_info[disp_num].last_rubberband_point.y = root_y;
	    XUngrabPointer(disp_info[disp_num].disp, CurrentTime);
	    disp_info[disp_num].input_mode = SELECTING;
	    define_super_selection(disp_num);
	    XUngrabServer(disp_info[disp_num].disp);
	    break;
	case PAINTING:
            break;
	case PENCILING:
            break;
	case TYPING:
            break;
	case ERASING:
            XUnmapWindow(disp_info[disp_num].disp,
                         disp_info[disp_num].eraser_win_id);
            break;
	case SPRAYPAINTING:
	    XtRemoveTimeOut(disp_info[disp_num].spraypaint_timer_id);
            break;
	case DRAWING_LINES:
            XDrawLine(disp_info[disp_num].disp, disp_info[disp_num].draw_win_id,
	            disp_info[disp_num].rubber_band_gc,
	            disp_info[disp_num].orig_rubberband_point.x,
	            disp_info[disp_num].orig_rubberband_point.y,
	            disp_info[disp_num].last_rubberband_point.x,
	            disp_info[disp_num].last_rubberband_point.y);
            for (i=0; i < num_disps; i++)
            {
                if (disp_info[i].in_session_bool)
                {
                    XDrawLine(disp_info[i].disp, disp_info[i].draw_win_id, 
    	               disp_info[disp_num].win_gc[i], 
    	               disp_info[disp_num].orig_rubberband_point.x, 
	               disp_info[disp_num].orig_rubberband_point.y,
    	               the_event->x, the_event->y);
                    XDrawLine(disp_info[i].disp, disp_info[i].backing_store, 
    	               disp_info[disp_num].win_gc[i], 
    	               disp_info[disp_num].orig_rubberband_point.x, 
	               disp_info[disp_num].orig_rubberband_point.y,
    	               the_event->x, the_event->y);
	            XFlush(disp_info[i].disp);
                }
            }
            break;
	case OUTLINING_SQUARES:
            draw_rectangle(disp_num, disp_info[disp_num].rubber_band_gc);
	    disp_info[disp_num].last_rubberband_point.x = the_event->x;
	    disp_info[disp_num].last_rubberband_point.y = the_event->y;

            x = (disp_info[disp_num].orig_rubberband_point.x < 
	         disp_info[disp_num].last_rubberband_point.x) ?
	         disp_info[disp_num].orig_rubberband_point.x :
	         disp_info[disp_num].last_rubberband_point.x;
        
            y = (disp_info[disp_num].orig_rubberband_point.y < 
	         disp_info[disp_num].last_rubberband_point.y) ?
	         disp_info[disp_num].orig_rubberband_point.y :
	         disp_info[disp_num].last_rubberband_point.y;
        
            width = abs(disp_info[disp_num].last_rubberband_point.x -
		        disp_info[disp_num].orig_rubberband_point.x);
        
            height = abs(disp_info[disp_num].last_rubberband_point.y -
		         disp_info[disp_num].orig_rubberband_point.y);

            for (i=0; i < num_disps; i++)
            {
                if (disp_info[i].in_session_bool)
                {
                    XDrawRectangle(disp_info[i].disp, disp_info[i].draw_win_id, 
    	                           disp_info[disp_num].win_gc[i], 
		                   x, y, width, height);
                    XDrawRectangle(disp_info[i].disp,disp_info[i].backing_store,
    	                           disp_info[disp_num].win_gc[i], 
		                   x, y, width, height);
	            XFlush(disp_info[i].disp);
                }
            }
	    break;
	case FILLING_SQUARES:
            draw_rectangle(disp_num, disp_info[disp_num].rubber_band_gc);
	    disp_info[disp_num].last_rubberband_point.x = the_event->x;
	    disp_info[disp_num].last_rubberband_point.y = the_event->y;

            x = (disp_info[disp_num].orig_rubberband_point.x < 
	         disp_info[disp_num].last_rubberband_point.x) ?
	         disp_info[disp_num].orig_rubberband_point.x :
	         disp_info[disp_num].last_rubberband_point.x;
        
            y = (disp_info[disp_num].orig_rubberband_point.y < 
	         disp_info[disp_num].last_rubberband_point.y) ?
	         disp_info[disp_num].orig_rubberband_point.y :
	         disp_info[disp_num].last_rubberband_point.y;
        
            width = abs(disp_info[disp_num].last_rubberband_point.x -
		        disp_info[disp_num].orig_rubberband_point.x);
        
            height = abs(disp_info[disp_num].last_rubberband_point.y -
		         disp_info[disp_num].orig_rubberband_point.y);

            for (i=0; i < num_disps; i++)
            {
                if (disp_info[i].in_session_bool)
                {
                    XFillRectangle(disp_info[i].disp, disp_info[i].draw_win_id, 
    	                           disp_info[disp_num].win_gc[i], 
		                   x, y, width, height);
                    XFillRectangle(disp_info[i].disp,disp_info[i].backing_store,
    	                           disp_info[disp_num].win_gc[i], 
		                   x, y, width, height);
	            XFlush(disp_info[i].disp);
                }
            }
	    break;
	case OUTLINING_OVALS:
            draw_oval(disp_num, disp_info[disp_num].rubber_band_gc);
	    disp_info[disp_num].last_rubberband_point.x = the_event->x;
	    disp_info[disp_num].last_rubberband_point.y = the_event->y;

            x = (disp_info[disp_num].orig_rubberband_point.x < 
	         disp_info[disp_num].last_rubberband_point.x) ?
	         disp_info[disp_num].orig_rubberband_point.x :
	         disp_info[disp_num].last_rubberband_point.x;
        
            y = (disp_info[disp_num].orig_rubberband_point.y < 
	         disp_info[disp_num].last_rubberband_point.y) ?
	         disp_info[disp_num].orig_rubberband_point.y :
	         disp_info[disp_num].last_rubberband_point.y;
        
            width = abs(disp_info[disp_num].last_rubberband_point.x -
		        disp_info[disp_num].orig_rubberband_point.x);
        
            height = abs(disp_info[disp_num].last_rubberband_point.y -
		         disp_info[disp_num].orig_rubberband_point.y);

            for (i=0; i < num_disps; i++)
            {
                if (disp_info[i].in_session_bool)
                {
                    XDrawArc(disp_info[i].disp, disp_info[i].draw_win_id, 
    	                     disp_info[disp_num].win_gc[i], 
		             x, y, width, height, 0, 23040);
                    XDrawArc(disp_info[i].disp,disp_info[i].backing_store,
    	                     disp_info[disp_num].win_gc[i], 
		             x, y, width, height, 0, 23040);
	            XFlush(disp_info[i].disp);
                }
            }
	    break;
	case FILLING_OVALS:
            draw_oval(disp_num, disp_info[disp_num].rubber_band_gc);
	    disp_info[disp_num].last_rubberband_point.x = the_event->x;
	    disp_info[disp_num].last_rubberband_point.y = the_event->y;

            x = (disp_info[disp_num].orig_rubberband_point.x < 
	         disp_info[disp_num].last_rubberband_point.x) ?
	         disp_info[disp_num].orig_rubberband_point.x :
	         disp_info[disp_num].last_rubberband_point.x;
        
            y = (disp_info[disp_num].orig_rubberband_point.y < 
	         disp_info[disp_num].last_rubberband_point.y) ?
	         disp_info[disp_num].orig_rubberband_point.y :
	         disp_info[disp_num].last_rubberband_point.y;
        
            width = abs(disp_info[disp_num].last_rubberband_point.x -
		        disp_info[disp_num].orig_rubberband_point.x);
        
            height = abs(disp_info[disp_num].last_rubberband_point.y -
		         disp_info[disp_num].orig_rubberband_point.y);

            for (i=0; i < num_disps; i++)
            {
                if (disp_info[i].in_session_bool)
                {
                    XFillArc(disp_info[i].disp, disp_info[i].draw_win_id, 
    	                     disp_info[disp_num].win_gc[i], 
		             x, y, width, height, 0, 23040);
                    XFillArc(disp_info[i].disp,disp_info[i].backing_store,
    	                     disp_info[disp_num].win_gc[i], 
		             x, y, width, height, 0, 23040);
	            XFlush(disp_info[i].disp);
                }
            }
	    break;
	case DRAGGING_SELECTION:
	    XUngrabPointer(disp_info[disp_num].disp, CurrentTime);
	    XSync(disp_info[disp_num].disp, False);
	    XSetInputFocus(disp_info[disp_num].disp,
			   disp_info[disp_num].draw_win_id,
			   RevertToPointerRoot, CurrentTime);
	    disp_info[disp_num].input_mode = SELECTING;
	    break;
	case GRABBING_COLOR:
	    XUngrabPointer(disp_info[disp_num].disp, CurrentTime);
	    disp_info[disp_num].input_mode = SELECTING;
	    boink_tool(disp_num, PAINTBRUSH_BUTTON);
	    break;
        default:
	    printf("ERROR: unknown input_mode, handle_ButtonRelease_event.\n");
	    break;
    }
}


/*
 * sync_drawing_areas_with - this function is called when the user at "disp_num"
 *                           has boinked a scrollbar, so that the drawing
 *                           areas are now "out of sync", and this function
 *                           is responsible for making all displays have
 *                           the exact same upper left corner of visible area.
 */
sync_drawing_areas_with(disp_num)
int disp_num;
{
    int n, i, max_hsbar_val, slider_size, max_vsbar_val;
    int new_hsbar_val, new_vsbar_val;
    Arg args[MAX_ARGS];

    n = 0;
    XtSetArg(args[n], XmNvalue, &new_hsbar_val); n++;
    XtGetValues(disp_info[disp_num].hsbar, args, n);
    n = 0;
    XtSetArg(args[n], XmNvalue, &new_vsbar_val); n++;
    XtGetValues(disp_info[disp_num].vsbar, args, n);

    /*
     * find least "maximum" value, so that we do not increment a scrollbar
     * past the least display's max value
     */
    for (i=0; i < num_disps; i++)
    {
        if (disp_info[i].in_session_bool)
        {
            n = 0;
            XtSetArg(args[n], XmNmaximum, &max_hsbar_val); n++;
            XtSetArg(args[n], XmNsliderSize, &slider_size); n++;
            XtGetValues(disp_info[i].hsbar, args, n);

	    if (new_hsbar_val > (max_hsbar_val - slider_size))
		new_hsbar_val = max_hsbar_val - slider_size;

            n = 0;
            XtSetArg(args[n], XmNmaximum, &max_vsbar_val); n++;
            XtSetArg(args[n], XmNsliderSize, &slider_size); n++;
            XtGetValues(disp_info[i].vsbar, args, n);

	    if (new_vsbar_val > (max_vsbar_val - slider_size))
		new_vsbar_val = max_vsbar_val - slider_size;
	}
    }

    for (i=0; i < num_disps; i++)
    {
        if (disp_info[i].in_session_bool)
        {
            XtMoveWidget(XtParent(disp_info[i].draw_area), 
			 -new_hsbar_val, -new_vsbar_val);
	    n = 0;
            XtSetArg(args[n], XmNvalue, new_vsbar_val); n++;
            XtSetValues(disp_info[i].vsbar, args, n);
	    n = 0;
            XtSetArg(args[n], XmNvalue, new_hsbar_val); n++;
            XtSetValues(disp_info[i].hsbar, args, n);
        }
    }
}


/*
 * vsbar_movement - this function handles all vertical scrollbar motion
 */
vsbar_movementCB(w, client_data, call_data)
Widget w;
caddr_t client_data;		/* this will be DECR, or INCR, etc */
caddr_t call_data;
{
    int the_action, disp_num;

    disp_num = ((int) client_data) >> 8;
    the_action = ((int) client_data) & 0x00FF;   /* not used right now */

    switch(the_action)
    {
        case INCR:			
        case DECR:
        case GENERIC_MOVE:
	    if (global_values.sync_scrollbars_bool == TRUE)
	        sync_drawing_areas_with(disp_num);
	    break;
	default:
	    printf("Unknown menu selected in vsbar_movement(), %d.\n", 
		   (int) client_data);
	    break;
    }
}


/*
 * hsbar_movement - this function handles all horizontal scrollbar motion
 */
hsbar_movementCB(w, client_data, call_data)
Widget w;
caddr_t client_data;		/* this will be DECR, or INCR, etc */
caddr_t call_data;
{
    int the_action, disp_num;

    disp_num = ((int) client_data) >> 8;
    the_action = ((int) client_data) & 0x00FF;

    switch(the_action)
    {
        case INCR:
        case DECR:
        case GENERIC_MOVE:
	    if (global_values.sync_scrollbars_bool == TRUE)
	        sync_drawing_areas_with(disp_num);
	    break;
	default:
	    printf("Unknown menu selected in hsbar_movement(), %d.\n", 
		   (int) client_data);
	    break;
    }
}

