/*
 * This File: wsfatbits.c
 *
 * This file contains the fatbits editor functions
 */

#include "wscrawl.h"
#include "selectb.h"
#include "pencil.h"
#include "eraser.h"

extern struct disp_info_struct disp_info[MAX_NUM_DISPS];
extern int num_disps;
static XmStringCharSet charset = (XmStringCharSet) XmSTRING_DEFAULT_CHARSET;
extern XImage *read_image_from_disk();


/*
 * handle_fatbits_dialog - this function is the callback for the fatbits
 *                              dialog.
 */
handle_fatbits_dialog(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int i, disp_num, which_button;
    int n, r_val, g_val, b_val;
    Arg args[MAX_ARGS];
    XColor the_color;
    XGCValues gcvalues;

    disp_num = ((int) client_data) >> 8;
    which_button = ((int) client_data) & 0x00FF;

    if ((disp_info[disp_num].fatbits_dialog.selected_region.region_is_selected)
        && (which_button != FATBITS_CLOSE))
    {
        place_fatbits_selection_permanently(disp_num);
    }

    if ((disp_info[disp_num].selected_region.region_is_selected) &&
        (which_button != FATBITS_CLOSE))
    {
	place_selection_permanently(disp_num);
    }

    switch(which_button)
    {
	case FATBITS_SELECTION_BUT:
            disp_info[disp_num].fatbits_dialog.input_mode = SELECTING;
            XDefineCursor(disp_info[disp_num].disp,
		  XtWindow(disp_info[disp_num].fatbits_dialog.draw_area),
	          XCreateFontCursor(disp_info[disp_num].disp, XC_crosshair));
	    break;
	case FATBITS_PENCIL_BUT:
            disp_info[disp_num].fatbits_dialog.input_mode = PENCILING;
            XDefineCursor(disp_info[disp_num].disp,
		  XtWindow(disp_info[disp_num].fatbits_dialog.draw_area),
	          XCreateFontCursor(disp_info[disp_num].disp, XC_pencil));
	    break;
	case FATBITS_ERASER_BUT:
            disp_info[disp_num].fatbits_dialog.input_mode = ERASING;
            XDefineCursor(disp_info[disp_num].disp,
		XtWindow(disp_info[disp_num].fatbits_dialog.draw_area),
                XCreateFontCursor(disp_info[disp_num].disp, XC_draped_box));
	    break;
	case FATBITS_CLOSE:
	    XtUnmapWidget(
		       disp_info[disp_num].fatbits_dialog.selected_region.win);
           disp_info[disp_num].fatbits_dialog.selected_region.region_is_selected
	    = FALSE;
	    XtUnmanageChild(disp_info[disp_num].fatbits_dialog.the_dialog);
	    jiggle_dialog(disp_info[disp_num].fatbits_dialog.the_dialog);
	    refresh_fatbits_area(disp_num);
	    break;
	default:
	    printf("ERROR: unknown event in handle_fatbits_dialog.\n");
	    break;
    }
}


/*
 * fatbits_draw_area_inputCB - this function handles all draw area input events
 *                     such as ButonPress or KeyPress in the fatbits window.
 */
fatbits_draw_area_inputCB(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int disp_num;
    XmDrawingAreaCallbackStruct *cb_struct;
    XEvent *the_event;
    XMotionEvent the_event2;
    XButtonPressedEvent *but_evt;
    int i, x, y, width, height;
    Window dummy;
    Widget dummy_w = (Widget)NULL;
    Boolean dummy_cbool;

    cb_struct = (XmDrawingAreaCallbackStruct *) call_data;
    the_event = (XEvent *) cb_struct->event;
    disp_num = (int) client_data;

    if ((disp_info[disp_num].fatbits_dialog.selected_region.region_is_selected)
	&& (the_event->type == ButtonPress))
    {
        place_fatbits_selection_permanently(disp_num);
    }

    if ((disp_info[disp_num].selected_region.region_is_selected) && 
	(the_event->type == ButtonPress))
    {
	place_selection_permanently(disp_num);
    }

    switch(the_event->type)
    {
        case ButtonPress:
            but_evt = (XButtonPressedEvent *) the_event;
	    disp_info[disp_num].fatbits_dialog.pointer_state = PRESSED;
	    switch(disp_info[disp_num].fatbits_dialog.input_mode)
	    {
                case SELECTING:
		    fatbits_start_selecting(disp_num, but_evt->x, but_evt->y);
		    break;
		case PENCILING:
		   if (XGetPixel(disp_info[disp_num].fatbits_dialog.local_image,
	                 but_evt->x / FATBITS_MAG, but_evt->y / FATBITS_MAG) != 
		       disp_info[disp_num].current_pixel)
		    {
			disp_info[disp_num].fatbits_dialog.pencil_bool = TRUE;
		    }
		    else
		    {
			disp_info[disp_num].fatbits_dialog.pencil_bool = FALSE;
		    }
	            handle_fatbits_pencil_event(disp_num, but_evt->x, 
						but_evt->y);
		    break;
                case ERASING:
                    XMoveResizeWindow(disp_info[disp_num].disp,
                       disp_info[disp_num].fatbits_dialog.eraser_win_id,
                       but_evt->x - 
                       disp_info[disp_num].fatbits_dialog.eraser_win_width/2,
                       but_evt->y -
                       disp_info[disp_num].fatbits_dialog.eraser_win_width/2,
                       disp_info[disp_num].fatbits_dialog.eraser_win_width - 4,
                       disp_info[disp_num].fatbits_dialog.eraser_win_width - 4);
                    XMapWindow(disp_info[disp_num].disp,
                         disp_info[disp_num].fatbits_dialog.eraser_win_id);
                    XFlush(disp_info[disp_num].disp);
	            handle_fatbits_eraser_event(disp_num, but_evt->x, 
						but_evt->y);
		    break;
	        default:
	            printf("ERROR: unknown mode fatbits_draw_area_inputCB.\n");
	            break;
            }
            break;
        case ButtonRelease:
            but_evt = (XButtonPressedEvent *) the_event;
	    disp_info[disp_num].fatbits_dialog.pointer_state = NOT_PRESSED;
	    switch(disp_info[disp_num].fatbits_dialog.input_mode)
	    {
                case SELECTING:
                    draw_fatbits_rectangle(disp_num);
		    draw_rectangle(disp_num,disp_info[disp_num].rubber_band_gc);

		    define_selection(disp_num);
		    define_fatbits_selection(disp_num);
		    /*
		     * now clear regular region
		     */
		    set_x_y_width_height(disp_num, &x, &y, &width, &height);
		    if (width == 0 || height == 0)
			return;    /* clicked to set region */

		    disp_info[disp_num].selected_region.region_has_moved = TRUE;
		    x = disp_info[disp_num].selected_region.cur_x;
	            y = disp_info[disp_num].selected_region.cur_y;
    		    width = disp_info[disp_num].selected_region.width;
    		    height = disp_info[disp_num].selected_region.height;

		    for (i=0; i < num_disps; i++)
		    {
		        if (disp_info[i].in_session_bool)
		        {
                            XClearArea(disp_info[i].disp,
                                     disp_info[i].draw_win_id,
                                    x, y, width, height, False);
                            XFillRectangle(disp_info[i].disp,
                                    disp_info[i].backing_store,
                                    disp_info[i].backing_store_gc, x, y,
                                    width, height);
			}
		    }
		    break;
		case PENCILING:
		    break;
                case ERASING:
                    XUnmapWindow(disp_info[disp_num].disp,
		             disp_info[disp_num].fatbits_dialog.eraser_win_id);
                    XFlush(disp_info[disp_num].disp);
		    break;
                case DRAGGING_SELECTION:
                    XUngrabPointer(disp_info[disp_num].disp, CurrentTime);
                    XSync(disp_info[disp_num].disp, False);
                    XSetInputFocus(disp_info[disp_num].disp,
                         XtWindow(disp_info[disp_num].fatbits_dialog.draw_area),
                         RevertToPointerRoot, CurrentTime);
                    disp_info[disp_num].fatbits_dialog.input_mode = SELECTING;
		    snap_fatbits_selection_to_grid(disp_num);
	            /*
	             * now fake the event to make the regular selection box move
	             */
	            the_event2.type = MotionNotify;
	            the_event2.x = 
	              disp_info[disp_num].fatbits_dialog.selected_region.cur_x +
		      disp_info[disp_num].selected_region.drag_offset.x +
		      disp_info[disp_num].fatbits_dialog.x;
	            the_event2.y = 
	              disp_info[disp_num].fatbits_dialog.selected_region.cur_y +
		      disp_info[disp_num].selected_region.drag_offset.y +
		      disp_info[disp_num].fatbits_dialog.y;
	            disp_info[disp_num].pointer_state = PRESSED;
	            selection_pointer_moveCB(dummy_w, disp_num << 8 | disp_num,
					 (XEvent *) &the_event2, &dummy_cbool);
		    disp_info[disp_num].pointer_state = NOT_PRESSED; 
		    break; 
		default: 
		    printf("ERROR: unknown mode fatbits_draw_area_inputCB.\n");
	            break;
            }
	    break;
        case KeyPress:
            handle_fatbits_KeyPress_event(disp_num, the_event);
	    break;
        default:
	    /* I think KeyRelease events occur here sometimes.
            printf("Unknown event type in fatbits_draw_area_inputCB().\n");
	    */
            break;
    }
}


/*
 * handle_fatbits_KeyPress_event - this function processes the keypress event
 */
handle_fatbits_KeyPress_event(disp_num, our_event)
int disp_num;
XEvent *our_event;
{
    XKeyPressedEvent *the_key_evt;
    char the_char[10];
    KeySym ksr;

    the_key_evt = (XKeyPressedEvent *) our_event;
    the_char[0] = 0;
    XLookupString((XKeyEvent *) (the_key_evt), the_char, 10, &ksr, NULL);

    switch((int) the_char[0])
    {
	case 0:		/* shift or some-such */
	    break;
	case 8:		/* 8 == Backspace */
	case 127:	/* 127 == Del or Rubout */
            if (
           disp_info[disp_num].fatbits_dialog.selected_region.region_is_selected
           == TRUE)
            {
	        fatbits_clear_region(disp_num);
	        clear_region(disp_num);
            }
	    break;
        default:
            break;
    }
}


/*
 * fatbits_start_selecting - this function is called on the mouse button down
 *                           while selecting a region in fatbits.
 */
fatbits_start_selecting(disp_num, x, y)
{
  int fox, foy, flx, fly;

  /*
   * first take care of fatbits
   */
  fox = disp_info[disp_num].fatbits_dialog.selected_region.fat_orig_rub_point.x=
	    x / FATBITS_MAG;
  foy = disp_info[disp_num].fatbits_dialog.selected_region.fat_orig_rub_point.y=
	    y / FATBITS_MAG;
  flx = disp_info[disp_num].fatbits_dialog.selected_region.fat_last_rub_point.x=
	    fox;
  fly = disp_info[disp_num].fatbits_dialog.selected_region.fat_last_rub_point.y=
	    foy;

  disp_info[disp_num].fatbits_dialog.selected_region.orig_rubberband_point.x =
	    fox * FATBITS_MAG + (FATBITS_MAG / 2);
  disp_info[disp_num].fatbits_dialog.selected_region.orig_rubberband_point.y =
	    foy * FATBITS_MAG + (FATBITS_MAG / 2);
  disp_info[disp_num].fatbits_dialog.selected_region.last_rubberband_point.x =
	    fox * FATBITS_MAG + (FATBITS_MAG / 2);
  disp_info[disp_num].fatbits_dialog.selected_region.last_rubberband_point.y =
	    foy * FATBITS_MAG + (FATBITS_MAG / 2);

  draw_fatbits_rectangle(disp_num);

    /*
     * now do the regular window
     */
    disp_info[disp_num].orig_rubberband_point.x = fox +
                               disp_info[disp_num].fatbits_dialog.x;
    disp_info[disp_num].orig_rubberband_point.y = foy +
                               disp_info[disp_num].fatbits_dialog.y;
    disp_info[disp_num].last_rubberband_point.x = fox +
			       disp_info[disp_num].fatbits_dialog.x;
    disp_info[disp_num].last_rubberband_point.y = foy +
			       disp_info[disp_num].fatbits_dialog.y;
    draw_rectangle(disp_num, disp_info[disp_num].rubber_band_gc);
}


/*
 * fatbits_cur_sel_change - this function is called on the mouse button move
 *                          while selecting a region in fatbits.  This function
 *                          returns TRUE if the fatbits point changed from
 *                          last time.
 */
fatbits_cur_sel_change(disp_num, x, y)
int disp_num, x, y;
{
    int flx, fly;

    flx = x / FATBITS_MAG;
    fly = y / FATBITS_MAG;

    bound_num(&flx, 0, disp_info[disp_num].fatbits_dialog.width - 1);
    bound_num(&fly, 0, disp_info[disp_num].fatbits_dialog.height - 1);
  
    if ((flx == 
       disp_info[disp_num].fatbits_dialog.selected_region.fat_last_rub_point.x)
       && (fly == 
       disp_info[disp_num].fatbits_dialog.selected_region.fat_last_rub_point.y))
    {
        return(FALSE);    /* no change */
    }
    else
    {
        disp_info[disp_num].fatbits_dialog.selected_region.fat_last_rub_point.x=
	    flx;
        disp_info[disp_num].fatbits_dialog.selected_region.fat_last_rub_point.y=
	    fly;
        return(TRUE);
    }
}


/*
 * fatbits_set_cur_sel_point - this function is called on the mouse button move
 *                          while selecting a region in fatbits.  This function
 *                          sets the current selection.
 */
fatbits_set_cur_sel_point(disp_num, x, y)
int disp_num, x, y;
{
    int flx, fly;

    flx = x / FATBITS_MAG;
    fly = y / FATBITS_MAG;

    bound_num(&flx, 0, disp_info[disp_num].fatbits_dialog.width - 1);
    bound_num(&fly, 0, disp_info[disp_num].fatbits_dialog.height - 1);
  
    disp_info[disp_num].fatbits_dialog.selected_region.last_rubberband_point.x =
              flx * FATBITS_MAG + (FATBITS_MAG / 2);
    disp_info[disp_num].fatbits_dialog.selected_region.last_rubberband_point.y =
              fly * FATBITS_MAG + (FATBITS_MAG / 2);

    /*
     * now do the regular window
     */
    draw_rectangle(disp_num,disp_info[disp_num].rubber_band_gc);
    disp_info[disp_num].last_rubberband_point.x = flx +
			       disp_info[disp_num].fatbits_dialog.x;
    disp_info[disp_num].last_rubberband_point.y = fly +
			       disp_info[disp_num].fatbits_dialog.y;
    draw_rectangle(disp_num,disp_info[disp_num].rubber_band_gc);
}


/*
 * draw_fatbits_rectangle - this function draws the current rectangle in the
 *                  fatbits dialog. The main idea of this function is to reverse
 *                  the coordinates if necessary for the XDrawRectangle call.
 */
draw_fatbits_rectangle(disp_num)
int disp_num;
{
    int x, y, width, height;

    x = 
   (disp_info[disp_num].fatbits_dialog.selected_region.orig_rubberband_point.x <
    disp_info[disp_num].fatbits_dialog.selected_region.last_rubberband_point.x)?
    disp_info[disp_num].fatbits_dialog.selected_region.orig_rubberband_point.x :
    disp_info[disp_num].fatbits_dialog.selected_region.last_rubberband_point.x;

    y = 
   (disp_info[disp_num].fatbits_dialog.selected_region.orig_rubberband_point.y <
    disp_info[disp_num].fatbits_dialog.selected_region.last_rubberband_point.y)?
    disp_info[disp_num].fatbits_dialog.selected_region.orig_rubberband_point.y :
    disp_info[disp_num].fatbits_dialog.selected_region.last_rubberband_point.y;

    width = abs(
    disp_info[disp_num].fatbits_dialog.selected_region.last_rubberband_point.x -
    disp_info[disp_num].fatbits_dialog.selected_region.orig_rubberband_point.x);

    height = abs(
    disp_info[disp_num].fatbits_dialog.selected_region.last_rubberband_point.y -
    disp_info[disp_num].fatbits_dialog.selected_region.orig_rubberband_point.y);

    XDrawRectangle(disp_info[disp_num].disp,
          XtWindow(disp_info[disp_num].fatbits_dialog.draw_area),
          disp_info[disp_num].fatbits_dialog.selected_region.rubber_band_gc, 
          x, y, width, height);
}


/*
 * handle_fatbits_pencil_event - this function takes an event X and Y and
 *                               and paints or unpaints that particular spot
 *                               on the fatbits drawing area.  It also
 *                               paints that pixel in the regular drawing 
 *                               area on all the displays.
 */
handle_fatbits_pencil_event(disp_num, real_x, real_y)
int disp_num, real_x, real_y;
{
    int x, y, i;
    unsigned long the_pixel;

    x = real_x / FATBITS_MAG;
    y = real_y / FATBITS_MAG;

    if ((x < 0) || (x > disp_info[disp_num].fatbits_dialog.width) ||
        (y < 0) || (y > disp_info[disp_num].fatbits_dialog.height))
    {
	return;		/* this pixel is outside the fatbits area */
    }

    the_pixel = XGetPixel(disp_info[disp_num].fatbits_dialog.local_image, x, y);

    /*
     * if we are erasing pixels, and this pixel is not already erased...
     */
    if ((the_pixel != disp_info[disp_num].fatbits_dialog.backgd_pixel) &&
	(disp_info[disp_num].fatbits_dialog.pencil_bool == FALSE))
    {
        XFillRectangle(disp_info[disp_num].disp,
                  disp_info[disp_num].fatbits_dialog.backing_store,
                  disp_info[disp_num].backing_store_gc, 
		  x*FATBITS_MAG + 1, y*FATBITS_MAG + 1, 
		  FATBITS_MAG - 1, FATBITS_MAG - 1);
        XFillRectangle(disp_info[disp_num].disp,
                  XtWindow(disp_info[disp_num].fatbits_dialog.draw_area),
                  disp_info[disp_num].backing_store_gc, 
		  x*FATBITS_MAG + 1, y*FATBITS_MAG + 1, 
		  FATBITS_MAG - 1, FATBITS_MAG - 1);
	XPutPixel(disp_info[disp_num].fatbits_dialog.local_image, x, y, 
	          disp_info[disp_num].fatbits_dialog.backgd_pixel);

        for (i=0; i < num_disps; i++)
        {
            if (disp_info[i].in_session_bool)
            {
                XDrawPoint(disp_info[i].disp, disp_info[i].draw_win_id,
                   disp_info[i].backing_store_gc, 
                   x + disp_info[disp_num].fatbits_dialog.x, 
                   y + disp_info[disp_num].fatbits_dialog.y);
                XDrawPoint(disp_info[i].disp, disp_info[i].backing_store,
                   disp_info[i].backing_store_gc, 
                   x + disp_info[disp_num].fatbits_dialog.x, 
                   y + disp_info[disp_num].fatbits_dialog.y);
                XFlush(disp_info[i].disp);
            }
        }
    }
    /*
     * else if we are drawing pixels in the foreground, and this pixel
     * is not already drawn
     */
    else if ((the_pixel != disp_info[disp_num].current_pixel) &&
   	     (disp_info[disp_num].fatbits_dialog.pencil_bool == TRUE))
    {
        XFillRectangle(disp_info[disp_num].disp,
                  disp_info[disp_num].fatbits_dialog.backing_store,
                  disp_info[disp_num].win_gc[disp_num], 
		  x*FATBITS_MAG + 1, y*FATBITS_MAG + 1, 
		  FATBITS_MAG - 1, FATBITS_MAG - 1);
        XFillRectangle(disp_info[disp_num].disp,
                  XtWindow(disp_info[disp_num].fatbits_dialog.draw_area),
                  disp_info[disp_num].win_gc[disp_num], 
		  x*FATBITS_MAG + 1, y*FATBITS_MAG + 1, 
		  FATBITS_MAG - 1, FATBITS_MAG - 1);
	XPutPixel(disp_info[disp_num].fatbits_dialog.local_image, x, y, 
	          disp_info[disp_num].current_pixel);

        for (i=0; i < num_disps; i++)
        {
            if (disp_info[i].in_session_bool)
            {
                XDrawPoint(disp_info[i].disp, disp_info[i].draw_win_id,
                   disp_info[disp_num].win_gc[i],
                   x + disp_info[disp_num].fatbits_dialog.x, 
                   y + disp_info[disp_num].fatbits_dialog.y);
                XDrawPoint(disp_info[i].disp, disp_info[i].backing_store,
                   disp_info[disp_num].win_gc[i],
                   x + disp_info[disp_num].fatbits_dialog.x, 
                   y + disp_info[disp_num].fatbits_dialog.y);
                XFlush(disp_info[i].disp);
            }
        }
    }
}


/*
 * handle_fatbits_eraser_event - this function takes an event X and Y and
 *                               and moves the eraser to that area and erases
 *                               the appropriate fatbit dots.  It also
 *                               erases these dots on the regular drawing
 *                               window.
 */
handle_fatbits_eraser_event(disp_num, real_x, real_y)
int disp_num, real_x, real_y;
{
    int i, j, k;
    int left, top, bottom, right;
    int fat_right, fat_bottom;
    int x, y, width, height;
    unsigned long the_pixel;

    XMoveWindow(disp_info[disp_num].disp,
	    disp_info[disp_num].fatbits_dialog.eraser_win_id,
	    real_x - disp_info[disp_num].fatbits_dialog.eraser_win_width/2,
	    real_y - disp_info[disp_num].fatbits_dialog.eraser_win_width/2);

    /* 
     * calculate the fatbits box that is to be erased
     */
    left = real_x - disp_info[disp_num].fatbits_dialog.eraser_win_width/2;
    top = real_y - disp_info[disp_num].fatbits_dialog.eraser_win_width/2;
    bottom = real_y + disp_info[disp_num].fatbits_dialog.eraser_win_width/2;
    right = real_x + disp_info[disp_num].fatbits_dialog.eraser_win_width/2;

    x = left / FATBITS_MAG;
    y = top / FATBITS_MAG;
    fat_right = (right / FATBITS_MAG);
    fat_bottom = (bottom / FATBITS_MAG);

    bound_num(&x, 0, disp_info[disp_num].fatbits_dialog.width - 1);
    bound_num(&y, 0, disp_info[disp_num].fatbits_dialog.height - 1);
    bound_num(&fat_right, 0, disp_info[disp_num].fatbits_dialog.width - 1);
    bound_num(&fat_bottom, 0, disp_info[disp_num].fatbits_dialog.height - 1);

    /*
     * Now erase this box of fatbits pixels
     */
    for (j=y; j <= fat_bottom; j++)
    {
        for (i=x; i <= fat_right; i++)
        {
            the_pixel =XGetPixel(disp_info[disp_num].fatbits_dialog.local_image,
				 i, j);
            if (the_pixel != disp_info[disp_num].fatbits_dialog.backgd_pixel)
	    {
                XFillRectangle(disp_info[disp_num].disp,
                      disp_info[disp_num].fatbits_dialog.backing_store,
                      disp_info[disp_num].backing_store_gc, 
		      i*FATBITS_MAG + 1, j*FATBITS_MAG + 1, 
		      FATBITS_MAG - 1, FATBITS_MAG - 1);
                XFillRectangle(disp_info[disp_num].disp,
                      XtWindow(disp_info[disp_num].fatbits_dialog.draw_area),
                      disp_info[disp_num].backing_store_gc, 
		      i*FATBITS_MAG + 1, j*FATBITS_MAG + 1, 
		      FATBITS_MAG - 1, FATBITS_MAG - 1);
	        XPutPixel(disp_info[disp_num].fatbits_dialog.local_image, i, j, 
	              disp_info[disp_num].fatbits_dialog.backgd_pixel);
    
                for (k=0; k < num_disps; k++)
                {
                    if (disp_info[k].in_session_bool)
                    {
                        XDrawPoint(disp_info[k].disp, disp_info[k].draw_win_id,
                           disp_info[k].backing_store_gc, 
                           i + disp_info[disp_num].fatbits_dialog.x, 
                           j + disp_info[disp_num].fatbits_dialog.y);
                        XDrawPoint(disp_info[k].disp,disp_info[k].backing_store,
                           disp_info[k].backing_store_gc, 
                           i + disp_info[disp_num].fatbits_dialog.x, 
                           j + disp_info[disp_num].fatbits_dialog.y);
                        XFlush(disp_info[k].disp);
                    }
                }
	    }
        }
    }
}


/*
 * bound_num - this function "bounds" the number to the min and max indicated.
 */
bound_num(num, min, max)
int *num, min, max;
{
    if (*num < min)
	*num = min;
    if (*num > max)
	*num = max;
}


/*
 * fatbits_pointer_moveCB - this function is called when the pointer moves 
 *                 in the drawing window.
 */
XtEventHandler fatbits_pointer_moveCB(w, client_data, our_event, cbool)
Widget w;
caddr_t client_data;
XEvent *our_event;
Boolean *cbool;
{
    int i, disp_num, x, y, root_x, root_y;
    XMotionEvent *the_event;
    Window dummy;
    Widget dummy_w = (Widget)NULL;
    Boolean dummy_cbool;

    if (our_event->type != MotionNotify)
        printf("ERROR: wrong event type in fatbits_pointer_moveCB.\n");

    the_event = (XMotionEvent *) our_event;
    disp_num = (int) client_data;

    if (disp_info[disp_num].fatbits_dialog.pointer_state != PRESSED)
        return;         /* not interested in this event */

    switch(disp_info[disp_num].fatbits_dialog.input_mode)
    {
        case SELECTING:
	    if (fatbits_cur_sel_change(disp_num, the_event->x, the_event->y))
	    {
                draw_fatbits_rectangle(disp_num);
	        fatbits_set_cur_sel_point(disp_num, the_event->x, the_event->y);
                draw_fatbits_rectangle(disp_num);
                XFlush(disp_info[disp_num].disp);
	    }
            break;
        case PENCILING:
            handle_fatbits_pencil_event(disp_num, the_event->x, the_event->y);
            break;
        case ERASING:
            handle_fatbits_eraser_event(disp_num, the_event->x, the_event->y);
            break;
        case DRAGGING_SELECTION:
            fatbits_selection_pointer_moveCB(dummy_w, disp_num << 8 | disp_num,
                           (XEvent *) the_event, &dummy_cbool);
	    /*
	     * now fake the event to make the regular selection box move
	     */
	    the_event->x = 
	            disp_info[disp_num].fatbits_dialog.selected_region.cur_x +
		    disp_info[disp_num].selected_region.drag_offset.x +
		    disp_info[disp_num].fatbits_dialog.x;
	    the_event->y = 
	            disp_info[disp_num].fatbits_dialog.selected_region.cur_y +
		    disp_info[disp_num].selected_region.drag_offset.y +
		    disp_info[disp_num].fatbits_dialog.y;
	    disp_info[disp_num].pointer_state = PRESSED;
	    selection_pointer_moveCB(dummy_w, disp_num << 8 | disp_num,
			             (XEvent *) the_event, &dummy_cbool);
	    disp_info[disp_num].pointer_state = NOT_PRESSED;
            break;
        default:
            printf("ERROR: unknown input mode in fatbits_pointer_moveCB.\n");
            break;
    }
}


/*
 * fatbits_draw_area_exposeCB - this function gets called when an expose event
 *                     occurs in the fatbits drawing area.
 */
fatbits_draw_area_exposeCB(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int disp_num;
    XmDrawingAreaCallbackStruct *cb_struct;
    XExposeEvent *the_event;

    cb_struct = (XmDrawingAreaCallbackStruct *) call_data;
    the_event = (XExposeEvent *) cb_struct->event;
    disp_num = (int) client_data;

    if (the_event->type != Expose)
        printf("ERROR: in fatbits_draw_area_exposeCB.\n");

    XCopyArea(disp_info[disp_num].disp, 
              disp_info[disp_num].fatbits_dialog.backing_store,
              XtWindow(disp_info[disp_num].fatbits_dialog.draw_area),
              disp_info[disp_num].win_gc[disp_num], the_event->x, the_event->y,
              the_event->width, the_event->height, the_event->x, the_event->y);
}


/*
 * prepare_fatbits_dialog - this function prepares the fatbits dialog
 * 			    and is called right before that dialog is mapped.
 */
prepare_fatbits_dialog(disp_num)
int disp_num;
{
    Pixmap button_pixmap;
    Pixel fg, bg;
    int n, width, height, fat_width, fat_height, x, y;
    Arg args[MAX_ARGS];
    XGCValues gcvalues;
    unsigned long the_pixel, the_last_pixel;

    if (XtIsManaged(disp_info[disp_num].fatbits_dialog.the_dialog))
    {
       if(disp_info[disp_num].fatbits_dialog.selected_region.region_is_selected)
	   place_fatbits_selection_permanently(disp_num);

       refresh_fatbits_area(disp_num);
    }

    set_paused_cursors();

    disp_info[disp_num].fatbits_dialog.x = 
				    disp_info[disp_num].selected_region.cur_x;
    disp_info[disp_num].fatbits_dialog.y = 
				    disp_info[disp_num].selected_region.cur_y;
    width = disp_info[disp_num].fatbits_dialog.width =
                                   disp_info[disp_num].selected_region.width;
    height = disp_info[disp_num].fatbits_dialog.height =
                                   disp_info[disp_num].selected_region.height;
    fat_width = (width * FATBITS_MAG) + 65;
    fat_height = (height * FATBITS_MAG) + 67;

    n = 0;
    XtSetArg(args[n], XmNwidth, fat_width); n++;
    XtSetArg(args[n], XmNheight, fat_height); n++;
    XtSetValues(disp_info[disp_num].fatbits_dialog.dialog_form, args, n);

    if (disp_info[disp_num].fatbits_dialog.backing_store != NULL)
    {
        XFreePixmap(disp_info[disp_num].disp,
                   disp_info[disp_num].fatbits_dialog.backing_store);
    }

    disp_info[disp_num].fatbits_dialog.backing_store =
                        XCreatePixmap(disp_info[disp_num].disp,
                            disp_info[disp_num].draw_win_id, fat_width, 
			    fat_height, DefaultDepth(disp_info[disp_num].disp,
                            DefaultScreen(disp_info[disp_num].disp)));

    XFillRectangle(disp_info[disp_num].disp,
                  disp_info[disp_num].fatbits_dialog.backing_store,
                  disp_info[disp_num].backing_store_gc, 0, 0, fat_width, 
		  fat_height);

    gcvalues.line_width = 0;
    XChangeGC(disp_info[disp_num].disp, disp_info[disp_num].menu_gc,
                      GCLineWidth, &gcvalues);
   
    for (x=0; x <= width; x++)            /* draw the grid */
    {
        for (y=0; y <= height; y++)
        {
            XDrawLine(disp_info[disp_num].disp, 
                    disp_info[disp_num].fatbits_dialog.backing_store,
                    disp_info[disp_num].menu_gc,
                    x*FATBITS_MAG, 0, x*FATBITS_MAG, fat_height);
            XDrawLine(disp_info[disp_num].disp, 
                    disp_info[disp_num].fatbits_dialog.backing_store,
                    disp_info[disp_num].menu_gc,
                    0, y*FATBITS_MAG, fat_width, y*FATBITS_MAG);
        }
    }

    gcvalues.line_width = 2;
    XChangeGC(disp_info[disp_num].disp, disp_info[disp_num].menu_gc,
                      GCLineWidth, &gcvalues);

    /* 
     * now get the image that needs to be painted into the fatbits editor
     */
    if ((disp_info[disp_num].fatbits_dialog.local_image = 
              XGetImage(disp_info[disp_num].disp,
                disp_info[disp_num].selected_region.backing_store[disp_num],
                0, 0, width, height, (~0), ZPixmap)) == NULL)
    {
        popup_error_dialog(disp_num, 
            "Could not get the_image for\nthe fatbits editor.\n");
        return;
    }
    
    the_last_pixel = (unsigned long) BlackPixelOfScreen(
	                    DefaultScreenOfDisplay(disp_info[disp_num].disp));
    XSetForeground(disp_info[disp_num].disp, disp_info[disp_num].menu_gc,
	            the_last_pixel);

    for (x=0; x < width; x++)            /* draw the image in the editor */
    {
        for (y=0; y < height; y++)
        {
	    the_pixel = XGetPixel(
		         disp_info[disp_num].fatbits_dialog.local_image, x, y);
		
	    if (the_pixel != disp_info[disp_num].fatbits_dialog.backgd_pixel)
	    {
	        if (the_pixel != the_last_pixel)
	        {
	            XSetForeground(disp_info[disp_num].disp, 
                               disp_info[disp_num].menu_gc, the_pixel);
	            the_last_pixel = the_pixel;
	        }
		    	
                XFillRectangle(disp_info[disp_num].disp,
                      disp_info[disp_num].fatbits_dialog.backing_store,
                      disp_info[disp_num].menu_gc, 
		      x*FATBITS_MAG + 1, y*FATBITS_MAG + 1, 
		      FATBITS_MAG - 1, FATBITS_MAG - 1);
	    }
        }
    }

    XSetForeground(disp_info[disp_num].disp, disp_info[disp_num].menu_gc,
	            (unsigned long) BlackPixelOfScreen(
	            DefaultScreenOfDisplay(disp_info[disp_num].disp)));

    unset_paused_cursors();
}


/*
 * refresh_fatbits_area - this function refreshes the area on the normal
 *                        drawing area with the image from the fatbits 
 *                        editor.
 */
refresh_fatbits_area(disp_num)
int disp_num;
{
    XImage *the_image;
    int width, height, depth, x, y, xhr, yhr, i;
    char tmp_name[50];
    int others;

    set_paused_cursors();                     /*this may take a while*/

    x = disp_info[disp_num].fatbits_dialog.x;
    y = disp_info[disp_num].fatbits_dialog.y;
    width = disp_info[disp_num].fatbits_dialog.width;
    height = disp_info[disp_num].fatbits_dialog.height;
    the_image = disp_info[disp_num].fatbits_dialog.local_image;

    /*
     * First put the image that is already prepared on the display that
     * had the fatbits editor running on it.  In the "best case", this
     * will be all that is necessary.
     */
    XPutImage(disp_info[disp_num].disp, disp_info[disp_num].backing_store,
              disp_info[disp_num].win_gc[disp_num], the_image,
              0, 0, x, y, width, height);
    XCopyArea(disp_info[disp_num].disp, disp_info[disp_num].backing_store,
              disp_info[disp_num].draw_win_id,
              disp_info[disp_num].win_gc[disp_num], x, y, width, height, x, y);

    /*
     * Now check if there are other active wscrawls running, and if so,
     * convert the image and pump it out to those displays.
     */
    for (i=0, others=FALSE; i < num_disps; i++)
    {
	if ((disp_info[i].in_session_bool == TRUE) && (i != disp_num))
	{
	    others = TRUE;
	    break;
	}
    }
    
    if (others != TRUE)   /* there are NO other people in session */
    {
        XDestroyImage(the_image);
    }
    else                  /* there ARE other people in session */
    {
        sprintf(tmp_name, "/tmp/ws.%d.%d", disp_num, getpid());

        save_image_on_disk(disp_info[disp_num].disp,
                   disp_info[disp_num].draw_win_id,
                   disp_info[disp_num].backing_store,
                   x, y, width, height, tmp_name);

        for (i=0; i < num_disps; i++)
        {
            if ((disp_info[i].in_session_bool == TRUE) && (disp_num != i))
            {
                depth = -1;
                the_image = read_image_from_disk(disp_info[i].disp,
                           disp_info[i].draw_win_id,
                           tmp_name, &width, &height, &depth);

                if ((depth == -1) || (the_image == NULL))  /*it failed*/
                {
                    popup_info_dialog(disp_num,
                         "Could not copy your fatbits\nto other person.");
                }
                else
                {
                    XPutImage(disp_info[i].disp, disp_info[i].backing_store,
                              disp_info[disp_num].win_gc[i], the_image,
                              0, 0, x, y, width, height);
                    XCopyArea(disp_info[i].disp, 
                              disp_info[i].backing_store,
                              disp_info[i].draw_win_id,
                              disp_info[i].win_gc[i], 
                              x, y, width, height, x, y);
                    XDestroyImage(the_image);
                }
            }
        }

	unlink(tmp_name);
    }

    unset_paused_cursors();
}


/*
 * put_bitmaps_into_fatbits_dialog - this function puts the bitmap tools
 * 				   into the brush fatbits dialog.
 */
put_bitmaps_into_fatbits_dialog(disp_num)
int disp_num;
{
    Pixmap button_pixmap;
    Pixel fg, bg;
    int n, width;
    Arg args[MAX_ARGS];
    XSetWindowAttributes set_attr;
    XGCValues gcvalues;

    /*
     * set the first cursor in the fatbits drawing area, and initialize
     * misc.
     */
    XDefineCursor(disp_info[disp_num].disp,
		  XtWindow(disp_info[disp_num].fatbits_dialog.draw_area),
		  XCreateFontCursor(disp_info[disp_num].disp, XC_pencil));
    disp_info[disp_num].fatbits_dialog.input_mode = PENCILING;
    disp_info[disp_num].fatbits_dialog.pointer_state = NOT_PRESSED;
    disp_info[disp_num].fatbits_dialog.backing_store = NULL;
    disp_info[disp_num].fatbits_dialog.selected_region.backing_store = NULL;
    disp_info[disp_num].fatbits_dialog.selected_region.region_is_selected=FALSE;

    /*
     * get the background color of the drawing area and save it forever
     */
    disp_info[disp_num].fatbits_dialog.backgd_pixel = (unsigned long) 
	 WhitePixelOfScreen(DefaultScreenOfDisplay(disp_info[disp_num].disp));

    /*
     * create the eraser window
     */
    set_attr.cursor = XCreateFontCursor(disp_info[disp_num].disp,XC_draped_box);
    set_attr.background_pixel =
	  WhitePixelOfScreen(DefaultScreenOfDisplay(disp_info[disp_num].disp));
    width = disp_info[disp_num].fatbits_dialog.eraser_win_width = 5*FATBITS_MAG;
    disp_info[disp_num].fatbits_dialog.eraser_win_id = 
	    XCreateWindow(disp_info[disp_num].disp,
      	          XtWindow(disp_info[disp_num].fatbits_dialog.draw_area), 
		  0, 0, width - 4, width - 4, 2, 0, 
		  CopyFromParent, CopyFromParent, 
		  CWCursor | CWBackPixel, &set_attr);
    disp_info[disp_num].fatbits_dialog.eraser_win_width/2,

    /*
     * create the rubberband gc
     */
    gcvalues.function = GXinvert;
    gcvalues.line_width = FATBITS_MAG;
    disp_info[disp_num].fatbits_dialog.selected_region.rubber_band_gc =
    XCreateGC(disp_info[disp_num].disp, 
	      XtWindow(disp_info[disp_num].fatbits_dialog.draw_area),
	      GCFunction | GCLineWidth, &gcvalues);
    gcvalues.line_style = LineOnOffDash;

    gcvalues.foreground = BlackPixelOfScreen(
			 DefaultScreenOfDisplay(disp_info[disp_num].disp));
    gcvalues.function = GXcopy;
    disp_info[disp_num].fatbits_dialog.selected_region.surround_line_gc =
    XCreateGC(disp_info[disp_num].disp, 
	      XtWindow(disp_info[disp_num].fatbits_dialog.draw_area),
	      GCFunction | GCLineWidth | GCForeground, &gcvalues);

    /* 
     * create the selection region
     */
    create_fatbits_region(disp_num);

    /*
     * put bitmaps into tool palette
     */
    n = 0;
    XtSetArg(args[n], XmNforeground, &fg); n++;
    XtSetArg(args[n], XmNbackground, &bg); n++;
    XtGetValues(disp_info[disp_num].fatbits_dialog.tool[0], args, n);

    button_pixmap = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      XtWindow(disp_info[disp_num].fatbits_dialog.draw_area), 
		      selectb_bits,
                      selectb_width, selectb_height, fg, bg,
                      DefaultDepth(disp_info[disp_num].disp,
                      DefaultScreen(disp_info[disp_num].disp)));
    n = 0;
    XtSetArg(args[n], XmNlabelType, XmPIXMAP); n++;
    XtSetArg(args[n], XmNlabelPixmap, button_pixmap); n++;
    XtSetArg(args[n], XmNwidth, 32); n++;
    XtSetArg(args[n], XmNheight, 32); n++;
    XtSetValues(disp_info[disp_num].fatbits_dialog.tool[0], args, n);

    button_pixmap = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      XtWindow(disp_info[disp_num].fatbits_dialog.draw_area), 
		      pencil_bits,
                      pencil_width, pencil_height, fg, bg,
                      DefaultDepth(disp_info[disp_num].disp,
                      DefaultScreen(disp_info[disp_num].disp)));
    n = 0;
    XtSetArg(args[n], XmNlabelType, XmPIXMAP); n++;
    XtSetArg(args[n], XmNlabelPixmap, button_pixmap); n++;
    XtSetArg(args[n], XmNwidth, 32); n++;
    XtSetArg(args[n], XmNheight, 32); n++;
    XtSetValues(disp_info[disp_num].fatbits_dialog.tool[1], args, n);

    button_pixmap = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      XtWindow(disp_info[disp_num].fatbits_dialog.draw_area), 
		      eraser_bits,
                      eraser_width, eraser_height, fg, bg,
                      DefaultDepth(disp_info[disp_num].disp,
                      DefaultScreen(disp_info[disp_num].disp)));
    n = 0;
    XtSetArg(args[n], XmNlabelType, XmPIXMAP); n++;
    XtSetArg(args[n], XmNlabelPixmap, button_pixmap); n++;
    XtSetArg(args[n], XmNwidth, 32); n++;
    XtSetArg(args[n], XmNheight, 32); n++;
    XtSetValues(disp_info[disp_num].fatbits_dialog.tool[2], args, n);
}


/*
 * create_fatbits_dialog - this function creates the popup dialog that
 *                          allows the user to fatbits edit a selection.
 */
Widget create_fatbits_dialog(disp_num)
int disp_num;
{
    Widget the_dialog;
    Widget the_separator;
    Widget the_frame, the_form;
    Widget draw_frame;
    Arg args[MAX_ARGS];
    int n;

    n = 0;
    XtSetArg(args[n], XmNautoUnmanage, False); n++;
    XtSetArg(args[n], XmNtraversalOn, True); n++;
    the_dialog = XmCreateFormDialog(disp_info[disp_num].app_shell, 
			   "Fatbits_Dialog", args, n);

    n = 0;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
    the_frame = XmCreateFrame(the_dialog, "dialog_frame", args, n);
    XtManageChild(the_frame);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNwidth, 64); n++;
    the_form = XmCreateForm(the_frame, "dialog_form", args, n);
    XtManageChild(the_form);
    disp_info[disp_num].fatbits_dialog.dialog_form = the_form;

    /*
     * create the tool palette
     */
    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 5); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    XtSetArg(args[n], XmNwidth, 32); n++;
    XtSetArg(args[n], XmNheight, 32); n++;
    disp_info[disp_num].fatbits_dialog.tool[0] = 
             XmCreatePushButton(the_form,"Selection", args,n);
    XtManageChild(disp_info[disp_num].fatbits_dialog.tool[0]);
    XtAddCallback(disp_info[disp_num].fatbits_dialog.tool[0], 
		  XmNactivateCallback, (XtCallbackProc)handle_fatbits_dialog,
        (XtPointer)(disp_num << 8 | FATBITS_SELECTION_BUT));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 5); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, 
	    disp_info[disp_num].fatbits_dialog.tool[0]); n++;
    XtSetArg(args[n], XmNtopOffset, 4); n++;
    XtSetArg(args[n], XmNwidth, 32); n++;
    XtSetArg(args[n], XmNheight, 32); n++;
    disp_info[disp_num].fatbits_dialog.tool[1] = 
             XmCreatePushButton(the_form,"Pencil", args,n);
    XtManageChild(disp_info[disp_num].fatbits_dialog.tool[1]);
    XtAddCallback(disp_info[disp_num].fatbits_dialog.tool[1], 
		  XmNactivateCallback, (XtCallbackProc)handle_fatbits_dialog,
        (XtPointer)(disp_num << 8 | FATBITS_PENCIL_BUT));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 5); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, 
	    disp_info[disp_num].fatbits_dialog.tool[1]); n++;
    XtSetArg(args[n], XmNtopOffset, 4); n++;
    XtSetArg(args[n], XmNwidth, 32); n++;
    XtSetArg(args[n], XmNheight, 32); n++;
    disp_info[disp_num].fatbits_dialog.tool[2] = 
             XmCreatePushButton(the_form,"Selection", args,n);
    XtManageChild(disp_info[disp_num].fatbits_dialog.tool[2]);
    XtAddCallback(disp_info[disp_num].fatbits_dialog.tool[2], 
		  XmNactivateCallback, (XtCallbackProc)handle_fatbits_dialog,
        (XtPointer)(disp_num << 8 | FATBITS_ERASER_BUT));

    /*
     * create the fatbits drawing area in a frame
     */
    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 50); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 10); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomOffset, 52); n++;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_IN); n++;
    draw_frame = XmCreateFrame(the_form, "draw_frame", args, n);
    XtManageChild(draw_frame);

    n = 0;
    XtSetArg(args[n], XmNwidth, 250); n++;
    XtSetArg(args[n], XmNheight, 250); n++;
    XtSetArg(args[n], XmNbackground, 
      WhitePixelOfScreen(DefaultScreenOfDisplay(disp_info[disp_num].disp)));n++;
    disp_info[disp_num].fatbits_dialog.draw_area =
                XmCreateDrawingArea(draw_frame, "fatbits_draw_area", args, n);
    XtManageChild(disp_info[disp_num].fatbits_dialog.draw_area);
    XtAddCallback(disp_info[disp_num].fatbits_dialog.draw_area, 
		  XmNexposeCallback, (XtCallbackProc)fatbits_draw_area_exposeCB, (XtPointer)disp_num);
    XtAddCallback(disp_info[disp_num].fatbits_dialog.draw_area, 
                  XmNinputCallback, (XtCallbackProc)fatbits_draw_area_inputCB, (XtPointer)disp_num);
    XtAddEventHandler(disp_info[disp_num].fatbits_dialog.draw_area, 
                  PointerMotionMask, False, (XtEventHandler)fatbits_pointer_moveCB, (XtPointer)disp_num);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 0); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 0); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, draw_frame); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    the_separator = XmCreateSeparatorGadget(the_form, "separator", args, n);
    XtManageChild(the_separator);

    n = 0;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget,  the_separator); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNleftPosition, 33); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNrightPosition, 62); n++;
    XtSetArg(args[n], XmNbottomOffset, 5); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    disp_info[disp_num].fatbits_dialog.close_button =
                XmCreatePushButtonGadget(the_form, "Close", args, n);
    XtManageChild(disp_info[disp_num].fatbits_dialog.close_button);
    XtAddCallback(disp_info[disp_num].fatbits_dialog.close_button, 
		  XmNactivateCallback, (XtCallbackProc)handle_fatbits_dialog, (XtPointer)(disp_num << 8 | FATBITS_CLOSE));

    set_dialog_wm_name(the_dialog, "Fatbits Editor");
    XtRealizeWidget(the_dialog);
    return(the_dialog);
}

