/*
 * This file:  wsfatreg.c
 *
 * This file contains all the functions for manipulating regions in fatbits.  
 */

#include "wscrawl.h"

extern int num_disps;
extern struct disp_info_struct disp_info[MAX_NUM_DISPS];


/* 
 * define_fatbits_selection - this function is called upon the release of the
 *                    mouse button when selecting a region.  This takes
 *                    care of setting the correct selected region, etc.
 */
define_fatbits_selection(disp_num)
int disp_num;
{
    int i, j, x, y, width, height, root_x, root_y;
    unsigned long the_pixel;

    set_fatbits_x_y_width_height(disp_num, &x, &y, &width, &height);

    if (width == 0 || height == 0)
    {
	return;  /* just clicked mouse to set a selection */
    }

    width++;	/* one wider and taller for good effect */
    height++;

    disp_info[disp_num].fatbits_dialog.selected_region.x = x;
    disp_info[disp_num].fatbits_dialog.selected_region.y = y;
    disp_info[disp_num].fatbits_dialog.selected_region.cur_x = x;
    disp_info[disp_num].fatbits_dialog.selected_region.cur_y = y;
    disp_info[disp_num].fatbits_dialog.selected_region.width = width;
    disp_info[disp_num].fatbits_dialog.selected_region.height = height;

    prepare_fatbits_region(disp_num, x, y, width, height);

    XCopyArea(disp_info[disp_num].disp, 
	      disp_info[disp_num].fatbits_dialog.backing_store,
	      disp_info[disp_num].fatbits_dialog.selected_region.backing_store,
	      disp_info[disp_num].win_gc[disp_num], x*FATBITS_MAG, 
	      y*FATBITS_MAG, width*FATBITS_MAG, height*FATBITS_MAG, 0, 0);

    /*
     * now get an image of the selection on the client side, and erase
     * the region image from local image and client side both
     */
    disp_info[disp_num].fatbits_dialog.selected_region.local_image = 
          XSubImage(disp_info[disp_num].fatbits_dialog.local_image,
		    x, y, width, height);

    for (j=y; j < y + height; j++)
    {
        for (i=x; i < x + width; i++)
        {
            the_pixel =XGetPixel(disp_info[disp_num].fatbits_dialog.local_image,
                                 i, j);
            if (the_pixel != disp_info[disp_num].fatbits_dialog.backgd_pixel)
            {
                XFillRectangle(disp_info[disp_num].disp,
                      disp_info[disp_num].fatbits_dialog.backing_store,
                      disp_info[disp_num].backing_store_gc,
                      i*FATBITS_MAG + 1, j*FATBITS_MAG + 1,
                      FATBITS_MAG - 1, FATBITS_MAG - 1);
                XPutPixel(disp_info[disp_num].fatbits_dialog.local_image, i, j,
			  disp_info[disp_num].fatbits_dialog.backgd_pixel);
            }
        }
    }

    /*
     * map it and away we go
     */
    XtMapWidget(disp_info[disp_num].fatbits_dialog.selected_region.win);
    disp_info[disp_num].fatbits_dialog.selected_region.region_is_selected =TRUE;
}


/*
 * draw_fatbits_selection_outline - this function draws the selection outline,
 */
draw_fatbits_selection_outline(disp_num)
int disp_num;
{
    int new_offset;
    XGCValues gcvalues;
    unsigned long the_pixel;

    gcvalues.foreground = WhitePixelOfScreen(
                            DefaultScreenOfDisplay(disp_info[disp_num].disp));
    gcvalues.line_width = FATBITS_MAG;
    XChangeGC(disp_info[disp_num].disp,
           disp_info[disp_num].fatbits_dialog.selected_region.surround_line_gc,
           GCLineWidth | GCForeground, &gcvalues);

    XDrawRectangle(disp_info[disp_num].disp,
     XtWindow(disp_info[disp_num].fatbits_dialog.selected_region.win),
     disp_info[disp_num].fatbits_dialog.selected_region.surround_line_gc, 
     FATBITS_MAG/2, FATBITS_MAG/2, 
     (disp_info[disp_num].fatbits_dialog.selected_region.width -1)*FATBITS_MAG,
     (disp_info[disp_num].fatbits_dialog.selected_region.height-1)*FATBITS_MAG);

    gcvalues.foreground = BlackPixelOfScreen(
                            DefaultScreenOfDisplay(disp_info[disp_num].disp));
    gcvalues.line_width = 0;
    XChangeGC(disp_info[disp_num].disp,
           disp_info[disp_num].fatbits_dialog.selected_region.surround_line_gc,
           GCLineWidth | GCForeground, &gcvalues);

    XDrawRectangle(disp_info[disp_num].disp,
     XtWindow(disp_info[disp_num].fatbits_dialog.selected_region.win),
     disp_info[disp_num].fatbits_dialog.selected_region.surround_line_gc, 
     0, 0, 
     (disp_info[disp_num].fatbits_dialog.selected_region.width*FATBITS_MAG)-1,
     (disp_info[disp_num].fatbits_dialog.selected_region.height*FATBITS_MAG)-1);

    XDrawRectangle(disp_info[disp_num].disp,
     XtWindow(disp_info[disp_num].fatbits_dialog.selected_region.win),
     disp_info[disp_num].fatbits_dialog.selected_region.surround_line_gc, 
     1, 1, 
     (disp_info[disp_num].fatbits_dialog.selected_region.width*FATBITS_MAG)-3,
     (disp_info[disp_num].fatbits_dialog.selected_region.height*FATBITS_MAG)-3);

    XDrawRectangle(disp_info[disp_num].disp,
     XtWindow(disp_info[disp_num].fatbits_dialog.selected_region.win),
     disp_info[disp_num].fatbits_dialog.selected_region.surround_line_gc, 
     2, 2, 
     (disp_info[disp_num].fatbits_dialog.selected_region.width*FATBITS_MAG)-4,
     (disp_info[disp_num].fatbits_dialog.selected_region.height*FATBITS_MAG)-4);

    XDrawRectangle(disp_info[disp_num].disp,
     XtWindow(disp_info[disp_num].fatbits_dialog.selected_region.win),
     disp_info[disp_num].fatbits_dialog.selected_region.surround_line_gc, 
     5, 5, 
     (disp_info[disp_num].fatbits_dialog.selected_region.width*FATBITS_MAG)-11,
    (disp_info[disp_num].fatbits_dialog.selected_region.height*FATBITS_MAG)-11);

    XDrawRectangle(disp_info[disp_num].disp,
     XtWindow(disp_info[disp_num].fatbits_dialog.selected_region.win),
     disp_info[disp_num].fatbits_dialog.selected_region.surround_line_gc, 
     6, 6, 
     (disp_info[disp_num].fatbits_dialog.selected_region.width*FATBITS_MAG)-13,
    (disp_info[disp_num].fatbits_dialog.selected_region.height*FATBITS_MAG)-13);

    XDrawRectangle(disp_info[disp_num].disp,
     XtWindow(disp_info[disp_num].fatbits_dialog.selected_region.win),
     disp_info[disp_num].fatbits_dialog.selected_region.surround_line_gc, 
     9, 9, 
     (disp_info[disp_num].fatbits_dialog.selected_region.width*FATBITS_MAG)-19,
    (disp_info[disp_num].fatbits_dialog.selected_region.height*FATBITS_MAG)-19);

    XDrawRectangle(disp_info[disp_num].disp,
     XtWindow(disp_info[disp_num].fatbits_dialog.selected_region.win),
     disp_info[disp_num].fatbits_dialog.selected_region.surround_line_gc, 
     10, 10, 
     (disp_info[disp_num].fatbits_dialog.selected_region.width*FATBITS_MAG)-21,
    (disp_info[disp_num].fatbits_dialog.selected_region.height*FATBITS_MAG)-21);

    XFlush(disp_info[disp_num].disp);
}


/*
 * set_fatbits_x_y_width_height - this function is used in several places.  It 
 * 			sorts the coordinates to give the proper "upper-left"
 *                      corner, and width and height.
 */
set_fatbits_x_y_width_height(disp_num, x, y, width, height)
int disp_num;
int *x, *y, *width, *height;
{
    *x = (
    disp_info[disp_num].fatbits_dialog.selected_region.fat_orig_rub_point.x <
    disp_info[disp_num].fatbits_dialog.selected_region.fat_last_rub_point.x) ?
    disp_info[disp_num].fatbits_dialog.selected_region.fat_orig_rub_point.x :
    disp_info[disp_num].fatbits_dialog.selected_region.fat_last_rub_point.x;

    *y = (
    disp_info[disp_num].fatbits_dialog.selected_region.fat_orig_rub_point.y <
    disp_info[disp_num].fatbits_dialog.selected_region.fat_last_rub_point.y) ?
    disp_info[disp_num].fatbits_dialog.selected_region.fat_orig_rub_point.y :
    disp_info[disp_num].fatbits_dialog.selected_region.fat_last_rub_point.y;

    *width = abs(
    disp_info[disp_num].fatbits_dialog.selected_region.fat_last_rub_point.x -
    disp_info[disp_num].fatbits_dialog.selected_region.fat_orig_rub_point.x);

    *height = abs(
    disp_info[disp_num].fatbits_dialog.selected_region.fat_last_rub_point.y -
    disp_info[disp_num].fatbits_dialog.selected_region.fat_orig_rub_point.y);
}


/*
 * fatbits_selection_area_inputCB - this function handles all selection area 
 *            input events such as ButtonPress or ButtonRelease.
 */
fatbits_selection_area_inputCB(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int disp_num, x, y, width, height;
    XmDrawingAreaCallbackStruct *cb_struct;
    XEvent *the_event;
    XButtonPressedEvent *but_event;

    cb_struct = (XmDrawingAreaCallbackStruct *) call_data;
    the_event = (XEvent *) cb_struct->event;
    disp_num = ((int) client_data) >> 8;

    switch(the_event->type)
    {
        case ButtonPress:
            if (XGrabPointer(disp_info[disp_num].disp,
                     XtWindow(disp_info[disp_num].fatbits_dialog.draw_area), 
		     False, PointerMotionMask | ButtonReleaseMask,
                     GrabModeAsync, GrabModeAsync, None, None,
                     CurrentTime) != GrabSuccess)
            {
                popup_info_dialog(disp_num,
                    "Couldn't grab the pointer.\nAborting Drag of Region.");
                return;
            }
	    disp_info[disp_num].fatbits_dialog.pointer_state = PRESSED;
	    disp_info[disp_num].fatbits_dialog.input_mode = DRAGGING_SELECTION;

            but_event = (XButtonPressedEvent *) the_event;
	    disp_info[disp_num].fatbits_dialog.selected_region.drag_offset.x = 
				    but_event->x;
	    disp_info[disp_num].fatbits_dialog.selected_region.drag_offset.y = 
				    but_event->y;
            disp_info[disp_num].selected_region.drag_offset.x = 0;
            disp_info[disp_num].selected_region.drag_offset.y = 0;
	    break;
        case ButtonRelease:
            disp_info[disp_num].fatbits_dialog.pointer_state = NOT_PRESSED;
	    break;
        case KeyPress:
	    handle_fatbits_KeyPress_event(disp_num, the_event);
	    break;
        case KeyRelease:
	    break;
	default:
	    printf("Unknown event type in fatbits_selection_area_inputCB.\n");
	    break;
    }
}


/*
 * fatbits_selection_pointer_moveCB - this function is called when the pointer 
 *                  moves in the selection window.  This probably means the
 *                  person is dragging the window, but maybe not.
 *
 *                  NOTE: this routine IS NOT called as a callback, but
 *                  instead from my own functions because of a hack for
 *                  smooth and speedy dragging.  I am leaving it defined as
 *                  an XtEventHandler just in case I want to revert.
 */
XtEventHandler fatbits_selection_pointer_moveCB(w,client_data, our_event, cbool)
Widget w;
caddr_t client_data;
XEvent *our_event;
Boolean *cbool;
{
    int i,  disp_num, which;
    XMotionEvent *the_event;
    int x, y, max_x, max_y;
    int real_x, real_y, fat_max_x, fat_max_y;

    if (our_event->type != MotionNotify)
    {
	printf("ERROR: in fatbits_selection_pointer_moveCB, invalid event.\n");
	return;
    }

    the_event = (XMotionEvent *) our_event;
    disp_num = ((int) client_data) >> 8;
    which = ((int) client_data) & 0x00FF;

    if (disp_info[disp_num].fatbits_dialog.pointer_state != PRESSED)
        return;		/* not interested in this event */

    x = the_event->x -
	      disp_info[disp_num].fatbits_dialog.selected_region.drag_offset.x;
    y = the_event->y -
	      disp_info[disp_num].fatbits_dialog.selected_region.drag_offset.y;

    max_x = FATBITS_MAG * (disp_info[disp_num].fatbits_dialog.width - 
	     disp_info[disp_num].fatbits_dialog.selected_region.width);
    bound_num(&x, 0, max_x);
    max_y = FATBITS_MAG * (disp_info[disp_num].fatbits_dialog.height - 
	     disp_info[disp_num].fatbits_dialog.selected_region.height);
    bound_num(&y, 0, max_y);

    XMoveWindow(disp_info[disp_num].disp, 
       XtWindow(disp_info[disp_num].fatbits_dialog.selected_region.win),
       x, y);
    XFlush(disp_info[disp_num].disp);

    real_x = disp_info[disp_num].fatbits_dialog.selected_region.cur_real_x = x;
    real_y = disp_info[disp_num].fatbits_dialog.selected_region.cur_real_y = y;

    fat_max_x = (disp_info[disp_num].fatbits_dialog.width - 
	     disp_info[disp_num].fatbits_dialog.selected_region.width);
    fat_max_y = (disp_info[disp_num].fatbits_dialog.height - 
	     disp_info[disp_num].fatbits_dialog.selected_region.height);

    x = (real_x + (FATBITS_MAG/2))/ FATBITS_MAG;
    y = (real_y + (FATBITS_MAG/2))/ FATBITS_MAG;

    bound_num(&x, 0, fat_max_x);
    bound_num(&y, 0, fat_max_y);

    disp_info[disp_num].fatbits_dialog.selected_region.cur_x = x;
    disp_info[disp_num].fatbits_dialog.selected_region.cur_y = y;
}


/*
 * snap_fatbits_selection_to_grid - this function snaps the selection 
 *                                  to be aligned with the grid.
 */
snap_fatbits_selection_to_grid(disp_num)
int disp_num;
{
    int i, real_x, real_y, x, y, width, height;
    int fat_max_x, fat_max_y;
    Window dummy;

    real_x = disp_info[disp_num].fatbits_dialog.selected_region.cur_real_x;
    real_y = disp_info[disp_num].fatbits_dialog.selected_region.cur_real_y;

    fat_max_x = (disp_info[disp_num].fatbits_dialog.width - 
	     disp_info[disp_num].fatbits_dialog.selected_region.width);
    fat_max_y = (disp_info[disp_num].fatbits_dialog.height - 
	     disp_info[disp_num].fatbits_dialog.selected_region.height);

    x = (real_x + (FATBITS_MAG/2))/ FATBITS_MAG;
    y = (real_y + (FATBITS_MAG/2))/ FATBITS_MAG;

    bound_num(&x, 0, fat_max_x);
    bound_num(&y, 0, fat_max_y);

    disp_info[disp_num].fatbits_dialog.selected_region.cur_x = x;
    disp_info[disp_num].fatbits_dialog.selected_region.cur_y = y;
    disp_info[disp_num].fatbits_dialog.selected_region.cur_real_x = 
					 x*FATBITS_MAG;
    disp_info[disp_num].fatbits_dialog.selected_region.cur_real_y = 
					 y*FATBITS_MAG;

    XMoveWindow(disp_info[disp_num].disp, 
       XtWindow(disp_info[disp_num].fatbits_dialog.selected_region.win),
       x*FATBITS_MAG, y*FATBITS_MAG);
    XFlush(disp_info[disp_num].disp);
}


/*
 * place_fatbits_selection_permanently - this function places the selection 
 *              back onto the screen, and sets all variables accordingly.
 */
place_fatbits_selection_permanently(disp_num)
int disp_num;
{
    int i, j, x, y, width, height;
    Window dummy;
    unsigned long the_pixel;

    disp_info[disp_num].fatbits_dialog.selected_region.region_is_selected =
								     FALSE;

    x = disp_info[disp_num].fatbits_dialog.selected_region.cur_x;
    y = disp_info[disp_num].fatbits_dialog.selected_region.cur_y;
    width = disp_info[disp_num].fatbits_dialog.selected_region.width;
    height = disp_info[disp_num].fatbits_dialog.selected_region.height;

    /*
     * first update the fatbits drawing area appropriately
     */
    XCopyArea(disp_info[disp_num].disp, 
	      disp_info[disp_num].fatbits_dialog.selected_region.backing_store,
	      disp_info[disp_num].fatbits_dialog.backing_store,
              disp_info[disp_num].win_gc[disp_num], 
	      0, 0, width*FATBITS_MAG, height*FATBITS_MAG, 
	      x*FATBITS_MAG, y*FATBITS_MAG);

    XCopyArea(disp_info[disp_num].disp, 
	      disp_info[disp_num].fatbits_dialog.selected_region.backing_store,
	      XtWindow(disp_info[disp_num].fatbits_dialog.draw_area),
              disp_info[disp_num].win_gc[disp_num], 
	      0, 0, width*FATBITS_MAG, height*FATBITS_MAG, 
	      x*FATBITS_MAG, y*FATBITS_MAG);

    /*
     * now copy the selection image on the client side
     */
    for (j=0; j < height; j++)
    {
        for (i=0; i < width; i++)
        {
            the_pixel = XGetPixel(
	        disp_info[disp_num].fatbits_dialog.selected_region.local_image,
                i, j);

            XPutPixel(disp_info[disp_num].fatbits_dialog.local_image, 
	              i + x, j + y, the_pixel);
        }
    }

    XtUnmapWidget(disp_info[disp_num].fatbits_dialog.selected_region.win);
    XFlush(disp_info[disp_num].disp);
}


/*
 * fatbits_selection_area_exposeCB - this function gets called when an 
 *               expose event occurs in the selection area.
 */
fatbits_selection_area_exposeCB(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int disp_num;
    XmDrawingAreaCallbackStruct *cb_struct;
    XGCValues gcvalues;
    XExposeEvent *the_event;

    cb_struct = (XmDrawingAreaCallbackStruct *) call_data;
    the_event = (XExposeEvent *) cb_struct->event;
    disp_num = ((int) client_data) >> 8;

    if (the_event->type != Expose)
	printf("ERROR: in select_area_exposeCB.\n");

    XCopyArea(disp_info[disp_num].disp, 
	 disp_info[disp_num].fatbits_dialog.selected_region.backing_store,
         XtWindow(disp_info[disp_num].fatbits_dialog.selected_region.win),
	 disp_info[disp_num].win_gc[disp_num], 0, 0, 
	 disp_info[disp_num].fatbits_dialog.selected_region.width*FATBITS_MAG, 
	 disp_info[disp_num].selected_region.height*FATBITS_MAG, 0, 0);

    /*
     * now draw the dashed line around the outside
     */
    draw_fatbits_selection_outline(disp_num);
}


/*
 * fatbits_clear_region - this function is called when "backspace" is hit
 *                while editing the fatbits region.
 */
fatbits_clear_region(disp_num)
int disp_num;
{
    int i, x, y, width, height;

    if (disp_info[disp_num].fatbits_dialog.selected_region.region_is_selected ==
	FALSE)
    {
        popup_error_dialog(disp_num, 
			   "ERROR: in fatbits clear region not selected");
    }

    /*
     * The user wants to clear this selected region.  Do it.
     */
    disp_info[disp_num].fatbits_dialog.selected_region.region_is_selected=FALSE;
    XtUnmapWidget(disp_info[disp_num].fatbits_dialog.selected_region.win);
}


/*
 * prepare_fatbits_region - this function prepares the region for an upcoming 
 *                  XCopyArea.  It resizes it, and moves it to the correct 
 *                  location, and allocates a pixmap for it.  The
 *                  "x" and "y" are relative to the disp_num's fat drawing area.
 */
prepare_fatbits_region(disp_num, x, y, width, height)
int disp_num, x, y, width, height;
{
    if (disp_info[disp_num].fatbits_dialog.selected_region.backing_store!= NULL)
    {
        XFreePixmap(disp_info[disp_num].disp,
            disp_info[disp_num].fatbits_dialog.selected_region.backing_store);
    }

    disp_info[disp_num].fatbits_dialog.selected_region.backing_store =
                        XCreatePixmap(disp_info[disp_num].disp,
                               disp_info[disp_num].draw_win_id, 
			       width*FATBITS_MAG, height*FATBITS_MAG,
                               DefaultDepth(disp_info[disp_num].disp,
                               DefaultScreen(disp_info[disp_num].disp)));
    XFillRectangle(disp_info[disp_num].disp, 
	  disp_info[disp_num].fatbits_dialog.selected_region.backing_store,
	  disp_info[disp_num].backing_store_gc, 0, 0,
	  width*FATBITS_MAG, height*FATBITS_MAG);

    XMoveResizeWindow(disp_info[disp_num].disp,
          XtWindow(disp_info[disp_num].fatbits_dialog.selected_region.win),
          x*FATBITS_MAG, y*FATBITS_MAG, width*FATBITS_MAG, height*FATBITS_MAG);

    XRaiseWindow(disp_info[disp_num].disp,
          XtWindow(disp_info[disp_num].fatbits_dialog.selected_region.win));
}


/*
 * create_fatbits_region - this function creates and initializes all the
 *                           data structures for the selected regions.
 *                           A selected region is the thing that you would
 *                           "cut" or "paste".
 */
create_fatbits_region(disp_num)
int disp_num;
{
    int n;
    Arg args[MAX_ARGS];

    n = 0;
    XtSetArg(args[n], XmNwidth, 10); n++;
    XtSetArg(args[n], XmNheight, 10); n++;
    XtSetArg(args[n], XmNmappedWhenManaged, False); n++;
    XtSetArg(args[n], XmNmarginWidth, 0); n++;
    XtSetArg(args[n], XmNmarginHeight, 0); n++;
    XtSetArg(args[n], XmNshadowThickness, 0); n++;
    disp_info[disp_num].fatbits_dialog.selected_region.win =
              XmCreateDrawingArea(disp_info[disp_num].fatbits_dialog.draw_area,
                                  "fatbits_selection_region", args, n);
    XtManageChild(disp_info[disp_num].fatbits_dialog.selected_region.win);
    XtRealizeWidget(disp_info[disp_num].fatbits_dialog.selected_region.win);
    XDefineCursor(disp_info[disp_num].disp,
	  XtWindow(disp_info[disp_num].fatbits_dialog.selected_region.win),
	  XCreateFontCursor(disp_info[disp_num].disp, XC_left_ptr));

    XtAddCallback(disp_info[disp_num].fatbits_dialog.selected_region.win,
		  XmNexposeCallback, (XtCallbackProc)fatbits_selection_area_exposeCB, 
		  (XtPointer)(disp_num << 8));
    XtAddCallback(disp_info[disp_num].fatbits_dialog.selected_region.win,
		  XmNinputCallback, (XtCallbackProc)fatbits_selection_area_inputCB, 
		  (XtPointer)(disp_num << 8));

}


