/*
 * This File: wshelp.c
 *
 * This file contains the help dialog creation and handling functions for 
 * wscrawl.
 */

#include "wscrawl.h"

static XmStringCharSet charset = (XmStringCharSet) XmSTRING_DEFAULT_CHARSET;
extern struct global_values_struct global_values;
extern struct disp_info_struct disp_info[MAX_NUM_DISPS];
Widget create_help_win();

/*
 * helpCB - this function handles the callbacks from the help dialog 
 */
helpCB(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int disp_num, which_action;
    int n;
    Arg args[MAX_ARGS];

    disp_num = ((int) client_data) >> 8;
    which_action = ((int) client_data) & 0x00FF;

    switch(which_action)
    {
	case OK:
	    XtUnmanageChild(disp_info[disp_num].help_dialog);
            jiggle_dialog(disp_info[disp_num].help_dialog);
            break;
	case ACTIVATE:
	    if (disp_info[disp_num].help_dialog == NULL)
	    {
		disp_info[disp_num].help_dialog = 
		       create_help_win(disp_num, disp_info[disp_num].app_shell);
	    }
            jiggle_dialog(disp_info[disp_num].help_dialog);
            XtManageChild(disp_info[disp_num].help_dialog);
            break;
	case ARM:
            n = 0;
            XtSetArg(args[n], XmNshadowThickness, 2); n++;
            XtSetArg(args[n], XmNmarginHeight, 2); n++;
            XtSetArg(args[n], XmNmarginWidth, 2); n++;
            XtSetValues(disp_info[disp_num].help_menu_bar_button, args, n);
            break;
	case DISARM:
            n = 0;
            XtSetArg(args[n], XmNshadowThickness, 0); n++;
            XtSetArg(args[n], XmNmarginHeight, 4); n++;
            XtSetArg(args[n], XmNmarginWidth, 4); n++;
            XtSetValues(disp_info[disp_num].help_menu_bar_button, args, n);
            break;
        default:
            popup_error_dialog(disp_num, "ERROR: unknown action in helpCB.");
	    break;
    }
}


/*
 * create_help_win - this function creates the help window
 */
Widget create_help_win(disp_num, parent)
int disp_num;
Widget parent;
{
    Widget the_dialog;
    Widget the_separator;
    Widget the_frame, the_form;
    Widget swindow;
    Widget the_text_widget;
    Widget title;
    Widget close_button;
    Widget message_box;
    Arg args[MAX_ARGS];
    register int n;

    static char message[500];       /* title text */
    XmString title_string = NULL;
    XmString message_string = NULL;
    XmString button_string = NULL;

    n = 0;
    XtSetArg(args[n], XmNautoUnmanage, False); n++;
    the_dialog = XmCreateFormDialog(disp_info[disp_num].app_shell,
                           "Help_Dialog", args, n);

    n = 0;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    the_frame = XmCreateFrame(the_dialog, "dialog_frame", args, n);
    XtManageChild(the_frame);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNwidth, 600); n++;
    the_form = XmCreateForm(the_frame, "dialog_form", args, n);
    XtManageChild(the_form);

    /*
     * Generate the message
     */
    sprintf(message, "%s", 
"\
                               Wscrawl 2.01 \n\
\n\
                           by Brian Wilson (brianw@apple.com) \n\
                              version: 10/26/92 \0");

    message_string = XmStringCreateLtoR(message, charset);
    n = 0;
    XtSetArg(args[n], XmNx, 15); n++;
    XtSetArg(args[n], XmNy, 15); n++;
    XtSetArg(args[n], XmNlabelString, message_string); n++;
    title = XmCreateLabelGadget(the_form, "title", args, n);
    XtManageChild(title);

    /*
     * Create the text widget containing the manual
     */
    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 10); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 10); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, title); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomOffset, 50); n++;
    XtSetArg(args[n], XmNheight, 200); n++;
    XtSetArg(args[n], XmNscrollingPolicy, XmAUTOMATIC); n++;
    XtSetArg(args[n], XmNeditable, False); n++;
    XtSetArg(args[n], XmNeditMode, XmMULTI_LINE_EDIT); n++;
    the_text_widget = XmCreateScrolledText(the_form, "manual", args, n);
    XtManageChild(the_text_widget);
    create_manual(the_text_widget);

    /*
     * Create the separator and the close button 
     */
    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 0); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 0); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, the_text_widget); n++;
    XtSetArg(args[n], XmNtopOffset, 10); n++;
    the_separator = XmCreateSeparatorGadget(the_form, "separator", args, n);
    XtManageChild(the_separator);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNleftPosition, 45); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNrightPosition, 55); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, the_text_widget); n++;
    XtSetArg(args[n], XmNtopOffset, 20); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
    close_button = XmCreatePushButtonGadget(the_form, "Close", args, n);
    XtManageChild(close_button);
    XtAddCallback(close_button, XmNactivateCallback,(XtCallbackProc)helpCB, (XtPointer)(disp_num << 8 | OK));

    set_dialog_wm_name(the_dialog, "Help");
    return(the_dialog);
}


/*
 * create_manual - this function fills the text widget with the manual text.
 */
create_manual(w)
Widget w;
{
    XmTextInsert(w, XmTextGetLastPosition(w), 
"\n\
                           TABLE OF CONTENTS\n\
\n\
                           1. Introduction\n\
                           2. Features and Shortcuts\n\
                           3. Wscrawl Resources\n\
                           4. Technical Information\n\
                           5. Credits\n\
\n\n\n");


    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
		  SECTION #1: Introduction\n\
\n\
");
	     
    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    The word 'wscrawl' stands for window-scrawl.  You can think of wscrawl\n\
as a paint program shared by any number of people at the same time.  In a\n\
normal use of wscrawl, a separate window is opened on each participant's \n\
display.  From that point onward, each participant sees the actions and\n\
ideas of every other participant as they occur.  Each individual may simply\n\
watch, or participate at any moment.  Any individual may exit out of the\n\
session at any time without affecting the other participants.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    Wscrawl 2.0 is public domain, so feel free to use/change/modify it in\n\
any way, just as long as you don't charge money for it and this message\n\
appears on any new versions.  Send bug reports and enhancement requests to\n\
the email address above and include the version date in your email.  I am\n\
always interested in comments, especially how you currently use wscrawl and\n\
what would make it more useful to you.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    Wscrawl 2.0 is not an Apple Computer product in any way.  Apple is\n\
not responsible for wscrawl nor any effect wscrawl might have, nor makes \n\
any claim to ownership of wscrawl.\n\n\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
		  SECTION #2: Features and Shortcuts\n\
\n\
");
	     
    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    The following is an overview of the functionality offered in wscrawl 2.0.\n\
Always remember that only the left button of the mouse has any effect in \n\
wscrawl.  There should never be any confusion about which mouse button does \n\
what.  The left mouse button does it all, and the other two do nothing.\n\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
THE TOOL PALETTE\n\
\n\
    The twelve buttons with pictures in them you see on the left side of \n\
wscrawl make up the 'tool palette'.  Clicking on one of the tools selects\n\
that tool for use in the drawing area.  Also, there is a 'shortcut' \n\
associated with each tool.  Double-click on the tool to invoke the shortcut.\n\
The shortcuts can also be invoked through the 'Options' menu.\n\
\n\
    TELEPOINTER:  The upper left corner of the tool palette holds the \n\
                 telepointer tool.  This tool allows you to point at places\n\
                 on the drawing area.  All members of the wscrawl session\n\
                 see your telepointer when you press and hold the mouse \n\
                 button.  The shortcut associated with this tool names\n\
                 the telepointer.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    SELECTION:  The dotted box tool directly to the right of the telepointer\n\
                 is the selection tool.  When you have selected this tool\n\
                 you can click and drag to select a region in the drawing \n\
                 area.  At that point you can drag the region by clicking and\n\
                 dragging it.  You can also save the region to disk, or cut\n\
                 the region out for pasting later.  The shortcut associated \n\
                 with this tool allows you to select regions of the screen \n\
                 that are outside your wscrawl program and drag them into \n\
                 your drawing area.  To do one of these 'super-selections'\n\
                 double-click the selection tool button, then click on the\n\
                 'Super-Select' button in the dialog that appears, then\n\
                 click and drag OUTSIDE your wscrawl program.  For example,\n\
                 on top of some other program like an 'xterm'.  After you\n\
                 release the mouse, wait until a shimmering selection box \n\
                 appears, then drag the selection into your wscrawl window.\n\
\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    PAINTBRUSH:  The paintbrush allows you to paint using different brush\n\
                 shapes.  The shortcut associated with this tool brings up\n\
                 a set of brush shapes to choose from.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    PENCIL:  The pencil is just like a very thin paintbrush.  The \n\
                 shortcut associated with this tool brings up a Fatbits\n\
                 Editor on the current selection.  To use the Fatbits editor,\n\
                 select a very small region on the drawing window and double\n\
                 click on the pencil.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    SPRAYCAN:  The spraycan tool allows you to spraypaint.  There is no\n\
                 shortcut associated with this tool.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    LETTER A:  The letter 'A' tool allows you to type characters.  Select\n\
                 this tool, then click to locate the insertion point in the\n\
                 drawing area, then type on the keyboard.  There is no \n\
                 shortcut on this tool.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    LINE:     The line tool allows you to draw straight lines.  The \n\
                 shortcut is to set the width of the line.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    ERASER:  The eraser allows you to erase parts of your drawing. Select\n\
                 the eraser tool, then click and drag in the drawing window \n\
                 to erase.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    SHAPE TOOLS:  The last four tools are shape tools. They allow you to\n\
                 draw various filled and hollow shapes.  The shortcut on the\n\
                 outline tools chooses the outline width.\n\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
THE MENUS\n\
\n\
    Most of the menus are self explanatory.  For example, the 'Colors' menu\n\
allows you to pick a color to paint with, and the 'Fonts' menu allows you\n\
to pick a font to type in.  Notice that any menu item that ends in three \n\
dots such as 'Add Display...' will pop up a dialog box.  Below are some \n\
explanations of some of the not-so-straight-forward menu items.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    Add Display...  This is how you add new people to your wscrawl session.\n\
                    Select this, then type a display name in the text box\n\
                    and then click on the 'OK' button.  The display name \n\
                    you type should look like 'machine_name:0.0'\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    Join Existing Session...  This enables you to join a wscrawl session\n\
                    that other people are holding.  The net effect is as if\n\
                    one of those people had selected their 'Add Display...'\n\
                    and typed in your display name.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    Cut/Copy/Paste...  These all operate on selections.  Select a region, \n\
                   then try selecting first 'Cut', then 'Paste', then 'Paste'\n\
                   again.  You may paste this selection several times.  \n\
                   NOTE: If you hold down the <cntl> key then drag a \n\
		   selection it will drag a copy away.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    Paste X-Selection...  If there is a program that copies graphics into \n\
                   a standard X-Selection, this allows you to paste it into\n\
                   wscrawl.  You must be in text insert mode to paste text\n\
                   and in any other insert mode to paste graphics.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    Scrollbar Settings...  If several people have their scrollbars 'locked \n\
                   together', then if one person scrolls, all the members of\n\
                   the session are scrolled along with that person.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    Custom Color...  This dialog allows you to create your own color hue.\n\
                   Use the R, G, and B sliders and watch the example color.\n\
                   When you wish to use the example color, click on the\n\
                   'Set Color' button, then draw in the drawing window.\n\
                   Additionally, if you see a color you wish to 'grab' that\n\
                   is anywhere on the display even OUTSIDE the wscrawl\n\
                   window, first click on the 'Grab Color' button, then\n\
                   click on a pixel displaying the color you wish.  The RGB\n\
                   sliders will slide themselves to the correct positions.\n\
\n\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
		  SECTION #3: Wscrawl Resources\n\
\n\
");
	     
    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    Many wscrawl settings can be changed through menu selections.  A feature\n\
of wscrawl is that it remembers all of your settings even if you quit wscrawl\n\
and start it again.  However, if you ever shut down your X-Server, wscrawl \n\
will forget all your settings.  Therefore, some items can be set through\n\
resources.  To use resources, place the indicated resource line in your \n\
'.Xdefaults' file and restart your X-Windows server.  Below are some useful\n\
resources.  Notice that some are duplicates to indicate possible values.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
     wscrawl.virtualWhiteboardWidth:                 1500\n\
     wscrawl.virtualWhiteboardHeight:               1500\n\
     wscrawl.interfaceLanguage:                       English\n\
     wscrawl.interfaceLanguage:                       German\n\
     wscrawl.telePointerName:                          default\n\
     wscrawl.telePointerName:                          Bob\n\
     wscrawl.syncScrollbars:                            False\n\
     wscrawl.syncScrollbars:                            True\n\
     wscrawl.continuousTelePointer:                  False\n\
     wscrawl.iconForeground:                           black\n\
     wscrawl.iconBackground:                          LightSteelBlue\n\
     wscrawl.showPointerCoordinates:              False\n\
     wscrawl*foreground:                                  black\n\
     wscrawl*background:                                 LightSteelBlue\n\
     wscrawl*swindow*foreground:                   LightSteelBlue\n\
     wscrawl*swindow*draw_area_frame*background: white\n\
     wscrawl*draw_area_frame*foreground:       blue\n\
     wscrawl*draw_area_frame*background:      white\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
     wscrawl*keyboardFocusPolicy:                   pointer\n\
     wscrawl*borderWidth:                                0\n\
     wscrawl*fontList:                                      variable\n\n\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
		  SECTION #4: Technical Information\n\
\n\
");
	     
    XmTextInsert(w, XmTextGetLastPosition(w), 
"\n\
    Wscrawl 2.0 is descended from a program I wrote called 'scrawl'.  Scrawl\n\
is a program that allows you to draw directly on other peoples' root \n\
windows without the victims knowing where it is coming from.  Right away \n\
people began 'scrawling' at each other to communicate, so to see if it\n\
was possible, I began wscrawl 1.0.  This was written entirely in Xlib, and \n\
is basically a scrawl session between two people in a window.  Flaws in the\n\
basic design of wscrawl 1.0 demanded a rewrite before I could go any further,\n\
so I took the opportunity to rewrite it in the Motif toolkit.  The result is\n\
the wscrawl 2.0 you are now using.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    Wscrawl 2.0 is part of a growing number of applications called real\n\
time, or 'synchronous', cooperative tools.  Researchers in the field of\n\
Computer Supported Cooperative Work, or 'CSCW', are examining how to build\n\
these tools efficiently, and more importantly how these tools can be of use\n\
to end users.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    Wscrawl 2.0 is unique in that it is built entirely on industry standard\n\
X-Windows with off-the-shelf industry standard Motif 1.1.  There are no \n\
custom communication protocols specified, except for the very simple 'Join\n\
Existing Session' feature which uses TCP/IP and a couple of packet formats.\n\
It is surprisingly simple to build a real-time cooperative application in \n\
this manner, and I would claim that after the general architecture of the \n\
application is set up, it only takes one and a half times as much effort and\n\
time to add some piece of functionality to the application than if you were\n\
adding that same piece of functionality to a non-cooperative application.\n\
With a decentralized architecture where you must specify a new custom \n\
communication protocol, thus throwing out all the work MIT did in creating\n\
the perfectly useable X-Windows protocol, it takes vastly more effort to \n\
add the smallest new piece of functionality to your application.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    The idea of any X-Windows client is that it runs on a host computer\n\
and takes input from and displays graphics on a display running an X-Server \n\
anywhere on the network.  Wscrawl simply takes input from and displays \n\
graphics on several displays at once.  Any X-Windows client COULD do this, \n\
but most choose not to because it normally doesn't make any sense.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    The one shortfall of using Motif 1.1 in this manner was in the menu \n\
convenience functions.  When you pull down a Motif menu, Motif no longer\n\
uses the main event loop, but instead it uses a small menu event loop.  If\n\
I had used the normal Motif menus, any member of the wscrawl session could\n\
'hang' the entire session by pulling down a menu.  This forced me to \n\
implement my own menus from scratch to avoid this problem.  In every\n\
other case such as dialog boxes, pushbuttons, and text-edit widgets Motif\n\
used the main event loop giving the desired illusion of multi-tasking to the\n\
members of the session.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    Wscrawl's strength of a centralized architecture is also its greatest \n\
weakness.  Although the single process hands off all graphics requests\n\
to the X-Servers, there is still the bottleneck of the one process.\n\
Although wscrawl has perfectly adequate performance on a local network,\n\
it runs unacceptably slow on current wide area networks.  When I used \n\
wscrawl between the United States and Germany, network traffic caused\n\
'hangs' for as long as 30 seconds at times.  A decentralized architecture\n\
could buffer the user from such hangs, still allowing him or her to draw,\n\
select menus, etc.  Then when the network clears itself, the buffered \n\
requests could be pushed through.  At that point the extremely complex\n\
problem of how to resolve race conditions becomes the programmer's number\n\
one concern.  In wscrawl, all race conditions are trivially eliminated by the\n\
centralized architecture.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    If you are interested in a more specific discussion of wscrawl 2.0's \n\
architecture, I have written a rough technical paper that goes into much\n\
greater depth.  Email me to get an ascii text version.\n\n\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
		  SECTION #5: Credits\n\
\n\
");
	     
    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    Wscrawl would not have been possible without the help, comments, \n\
contributed code and encouragement from more people than I can list\n\
here.  The following deserve special mention:\n\n");
	     
    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    Harry Phinney for answering ten bazillion X-Windows low level technical\n\
questions and as far as I can tell never getting one wrong.  Pam Raby my\n\
Motif guru.  Mark Cook for his xwd file I/O functions.  Craig Durland, \n\
Keith Marchington, Charlie Amacher for technical help.  Sven Dack, Klaus\n\
Reichenberger for German translations and German beer.  Randy Wilson, \n\
trans-Atlantic test base.\n\n");

    XmTextInsert(w, XmTextGetLastPosition(w), 
"\
    Finally, Dr. Norbert Streitz and the German Institute GMD-IPSI who \n\
supported the development of much of version 2.0.  Thank you for providing\n\
an excellent environment to do research in.\n");
}
