/*
 * This File: wsinit.c
 *
 * This file contains the initialization functions to get the windows up and
 * going.
 */

#include "wscrawl.h"
#include "languages.h"

extern int num_disps;
extern XtAppContext appCon;
extern struct disp_info_struct disp_info[MAX_NUM_DISPS];
extern struct global_values_struct global_values;
static XmStringCharSet charset = (XmStringCharSet) XmSTRING_DEFAULT_CHARSET;

Widget create_application();
Widget create_menus();
Widget create_a_menu();
extern int draw_area_inputCB();
extern int draw_area_exposeCB();
extern int status_area_exposeCB();
extern int menu_armCB();
extern int menu_disarmCB();
extern int menu_pointer_moveCB();
extern int tool_activateCB();
extern int helpCB();
extern int selection_area_inputCB();
extern int wm_kill_window();
extern int selection_area_exposeCB();
extern XtEventHandler pointer_moveCB();
extern XtEventHandler selection_pointer_moveCB();
extern XtEventHandler handle_selection_keypressCB();
extern XtEventHandler handle_wscrawl_window_enterCB();
extern XtEventHandler handle_enter_or_leave_draw_area();
extern Widget create_brush_dialog();
extern Widget create_color_picker_dialog();

extern Widget create_fatbits_dialog();
extern Widget create_eraser_dialog();
extern Widget create_dirty_add_display_dialog();
extern Widget create_scrollbar_dialog();
extern Widget create_select_dialog();
extern Widget create_read_file_dialog();
extern Widget create_save_file_dialog();
extern Widget create_info_dialog();
extern Widget create_error_dialog();
extern Widget create_help_win();
extern Widget create_line_dialog();
extern Widget create_square_dialog();
extern Widget create_oval_dialog();
extern Widget create_change_pointer_dialog();
extern Widget create_mouse_dialog();


/*
 * create_application - this function creates the applications window's 
 *                      children
 */
Widget create_application(parent, disp_num)
Widget parent;
int disp_num;
{
    Widget main_window;
    Widget menu_bar;
    Widget frame;
    Widget command_form, command_frame, command_frame2;
    Widget form, tool_form1, tool_form, big_tool_form;
    Widget state_frame, state_form;
    Widget draw_area_frame;
    Widget tool_frame;
    Widget spacer;
    Widget swinframe;
    Widget button1, button2;
    Arg args[MAX_ARGS];
    int n;
    XmString label_string;
    Pixmap button_pixmap;
    Atom xa_WM_DELETE_WINDOW;

    /*
     * Create MainWindow.
     */
    n = 0;
    main_window = XmCreateMainWindow(parent, "main_win", args, n);
    XtManageChild(main_window);
    XtAddEventHandler(main_window, EnterWindowMask | LeaveWindowMask, False, 
		   (XtEventHandler)handle_wscrawl_window_enterCB, (XtPointer)disp_num);
    
    disp_info[disp_num].main_window = main_window;

    menu_bar = create_menus(main_window, parent, disp_num); /* create menus */

    /*
     * Create Frame MainWindow and ScrolledWindow in Frame, and scroll bars
     * in the scrolled window
     */
    n = 0;
    form = XmCreateForm(main_window, "work_area_form", args, n); 
    XtManageChild(form);

    n = 0;
    XtSetArg(args[n], XmNmarginWidth, 4); n++;
    XtSetArg(args[n], XmNmarginHeight, 4); n++;
    XtSetArg(args[n], XmNshadowThickness, 2); n++;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    tool_frame = XmCreateFrame(form, "tool_frame", args, n);
    XtManageChild(tool_frame);

    n = 0;
    XtSetArg(args[n], XmNwidth, 64); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    tool_form = XmCreateForm(tool_frame, "tool_form", args, n); 
    XtManageChild(tool_form);

    /*
     * create all the pushbuttons in the tool palette 
     */
    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNhighlightThickness, 0); n++;
    XtSetArg(args[n], XmNwidth, 32); n++;
    XtSetArg(args[n], XmNheight, 32); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    disp_info[disp_num].tool[0] = XmCreatePushButton(tool_form,"Pointer",
						     args,n);
    XtManageChild(disp_info[disp_num].tool[0]);
    XtAddCallback(disp_info[disp_num].tool[0], XmNactivateCallback,
		  (XtCallbackProc)tool_activateCB, (XtPointer)(disp_num << 8 | POINTER_BUTTON));

    n = 0;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftWidget, disp_info[disp_num].tool[0]); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNwidth, 32); n++;
    XtSetArg(args[n], XmNheight, 32); n++;
    XtSetArg(args[n], XmNhighlightThickness, 0); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    disp_info[disp_num].tool[1] =XmCreatePushButton(tool_form,"Selection",
						    args,n); 
    XtManageChild(disp_info[disp_num].tool[1]);
    XtAddCallback(disp_info[disp_num].tool[1], XmNactivateCallback,
		  (XtCallbackProc)tool_activateCB, (XtPointer)(disp_num << 8 | SELECTION_BUTTON));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, disp_info[disp_num].tool[0]); n++;
    XtSetArg(args[n], XmNwidth, 32); n++;
    XtSetArg(args[n], XmNheight, 32); n++;
    XtSetArg(args[n], XmNhighlightThickness, 0); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    disp_info[disp_num].tool[2] = XmCreatePushButton(tool_form,"Paint",args,n);
    XtManageChild(disp_info[disp_num].tool[2]);
    XtAddCallback(disp_info[disp_num].tool[2], XmNactivateCallback,
		  (XtCallbackProc)tool_activateCB, (XtPointer)(disp_num << 8 | PAINTBRUSH_BUTTON));

    n = 0;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftWidget, disp_info[disp_num].tool[2]); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, disp_info[disp_num].tool[1]); n++;
    XtSetArg(args[n], XmNwidth, 32); n++;
    XtSetArg(args[n], XmNheight, 32); n++;
    XtSetArg(args[n], XmNhighlightThickness, 0); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    disp_info[disp_num].tool[3] =XmCreatePushButton(tool_form,"Pencil",args,n); 
    XtManageChild(disp_info[disp_num].tool[3]);
    XtAddCallback(disp_info[disp_num].tool[3], XmNactivateCallback,
		  (XtCallbackProc)tool_activateCB, (XtPointer)(disp_num << 8 | PENCIL_BUTTON));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, disp_info[disp_num].tool[2]); n++;
    XtSetArg(args[n], XmNwidth, 32); n++;
    XtSetArg(args[n], XmNheight, 32); n++;
    XtSetArg(args[n], XmNhighlightThickness, 0); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    disp_info[disp_num].tool[4] = XmCreatePushButton(tool_form,"spraycan",
						     args,n);
    XtManageChild(disp_info[disp_num].tool[4]);
    XtAddCallback(disp_info[disp_num].tool[4], XmNactivateCallback,
		  (XtCallbackProc)tool_activateCB, (XtPointer)(disp_num << 8 | SPRAYCAN_BUTTON));

    n = 0;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftWidget, disp_info[disp_num].tool[4]); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, disp_info[disp_num].tool[3]); n++;
    XtSetArg(args[n], XmNwidth, 32); n++;
    XtSetArg(args[n], XmNheight, 32); n++;
    XtSetArg(args[n], XmNhighlightThickness, 0); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    disp_info[disp_num].tool[5] = XmCreatePushButton(tool_form,"Type",args,n); 
    XtManageChild(disp_info[disp_num].tool[5]);
    XtAddCallback(disp_info[disp_num].tool[5], XmNactivateCallback,
		  (XtCallbackProc)tool_activateCB, (XtPointer)(disp_num << 8 | TYPE_BUTTON));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, disp_info[disp_num].tool[4]); n++;
    XtSetArg(args[n], XmNwidth, 32); n++;
    XtSetArg(args[n], XmNheight, 32); n++;
    XtSetArg(args[n], XmNhighlightThickness, 0); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    disp_info[disp_num].tool[6]=XmCreatePushButton(tool_form,"linesym",args,n);
    XtManageChild(disp_info[disp_num].tool[6]);
    XtAddCallback(disp_info[disp_num].tool[6], XmNactivateCallback,
		  (XtCallbackProc)tool_activateCB, (XtPointer)(disp_num << 8 | LINE_BUTTON));

    n = 0;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftWidget, disp_info[disp_num].tool[6]); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, disp_info[disp_num].tool[5]); n++;
    XtSetArg(args[n], XmNwidth, 32); n++;
    XtSetArg(args[n], XmNheight, 32); n++;
    XtSetArg(args[n], XmNhighlightThickness, 0); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    disp_info[disp_num].tool[7] =XmCreatePushButton(tool_form,"eraser",args,n);
    XtManageChild(disp_info[disp_num].tool[7]);
    XtAddCallback(disp_info[disp_num].tool[7], XmNactivateCallback,
		  (XtCallbackProc)tool_activateCB, (XtPointer)(disp_num << 8 | ERASER_BUTTON));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, disp_info[disp_num].tool[6]); n++;
    XtSetArg(args[n], XmNwidth, 32); n++;
    XtSetArg(args[n], XmNheight, 32); n++;
    XtSetArg(args[n], XmNhighlightThickness, 0); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    disp_info[disp_num].tool[8]=XmCreatePushButton(tool_form,"square",args,n);
    XtManageChild(disp_info[disp_num].tool[8]);
    XtAddCallback(disp_info[disp_num].tool[8], XmNactivateCallback,
		  (XtCallbackProc)tool_activateCB, (XtPointer)(disp_num << 8 | OUTLINESQUARE_BUTTON));

    n = 0;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftWidget, disp_info[disp_num].tool[8]); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, disp_info[disp_num].tool[7]); n++;
    XtSetArg(args[n], XmNwidth, 32); n++;
    XtSetArg(args[n], XmNheight, 32); n++;
    XtSetArg(args[n], XmNhighlightThickness, 0); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    disp_info[disp_num].tool[9] =XmCreatePushButton(tool_form,"fillsquare",
					args,n);
    XtManageChild(disp_info[disp_num].tool[9]);
    XtAddCallback(disp_info[disp_num].tool[9], XmNactivateCallback,
		  (XtCallbackProc)tool_activateCB, (XtPointer)(disp_num << 8 | FILLSQUARE_BUTTON));

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, disp_info[disp_num].tool[8]); n++;
    XtSetArg(args[n], XmNwidth, 32); n++;
    XtSetArg(args[n], XmNheight, 32); n++;
    XtSetArg(args[n], XmNhighlightThickness, 0); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    disp_info[disp_num].tool[10]=XmCreatePushButton(tool_form,"outlineoval",
			    args,n);
    XtManageChild(disp_info[disp_num].tool[10]);
    XtAddCallback(disp_info[disp_num].tool[10], XmNactivateCallback,
		  (XtCallbackProc)tool_activateCB, (XtPointer)(disp_num << 8 | OUTLINEOVAL_BUTTON));

    n = 0;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftWidget, disp_info[disp_num].tool[10]); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, disp_info[disp_num].tool[9]); n++;
    XtSetArg(args[n], XmNwidth, 32); n++;
    XtSetArg(args[n], XmNheight, 32); n++;
    XtSetArg(args[n], XmNhighlightThickness, 0); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    disp_info[disp_num].tool[11] =XmCreatePushButton(tool_form,"filloval",
					 args,n);
    XtManageChild(disp_info[disp_num].tool[11]);
    XtAddCallback(disp_info[disp_num].tool[11], XmNactivateCallback,
		  (XtCallbackProc)tool_activateCB, (XtPointer)(disp_num << 8 | FILLOVAL_BUTTON));

    /*
     * Create the scrolled window in a frame
     */
    n = 0;
    XtSetArg(args[n], XmNmarginWidth, 4); n++;
    XtSetArg(args[n], XmNmarginHeight, 4); n++;
    XtSetArg(args[n], XmNshadowThickness, 2); n++;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftWidget, tool_frame); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    swinframe = XmCreateFrame(form, "swinframe", args, n);
    XtManageChild(swinframe);

    n = 0;
    XtSetArg(args[n], XmNscrolledWindowMarginHeight, 2); n++;
    XtSetArg(args[n], XmNscrolledWindowMarginWidth, 2); n++;
    XtSetArg(args[n], XmNscrollBarDisplayPolicy, XmAS_NEEDED); n++;
    XtSetArg(args[n], XmNscrollingPolicy, XmAUTOMATIC); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    disp_info[disp_num].swindow = XmCreateScrolledWindow(swinframe, "swindow", 
	      args, n); 
    XtManageChild(disp_info[disp_num].swindow);

    /*
     * Create Drawing Area in a Frame in the Scrolled Window
     */
    n = 0;
    XtSetArg(args[n], XmNshadowThickness, 0); n++;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_IN); n++;
    draw_area_frame = XmCreateFrame(disp_info[disp_num].swindow, 
				    "draw_area_frame", args, n); 
    XtManageChild(draw_area_frame);

    n = 0;
    XtSetArg(args[n], XmNmarginWidth, 10); n++;
    XtSetArg(args[n], XmNmarginHeight, 10); n++; 
    XtSetArg(args[n], XmNshadowThickness, 0); n++;
    XtSetArg(args[n], XmNwidth, global_values.draw_area_width); n++;
    XtSetArg(args[n], XmNheight, global_values.draw_area_height); n++;
    XtSetArg(args[n], XmNresizePolicy, XmRESIZE_NONE); n++;
    disp_info[disp_num].draw_area = XmCreateDrawingArea(draw_area_frame, 
					 "draw_area", args, n); 
    XtManageChild(disp_info[disp_num].draw_area);
    XtAddCallback(disp_info[disp_num].draw_area, XmNexposeCallback, 
		  (XtCallbackProc)draw_area_exposeCB, (XtPointer)disp_num);
    XtAddCallback(disp_info[disp_num].draw_area, XmNinputCallback, 
		  (XtCallbackProc)draw_area_inputCB, (XtPointer)disp_num);
    XtAddEventHandler(disp_info[disp_num].draw_area,
		   EnterWindowMask | LeaveWindowMask, False, 
		   (XtEventHandler)handle_enter_or_leave_draw_area, (XtPointer)disp_num);
    XtAddEventHandler(disp_info[disp_num].draw_area, PointerMotionMask, False,
		   (XtEventHandler)pointer_moveCB, (XtPointer)disp_num);

    n = 0;
    XtSetArg(args[n], XmNworkWindow, draw_area_frame); n++;
    XtSetValues(disp_info[disp_num].swindow, args, n);

    /*
     * Create status window at the bottom (in the command window region)
     * in the scrolled window.  I'd sell my soul if I could
     * get this to appear at the bottom.  Right now it appears at the
     * top.
     */
    n = 0;
    command_form = XmCreateForm(main_window, "status_area_form", args, n); 
    XtManageChild(command_form);

    n = 0;
    XtSetArg(args[n], XmNmarginWidth, 4); n++;
    XtSetArg(args[n], XmNmarginHeight, 4); n++;
    XtSetArg(args[n], XmNshadowThickness, 2); n++;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNleftPosition, 0); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNrightPosition, 100); n++;
    command_frame = XmCreateFrame(command_form, "status_frame", args, n);
    XtManageChild(command_frame);

    n = 0;
    XtSetArg(args[n], XmNmarginWidth, 4); n++;
    XtSetArg(args[n], XmNmarginHeight, 4); n++;
    XtSetArg(args[n], XmNshadowThickness, 2); n++;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_IN); n++;
    command_frame2 = XmCreateFrame(command_frame, "status_frame", args, n);
    XtManageChild(command_frame2);

    n = 0;
    XtSetArg(args[n], XmNshadowThickness, 0); n++;
    disp_info[disp_num].status_area =
         XmCreateDrawingArea(command_frame2, "status_area", args, n);
    XtManageChild(disp_info[disp_num].status_area);
    XtAddCallback(disp_info[disp_num].status_area, XmNexposeCallback, 
		  (XtCallbackProc)status_area_exposeCB, (XtPointer)disp_num);

    /*
     * add the "state area" that is underneath the tool palette
     */
    n = 0;
    XtSetArg(args[n], XmNmarginWidth, 4); n++;
    XtSetArg(args[n], XmNmarginHeight, 4); n++;
    XtSetArg(args[n], XmNshadowThickness, 2); n++;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
    XtSetArg(args[n], XmNleftPosition, 0); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, tool_frame); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNrightWidget, swinframe); n++;
    state_frame = XmCreateFrame(form, "state_frame", args, n);
    XtManageChild(state_frame);

    n = 0;
    XtSetArg(args[n], XmNwidth, 80); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNrightWidget, swinframe); n++;
    state_form = XmCreateForm(state_frame, "state_form", args, n); 
    XtManageChild(state_form);

    n = 0;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 15); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 15); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 0); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
    disp_info[disp_num].pointer_x_coords_label = 
	XmCreateLabelGadget(state_form, "x=0", args, n); 

    n = 0;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 30); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftOffset, 15); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 0); n++;
    XtSetArg(args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
    disp_info[disp_num].pointer_y_coords_label = 
	XmCreateLabelGadget(state_form, "y=0", args, n); 

    /*
     * set it all together
     */
    XmMainWindowSetAreas(main_window, menu_bar, command_form, NULL, NULL, form);

    xa_WM_DELETE_WINDOW = XmInternAtom(disp_info[disp_num].disp, 
				       "WM_DELETE_WINDOW", False);
    XmAddWMProtocolCallback(parent, xa_WM_DELETE_WINDOW, wm_kill_window,
			    disp_num);

    /*
     * create some dialogs, or mark them as NOT created with NULL.
     */
    disp_info[disp_num].add_display_dialog = NULL;
    disp_info[disp_num].attract_attention_dialog.the_dialog = NULL;
    disp_info[disp_num].change_pointer_dialog = 
		create_change_pointer_dialog(disp_num);
    disp_info[disp_num].brush_dialog.the_dialog = create_brush_dialog(disp_num);
    disp_info[disp_num].join_session_dialog = NULL;
    disp_info[disp_num].language_dialog.the_dialog = NULL;
    disp_info[disp_num].mouse_dialog.the_dialog = create_mouse_dialog(disp_num);
    disp_info[disp_num].color_picker_dialog.the_dialog = 
              			create_color_picker_dialog(disp_num);
    disp_info[disp_num].fatbits_dialog.the_dialog =
				 create_fatbits_dialog(disp_num);
    disp_info[disp_num].line_dialog.the_dialog = NULL;
    disp_info[disp_num].square_dialog.the_dialog = NULL;
    disp_info[disp_num].oval_dialog.the_dialog = NULL;
    disp_info[disp_num].eraser_dialog.the_dialog = NULL;
    disp_info[disp_num].dirty_add_display_dialog.the_dialog =
				 create_dirty_add_display_dialog(disp_num);
    disp_info[disp_num].scrollbar_dialog.the_dialog = NULL;
    disp_info[disp_num].select_dialog.the_dialog = NULL;
    disp_info[disp_num].read_file_dialog.the_dialog = NULL;
    disp_info[disp_num].save_file_dialog.the_dialog = NULL;

    disp_info[disp_num].info_dialog = create_info_dialog(disp_num);
    disp_info[disp_num].error_dialog = create_error_dialog(disp_num);
    disp_info[disp_num].help_dialog = NULL;

    /*
     * return the result
     */
    return(main_window);
}


/*
 * create_menus - this function creates all the pull down menus and attaches
 *                the callbacks to them.  It returns the menu bar widget.
 */
Widget create_menus(main_window, parent, disp_num)
Widget main_window;
Widget parent;
int disp_num;
{
    int mn;
    int n, i, j;
    int max_width, width, height, max_item;
    char **menu_bar_labels;
    char *str_typ_return;
    char tmp_str[250];
    XrmValue val_ret;
    Arg args[MAX_ARGS];
    Widget menu_bar;
    Widget menu_pane;
    Widget cascade;
    Widget button;
    Widget separator;
    Widget edit_field;
    Widget menu_bar_frame;
    Widget menu_bar_form;
    Widget menu_form;
    Widget menu_frame;
    XFontStruct *the_font_struct;

    /*
     * initialize some important constants
     */
    disp_info[disp_num].menu[0].num_menu_items = 0;
    disp_info[disp_num].menu[1].num_menu_items = 0;
    disp_info[disp_num].menu[2].num_menu_items = 0;
    disp_info[disp_num].menu[3].num_menu_items = 0;
    disp_info[disp_num].menu[4].num_menu_items = 0;
    disp_info[disp_num].menu[5].num_menu_items = 8;

    for (i=0; strcmp("****", english_menu0[i]) != 0; i++)
        (disp_info[disp_num].menu[0].num_menu_items)++;

    for (i=0; strcmp("****", english_menu1[i]) != 0; i++)
        (disp_info[disp_num].menu[1].num_menu_items)++;

    for (i=0; strcmp("****", english_menu2[i]) != 0; i++)
        (disp_info[disp_num].menu[2].num_menu_items)++;

    for (i=0; strcmp("****", english_menu3[i]) != 0; i++)
        (disp_info[disp_num].menu[3].num_menu_items)++;

    for (i=0; strcmp("****", english_menu4[i]) != 0; i++)
        (disp_info[disp_num].menu[4].num_menu_items)++;

    for (i=0; i < disp_info[disp_num].NUM_EFF_MENUS; i++)
        for (j=0; j < disp_info[disp_num].menu[i].num_menu_items; j++)
            disp_info[disp_num].menu[i].selectable[j] = TRUE;

    disp_info[disp_num].menu[0].selectable[3] = FALSE;	/* separators */
    disp_info[disp_num].menu[0].selectable[4] = FALSE;
    disp_info[disp_num].menu[0].selectable[7] = FALSE;
    disp_info[disp_num].menu[1].selectable[6] = FALSE;
    disp_info[disp_num].menu[1].selectable[9] = FALSE;
    disp_info[disp_num].menu[2].selectable[3] = FALSE;
    disp_info[disp_num].menu[3].selectable[14] = FALSE;

    if (XrmGetResource(XtDatabase(disp_info[disp_num].disp),
           "wscrawl.interfaceLanguage", "wscrawl.InterfaceLanguage",
           &str_typ_return, &val_ret) == TRUE)
    {
        strncpy(tmp_str, val_ret.addr, 248);

        if (strncmp(tmp_str, "German", 6) == 0)
        {
            menu_bar_labels = german_menu_bar_labels;
	    strcpy(disp_info[disp_num].status1_line_format, german_status1);
	    strcpy(disp_info[disp_num].status_line_format, german_status);
        }
        else if (strncmp(tmp_str, "English", 7) == 0)
        {
            menu_bar_labels = english_menu_bar_labels;
	    strcpy(disp_info[disp_num].status1_line_format, english_status1);
	    strcpy(disp_info[disp_num].status_line_format, english_status);
        }
	else
	{
	    printf("WARNING: unknown interfaceLanguage resource setting: \n");
	    printf("%s.\n", tmp_str);
	}
    }
    else
    {
        printf("WARNING: could not find interface_language resource.\n");
        menu_bar_labels = english_menu_bar_labels;
	strcpy(disp_info[disp_num].status1_line_format, english_status1);
	strcpy(disp_info[disp_num].status_line_format, english_status);
    }

    /*
     * Create MenuBar in MainWindow.
     */
    n = 0;
    XtSetArg(args[n], XmNshadowThickness, 2); n++;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    menu_bar_frame = XmCreateFrame(main_window, "menu_bar_frame", args, n); 
    XtManageChild(menu_bar_frame);

    n = 0;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    menu_bar_form = XmCreateForm(menu_bar_frame, "menu_bar_form", args, n); 
    XtManageChild(menu_bar_form);

    for (i=0; i < disp_info[disp_num].NUM_EFF_MENUS; i++)
    {
        n = 0;
	if (i == 0)
	{
            XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
            XtSetArg(args[n], XmNleftOffset, 3); n++;
	}
	else
	{
            XtSetArg(args[n],XmNleftAttachment, XmATTACH_WIDGET); n++;
            XtSetArg(args[n], XmNleftWidget, 
		     disp_info[disp_num].menu[i-1].menu_bar_button); n++;
	}
        XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
        XtSetArg(args[n], XmNtopOffset, 3); n++;
        XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
        XtSetArg(args[n], XmNbottomOffset, 2); n++;
        XtSetArg(args[n], XmNmarginLeft, 5); n++;
        XtSetArg(args[n], XmNmarginRight, 5); n++;
            XtSetArg(args[n], XmNfillOnArm, False); n++;
        XtSetArg(args[n], XmNshadowThickness, 0); n++;
        XtSetArg(args[n], XmNmarginHeight, 4); n++;
        XtSetArg(args[n], XmNmarginWidth, 4); n++;
        XtSetArg(args[n], XmNtraversalOn, False); n++;
        disp_info[disp_num].menu[i].menu_bar_button = 
	     XmCreatePushButton(menu_bar_form, menu_bar_labels[i], args, n); 
        XtManageChild(disp_info[disp_num].menu[i].menu_bar_button);
        XtAddCallback(disp_info[disp_num].menu[i].menu_bar_button, 
		XmNarmCallback, (XtCallbackProc)menu_armCB, (XtPointer)(disp_num << 8 | i));
        XtAddCallback(disp_info[disp_num].menu[i].menu_bar_button, 
		  XmNdisarmCallback, (XtCallbackProc)menu_disarmCB, (XtPointer)(disp_num << 8 | i));
        XtAddEventHandler(disp_info[disp_num].menu[i].menu_bar_button, 
		      PointerMotionMask, False,
		      (XtEventHandler)menu_pointer_moveCB, (XtPointer)(disp_num << 8 | i));
    }

    n = 0;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightOffset, 3); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopOffset, 3); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomOffset, 2); n++;
    XtSetArg(args[n], XmNmarginLeft, 5); n++;
    XtSetArg(args[n], XmNmarginRight, 5); n++;
    XtSetArg(args[n], XmNfillOnArm, False); n++;
    XtSetArg(args[n], XmNshadowThickness, 0); n++;
    XtSetArg(args[n], XmNmarginHeight, 4); n++;
    XtSetArg(args[n], XmNmarginWidth, 4); n++;
    XtSetArg(args[n], XmNtraversalOn, False); n++;
    disp_info[disp_num].help_menu_bar_button = 
	     XmCreatePushButton(menu_bar_form, "Help", args, n); 
    XtManageChild(disp_info[disp_num].help_menu_bar_button);
    XtAddCallback(disp_info[disp_num].help_menu_bar_button, XmNactivateCallback,
		  (XtCallbackProc)helpCB, (XtPointer)(disp_num << 8 | ACTIVATE));
    XtAddCallback(disp_info[disp_num].help_menu_bar_button, XmNarmCallback, 
		  (XtCallbackProc)helpCB, (XtPointer)(disp_num << 8 | ARM));
    XtAddCallback(disp_info[disp_num].help_menu_bar_button, 
		  XmNdisarmCallback, (XtCallbackProc)helpCB, (XtPointer)(disp_num << 8 | DISARM));

    /*
     * Create all the items in the pulldown menus
     */
    create_a_menu(disp_num, 0, english_menu0, main_window);  /* File */
    create_a_menu(disp_num, 1, english_menu1, main_window);  /* Edit */
    create_a_menu(disp_num, 2, english_menu2, main_window);  /* Options */
    create_a_menu(disp_num, 3, english_menu3, main_window);  /* Color */
    create_a_menu(disp_num, 4, english_menu4, main_window);  /* Font */

    return(menu_bar_frame);
}


/*
 * create_a_menu - this function creates the menu pane and all the buttons
 *                 for this menu. "mn" is the "menu_number" where 0 == FILE,
 *                 1 == EDIT, etc.
 */
Widget create_a_menu(disp_num, mn, item_text, main_window)
int disp_num, mn;
char *item_text[];
Widget main_window;
{
    int n, i, max_width, width, max_item;
    Arg args[MAX_ARGS];
    Widget menu_frame, menu_form;
    XFontStruct *the_font_struct;

    /*
     * create the menu_pane
     */
    n = 0;
    XtSetArg(args[n], XmNmappedWhenManaged, False); n++;
    disp_info[disp_num].menu[mn].menu_shell = XtCreatePopupShell("menu_shell",
	 overrideShellWidgetClass, main_window, args, n); 
    XtManageChild(disp_info[disp_num].menu[mn].menu_shell);

    n = 0;
    XtSetArg(args[n],XmNleftAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
    XtSetArg(args[n], XmNleftWidget, 
	     disp_info[disp_num].menu[mn].menu_bar_button); n++;
    XtSetArg(args[n],XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNshadowThickness, 2); n++;
    XtSetArg(args[n], XmNshadowType, XmSHADOW_OUT); n++;
    menu_frame = XmCreateFrame(disp_info[disp_num].menu[mn].menu_shell, 
				"menu_frame", args, n); 
    XtManageChild(menu_frame);

    n = 0;
    menu_form = XmCreateForm(menu_frame, "menu_form", args, n); 
    XtManageChild(menu_form);

    /* 
     * find the widest item in the menu
     */
    the_font_struct = XLoadQueryFont(disp_info[disp_num].disp, "variable");
    max_width = 0;
    for (i=0; i < disp_info[disp_num].menu[mn].num_menu_items; i++)
    {
	if (strcmp(item_text[i], "----") != 0)    /* it is a separator */
	{
	    width = XTextWidth(the_font_struct, item_text[i],
			       strlen(item_text[i]));

	    if (width > max_width)
	    {
		max_width = width;
		max_item = i;
	    }
        }
    }

    for (i=0; i < disp_info[disp_num].menu[mn].num_menu_items; i++)
    {
	if (strcmp(item_text[i], "----") == 0)    /* it is a separator */
	{
            n = 0;
            XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
            XtSetArg(args[n], XmNleftOffset, 5); n++;
            XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
            XtSetArg(args[n], XmNrightOffset, 5); n++;
            XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
            XtSetArg(args[n], XmNtopWidget, 
	             disp_info[disp_num].menu[mn].menu_item[i-1]); n++;
            disp_info[disp_num].menu[mn].menu_item[i] = 
	        XmCreateSeparatorGadget(menu_form, "separator", args, n); 
            XtManageChild(disp_info[disp_num].menu[mn].menu_item[i]);
	}
	else
	{
            n = 0;
            XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
            XtSetArg(args[n], XmNleftOffset, 5); n++;
            if (max_item != i)
	    {
                XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
	    }
            XtSetArg(args[n], XmNrightOffset, 5); n++;
	    if (i == 0)
	    {
                XtSetArg(args[n], XmNtopOffset, 3); n++;
                XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
	    }
	    else
	    {
                XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
                XtSetArg(args[n], XmNtopWidget, 
	             disp_info[disp_num].menu[mn].menu_item[i-1]); n++;
	    }
            XtSetArg(args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
            XtSetArg(args[n], XmNshadowThickness, 0); n++;
            XtSetArg(args[n], XmNmarginHeight, 4); n++;
            XtSetArg(args[n], XmNmarginWidth, 4); n++;
            XtSetArg(args[n], XmNbottomOffset, 3); n++;
            disp_info[disp_num].menu[mn].menu_item[i] = 
                XmCreatePushButtonGadget(menu_form, item_text[i], args, n); 
            XtManageChild(disp_info[disp_num].menu[mn].menu_item[i]);
	}
    }
    return(menu_form);
}


/*
 * read_one_time_only_global_resources - this function reads the resource 
 *                          database of the first display for global values
 *                          that are shared amongst all the displays. 
 */
read_one_time_only_global_resources(disp, global_values)
Display *disp;
struct global_values_struct *global_values;
{
    char *str_typ_return;
    XrmValue val_ret;
    char tmp_str[250];
    int i, tmp_num;

    if (XrmGetResource(XtDatabase(disp), "wscrawl.syncScrollbars", 
	"wscrawl.SyncScrollbars", &str_typ_return, &val_ret) == TRUE)
    {
	strncpy(tmp_str, val_ret.addr, 248);

	if ((strncmp(tmp_str, "True", 4) == 0) ||
	    (strncmp(tmp_str, "TRUE", 4) == 0))
	{
	    global_values->sync_scrollbars_bool = TRUE;
	}
	else if ((strncmp(tmp_str, "False", 5) == 0) ||
	         (strncmp(tmp_str, "FALSE", 5) == 0))
	{
	    global_values->sync_scrollbars_bool = FALSE;
	}
	else
	{
	    printf("WARNING: unknown syncScrollbars resource setting: \n");
	    printf("%s.\n", tmp_str);
        }
    }

    if (XrmGetResource(XtDatabase(disp), "wscrawl.virtualWhiteboardWidth", 
	"wscrawl.VirtualWhiteboardWidth", &str_typ_return, &val_ret) == TRUE)
    {
	strncpy(tmp_str, val_ret.addr, 248);

	tmp_num = 0;
	sscanf(tmp_str, "%d", &tmp_num);

	if (tmp_num > 10)
	{
	    global_values->draw_area_width = tmp_num;
	}
	else
	{
	    printf("WARNING: invalid virtualWhitboardWidth resource.\n");
	    global_values->draw_area_width = 600;
	}
    }

    if (XrmGetResource(XtDatabase(disp), "wscrawl.virtualWhiteboardHeight", 
	"wscrawl.VirtualWhiteboardHeight", &str_typ_return, &val_ret) == TRUE)
    {
	strncpy(tmp_str, val_ret.addr, 248);

	tmp_num = 0;
	sscanf(tmp_str, "%d", &tmp_num);

	if (tmp_num > 10)
	{
	    global_values->draw_area_height = tmp_num;
	}
	else
	{
	    printf("WARNING: invalid virtualWhitboardHeight resource.\n");
	    global_values->draw_area_height = 600;
	}
    }
}


/*
 * read_special_resources - this function reads the resource database
 *                          for the special resources that wscrawl uses
 *                          BESIDES the standard win_width, etc for Motif.
 */
read_special_resources(disp_num)
int disp_num;
{
    char *str_typ_return;
    XrmValue val_ret;
    XFontStruct *the_font_struct;
    char tmp_str[250];
    int i;

    if (XrmGetResource(XtDatabase(disp_info[disp_num].disp), 
           "wscrawl.interfaceLanguage", "wscrawl.InterfaceLanguage",
	   &str_typ_return, &val_ret) == TRUE)
    {
	strncpy(tmp_str, val_ret.addr, 248);

	if (strncmp(tmp_str, "German", 6) == 0)
	{
            set_language_of_interface(disp_num, GERMAN);
	    disp_info[disp_num].current_language = GERMAN;
	}
	else if (strncmp(tmp_str, "English", 7) == 0)
	{
            set_language_of_interface(disp_num, ENGLISH);
	    disp_info[disp_num].current_language = ENGLISH;
	}
        else
	{
	    printf("WARNING: unknown interface_language resource setting: \n");
	    printf("%s.\n", tmp_str);
        }
    }
    else
    {
	printf("WARNING: could not find interface_language resource.\n");
        set_language_of_interface(disp_num, ENGLISH);
	disp_info[disp_num].current_language = ENGLISH;
    }

    if (XrmGetResource(XtDatabase(disp_info[disp_num].disp), 
           "wscrawl.telePointerName", "wscrawl.TelePointerName",
	   &str_typ_return, &val_ret) == TRUE)
    {
	strncpy(tmp_str, val_ret.addr, 248);

	if (strncmp(tmp_str, "default", 6) != 0)
	{
	    strncpy(disp_info[disp_num].rubber_pointer.pointer_str, 
	            tmp_str, 48);
	    the_font_struct = XLoadQueryFont(disp_info[disp_num].disp, "fixed");
	    disp_info[disp_num].rubber_pointer.height = 
                                   15 + the_font_struct->max_bounds.ascent +
                                   the_font_struct->max_bounds.descent;
            disp_info[disp_num].rubber_pointer.width = 15 + 
                  XTextWidth(the_font_struct,
                  disp_info[disp_num].rubber_pointer.pointer_str,
                  strlen(disp_info[disp_num].rubber_pointer.pointer_str));
	}
    }

    if (XrmGetResource(XtDatabase(disp_info[disp_num].disp), 
	 "wscrawl.continuousTelePointer", "wscrawl.ContinuousTelePointer", 
	 &str_typ_return, &val_ret) == TRUE)
    {
	strncpy(tmp_str, val_ret.addr, 248);

	if ((strncmp(tmp_str, "True", 4) == 0) ||
	    (strncmp(tmp_str, "TRUE", 4) == 0))
	{
	    disp_info[disp_num].show_telepointer_always = TRUE;
	    XmToggleButtonGadgetSetState(
		disp_info[disp_num].mouse_dialog.telepointer_toggle, True,
		False);
	}
	else if ((strncmp(tmp_str, "False", 5) == 0) ||
	         (strncmp(tmp_str, "FALSE", 5) == 0))
	{
	    disp_info[disp_num].show_telepointer_always = FALSE;
	    XmToggleButtonGadgetSetState(
		disp_info[disp_num].mouse_dialog.telepointer_toggle, False,
		False);
	}
        else
	{
	    printf("WARNING:unknown continuousTelePointer resource setting: ");
	    printf("%s.\n", tmp_str);
        }
    }
    else
    {
	disp_info[disp_num].show_telepointer_always = FALSE;
    }

    if (XrmGetResource(XtDatabase(disp_info[disp_num].disp), 
	 "wscrawl.showPointerCoordinates", "wscrawl.ShowPointerCoordinates", 
	 &str_typ_return, &val_ret) == TRUE)
    {
	strncpy(tmp_str, val_ret.addr, 248);

	if ((strncmp(tmp_str, "True", 4) == 0) ||
	    (strncmp(tmp_str, "TRUE", 4) == 0))
	{
	    disp_info[disp_num].show_pointer_coords = TRUE;
	    XtManageChild(disp_info[disp_num].pointer_x_coords_label);
	    XtManageChild(disp_info[disp_num].pointer_y_coords_label);
	    XmToggleButtonGadgetSetState(
		disp_info[disp_num].mouse_dialog.coordinates_toggle, True,
		False);
	}
	else if ((strncmp(tmp_str, "False", 5) == 0) ||
	         (strncmp(tmp_str, "FALSE", 5) == 0))
	{
	    disp_info[disp_num].show_pointer_coords = FALSE;
	    XmToggleButtonGadgetSetState(
		disp_info[disp_num].mouse_dialog.coordinates_toggle, False,
		False);
	}
        else
	{
	    printf("WARNING:unknown showPointerCoordinates resource setting: ");
	    printf("%s.\n", tmp_str);
        }
    }
    else
    {
        disp_info[disp_num].show_pointer_coords = FALSE;
    }

    if (XrmGetResource(XtDatabase(disp_info[disp_num].disp), 
	 "wscrawl.iconForeground", "wscrawl.IconForeground", 
	 &str_typ_return, &val_ret) == TRUE)
    {
	strncpy(disp_info[disp_num].icon_fg_str, val_ret.addr, 248);
    }
    else
    {
	strncpy(disp_info[disp_num].icon_fg_str, "black", 248);
    }

    if (XrmGetResource(XtDatabase(disp_info[disp_num].disp), 
	 "wscrawl.iconBackground", "wscrawl.IconBackground", 
	 &str_typ_return, &val_ret) == TRUE)
    {
	strncpy(disp_info[disp_num].icon_bg_str, val_ret.addr, 248);
    }
    else
    {
	strncpy(disp_info[disp_num].icon_bg_str, "white", 248);
    }
}


/*
 * create_selected_regions - this function creates and initializes all the 
 *                           data structures for the selected regions.
 *			     A selected region is the thing that you would
 *                           "cut" or "paste".
 */
create_selected_regions(disp_num)
int disp_num;
{
    int mn;
    int n, i;
    Arg args[MAX_ARGS];
    Widget menu_bar;
    XWindowChanges win_change;
    XGCValues gcvalues;
    XWMHints wmhints;

    /*
     * create this new display's selection regions
     */
    for (i=0; i < num_disps; i++)
    {
	if (disp_info[i].in_session_bool == TRUE)
	{
            n = 0;
            XtSetArg(args[n], XmNwidth, 10); n++;
            XtSetArg(args[n], XmNheight, 10); n++;
            XtSetArg(args[n], XmNmappedWhenManaged, False); n++;
	    if (i == disp_num)
	    {
                disp_info[disp_num].selected_region.selection_shell[i] = 
	            XtCreatePopupShell("selection_shell", 
		    overrideShellWidgetClass, 
                    disp_info[i].main_window, args, n); 
	    }
	    else
	    {
                XtSetArg(args[n], XmNmarginWidth, 0); n++;
                XtSetArg(args[n], XmNmarginHeight, 0); n++;
                XtSetArg(args[n], XmNshadowThickness, 0); n++;
                disp_info[disp_num].selected_region.selection_shell[i] = 
		    XmCreateDrawingArea(disp_info[i].draw_area,
		                     "selection_shell2", args, n);
	    }
            XtManageChild(
		   disp_info[disp_num].selected_region.selection_shell[i]);

            n = 0;
            XtSetArg(args[n], XmNshadowThickness, 0); n++;
            disp_info[disp_num].selected_region.selection_area[i] = 
		    XmCreateDrawingArea(
		     disp_info[disp_num].selected_region.selection_shell[i], 
		     "region_area", args, n);
           XtManageChild(disp_info[disp_num].selected_region.selection_area[i]);
            XtAddCallback(disp_info[disp_num].selected_region.selection_area[i],
			 XmNinputCallback, (XtCallbackProc)selection_area_inputCB, 
		         (XtPointer)(disp_num << 8 | i));
            XtAddCallback(disp_info[disp_num].selected_region.selection_area[i],
                   XmNexposeCallback, (XtCallbackProc)selection_area_exposeCB, 
		         (XtPointer)(disp_num << 8 | i));

	    if (i == disp_num)
	    {
                XtAddEventHandler(
		   disp_info[disp_num].selected_region.selection_area[i],
		   EnterWindowMask | LeaveWindowMask, False, 
		   (XtEventHandler)handle_selection_keypressCB, (XtPointer)disp_num);
		/*
                XtAddEventHandler(
	          disp_info[disp_num].selected_region.selection_area[i], 
	          PointerMotionMask, False, (XtEventHandler)selection_pointer_moveCB, 
		  (XtPointer)(disp_num << 8 | i));
		 */
	    }

            XtRealizeWidget(
		 disp_info[disp_num].selected_region.selection_shell[i]);

	    win_change.border_width = 0;
	    XConfigureWindow(disp_info[i].disp,
               XtWindow(disp_info[disp_num].selected_region.selection_shell[i]),
	       CWBorderWidth, &win_change);

	    disp_info[disp_num].selected_region.backing_store[i] = NULL;
	    disp_info[disp_num].selected_region.clip_board[i] = NULL;
	    
	    gcvalues.line_width = 0;
	    gcvalues.function = GXinvert;
	    gcvalues.line_style = LineOnOffDash;
	    gcvalues.cap_style = CapButt;
	    disp_info[disp_num].selected_region.surround_line_gc[i] =
	      XCreateGC(disp_info[i].disp,
		XtWindow(disp_info[disp_num].selected_region.selection_area[i]),
	        GCFunction | GCLineWidth | GCLineStyle | GCCapStyle, 
		&gcvalues);
	}
    }

    /*
     * now add this new display to all the OTHER display's selection regions
     */
    for (i=0; i < num_disps; i++)
    {
	if ((disp_info[i].in_session_bool == TRUE) && (i != disp_num))
	{
            n = 0;
            XtSetArg(args[n], XmNwidth, 10); n++;
            XtSetArg(args[n], XmNheight, 10); n++;
            XtSetArg(args[n], XmNmappedWhenManaged, False); n++;
            XtSetArg(args[n], XmNmarginWidth, 0); n++;
            XtSetArg(args[n], XmNmarginHeight, 0); n++;
            XtSetArg(args[n], XmNshadowThickness, 0); n++;
            disp_info[i].selected_region.selection_shell[disp_num] = 
		    XmCreateDrawingArea(disp_info[disp_num].draw_area,
		                     "selection_shell2", args, n);
            XtManageChild(
		   disp_info[i].selected_region.selection_shell[disp_num]);

            n = 0;
            XtSetArg(args[n], XmNwidth, 10); n++;
            XtSetArg(args[n], XmNheight, 10); n++;
            XtSetArg(args[n], XmNshadowThickness, 0); n++;
            disp_info[i].selected_region.selection_area[disp_num] = 
		    XmCreateDrawingArea(
		     disp_info[i].selected_region.selection_shell[disp_num], 
		     "region_area", args, n);
           XtManageChild(disp_info[i].selected_region.selection_area[disp_num]);
            XtAddCallback(disp_info[i].selected_region.selection_area[disp_num],
			 XmNinputCallback, (XtCallbackProc)selection_area_inputCB, 
		         (XtPointer)(i << 8 | disp_num));
            XtAddCallback(disp_info[i].selected_region.selection_area[disp_num],
                   XmNexposeCallback, (XtCallbackProc)selection_area_exposeCB, 
		         (XtPointer)(i << 8 | disp_num));

            XtRealizeWidget(
		 disp_info[i].selected_region.selection_shell[disp_num]);

	    win_change.border_width = 0;
	    XConfigureWindow(disp_info[disp_num].disp,
               XtWindow(disp_info[i].selected_region.selection_shell[disp_num]),
	       CWBorderWidth, &win_change);

	    disp_info[i].selected_region.backing_store[disp_num] = NULL;
	    disp_info[disp_num].selected_region.clip_board[disp_num] = NULL;

	    gcvalues.line_width = 0;
	    gcvalues.function = GXinvert;
	    gcvalues.line_style = LineOnOffDash;
	    gcvalues.cap_style = CapButt;
	    disp_info[i].selected_region.surround_line_gc[disp_num] =
	      XCreateGC(disp_info[disp_num].disp,
		XtWindow(disp_info[i].selected_region.selection_area[disp_num]),
	        GCFunction | GCLineWidth | GCLineStyle | GCCapStyle, 
		&gcvalues);
	}
    }
}


/*
 * init_communication - this function initializes the TCP/IP communication
 *                      for the future.  If this wscrawl program is a "post"
 *                      or a "knock", then it never continues after this
 *                      function.
 */
init_communication(disp_num, knock, location)
int disp_num;
int knock;
char *location;
{
    int client_len, n;
    struct knock_knock_msg_struct knock_msg;

    if (knock == TRUE)		/* this is a "knock-knock, let me in..." */
    {
	knock_knock(disp_num, location);
	exit(0);
    }
    else			/* this is the server waiting for the knock */
    {
	disp_info[disp_num].my_socket = 
			 open_socket_to_receive_data(SERV_TCP_PORT);
    }
}


/*
 * open_socket_to_receive_data - this function starts up the wscrawl socket
 *                               listening for anyone "knocking on the door"
 *                               to let them into this session.  This is
 *                               a TCP/IP communication point.
 */
open_socket_to_receive_data(tcp_port_num)
int tcp_port_num;
{
    int s, cli_s;
    struct sockaddr_in serv_addr, cli_addr;
    struct hostent *dest_machine_info;

    if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
        perror("ERROR: socket call failed for server");
        exit(0);
    }

    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;            /* address family */
    serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    serv_addr.sin_port = htons(tcp_port_num);  /* internet TCP port number */

    if (bind(s, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
    {
        perror("WARNING: bind call failed in receiver");
        printf("Running two copies of wscrawl on the same machine will ");
	printf("cause this.\n");
        return(0);
    }

    listen(s, 5);

    return(s);
}


/*
 * my_open_socket_to_send_data - this function opens the socket and initializes
 *                            it properly for sending data through.
 */
my_open_socket_to_send_data(dest_machine, tcp_port_num)
char *dest_machine;
int tcp_port_num;
{
    int s;
    struct hostent *hp;
    struct sockaddr_in serv_addr;

    /* 
     * Modified to support dotted IP addresses as well as hostnames, fix
     * by Bruce Barnett
     */
    bzero((char *) &serv_addr, sizeof (struct sockaddr_in));
    serv_addr.sin_family = AF_INET;            /* address family */
    serv_addr.sin_addr.s_addr = inet_addr(dest_machine); /* dotted address? */

    if (serv_addr.sin_addr.s_addr == (u_int) -1)  /* not a dotted IP address */
    {
        if ((hp = gethostbyname(dest_machine)) == NULL)
        {
            perror("ERROR: in my_open_socket_to_send_data, gethostbyname fail");
            exit(0);
        } 
        else 
        {
             /* copy the address to serv_addr */
             bcopy(hp->h_addr, (char *) &(serv_addr.sin_addr), hp->h_length);
        }
    }

    serv_addr.sin_port = htons(tcp_port_num);  /* internet TCP port number */

    if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	printf("ERROR: in my_open_socket_to_send_data, socket failed.\n");

    if (connect(s, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
	perror("ERROR: in my_open_socket_to_send_data, connect failed");
    
    return(s);
}


/*
 * parse_command_line - this function parses the command line and sets the
 *             various global variables for the opening display
 *             and color party coming up.  Sorry that this routine is so
 *             ugly, but just go slow, and it's not too bad.  I'd look at
 *             the "command_line_err()" routine to understand what each
 *             of the command line options mean.
 *             (ex. % wscrawl -display hpcvxbw:0 -knock)
 */
parse_command_line(argv, argc, knock, location, global_values)
char *argv[];
int argc;
int *knock;
char *location;
struct global_values_struct *global_values;
{
     int i, current_arg;

     *knock = FALSE;            	/*defaults set here*/
     strncpy(location, "airrs2", 50);
     global_values->edit_file_bool = FALSE;    /* edit a file on startup */

     for (current_arg = 1; current_arg < argc; current_arg+=2)
     {
         if (strcmp(argv[current_arg],"-knock") == 0)
         {
	     *knock = TRUE;
	     current_arg--;	/* hack to have arguments come out right */
         }
         else if (strcmp(argv[current_arg],"-loc") == 0)
         {
             if (current_arg != argc-1)
                 strncpy(location, argv[current_arg+1], 50);
             else
                 command_line_err();     /* no location was specified */
         }
         else if (strcmp(argv[current_arg],"-editfile") == 0)
         {
             if (current_arg != argc-1)
             {
                 global_values->edit_file_bool = TRUE;
                 strncpy(global_values->edit_filename, argv[current_arg+1], 95);
             }
             else
                 command_line_err();     /* no filename was specified */
         }
         else
         {
             command_line_err();
         }
     }
}


/*
 * command_line_err - this function spits out a standard error message and
 *                    exits the program.
 */
command_line_err()
{
    printf("\nBad command line option dude.\n\n");
    printf("Try: wscrawl\n");
    printf("  [-display displayname]\n");
    printf("  [-loc location]\n");
    printf("  [-knock]            (add me to the session at \"location\")\n");
    printf("  [-editfile xwd_filename]\n\n");
    exit(0);
}


/*
 * set_default_resources - this function sees if the regular defaults are set
 *                         up, and if not, does the equivalent of an "xrdb -m"
 *                         to set them up correctly.
 */
set_default_resources(source_disp, disp_num, disp_name, argc, argv)
int source_disp;
int disp_num;
char *disp_name;
int argc;
char *argv[];
{
    char *str_typ_return;
    XrmValue val_ret;
    char tmp_str[250];
    char *buf;
    int fake_argc;
    char *fake_argv[100];
    int i;

    if (XrmGetResource(XtDatabase(disp_info[disp_num].disp), 
           "wscrawl.interfaceLanguage", "wscrawl.InterfaceLanguage",
	   &str_typ_return, &val_ret) == TRUE)
    {
	return;		/* the resources are set up properly */
    }
    else
    {
	buf = (char *) malloc(10000 * sizeof(char));
	init_resource_buf(buf);
	merge_resources(disp_info[disp_num].disp, buf);
	free(buf);
	
	/*
	 * now re-open the display to get the new resources into
	 * the XtDatabase.
	 */
	XtCloseDisplay(disp_info[disp_num].disp);

	if (source_disp == disp_num)     /* it is the first display */
	{
            if ((disp_info[disp_num].disp = XtOpenDisplay(appCon, disp_name, 
	          "wscrawl", "WScrawl", NULL, 0, &argc, argv)) == NULL)
            {
	        printf("ERROR: for unknown reason cannot re-open display.\n");
	        exit(0);
            }
	}
        else		            /* make background of drawing area white */
	{
	    fake_argc = 17; 
	    fake_argv[0] = "wscrawl";
	    fake_argv[1] = "-xrm";
	    fake_argv[2] = "wscrawl*swindow*draw_area_frame*background: white";
	    fake_argv[3] = "-xrm";
	    fake_argv[4] = "wscrawl*foreground: black";
	    fake_argv[5] = "-xrm";
	    fake_argv[6] = "wscrawl*keyboardFocusPolicy: pointer";
	    fake_argv[7] = "-xrm";
	    fake_argv[8] = "wscrawl*background: LightSteelBlue";
	    fake_argv[9] = "-xrm";
	    fake_argv[10]= "wscrawl*swindow*foreground: LightSteelBlue";
	    fake_argv[11]= "-xrm";
	    fake_argv[12]= "wscrawl*fontList: variable";
	    fake_argv[13]= "-xrm";
	    fake_argv[14]= "wscrawl*borderWidth: 0";
	    fake_argv[15]= "-xrm";
	    fake_argv[16]= "wscrawl.interfaceLanguage: English";
            if ((disp_info[disp_num].disp = XtOpenDisplay(appCon, disp_name, 
	         "wscrawl", "WScrawl", NULL, 0, &fake_argc, fake_argv)) == NULL)
            {
	        printf("ERROR: for unknown reason cannot re-open display.\n");
	        exit(0);
            }
	}
    }
}


/*
 * init_resource_buf - this function fills out "buf" with the correct default
 *                     resources.
 */
init_resource_buf(buf)
char *buf;
{
    sprintf(buf, "%s", 
"wscrawl.virtualWhiteboardWidth:		1500\n\
wscrawl.virtualWhiteboardHeight:		1500\n\
wscrawl.interfaceLanguage:			English\n\
wscrawl.telePointerName:			default\n\
wscrawl.syncScrollbars:				False\n\
wscrawl.continuousTelePointer:			False\n\
wscrawl.showPointerCoordinates:			False\n\
wscrawl.iconForeground:				black\n\
wscrawl.iconBackground:				LightSteelBlue\n\
wscrawl*foreground:				black\n\
wscrawl*background:				LightSteelBlue\n\
wscrawl*swindow*draw_area_frame*background:     white\n\
wscrawl*draw_area_frame*foreground:             blue\n\
wscrawl*draw_area_frame*background:             white\n\
wscrawl*keyboardFocusPolicy:                    pointer\n\
wscrawl*fontList:				variable\n\
wscrawl*borderWidth:				0\n\
wscrawl*swindow*foreground:			LightSteelBlue\n");
}

