/*
 * This File: wslang.c
 *
 * This file contains the language interface changing functions.
 */

#include "wscrawl.h"
#include "languages.h"
#include "pause_curs.h"

extern struct disp_info_struct disp_info[MAX_NUM_DISPS];
static XmStringCharSet charset = (XmStringCharSet) XmSTRING_DEFAULT_CHARSET;
char **menu_bar_labels;
char **menu_items[NUM_MENUS];
char *add_display_prompt;
char *join_session_prompt;
char *language_prompt;

char *dirty_add_display_prompt;
char *dirty_add_display_but1;
char *dirty_add_display_but2;
char *dirty_add_display_but3;

char *scrollbar_prompt;
char *scrollbar_but1;
char *scrollbar_but2;

char *mouse_dialog_prompt;
char *mouse_dialog_toggle1;
char *mouse_dialog_toggle2;

char *line_width_prompt;
char *square_width_prompt;
char *oval_width_prompt;

char *save_file_format_prompt;
char *save_file_format_but[5];

char *read_file_format_prompt;
char *read_file_format_but[5];

char *color_picker_prompt;
char *color_picker_but1;
char *color_picker_but2;
char *color_picker_but3;

char *ok_str;
char *cancel_str;
char *close_str;
char *help_str;
char *english_str;
char *german_str;
char *prog_name;

int pause_cursor_width, pause_cursor_height;
int pause_cursor_x_hot, pause_cursor_y_hot;
char *pause_cursor_bits;

XmString ok_label_str = NULL;
XmString cancel_label_str = NULL;


/*
 * set_language_of_interface - this function is the entry point to the 
 *                             functions in this file.
 */
set_language_of_interface(disp_num, language)
int disp_num, language;
{
    int n;
    Arg args[MAX_ARGS];
    char buf[256];

    if (disp_info[disp_num].language_dialog.current_language == language)
	return;

    switch(language)
    {
	case ENGLISH:
	    strcpy(disp_info[disp_num].status1_line_format, english_status1);
	    strcpy(disp_info[disp_num].status_line_format, english_status);
	    menu_bar_labels = english_menu_bar_labels;
	    menu_items[0] = english_menu0;
	    menu_items[1] = english_menu1;
	    menu_items[2] = english_menu2;
	    menu_items[3] = english_menu3;
	    menu_items[4] = english_menu4;
	    ok_str = english_ok;
	    cancel_str = english_cancel;
	    close_str = english_color_picker_close_but;
	    english_str = english_english;
	    german_str = english_german;
	    help_str = english_help;
	    add_display_prompt = english_add_display_prompt;
            join_session_prompt = english_join_session_prompt;
	    language_prompt = english_language_prompt;
	    prog_name = english_name;
	    pause_cursor_width = pause_curs_width;
	    pause_cursor_height = pause_curs_height;
	    pause_cursor_x_hot = pause_curs_x_hot;
	    pause_cursor_y_hot = pause_curs_y_hot;
	    pause_cursor_bits = pause_curs_bits;
	    /* 
	     * some dialogs 
	     */
	    dirty_add_display_prompt = english_dirty_add_display_prompt;
            dirty_add_display_but1 = english_dirty_add_display_but1;
            dirty_add_display_but2 = english_dirty_add_display_but2;
            dirty_add_display_but3 = english_dirty_add_display_but3;

            scrollbar_prompt = english_scrollbar_prompt;
            scrollbar_but1 = english_scrollbar_but1;
            scrollbar_but2 = english_scrollbar_but2;

            save_file_format_prompt = english_save_file_format_prompt;
            save_file_format_but[1] = english_save_file_format_but1;
            save_file_format_but[2] = english_save_file_format_but2;
            save_file_format_but[3] = english_save_file_format_but3;
            save_file_format_but[4] = english_save_file_format_but4;

            read_file_format_prompt = english_read_file_format_prompt;
            read_file_format_but[1] = english_read_file_format_but1;
            read_file_format_but[2] = english_read_file_format_but2;
            read_file_format_but[3] = english_read_file_format_but3;
            read_file_format_but[4] = english_read_file_format_but4;

            color_picker_prompt = english_color_picker_prompt;
            color_picker_but1 = english_color_picker_set_col_but;
            color_picker_but2 = english_color_picker_close_but;
            color_picker_but3 = english_color_picker_grab_color_but;

            mouse_dialog_prompt = english_mouse_dialog_prompt;
            mouse_dialog_toggle1 = english_mouse_dialog_toggle1;
            mouse_dialog_toggle2 = english_mouse_dialog_toggle2;

            line_width_prompt = english_line_width_prompt;
            square_width_prompt = english_square_width_prompt;
            oval_width_prompt = english_oval_width_prompt;

	    sprintf(buf, "wscrawl.interfaceLanguage: English\n");
	    merge_resources(disp_info[disp_num].disp, buf);
	    break;
	case GERMAN:
	    strcpy(disp_info[disp_num].status1_line_format, german_status1);
	    strcpy(disp_info[disp_num].status_line_format, german_status);
	    menu_bar_labels = german_menu_bar_labels;
	    menu_items[0] = german_menu0;
	    menu_items[1] = german_menu1;
	    menu_items[2] = german_menu2;
	    menu_items[3] = german_menu3;
	    menu_items[4] = german_menu4;
	    ok_str = german_ok;
	    cancel_str = german_cancel;
	    close_str = german_color_picker_close_but;
	    english_str = german_english;
	    help_str = german_help;
	    german_str = german_german;
	    add_display_prompt = german_add_display_prompt;
            join_session_prompt = german_join_session_prompt;
	    language_prompt = german_language_prompt;
	    prog_name = german_name;
	    pause_cursor_width = german_pause_cur_width;
	    pause_cursor_height = german_pause_cur_height;
	    pause_cursor_x_hot = german_pause_cur_x_hot;
	    pause_cursor_y_hot = german_pause_cur_y_hot;
	    pause_cursor_bits = german_pause_cur_bits;
	    /* 
	     * some dialogs 
	     */
	    dirty_add_display_prompt = german_dirty_add_display_prompt;
            dirty_add_display_but1 = german_dirty_add_display_but1;
            dirty_add_display_but2 = german_dirty_add_display_but2;
            dirty_add_display_but3 = german_dirty_add_display_but3;

            scrollbar_prompt = german_scrollbar_prompt;
            scrollbar_but1 = german_scrollbar_but1;
            scrollbar_but2 = german_scrollbar_but2;

            save_file_format_prompt = german_save_file_format_prompt;
            save_file_format_but[1] = german_save_file_format_but1;
            save_file_format_but[2] = german_save_file_format_but2;
            save_file_format_but[3] = german_save_file_format_but3;
            save_file_format_but[4] = german_save_file_format_but4;

            read_file_format_prompt = german_read_file_format_prompt;
            read_file_format_but[1] = german_read_file_format_but1;
            read_file_format_but[2] = german_read_file_format_but2;
            read_file_format_but[3] = german_read_file_format_but3;
            read_file_format_but[4] = german_read_file_format_but4;

            color_picker_prompt = german_color_picker_prompt;
            color_picker_but1 = german_color_picker_set_col_but;
            color_picker_but2 = german_color_picker_close_but;
            color_picker_but3 = german_color_picker_grab_color_but;

            mouse_dialog_prompt = german_mouse_dialog_prompt;
            mouse_dialog_toggle1 = german_mouse_dialog_toggle1;
            mouse_dialog_toggle2 = german_mouse_dialog_toggle2;

            line_width_prompt = german_line_width_prompt;
            square_width_prompt = german_square_width_prompt;
            oval_width_prompt = german_oval_width_prompt;

	    sprintf(buf, "wscrawl.interfaceLanguage: German\n");
	    merge_resources(disp_info[disp_num].disp, buf);
	    break;
	default:
	    printf("ERROR: unknown language in set_language_of_interface.\n");
	    break;
    }

    change_interface(disp_num, language);
    fill_out_menu_data_structures(disp_num);  /* menus have changed dimension */
    change_pause_cursor(disp_num);
}


/*
 * change_pause_cursor - this function changes the pause cursor
 */
change_pause_cursor(disp_num)
int disp_num;
{
    Pixel fg, bg;
    XColor fg1, bg1;
    Pixmap source, mask;
    
    XFreeCursor(disp_info[disp_num].disp, disp_info[disp_num].pause_cursor);

    /*
     * create the "Please Wait" cursor
     */
    fg1.flags = DoRed | DoGreen | DoBlue;
    bg1.flags = DoRed | DoGreen | DoBlue;
    fg1.red = fg1.blue = fg1.green = 0;
    bg1.red = bg1.blue = bg1.green = 0;

    fg = WhitePixelOfScreen(DefaultScreenOfDisplay(disp_info[disp_num].disp));
    bg = BlackPixelOfScreen(DefaultScreenOfDisplay(disp_info[disp_num].disp));

    source = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, pause_cursor_bits,
                      pause_cursor_width, pause_cursor_height, 1, 0, 1);
    mask = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, pause_cursor_bits,
                      pause_cursor_width, pause_cursor_height, 1, 0, 1);

    disp_info[disp_num].pause_cursor = XCreatePixmapCursor(
                   disp_info[disp_num].disp, source, mask, &fg1, &bg1,
                   pause_cursor_x_hot, pause_cursor_y_hot);
}


/*
 * change_interface - this function changes the interface to
 *                    the globally prepared values.
 */
change_interface(disp_num, language)
int disp_num, language;
{
    int i, j, n, max_width_item, max_width;
    Arg args[MAX_ARGS];
    XmString label_str = NULL;
    Widget button;
    XSizeHints size_hints;
    XFontStruct *the_font_struct;

    /*
     * prepare some general terms for later use
     */
    ok_label_str = XmStringCreateLtoR(ok_str, charset);
    cancel_label_str = XmStringCreateLtoR(cancel_str, charset);

    /*
     * change the menu bar labels
     */
    for (i=0; i < disp_info[disp_num].NUM_EFF_MENUS; i++)
    {
        n = 0;
	label_str = XmStringCreateLtoR(menu_bar_labels[i], charset);
        XtSetArg(args[n], XmNlabelString, label_str); n++;
        XtSetValues(disp_info[disp_num].menu[i].menu_bar_button, args, n);
	if (label_str)
	    XtFree((char*)label_str);
    }

    /*
     * change all the menu item labels
     */
    the_font_struct = XLoadQueryFont(disp_info[disp_num].disp, "variable");

    for (j=0; j < NUM_MENUS - 1; j++)
    {
	max_width = find_max_width_item(disp_num, j, the_font_struct);

        for (i=0; i < disp_info[disp_num].menu[j].num_menu_items; i++)
        {
            n = 0;
	    label_str = XmStringCreateLtoR(menu_items[j][i], charset);
            XtSetArg(args[n], XmNlabelString, label_str); n++;
            XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
            XtSetValues(disp_info[disp_num].menu[j].menu_item[i], args, n);
	    if (label_str)
	        XtFree((char*)label_str);
        }

        n = 0;
        XtSetArg(args[n], XmNwidth, (Dimension) (max_width + 22)); n++;
        XtSetValues(disp_info[disp_num].menu[j].menu_shell, args, n);
    }

    n = 0;
    label_str = XmStringCreateLtoR(help_str, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].help_menu_bar_button, args, n);
    if (label_str)
        XtFree((char*)label_str);

    draw_status_win_message(disp_num);

    change_language_dialog(disp_num);
    change_add_display_dialog(disp_num);
    change_join_session_dialog(disp_num);
    change_dirty_add_dialog(disp_num);
    change_scrollbar_dialog(disp_num);
    change_save_file_dialog(disp_num);
    change_read_file_dialog(disp_num);
    change_color_picker_dialog(disp_num);
    change_mouse_dialog(disp_num);
    change_line_width_dialog(disp_num);
    
    size_hints.flags = NULL;
    XSetStandardProperties(disp_info[disp_num].disp,
			   XtWindow(disp_info[disp_num].app_shell),
		           prog_name, "wscrawl", None, NULL, 0, &size_hints);
}


/*
 * find_max_width_item - this function returns the widest element of the
 *                       menu passed in the current language.
 */
find_max_width_item(disp_num, menu_num, the_font_struct)
int disp_num, menu_num;
XFontStruct *the_font_struct;
{
    int i, max_width, width, max_item;

    max_width = 0;
    for (i=0; i < disp_info[disp_num].menu[menu_num].num_menu_items; i++)
    {
        if (strcmp(menu_items[menu_num][i], "----") != 0)
        {
            width = XTextWidth(the_font_struct, menu_items[menu_num][i],
                               strlen(menu_items[menu_num][i]));

            if (width > max_width)
            {
                max_width = width;
                max_item = i;
            }
        }
    }
    
    return(max_width);
}


/*
 * change_language_dialog - this function changes the language dialog to the
 *                          correct language.
 */
change_language_dialog(disp_num)
int disp_num;
{
    int i, j, n;
    Arg args[MAX_ARGS];
    XmString label_str = NULL;
    Widget button;

    n = 0;
    XtSetArg(args[n], XmNlabelString, ok_label_str); n++;
    XtSetValues(disp_info[disp_num].language_dialog.ok_button, args, n);

    n = 0;
    label_str = XmStringCreateLtoR(language_prompt, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].language_dialog.label, args, n);
    if (label_str)
        XtFree((char*)label_str);

    n = 0;
    label_str = XmStringCreateLtoR(english_str, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].language_dialog.english_button, args, n);
    if (label_str)
        XtFree((char*)label_str);

    n = 0;
    label_str = XmStringCreateLtoR(german_str, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].language_dialog.german_button, args, n);
    if (label_str)
        XtFree((char*)label_str);
}


/*
 * change_ok_and_cancel_buttons - this function changes the selection box
 *                                ok and cancel buttons.
 */
change_ok_and_cancel_buttons(the_dialog)
Widget the_dialog;
{
    int n;
    Arg args[MAX_ARGS];
    Widget button;

    button = XmSelectionBoxGetChild(the_dialog, XmDIALOG_OK_BUTTON);
    n = 0;
    XtSetArg(args[n], XmNlabelString, ok_label_str); n++;
    XtSetValues(button, args, n);

    button = XmSelectionBoxGetChild(the_dialog, XmDIALOG_CANCEL_BUTTON);
    n = 0;
    XtSetArg(args[n], XmNlabelString, cancel_label_str); n++;
    XtSetValues(button, args, n);
}


/*
 * change_add_display_dialog - this function changes the add_display dialog
 */
change_add_display_dialog(disp_num)
int disp_num;
{
    int n;
    Arg args[MAX_ARGS];
    XmString label_str = NULL;

    change_ok_and_cancel_buttons(disp_info[disp_num].add_display_dialog);

    n = 0;
    label_str = XmStringCreateLtoR(add_display_prompt, charset);
    XtSetArg(args[n], XmNselectionLabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].add_display_dialog, args, n);
    if (label_str)
        XtFree((char*)label_str);
}


/*
 * change_join_session_dialog - this function changes the join_session dialog
 */
change_join_session_dialog(disp_num)
int disp_num;
{
    int n;
    Arg args[MAX_ARGS];
    XmString label_str = NULL;

    change_ok_and_cancel_buttons(disp_info[disp_num].join_session_dialog);

    n = 0;
    label_str = XmStringCreateLtoR(join_session_prompt, charset);
    XtSetArg(args[n], XmNselectionLabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].join_session_dialog, args, n);
    if (label_str)
        XtFree((char*)label_str);
}


/*
 * change_dirty_add_dialog - this function changes the dirty_add_display dialog
 */
change_dirty_add_dialog(disp_num)
int disp_num;
{
    int n;
    Arg args[MAX_ARGS];
    XmString label_str = NULL;

    n = 0;
    label_str = XmStringCreateLtoR(dirty_add_display_prompt, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].dirty_add_display_dialog.label, args, n);
    if (label_str)
        XtFree((char*)label_str);

    n = 0;
    label_str = XmStringCreateLtoR(dirty_add_display_but1, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].dirty_add_display_dialog.copy_all_button,
		args, n);
    if (label_str)
        XtFree((char*)label_str);

    n = 0;
    label_str = XmStringCreateLtoR(dirty_add_display_but2, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].dirty_add_display_dialog.copy_vis_button,
		args, n);
    if (label_str)
        XtFree((char*)label_str);

    n = 0;
    label_str = XmStringCreateLtoR(dirty_add_display_but3, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].dirty_add_display_dialog.clear_all_button,
		args, n);
    if (label_str)
        XtFree((char*)label_str);

    n = 0;
    XtSetArg(args[n], XmNlabelString, cancel_label_str); n++;
    XtSetValues(disp_info[disp_num].dirty_add_display_dialog.cancel_button,
		args, n);
}


/*
 * change_scrollbar_dialog - this function changes the scrollbar dialog
 */
change_scrollbar_dialog(disp_num)
int disp_num;
{
    int n;
    Arg args[MAX_ARGS];
    XmString label_str = NULL;

    n = 0;
    label_str = XmStringCreateLtoR(scrollbar_prompt, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].scrollbar_dialog.label, args, n);
    if (label_str)
        XtFree((char*)label_str);

    n = 0;
    label_str = XmStringCreateLtoR(scrollbar_but1, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].scrollbar_dialog.scroll_sync_button,
		args, n);
    if (label_str)
        XtFree((char*)label_str);

    n = 0;
    label_str = XmStringCreateLtoR(scrollbar_but2, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].scrollbar_dialog.scroll_a_sync_button,
		args, n);
    if (label_str)
        XtFree((char*)label_str);

    n = 0;
    XtSetArg(args[n], XmNlabelString, cancel_label_str); n++;
    XtSetValues(disp_info[disp_num].scrollbar_dialog.cancel_button,
		args, n);
}


/*
 * change_color_picker_dialog - this function changes the color picker dialog
 */
change_color_picker_dialog(disp_num)
int disp_num;
{
    int n;
    Arg args[MAX_ARGS];
    XmString label_str = NULL;

    n = 0;
    label_str = XmStringCreateLtoR(color_picker_prompt, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].color_picker_dialog.label, args, n);
    if (label_str)
        XtFree((char*)label_str);

    n = 0;
    label_str = XmStringCreateLtoR(color_picker_but1, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].color_picker_dialog.alloc_button, args, n);
    if (label_str)
        XtFree((char*)label_str);

    n = 0;
    label_str = XmStringCreateLtoR(color_picker_but2, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].color_picker_dialog.cancel_button,args, n);
    if (label_str)
        XtFree((char*)label_str);

    n = 0;
    label_str = XmStringCreateLtoR(color_picker_but3, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].color_picker_dialog.grab_button,args, n);
    if (label_str)
        XtFree((char*)label_str);
}


/*
 * change_mouse_dialog - this function changes the mouse_options dialog
 */
change_mouse_dialog(disp_num)
int disp_num;
{
    int n;
    Arg args[MAX_ARGS];
    XmString label_str = NULL;

    n = 0;
    label_str = XmStringCreateLtoR(mouse_dialog_prompt, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].mouse_dialog.label, args, n);
    if (label_str)
        XtFree((char*)label_str);

    n = 0;
    label_str = XmStringCreateLtoR(mouse_dialog_toggle1, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].mouse_dialog.coordinates_toggle, args, n);
    if (label_str)
        XtFree((char*)label_str);

    n = 0;
    label_str = XmStringCreateLtoR(mouse_dialog_toggle2, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].mouse_dialog.telepointer_toggle, args, n);
    if (label_str)
        XtFree((char*)label_str);

    n = 0;
    label_str = XmStringCreateLtoR(close_str, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].mouse_dialog.close_button, args, n);
    if (label_str)
        XtFree((char*)label_str);
}


/*
 * change_line_width_dialog - this function changes the line width prompts
 *                            in several dialogs
 */
change_line_width_dialog(disp_num)
int disp_num;
{
    int n;
    Arg args[MAX_ARGS];
    XmString label_str = NULL;
    Widget button, prompt;

    prompt = XmSelectionBoxGetChild(disp_info[disp_num].line_dialog.the_dialog,
				    XmDIALOG_SELECTION_LABEL);
    n = 0;
    label_str = XmStringCreateLtoR(line_width_prompt, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(prompt, args, n);
    if (label_str)
        XtFree((char*)label_str);
    change_ok_and_cancel_buttons(disp_info[disp_num].line_dialog.the_dialog);

    prompt = XmSelectionBoxGetChild(
	     disp_info[disp_num].square_dialog.the_dialog, 
	     XmDIALOG_SELECTION_LABEL);
    n = 0;
    label_str = XmStringCreateLtoR(square_width_prompt, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(prompt, args, n);
    if (label_str)
        XtFree((char*)label_str);
    change_ok_and_cancel_buttons(disp_info[disp_num].square_dialog.the_dialog);

    prompt = XmSelectionBoxGetChild(disp_info[disp_num].oval_dialog.the_dialog,
				    XmDIALOG_SELECTION_LABEL);
    n = 0;
    label_str = XmStringCreateLtoR(oval_width_prompt, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(prompt, args, n);
    if (label_str)
        XtFree((char*)label_str);
    change_ok_and_cancel_buttons(disp_info[disp_num].oval_dialog.the_dialog);
}


/*
 * change_save_file_dialog - this function changes the save_file_dialog
 */
change_save_file_dialog(disp_num)
int disp_num;
{
    int n, i;
    Arg args[MAX_ARGS];
    XmString label_str = NULL;
    Widget but;

    n = 0;
    label_str = XmStringCreateLtoR(save_file_format_prompt, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].save_file_dialog.label, args, n);
    if (label_str)
        XtFree((char*)label_str);

    for (i=1; i<=4; i++)
    {
        n = 0;
        label_str = XmStringCreateLtoR(save_file_format_but[i], charset);
        XtSetArg(args[n], XmNlabelString, label_str); n++;
	switch(i)
	{
	    case 1:   
		but = disp_info[disp_num].save_file_dialog.bitmap_button;
		break;
	    case 2:   
		but = disp_info[disp_num].save_file_dialog.image_button;
		break;
	    case 3:   
		but = disp_info[disp_num].save_file_dialog.greyscale_ps_button;
		break;
	    case 4:   
		but = disp_info[disp_num].save_file_dialog.ppm_button;
		break;
	    default:
		printf("ERROR: in change_save_file_dialog.\n");
		break;
	}
        XtSetValues(but, args, n);
        if (label_str)
            XtFree((char*)label_str);
    }

    n = 0;
    XtSetArg(args[n], XmNlabelString, cancel_label_str); n++;
    XtSetValues(disp_info[disp_num].save_file_dialog.cancel_button, args, n);
}


/*
 * change_read_file_dialog - this function changes the read_file_dialog
 */
change_read_file_dialog(disp_num)
int disp_num;
{
    int n, i;
    Arg args[MAX_ARGS];
    XmString label_str = NULL;
    Widget but;

    n = 0;
    label_str = XmStringCreateLtoR(read_file_format_prompt, charset);
    XtSetArg(args[n], XmNlabelString, label_str); n++;
    XtSetValues(disp_info[disp_num].read_file_dialog.label, args, n);
    if (label_str)
        XtFree((char*)label_str);

    for (i=1; i<=4; i++)
    {
        n = 0;
        label_str = XmStringCreateLtoR(read_file_format_but[i], charset);
        XtSetArg(args[n], XmNlabelString, label_str); n++;
	switch(i)
	{
	    case 1:   
		but = disp_info[disp_num].read_file_dialog.bitmap_button;
		break;
	    case 2:   
		but = disp_info[disp_num].read_file_dialog.image_button;
		break;
	    case 3:   
		but = disp_info[disp_num].read_file_dialog.text_file_button;
		break;
	    case 4:   
		but = disp_info[disp_num].read_file_dialog.ppm_button;
		break;
	    default:
		printf("ERROR: in change_read_file_dialog.\n");
		break;
	}
        XtSetValues(but, args, n);
        if (label_str)
            XtFree((char*)label_str);
    }

    n = 0;
    XtSetArg(args[n], XmNlabelString, cancel_label_str); n++;
    XtSetValues(disp_info[disp_num].read_file_dialog.cancel_button, args, n);
}

