/*
 * This File: wsmenu.c
 *
 * This file contains the menu functions to make the menus behave correctly.
 * It also contains the tool palette code.
 */

#include "wscrawl.h"
#include "rubpoint.h"
#include "selectb.h"
#include "paintb.h"
#include "pencil.h"
#include "spraycan.h"
#include "typesym.h"
#include "linesym.h"
#include "eraser.h"
#include "square.h"
#include "fillsquare.h"
#include "oval.h"
#include "filloval.h"

extern struct disp_info_struct disp_info[MAX_NUM_DISPS];
extern int num_disps;
extern int DOUBLE_CLICK_THRESHOLD[MAX_NUM_DISPS];         /* in milliseconds */
static XmStringCharSet charset = (XmStringCharSet) XmSTRING_DEFAULT_CHARSET;
extern struct global_values_struct global_values;
XtEventHandler menu_buttonExposeCB();

extern Widget create_add_display_dialog();
extern Widget create_line_dialog();
extern Widget create_square_dialog();
extern Widget create_oval_dialog();
extern Widget create_change_pointer_dialog();
extern Widget create_join_session_dialog();
extern Widget create_language_dialog();
extern Widget create_attract_attention_dialog();
extern Widget create_mouse_dialog();
extern Widget create_color_picker_dialog();
extern Widget create_brush_dialog();
extern Widget create_eraser_dialog();
extern Widget create_dirty_add_display_dialog();
extern Widget create_scrollbar_dialog();
extern Widget create_select_dialog();
extern Widget create_read_file_dialog();
extern Widget create_save_file_dialog();


/*
 * boink_tool - this function artificially selects one of the tools on
 *              the tool palette from within this program.  It does
 *              this by filling out the correct structures and calling
 *              "tool_activateCB()".
 */
boink_tool(disp_num, which_tool)
int disp_num, which_tool;
{
    Widget dummy = (Widget)NULL;
    XEvent tmp_event;
    XmDrawingAreaCallbackStruct cb_struct;
    XButtonReleasedEvent *the_event;

    cb_struct.event = &tmp_event;
    the_event = (XButtonReleasedEvent *) cb_struct.event;
    the_event->time = ~0;    /* this is so that it is not a double click */

    tool_activateCB(dummy, disp_num << 8 | which_tool, &cb_struct);
}


/*
 * tool_activateCB - this function is called when a tool on the tool palette
 *                   get's selected.  This function then takes care of 
 *                   changing the tool, changing the cursor, etc.
 */
tool_activateCB(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int i;
    int which, which_tool, disp_num;
    Cursor the_cursor;
    XGCValues gcvalues;
    XmDrawingAreaCallbackStruct *cb_struct;
    XButtonReleasedEvent *the_event;
    static long last_click_time[MAX_NUM_DISPS];
    static int last_tool_selected[MAX_NUM_DISPS];
    static int pencil_click = 0;
    char tmp_str[50], *the_text;
    XtWidgetGeometry geom_req, reply_ret;

    disp_num = ((int) client_data) >> 8;
    which_tool = ((int) client_data) & 0x00FF;
  
    cb_struct = (XmDrawingAreaCallbackStruct *) call_data;
    the_event = (XButtonReleasedEvent *) cb_struct->event;

    if (disp_info[disp_num].selected_region.region_is_selected == TRUE)
    {
        place_selection_permanently(disp_num);

	if (which_tool == PENCIL_BUTTON) /* this is to detect fatbits activate*/
	    pencil_click = 1;
	else
	    pencil_click = 0;
    }
    else
    {
	if ((which_tool == PENCIL_BUTTON) && (pencil_click == 1))
	    pencil_click = 2;  /* activate fatbits */
	else
	    pencil_click = 0; 
    }


    if (disp_info[disp_num].input_mode == TYPING)
	erase_type_cursor(disp_num);	/* erase type cursor */

    disp_info[disp_num].pointer_state = NOT_PRESSED;

    switch(which_tool)
    {
	case POINTER_BUTTON:
 	    the_cursor = XC_target;
	    disp_info[disp_num].input_mode = POINTING;
	    if (double_click(disp_num, the_event, last_click_time, 
			     last_tool_selected, which_tool))
            {
                process_menu_selection(disp_num,OPTIONS_MENU,TELEPOINTER_LABEL);
	    }
	    break;
	case SELECTION_BUTTON:
 	    the_cursor = XC_crosshair;
	    disp_info[disp_num].input_mode = SELECTING;
            gcvalues.line_width = 0;
            XChangeGC(disp_info[disp_num].disp, 
		      disp_info[disp_num].rubber_band_gc,
                      GCLineWidth, &gcvalues);
	    if (double_click(disp_num, the_event, last_click_time, 
			     last_tool_selected, which_tool))
            {
                process_menu_selection(disp_num, OPTIONS_MENU, SUPER_SELECTION);
	    }
	    break;
	case PENCIL_BUTTON:
 	    the_cursor = XC_pencil;
	    disp_info[disp_num].input_mode = PENCILING;
            for (i=0; i < num_disps; i++)
            {
                if (disp_info[i].in_session_bool)
                {
                    gcvalues.line_width = 0;
                    XChangeGC(disp_info[i].disp, disp_info[disp_num].win_gc[i],
                              GCLineWidth, &gcvalues);
                }
            }

	    if (double_click(disp_num, the_event, last_click_time, 
			     last_tool_selected, which_tool))
            {
	        if (pencil_click == 2)	/* fake a selection */
	        {
		    disp_info[disp_num].orig_rubberband_point.x = 
			         disp_info[disp_num].selected_region.cur_x;
		    disp_info[disp_num].last_rubberband_point.x = 
				 disp_info[disp_num].selected_region.cur_x +
				 disp_info[disp_num].selected_region.width;
		    disp_info[disp_num].orig_rubberband_point.y = 
			         disp_info[disp_num].selected_region.cur_y;
		    disp_info[disp_num].last_rubberband_point.y = 
				 disp_info[disp_num].selected_region.cur_y +
				 disp_info[disp_num].selected_region.height;
	            define_selection(disp_num);
	        }
                process_menu_selection(disp_num, EDIT_MENU, FATBITS_SELECTION);
	        if (pencil_click == 2)
		{
		    boink_tool(disp_num, SELECTION_BUTTON);
		    return;
		}
	    }
	    break;
	case TYPE_BUTTON:
	    if (double_click(disp_num, the_event, last_click_time, 
			     last_tool_selected, which_tool))
            {
		popup_info_dialog(disp_num, 
		      "There is no double-click\noption on this tool.");
	    }
	    if (disp_info[disp_num].input_mode == TYPING)
	    {
	        draw_type_cursor(disp_num);
		return;			/* no sense in resetting everything */
	    }
 	    the_cursor = XC_xterm;
	    disp_info[disp_num].input_mode = TYPING;
            disp_info[disp_num].type_info.orig_pos.x = 40;
            disp_info[disp_num].type_info.orig_pos.y = 40;
            disp_info[disp_num].type_info.cur_pos.x = 40;
            disp_info[disp_num].type_info.cur_pos.y = 40;
            disp_info[disp_num].type_info.prompt_width =
             disp_info[disp_num].type_info.the_font_struct->max_bounds.width;
            disp_info[disp_num].type_info.char_height =
         (disp_info[disp_num].type_info.the_font_struct)->max_bounds.ascent+
         (disp_info[disp_num].type_info.the_font_struct)->max_bounds.descent;
            disp_info[disp_num].type_info.type_history = NULL;
            /*
             * draw the cursor
             */
	    draw_type_cursor(disp_num);
	    break;
	case PAINTBRUSH_BUTTON:
	    which = disp_info[disp_num].paintbrush.which_paintb;
            XDefineCursor(disp_info[disp_num].disp, 
			  disp_info[disp_num].draw_win_id,
			  disp_info[disp_num].paintbrush.paintb_curs[which]);

	    disp_info[disp_num].input_mode = PAINTING;

	    if (disp_info[disp_num].paintbrush.paint_mode == NORMAL)
	    {
                for (i=0; i < num_disps; i++)
                {
                    if (disp_info[i].in_session_bool)
                    {
                        gcvalues.line_width = 
			             disp_info[disp_num].paintbrush.brush_width;
                        XChangeGC(disp_info[i].disp, 
			      disp_info[disp_num].win_gc[i],
                              GCLineWidth, &gcvalues);
                    }
                }
	    }

	    if (double_click(disp_num, the_event, last_click_time, 
			     last_tool_selected, which_tool))
            {
                process_menu_selection(disp_num, OPTIONS_MENU, BRUSH_SHAPES);
	    }
	    break;
	case SPRAYCAN_BUTTON:
 	    the_cursor = XC_spraycan;
	    disp_info[disp_num].input_mode = SPRAYPAINTING;
	    if (double_click(disp_num, the_event, last_click_time, 
			     last_tool_selected, which_tool))
            {
		popup_info_dialog(disp_num, 
		      "There is no double-click\noption on this tool.");
	    }
	    break;
	case ERASER_BUTTON:
 	    the_cursor = XC_draped_box;
	    disp_info[disp_num].input_mode = ERASING;
	    if (double_click(disp_num, the_event, last_click_time, 
			     last_tool_selected, which_tool))
            {
                process_menu_selection(disp_num, OPTIONS_MENU, ERASER_OPTIONS);
	    }
	    break;
	case LINE_BUTTON:
 	    the_cursor = XC_crosshair;
	    disp_info[disp_num].input_mode = DRAWING_LINES;
            for (i=0; i < num_disps; i++)
            {
                if (disp_info[i].in_session_bool)
                {
                    gcvalues.line_width = 
				    disp_info[disp_num].line_dialog.line_width;
                    XChangeGC(disp_info[i].disp, disp_info[disp_num].win_gc[i],
                              GCLineWidth, &gcvalues);
                }
            }
            gcvalues.line_width = disp_info[disp_num].line_dialog.line_width;
            XChangeGC(disp_info[disp_num].disp, 
		      disp_info[disp_num].rubber_band_gc,
                      GCLineWidth, &gcvalues);
	    if (double_click(disp_num, the_event, last_click_time, 
			     last_tool_selected, which_tool))
            {
                process_menu_selection(disp_num, OPTIONS_MENU, LINE_SETTINGS);
	    }
	    break;
	case OUTLINESQUARE_BUTTON:
 	    the_cursor = XC_crosshair;
	    disp_info[disp_num].input_mode = OUTLINING_SQUARES;
            for (i=0; i < num_disps; i++)
            {
                if (disp_info[i].in_session_bool)
                {
                    gcvalues.line_width =
			         disp_info[disp_num].square_dialog.line_width;
                    XChangeGC(disp_info[i].disp, disp_info[disp_num].win_gc[i],
                              GCLineWidth, &gcvalues);
                }
            }
            gcvalues.line_width = disp_info[disp_num].square_dialog.line_width;
            XChangeGC(disp_info[disp_num].disp, 
		      disp_info[disp_num].rubber_band_gc,
                      GCLineWidth, &gcvalues);
	    if (double_click(disp_num, the_event, last_click_time, 
			     last_tool_selected, which_tool))
            {
                process_menu_selection(disp_num, OPTIONS_MENU, SQUARE_SETTINGS);
	    }
	    break;
	case FILLSQUARE_BUTTON:
 	    the_cursor = XC_crosshair;
	    disp_info[disp_num].input_mode = FILLING_SQUARES;
            gcvalues.line_width = 0;
            XChangeGC(disp_info[disp_num].disp, 
		      disp_info[disp_num].rubber_band_gc,
                      GCLineWidth, &gcvalues);
	    if (double_click(disp_num, the_event, last_click_time, 
			     last_tool_selected, which_tool))
            {
		popup_info_dialog(disp_num, 
		      "There is no double-click\noption on this tool.");
	    }
	    break;
	case OUTLINEOVAL_BUTTON:
 	    the_cursor = XC_crosshair;
	    disp_info[disp_num].input_mode = OUTLINING_OVALS;
            gcvalues.line_width = disp_info[disp_num].oval_dialog.line_width;
            XChangeGC(disp_info[disp_num].disp, 
		      disp_info[disp_num].rubber_band_gc,
                      GCLineWidth, &gcvalues);
            for (i=0; i < num_disps; i++)
            {
                if (disp_info[i].in_session_bool)
                {
                    gcvalues.line_width = 
				  disp_info[disp_num].oval_dialog.line_width;
                    XChangeGC(disp_info[i].disp, disp_info[disp_num].win_gc[i],
                              GCLineWidth, &gcvalues);
                }
            }
	    if (double_click(disp_num, the_event, last_click_time, 
			     last_tool_selected, which_tool))
            {
                process_menu_selection(disp_num, OPTIONS_MENU, OVAL_SETTINGS);
	    }
	    break;
	case FILLOVAL_BUTTON:
 	    the_cursor = XC_crosshair;
	    disp_info[disp_num].input_mode = FILLING_OVALS;
            gcvalues.line_width = 0;
            XChangeGC(disp_info[disp_num].disp, 
		      disp_info[disp_num].rubber_band_gc,
                      GCLineWidth, &gcvalues);
	    if (double_click(disp_num, the_event, last_click_time, 
			     last_tool_selected, which_tool))
            {
		popup_info_dialog(disp_num, 
		      "There is no double-click\noption on this tool.");
	    }
	    break;
	default:
	    printf("ERROR: unknown tool selected in tool_activateCB.\n");
	    break;
    }

    if (which_tool != PAINTBRUSH_BUTTON)
    {
         XDefineCursor(disp_info[disp_num].disp,disp_info[disp_num].draw_win_id,
		 XCreateFontCursor(disp_info[disp_num].disp, the_cursor));
    }

    last_click_time[disp_num] = the_event->time;
    last_tool_selected[disp_num] = which_tool;
}


/*
 * double_click - this function returns TRUE if this tool has just been clicked
 *                for the second time in a double click sequence.  Otherwise
 *                it returns FALSE.
 */
double_click(disp_num, the_event, last_clk_time, last_tl_slcted, which_tool)
int disp_num;
XButtonReleasedEvent *the_event;
long *last_clk_time;
int *last_tl_slcted, which_tool;
{
    if (last_tl_slcted[disp_num] != which_tool)
    {
	return(FALSE);
    }
    else if ((last_clk_time[disp_num] + DOUBLE_CLICK_THRESHOLD[disp_num]) < 
	     the_event->time)
    {
	return(FALSE);
    }
    else
    {
	return(TRUE);
    }
}


/*
 * swap_shadow_colors - this function swaps the indicated widgets top and
 *                        bottom shadow colors.  This is used to uninvert
 *                        the initial menu bar button pressed.
 */
swap_shadow_colors(disp_num, which_menu)
int disp_num, which_menu;
{
    int n, i, j;
    Arg args[MAX_ARGS];
    Pixel top_color, bottom_color;
    Widget w;

    w = disp_info[disp_num].menu[which_menu].menu_bar_button;

    n = 0;
    XtSetArg(args[n], XmNtopShadowColor, &top_color); n++;
    XtSetArg(args[n], XmNbottomShadowColor, &bottom_color); n++;
    XtGetValues(w, args, n);

    n = 0;
    XtSetArg(args[n], XmNtopShadowColor, bottom_color); n++;
    XtSetArg(args[n], XmNbottomShadowColor, top_color); n++;
    XtSetValues(w, args, n);

    if (disp_info[disp_num].menu[which_menu].menu_bar_inverted == TRUE)
        disp_info[disp_num].menu[which_menu].menu_bar_inverted = FALSE;
    else
        disp_info[disp_num].menu[which_menu].menu_bar_inverted = TRUE;
}


/*
 * set_orig_shadow_colors - this function sets the indicated widgets top and
 *                        bottom shadow colors to the original settings.  This 
 *                        is used to uninvert the initial menu bar button 
 *                        pressed.
 */
set_orig_shadow_colors(disp_num, which_menu)
int disp_num, which_menu;
{
    int n, i, j;
    Arg args[MAX_ARGS];
    Pixel top_color, bottom_color;

    n = 0;
    XtSetArg(args[n], XmNtopShadowColor, 
	disp_info[disp_num].menu[which_menu].top_shadow_color); n++;
    XtSetArg(args[n], XmNbottomShadowColor, 
	disp_info[disp_num].menu[which_menu].bottom_shadow_color); n++;
    XtSetValues(disp_info[disp_num].menu[which_menu].menu_bar_button, args, n);

    disp_info[disp_num].menu[which_menu].menu_bar_inverted = FALSE;
}


/*
 * set_alternate_shadow_colors - this function sets the indicated widgets top 
 *                      and bottom shadow colors to the original settings. This
 *                      is used to uninvert the initial menu bar button 
 *                      pressed.
 */
set_alternate_shadow_colors(disp_num, which_menu)
int disp_num, which_menu;
{
    int n, i, j;
    Arg args[MAX_ARGS];
    Pixel top_color, bottom_color;

    n = 0;
    XtSetArg(args[n], XmNtopShadowColor, 
	disp_info[disp_num].menu[which_menu].bottom_shadow_color); n++;
    XtSetArg(args[n], XmNbottomShadowColor, 
	disp_info[disp_num].menu[which_menu].top_shadow_color); n++;
    XtSetValues(disp_info[disp_num].menu[which_menu].menu_bar_button, args, n);

    disp_info[disp_num].menu[which_menu].menu_bar_inverted = TRUE;
}


/*
 * paint_button_shadow - this function paints the button shadow in the menu
 *                       bar.  It's all smoke and mirrors, dude.  Motif
 *                       cannot take us where we want to go.
 */
paint_button_shadow(disp_num, which_menu)
int disp_num, which_menu;
{
    Widget w;
    Dimension d_width, d_height;
    int x, y, width, height;
    int n;
    Arg args[MAX_ARGS];

    w = disp_info[disp_num].menu[which_menu].menu_bar_button;
    n = 0;
    XtSetArg(args[n], XmNwidth, &d_width); n++;
    XtSetArg(args[n], XmNheight, &d_height); n++;
    XtGetValues(w, args, n);

    x = 2;
    y = 3;
    width = ((int) d_width) - 4 ;
    height = ((int) d_height) - 6;

    /* top */
    XDrawLine(disp_info[disp_num].disp, XtWindow(w),
              disp_info[disp_num].menu_top_shadow_gc, x, y, x + width, y);
    XDrawLine(disp_info[disp_num].disp, XtWindow(w),
              disp_info[disp_num].menu_top_shadow_gc, x, y+1, x + width, y+1);

    /* left */
    XDrawLine(disp_info[disp_num].disp, XtWindow(w),
              disp_info[disp_num].menu_top_shadow_gc, x, y, x, y + height);
    XDrawLine(disp_info[disp_num].disp, XtWindow(w),
              disp_info[disp_num].menu_top_shadow_gc, x+1, y, x+1, y + height);

    /* bottom */
    XDrawLine(disp_info[disp_num].disp, XtWindow(w),
              disp_info[disp_num].menu_bottom_shadow_gc, 
	      x+1, y + height - 1, x + width, y + height - 1);
    XDrawLine(disp_info[disp_num].disp, XtWindow(w),
              disp_info[disp_num].menu_bottom_shadow_gc, 
	      x, y + height, x + width, y + height);

    /* right */
    XDrawLine(disp_info[disp_num].disp, XtWindow(w),
              disp_info[disp_num].menu_bottom_shadow_gc, 
	      x+width - 1 , y+1, x+width - 1, y + height);
    XDrawLine(disp_info[disp_num].disp, XtWindow(w),
              disp_info[disp_num].menu_bottom_shadow_gc, 
	      x+width, y, x+width, y + height);

    XFlush(disp_info[disp_num].disp);
}


/*
 * erase_button_shadow - this function paints the button shadow in the menu
 *                       bar.  It's all smoke and mirrors, folks.
 */
erase_button_shadow(disp_num, which_menu)
int disp_num, which_menu;
{
    Widget w;

    w = disp_info[disp_num].menu[which_menu].menu_bar_button;
    XClearArea(disp_info[disp_num].disp, XtWindow(w), 0, 0, 200, 50, True);
    XFlush(disp_info[disp_num].disp);
}


/*
 * menu_armCB - this function is called (for instance), when the "File"
 *              menu button has been clicked on, so that we should extend
 *              this menu.
 */
menu_armCB(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int i, x, y;
    int which_menu, disp_num;
    Pixel top_color, bottom_color;
    int n;
    Arg args[MAX_ARGS];

    disp_num = ((int) client_data) >> 8;
    which_menu = ((int) client_data) & 0x00FF;
    disp_info[disp_num].pointer_state = PRESSED;

    paint_button_shadow(disp_num, which_menu);

    disp_info[disp_num].cur_menu_num = which_menu;
    disp_info[disp_num].orig_menu_num = which_menu;

    move_menu(disp_num, which_menu);
    XtMapWidget(disp_info[disp_num].menu[which_menu].menu_shell);
    disp_info[disp_num].cur_item_num = -1;
}


/*
 * move_menu - this function moves the menu to the proper location
 */
move_menu(disp_num, which_menu)
int disp_num, which_menu;
{
    int x, y;
    XWindowAttributes attr_ret;
    XWindowChanges win_change;
    Window dummy;

    XGetWindowAttributes(disp_info[disp_num].disp, 
		XtWindow(disp_info[disp_num].menu[which_menu].menu_bar_button),
		&attr_ret);

    /*
    XtTranslateCoords(disp_info[disp_num].app_shell, attr_ret.x, attr_ret.y,
		      &x, &y);
    */
    myXtTranslateCoords(disp_num, disp_info[disp_num].app_shell, 
			attr_ret.x, attr_ret.y, &x, &y);

    y += attr_ret.height;
    x += 4;
    XtMoveWidget(disp_info[disp_num].menu[which_menu].menu_shell, x, y);
    XRaiseWindow(disp_info[disp_num].disp, 
              XtWindow(disp_info[disp_num].menu[which_menu].menu_shell));
    XFlush(disp_info[disp_num].disp);
}


/*
 * myXtTranslateCoords - this call has the exact behavior that XtTranslateCoords
 *                       SHOULD have but does not because of toolkit bugs.
 *                       This call uses Xlib calls and asks the server for
 *                       the info.  It is much slower than the Xt call would
 *                       have been.
 */
myXtTranslateCoords(disp_num, w, x, y, rootx_return, rooty_return)
int disp_num;
Widget w;
int x, y;
int *rootx_return, *rooty_return;
{
    Window win_id, dummy;

    win_id = XtWindow(w);

    XTranslateCoordinates(disp_info[disp_num].disp, win_id,
			  RootWindow(disp_info[disp_num].disp,
			  DefaultScreen(disp_info[disp_num].disp)),
			  x, y, rootx_return, rooty_return, &dummy);
}


/*
 * menu_disarmCB - this function is called (for instance), when the "File"
 *              menu button becomes disarmed, so that we should be retracted.
 */
menu_disarmCB(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int i;
    int which_menu, item_num, menu_num, disp_num, orig_menu_num, cur_menu_num;
    XmDrawingAreaCallbackStruct *cb_struct;
    XButtonReleasedEvent *the_event;
    int n;
    Arg args[MAX_ARGS];

    disp_num = ((int) client_data) >> 8;
    which_menu = ((int) client_data) & 0x00FF;
    disp_info[disp_num].pointer_state = NOT_PRESSED;

    cb_struct = (XmDrawingAreaCallbackStruct *) call_data;
    the_event = (XButtonReleasedEvent *) cb_struct->event;

    orig_menu_num = disp_info[disp_num].orig_menu_num;
    cur_menu_num = disp_info[disp_num].cur_menu_num;
    XtUnmapWidget(disp_info[disp_num].menu[cur_menu_num].menu_shell);

    if (cur_menu_num == orig_menu_num)
    {
	erase_button_shadow(disp_num, orig_menu_num);
    }
    else
    {
        n = 0;
        XtSetArg(args[n], XmNshadowThickness, 0); n++;
        XtSetArg(args[n], XmNmarginHeight, 4); n++;
        XtSetArg(args[n], XmNmarginWidth, 4); n++;
        XtSetValues(disp_info[disp_num].menu[
	        disp_info[disp_num].cur_menu_num].menu_bar_button, args, n);
    }

    if (disp_info[disp_num].cur_item_num != -1)
    {
        n = 0;
        XtSetArg(args[n], XmNshadowThickness, 0); n++;
        XtSetArg(args[n], XmNmarginHeight, 4); n++;
        XtSetArg(args[n], XmNmarginWidth, 4); n++;
        XtSetValues(disp_info[disp_num].menu[
			disp_info[disp_num].cur_menu_num].menu_item[
	                disp_info[disp_num].cur_item_num], args, n);
        
	item_num = disp_info[disp_num].cur_item_num;
        menu_num = disp_info[disp_num].cur_menu_num;

	process_menu_selection(disp_num, menu_num, item_num);
    }
}


/*
 * fill_out_menu_data_structures - this function fills out the data structures
 *       			   necessary to drag in the menus, such as
 * 				   the menu dimensions, etc.
 */
fill_out_menu_data_structures(disp_num)
int disp_num;
{
    int n, i, j;
    Arg args[MAX_ARGS];
    Dimension width, height;
    Pixel top_color, bottom_color;
    XGCValues gcvalues;

    disp_info[disp_num].menu[0].menu_bar_width = 0;

    for (i=0; i < disp_info[disp_num].NUM_EFF_MENUS; i++)
    {
        n = 0;
        XtSetArg(args[n], XmNtopShadowColor, &top_color); n++;
        XtSetArg(args[n], XmNbottomShadowColor, &bottom_color); n++;
        XtGetValues(disp_info[disp_num].tool[0], args, n);
        disp_info[disp_num].menu[i].top_shadow_color = top_color;
        disp_info[disp_num].menu[i].bottom_shadow_color = bottom_color;

	gcvalues.line_width = 0;
	gcvalues.foreground = (unsigned long) top_color;
        disp_info[disp_num].menu_top_shadow_gc = 
	       XCreateGC(disp_info[disp_num].disp,
		         disp_info[disp_num].draw_win_id, 
			 GCForeground | GCLineWidth, &gcvalues);
	gcvalues.line_width = 0;
	gcvalues.foreground = (unsigned long) bottom_color;
        disp_info[disp_num].menu_bottom_shadow_gc = 
	       XCreateGC(disp_info[disp_num].disp,
		         disp_info[disp_num].draw_win_id, 
			 GCForeground | GCLineWidth, &gcvalues);

        n = 0;
        XtSetArg(args[n], XmNwidth, &width); n++;
        XtSetArg(args[n], XmNheight, &height); n++;
        XtGetValues(disp_info[disp_num].menu[i].menu_bar_button, args, n);

        disp_info[disp_num].menu[i+1].menu_bar_width = 
	   disp_info[disp_num].menu[i].menu_bar_width + (int) width;

        disp_info[disp_num].menu[i].menu_bar_height = (int) height;

        n = 0;
        XtSetArg(args[n], XmNwidth, &width); n++;
        XtGetValues(disp_info[disp_num].menu[i].menu_shell, args, n);
        disp_info[disp_num].menu[i].menu_width = (int) width;

        disp_info[disp_num].menu[i].menu_item_dim[0] = 5;
        for (j=0; j < disp_info[disp_num].menu[i].num_menu_items; j++)
        {
	    if ((i == 0 && j == 3) || 
		(i == 0 && j == 7) ||
		(i == 1 && j == 6) ||
		(i == 1 && j == 9) ||
		(i == 2 && j == 3) ||
		(i == 3 && j == 14))       /* separator gadgets */
	    {
                disp_info[disp_num].menu[i].menu_item_dim[j+1] = 
	           disp_info[disp_num].menu[i].menu_item_dim[j] + 2;
		continue;
	    }
            n = 0;
            XtSetArg(args[n], XmNheight, &height); n++;
            XtGetValues(disp_info[disp_num].menu[i].menu_item[j], args, n);
    
            disp_info[disp_num].menu[i].menu_item_dim[j+1] = 
	       disp_info[disp_num].menu[i].menu_item_dim[j] + (int) height;
        }
    }
}


/*
 * make_selection_dependant_menus_stippled - this function makes all the
 *                                 selection dependant menus such as "Cut",
 *                                 stippled, so that you cannot select them.
 */
make_selection_dependant_menus_stippled(disp_num)
int disp_num;
{
    int n, i;
    Arg args[MAX_ARGS];

    n = 0;					/* save selection */
    XtSetArg(args[n], XmNsensitive, False); n++;
    XtSetValues(disp_info[disp_num].menu[0].menu_item[4], args, n);
    disp_info[disp_num].menu[0].selectable[4] = FALSE;

    for (i=0; i < 6; i++)	/* many things under "Edit" */
    {
	if (i == 2 || i == 3)
	    continue;    /* Paste and Paste X-Selection is not stippled */
        n = 0;
        XtSetArg(args[n], XmNsensitive, False); n++;
        XtSetValues(disp_info[disp_num].menu[1].menu_item[i], args, n);
        disp_info[disp_num].menu[1].selectable[i] = FALSE;
    }
}


/*
 * make_selection_dependant_menus_unstippled - this function makes all the
 *                                 selection dependant menus such as "Cut",
 *                                 unstippled, so that you can select them.
 */
make_selection_dependant_menus_unstippled(disp_num)
int disp_num;
{
    int n, i;
    Arg args[MAX_ARGS];

    n = 0;					/* save selection */
    XtSetArg(args[n], XmNsensitive, True); n++;
    XtSetValues(disp_info[disp_num].menu[0].menu_item[4], args, n);
    disp_info[disp_num].menu[0].selectable[4] = TRUE;

    for (i=0; i < 6; i++)	/* many things under "Edit" */
    {
        n = 0;
        XtSetArg(args[n], XmNsensitive, True); n++;
        XtSetValues(disp_info[disp_num].menu[1].menu_item[i], args, n);
        disp_info[disp_num].menu[1].selectable[i] = TRUE;
    }
}


/*
 * menu_pointer_moveCB - this function is called when the pointer moves in the
 *                       menu window.
 */
XtEventHandler menu_pointer_moveCB(w, client_data, our_event, cbool)
Widget w;
caddr_t client_data;
XEvent *our_event;
Boolean *cbool;
{
    int initial_menu, disp_num, old_menu_num, new_cur_menu, new_cur_item;
    int old_item_num, orig_menu_width, orig_menu_height;
    XMotionEvent *the_event;
    XEvent gen_event;
    int n;
    Arg args[MAX_ARGS];

    disp_num = ((int) client_data) >> 8;
    initial_menu = ((int) client_data) & 0x00FF;
    the_event = (XMotionEvent *) our_event;

    if (disp_info[disp_num].pointer_state != PRESSED)
        return;         /* not interested in this event */

    new_cur_menu = pointer_is_in_menu_num(disp_num, initial_menu,
			                  the_event->x, the_event->y);

    old_menu_num = disp_info[disp_num].cur_menu_num;
    old_item_num = disp_info[disp_num].cur_item_num;

    if ((new_cur_menu == old_menu_num) &&
        (new_cur_menu == disp_info[disp_num].orig_menu_num))
    {
        paint_button_shadow(disp_num, new_cur_menu);
    }

    if (new_cur_menu != old_menu_num)
    {
        if (old_menu_num == initial_menu)
        {
	    erase_button_shadow(disp_num, initial_menu);
        }
        else
        {
            n = 0;
            XtSetArg(args[n], XmNshadowThickness, 0); n++;
            XtSetArg(args[n], XmNmarginHeight, 4); n++;
            XtSetArg(args[n], XmNmarginWidth, 4); n++;
            XtSetValues(disp_info[disp_num].menu[old_menu_num].menu_bar_button, 
		        args, n);
        }

        XtUnmapWidget(disp_info[disp_num].menu[old_menu_num].menu_shell);
        move_menu(disp_num, new_cur_menu);
	XtMapWidget(disp_info[disp_num].menu[new_cur_menu].menu_shell);
	disp_info[disp_num].cur_menu_num = new_cur_menu;

        if (new_cur_menu == initial_menu)
        {
	    paint_button_shadow(disp_num, initial_menu);
        }
        else
        {
            n = 0;
            XtSetArg(args[n], XmNshadowThickness, 2); n++;
            XtSetArg(args[n], XmNmarginHeight, 2); n++;
            XtSetArg(args[n], XmNmarginWidth, 2); n++;
            XtSetValues(disp_info[disp_num].menu[
	            disp_info[disp_num].cur_menu_num].menu_bar_button, args, n);
        }

	if (old_item_num != -1)
	{
            n = 0;
            XtSetArg(args[n], XmNshadowThickness, 0); n++;
            XtSetArg(args[n], XmNmarginHeight, 4); n++;
            XtSetArg(args[n], XmNmarginWidth, 4); n++;
            XtSetValues(disp_info[disp_num].menu[old_menu_num].menu_item[
	                old_item_num], args, n);
	    disp_info[disp_num].cur_item_num = -1;
	    old_item_num = -1;
	}
    }

    new_cur_item = pointer_is_in_item_num(disp_num, new_cur_menu, initial_menu,
			                  the_event->x, the_event->y);

    if (new_cur_item != disp_info[disp_num].cur_item_num)
    {

	if (old_item_num != -1)
	{
            n = 0;
            XtSetArg(args[n], XmNshadowThickness, 0); n++;
            XtSetArg(args[n], XmNmarginHeight, 4); n++;
            XtSetArg(args[n], XmNmarginWidth, 4); n++;
            XtSetValues(disp_info[disp_num].menu[old_menu_num].menu_item[
	            disp_info[disp_num].cur_item_num], args, n);
	}

	disp_info[disp_num].cur_item_num = new_cur_item;

	if (new_cur_item != -1)
	{
            n = 0;
            XtSetArg(args[n], XmNshadowThickness, 2); n++;
            XtSetArg(args[n], XmNmarginHeight, 2); n++;
            XtSetArg(args[n], XmNmarginWidth, 2); n++;
            XtSetValues(disp_info[disp_num].menu[
	            disp_info[disp_num].cur_menu_num].menu_item[
	            disp_info[disp_num].cur_item_num], args, n);
	}
    }
}


/*
 * pointer_is_in_menu_num - this function returns the menu index that should
 *                          be extended.
 */
pointer_is_in_menu_num(disp_num, initial_menu, pointer_x, pointer_y)
int disp_num, initial_menu, pointer_x, pointer_y;
{
    int i, x;

    if (pointer_y > disp_info[disp_num].menu[initial_menu].menu_bar_height)
	return(disp_info[disp_num].cur_menu_num);

    x = pointer_x;
    x += disp_info[disp_num].menu[initial_menu].menu_bar_width;

    for (i=0; i < disp_info[disp_num].NUM_EFF_MENUS; i++)
    {
        if (x < disp_info[disp_num].menu[i+1].menu_bar_width)
	    return(i);
    }

    return(disp_info[disp_num].NUM_EFF_MENUS - 1);
}


/*
 * pointer_is_in_item_num - this function returns the item number that this
 *                          pointer position is in.
 */
pointer_is_in_item_num(disp_num, which_menu, initial_menu, pointer_x, pointer_y)
int disp_num, which_menu, initial_menu, pointer_x, pointer_y;
{
    int n, i, height_item, menu_width, left_edge;

    menu_width = disp_info[disp_num].menu[which_menu].menu_width;
    left_edge = disp_info[disp_num].menu[which_menu].menu_bar_width -
                disp_info[disp_num].menu[initial_menu].menu_bar_width;

    /*printf("wm = %d, px = %d, py = %d\n", which_menu, pointer_x, pointer_y);*/
    if (pointer_y < disp_info[disp_num].menu[which_menu].menu_bar_height)
	return(-1);
    else if ((pointer_x < left_edge) || (pointer_x > left_edge + menu_width))
	return(-1);
    else
    {
        pointer_y -= disp_info[disp_num].menu[which_menu].menu_bar_height;

        for (i=0; i < disp_info[disp_num].menu[which_menu].num_menu_items; i++)
        {
	    if (pointer_y < 
		disp_info[disp_num].menu[which_menu].menu_item_dim[i+1])
	    {
		if (disp_info[disp_num].menu[which_menu].selectable[i] == FALSE)
		    return(-1);    /* this item is a separator gadget, or ... */
		else
	            return(i);
	    }
        }

	return(-1);
    }
}


/*
 * put_bitmaps_into_tool_palette - this funcion stuffs the bitmaps into the
 *                                 tool palette.
 */
put_bitmaps_into_tool_palette(disp_num)
{
    Pixmap button_pixmap;
    Pixel fg, bg;
    int n;
    Arg args[MAX_ARGS];

    n = 0;
    XtSetArg(args[n], XmNforeground, &fg); n++;
    XtSetArg(args[n], XmNbackground, &bg); n++;
    XtGetValues(disp_info[disp_num].tool[0], args, n);

    button_pixmap = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, rubpoint_bits,
                      rubpoint_width, rubpoint_height, fg, bg,
                      DefaultDepth(disp_info[disp_num].disp,
                      DefaultScreen(disp_info[disp_num].disp)));
    n = 0;
    XtSetArg(args[n], XmNlabelType, XmPIXMAP); n++;
    XtSetArg(args[n], XmNlabelPixmap, button_pixmap); n++;
    XtSetValues(disp_info[disp_num].tool[0], args, n);

    button_pixmap = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, selectb_bits,
                      selectb_width, selectb_height, fg, bg,
                      DefaultDepth(disp_info[disp_num].disp,
                      DefaultScreen(disp_info[disp_num].disp)));
    n = 0;
    XtSetArg(args[n], XmNlabelType, XmPIXMAP); n++;
    XtSetArg(args[n], XmNlabelPixmap, button_pixmap); n++;
    XtSetValues(disp_info[disp_num].tool[1], args, n);

    button_pixmap = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, paintb_bits,
                      paintb_width, paintb_height, fg, bg,
                      DefaultDepth(disp_info[disp_num].disp,
                      DefaultScreen(disp_info[disp_num].disp)));
    n = 0;
    XtSetArg(args[n], XmNlabelType, XmPIXMAP); n++;
    XtSetArg(args[n], XmNlabelPixmap, button_pixmap); n++;
    XtSetValues(disp_info[disp_num].tool[2], args, n);

    button_pixmap = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, pencil_bits,
                      pencil_width, pencil_height, fg, bg,
                      DefaultDepth(disp_info[disp_num].disp,
                      DefaultScreen(disp_info[disp_num].disp)));
    n = 0;
    XtSetArg(args[n], XmNlabelType, XmPIXMAP); n++;
    XtSetArg(args[n], XmNlabelPixmap, button_pixmap); n++;
    XtSetValues(disp_info[disp_num].tool[3], args, n);

    button_pixmap = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, spraycan_bits,
                      spraycan_width, spraycan_height, fg, bg,
                      DefaultDepth(disp_info[disp_num].disp,
                      DefaultScreen(disp_info[disp_num].disp)));
    n = 0;
    XtSetArg(args[n], XmNlabelType, XmPIXMAP); n++;
    XtSetArg(args[n], XmNlabelPixmap, button_pixmap); n++;
    XtSetValues(disp_info[disp_num].tool[4], args, n);

    button_pixmap = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, typesym_bits,
                      typesym_width, typesym_height, fg, bg,
                      DefaultDepth(disp_info[disp_num].disp,
                      DefaultScreen(disp_info[disp_num].disp)));
    n = 0;
    XtSetArg(args[n], XmNlabelType, XmPIXMAP); n++;
    XtSetArg(args[n], XmNlabelPixmap, button_pixmap); n++;
    XtSetValues(disp_info[disp_num].tool[5], args, n);

    button_pixmap = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, linesym_bits,
                      linesym_width, linesym_height, fg, bg,
                      DefaultDepth(disp_info[disp_num].disp,
                      DefaultScreen(disp_info[disp_num].disp)));
    n = 0;
    XtSetArg(args[n], XmNlabelType, XmPIXMAP); n++;
    XtSetArg(args[n], XmNlabelPixmap, button_pixmap); n++;
    XtSetValues(disp_info[disp_num].tool[6], args, n);

    button_pixmap = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, eraser_bits,
                      eraser_width, eraser_height, fg, bg,
                      DefaultDepth(disp_info[disp_num].disp,
                      DefaultScreen(disp_info[disp_num].disp)));
    n = 0;
    XtSetArg(args[n], XmNlabelType, XmPIXMAP); n++;
    XtSetArg(args[n], XmNlabelPixmap, button_pixmap); n++;
    XtSetValues(disp_info[disp_num].tool[7], args, n);

    button_pixmap = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, square_bits,
                      square_width, square_height, fg, bg,
                      DefaultDepth(disp_info[disp_num].disp,
                      DefaultScreen(disp_info[disp_num].disp)));
    n = 0;
    XtSetArg(args[n], XmNlabelType, XmPIXMAP); n++;
    XtSetArg(args[n], XmNlabelPixmap, button_pixmap); n++;
    XtSetValues(disp_info[disp_num].tool[8], args, n);

    button_pixmap = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, fillsquare_bits,
                      fillsquare_width, fillsquare_height, fg, bg,
                      DefaultDepth(disp_info[disp_num].disp,
                      DefaultScreen(disp_info[disp_num].disp)));
    n = 0;
    XtSetArg(args[n], XmNlabelType, XmPIXMAP); n++;
    XtSetArg(args[n], XmNlabelPixmap, button_pixmap); n++;
    XtSetValues(disp_info[disp_num].tool[9], args, n);

    button_pixmap = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, oval_bits,
                      oval_width, oval_height, fg, bg,
                      DefaultDepth(disp_info[disp_num].disp,
                      DefaultScreen(disp_info[disp_num].disp)));
    n = 0;
    XtSetArg(args[n], XmNlabelType, XmPIXMAP); n++;
    XtSetArg(args[n], XmNlabelPixmap, button_pixmap); n++;
    XtSetValues(disp_info[disp_num].tool[10], args, n);

    button_pixmap = XCreatePixmapFromBitmapData(disp_info[disp_num].disp,
                      disp_info[disp_num].draw_win_id, filloval_bits,
                      filloval_width, filloval_height, fg, bg,
                      DefaultDepth(disp_info[disp_num].disp,
                      DefaultScreen(disp_info[disp_num].disp)));
    n = 0;
    XtSetArg(args[n], XmNlabelType, XmPIXMAP); n++;
    XtSetArg(args[n], XmNlabelPixmap, button_pixmap); n++;
    XtSetValues(disp_info[disp_num].tool[11], args, n);
}


/*
 * process_menu_selection - this function is called when a menu item was
 *                          selected, and is responsible for initiating
 *                          the correct behavior from this point.
 */
process_menu_selection(disp_num, menu_num, item_num)
int disp_num, menu_num, item_num;
{
    XColor scrn_def_ret, exact_def_ret;
    XFontStruct *the_font_struct;
    int i;
    XGCValues gcvalues;
    char tmp_str[200];
    char *the_text;

    switch(menu_num)
    {
        case FILE_MENU:                       /* FILE menu */
            switch(item_num)
	    {
	        case ADD_DISPLAY:
		    if (disp_info[disp_num].add_display_dialog == NULL)
		    {
		        disp_info[disp_num].add_display_dialog =
					  create_add_display_dialog(disp_num);
		    }
		    XtManageChild(disp_info[disp_num].add_display_dialog);
		    break;
	        case JOIN_EXISTING_SESSION:
		    if (disp_info[disp_num].join_session_dialog == NULL)
		    {
			disp_info[disp_num].join_session_dialog =
					create_join_session_dialog(disp_num);
		    }
		    XtManageChild(disp_info[disp_num].join_session_dialog);
		    break;
	        case ATTRACT_ATTENTION:
		    if (disp_info[disp_num].attract_attention_dialog.the_dialog
			== NULL)
		    {
			disp_info[disp_num].attract_attention_dialog.the_dialog
			 = create_attract_attention_dialog(disp_num);
		    }
		    XtManageChild(
		      disp_info[disp_num].attract_attention_dialog.the_dialog);
		    break;
		case 3:		/* SEPARATOR */
		    break;
		case SAVE_SELECTION:
		    if (disp_info[disp_num].save_file_dialog.the_dialog == NULL)
		    {
			disp_info[disp_num].save_file_dialog.the_dialog =
				create_save_file_dialog(disp_num);
		    }
		    XtManageChild(
		         disp_info[disp_num].save_file_dialog.the_dialog);
		    break;
		case SAVE_ENTIRE_WHITEBOARD:
		    process_menu_selection(disp_num, EDIT_MENU, SELECT_ALL);
		    process_menu_selection(disp_num, FILE_MENU, SAVE_SELECTION);
		    break;
		case READ_IN_A_FILE:
		    if (disp_info[disp_num].read_file_dialog.the_dialog == NULL)
		    {
			disp_info[disp_num].read_file_dialog.the_dialog =
					create_read_file_dialog(disp_num);
		    }
		    XtManageChild(
		         disp_info[disp_num].read_file_dialog.the_dialog);
		    break;
		case 7:		/* SEPARATOR */
		    break;
		case QUIT:
		    if (disp_info[disp_num].selected_region.region_is_selected)
		    {
			sprintf(tmp_str, "%s\n%s\n%s",
			   "You cannot quit with a region selected.",
			   "Please unselect the region, then attempt",
			   "the 'quit' menu again.");
			popup_info_dialog(disp_num, tmp_str);
			return;
		    }

		    /*
		     * remove selections from other displays on this one
		     */
		    for (i=0; i < num_disps; i++)
		    {
		        if ((disp_info[i].in_session_bool == TRUE) &&
			    (disp_info[i].selected_region.region_is_selected
							 == TRUE))
			{
	                    XtRemoveTimeOut(
			      disp_info[i].selected_region.timer_id[disp_num]);
			}
		    }

		    disp_info[disp_num].in_session_bool = FALSE;

		    if ((global_values.edit_file_bool == TRUE) &&
			(global_values.screen_dirty_bit == TRUE)) /*sepia hack*/
		    {
		        save_xwd_from_drawing_area(disp_num, 
			                           global_values.edit_filename);
		    }

		    XtDestroyWidget(disp_info[disp_num].app_shell);

		    for (i=0; i < num_disps; i++)
		        if (disp_info[i].in_session_bool)
		            draw_status_win_message(i);
				     
		    for (i=0; i < num_disps; i++)
		    {
		        if (disp_info[i].in_session_bool == TRUE)
		            return;
		    }

		    exit(0);	/* no one left in session */
		    break;
		default:
		    printf("ERROR: unknown menu under FILE.\n");
		    break;
            }
	    break;
        case EDIT_MENU:                       /* EDIT menu */
            switch(item_num)
	    {
	        case CUT:
		    cut_region(disp_num);
		    break;
	        case COPY:
		    copy_region(disp_num);
		    break;
	        case PASTE:
		    paste_region(disp_num);
		    break;
	        case PASTE_X_SELECTION:
		    paste_x_selection_region(disp_num);
		    break;
	        case CLEAR:
		    clear_region(disp_num);
		    break;
	        case FATBITS_SELECTION:
		    if (disp_info[disp_num].selected_region.region_is_selected)
		    {
		        prepare_fatbits_dialog(disp_num);
		        place_selection_permanently(disp_num);
	                XtManageChild(
		           disp_info[disp_num].fatbits_dialog.the_dialog);
		    }
		    else
		    {
			popup_info_dialog(disp_num, 
			   "Select a region,\nthen try this again.");
		    }
		    break;
	        case 6:		/* separator */
		    break;
	        case SELECT_VIS_AREA:
		    select_visible_area(disp_num);
		    break;
	        case SELECT_ALL:
		    select_all(disp_num);
		    break;
	        case 9:		/* separator */
		    break;
	        case CLEAR_WHITEBOARD:
		    if (disp_info[disp_num].input_mode == TYPING)
			erase_type_cursor(disp_num);	/* erase it */
                    for (i=0; i < num_disps; i++)
		    {
		       if (disp_info[i].in_session_bool == TRUE)
		       {
			    XClearWindow(disp_info[i].disp, 
			                 disp_info[i].draw_win_id);
		            XFillRectangle(disp_info[i].disp, 
					 disp_info[i].backing_store,
					 disp_info[i].backing_store_gc, 0, 0, 
					 disp_info[i].draw_win_width,
					 disp_info[i].draw_win_height);
		        }
		    }
		    global_values.screen_dirty_bit = FALSE;  /* clean screen */
		    if (disp_info[disp_num].input_mode == TYPING)
			draw_type_cursor(disp_num);	/* redraw it */
                    break;
		default:
		    printf("ERROR: unknown menu under EDIT.\n");
		    break;
            }
	    break;
        case OPTIONS_MENU:                    /* OPTION menu */
            switch(item_num)
	    {
	        case LANGUAGE:	 	      /* language dialog */
		    disp_info[disp_num].language_dialog.current_language =
		    disp_info[disp_num].current_language;
		    if (disp_info[disp_num].language_dialog.the_dialog == NULL)
		    {
			disp_info[disp_num].language_dialog.the_dialog =
				create_language_dialog(disp_num);
		    }
		    XtManageChild(
			    disp_info[disp_num].language_dialog.the_dialog);
		    XRaiseWindow(disp_info[disp_num].disp,
		      XtWindow(disp_info[disp_num].language_dialog.the_dialog));
		    break;
		case SCROLLBAR_SETTINGS:
		    if (disp_info[disp_num].scrollbar_dialog.the_dialog == NULL)
		    {
			disp_info[disp_num].scrollbar_dialog.the_dialog =
				create_scrollbar_dialog(disp_num);
		    }
		    XtManageChild(
			    disp_info[disp_num].scrollbar_dialog.the_dialog);
		    break;
		case MOUSE_OPTIONS:
		    if (disp_info[disp_num].mouse_dialog.the_dialog == NULL)
		    {
			disp_info[disp_num].mouse_dialog.the_dialog =
				create_mouse_dialog(disp_num);
		    }
		    XtManageChild(
			    disp_info[disp_num].mouse_dialog.the_dialog);
		    break;
		case 3:		              /* separator */
		    break;
		case TELEPOINTER_LABEL:
		    if (disp_info[disp_num].change_pointer_dialog == NULL)
		    {
			disp_info[disp_num].change_pointer_dialog =
				create_change_pointer_dialog(disp_num);
		    }
		    XtManageChild(disp_info[disp_num].change_pointer_dialog);
		    break;
		case SUPER_SELECTION:
		    if (disp_info[disp_num].select_dialog.the_dialog == NULL)
		    {
			disp_info[disp_num].select_dialog.the_dialog =
					create_select_dialog(disp_num);
		    }
		    XtManageChild(disp_info[disp_num].select_dialog.the_dialog);
		    break;
	        case BRUSH_SHAPES:
		    if (disp_info[disp_num].brush_dialog.the_dialog == NULL)
		    {
			disp_info[disp_num].brush_dialog.the_dialog =
				create_brush_dialog(disp_num);
		    }
		    XtManageChild(disp_info[disp_num].brush_dialog.the_dialog);
		    XRaiseWindow(disp_info[disp_num].disp,
		      XtWindow(disp_info[disp_num].brush_dialog.the_dialog));
		    break;
		case LINE_SETTINGS:
		    if (disp_info[disp_num].line_dialog.the_dialog == NULL)
		    {
			disp_info[disp_num].line_dialog.the_dialog =
				create_line_dialog(disp_num);
		    }
		    XtManageChild(disp_info[disp_num].line_dialog.the_dialog);
		    break;
		case ERASER_OPTIONS:
		    if (disp_info[disp_num].eraser_dialog.the_dialog == NULL)
		    {
			disp_info[disp_num].eraser_dialog.the_dialog =
				create_eraser_dialog(disp_num);
		    }
		    XtManageChild(disp_info[disp_num].eraser_dialog.the_dialog);
		    the_text = XmTextGetString(
				  disp_info[disp_num].eraser_dialog.textfield);
		    sprintf(tmp_str,"%d", disp_info[disp_num].eraser_win_width);
		    XmTextReplace(disp_info[disp_num].eraser_dialog.textfield,
			          0, strlen(tmp_str), tmp_str);
                    XmTextSetSelection(
			      disp_info[disp_num].eraser_dialog.textfield,
			      0, strlen(tmp_str), 0);
		    XRaiseWindow(disp_info[disp_num].disp,
		      XtWindow(disp_info[disp_num].eraser_dialog.the_dialog));
		    break;
		case SQUARE_SETTINGS:
		    if (disp_info[disp_num].square_dialog.the_dialog == NULL)
		    {
			disp_info[disp_num].square_dialog.the_dialog =
				create_square_dialog(disp_num);
		    }
		    XtManageChild(disp_info[disp_num].square_dialog.the_dialog);
		    break;
		case OVAL_SETTINGS:
		    if (disp_info[disp_num].oval_dialog.the_dialog == NULL)
		    {
			disp_info[disp_num].oval_dialog.the_dialog = 
				  create_oval_dialog(disp_num);
		    }
		    XtManageChild(disp_info[disp_num].oval_dialog.the_dialog);
		    break;
		default:
		    printf("ERROR: unknown menu under OPTIONS.\n");
		    break;
	    }
	    break;
        case COLOR_MENU:                      /* COLOR menu */
	    if (item_num == 
		(disp_info[disp_num].menu[COLOR_MENU].num_menu_items - 1))
	    {
		if (disp_info[disp_num].color_picker_dialog.the_dialog == NULL)
		{
		    disp_info[disp_num].color_picker_dialog.the_dialog =
			create_color_picker_dialog(disp_num);
		}
		XtManageChild(
			  disp_info[disp_num].color_picker_dialog.the_dialog);
		return;
	    }

            strncpy(disp_info[disp_num].cur_color_str, colors[item_num], 40);
            for (i=0; i < num_disps; i++)
            {
                if (disp_info[i].in_session_bool)
                {
                    if (!XAllocNamedColor(disp_info[i].disp, DefaultColormap(
                         disp_info[i].disp, DefaultScreen(disp_info[i].disp)),
                         colors[item_num], &scrn_def_ret, &exact_def_ret))
                    {
                        continue;        /*sorry, no good. No colors left.*/
                    }
		    if (disp_num == i)
		        disp_info[disp_num].current_pixel = scrn_def_ret.pixel;
                    gcvalues.foreground = scrn_def_ret.pixel;
                    XChangeGC(disp_info[i].disp, disp_info[disp_num].win_gc[i],
                              GCForeground, &gcvalues);
                }
            }
	    break;
        case FONT_MENU:                      /* FONT menu */
            strncpy(disp_info[disp_num].cur_font_str, fonts[item_num], 250);
            for (i=0; i < num_disps; i++)
            {
                if (disp_info[i].in_session_bool)
                {
                    if ((the_font_struct = XLoadQueryFont(disp_info[i].disp,
                              fonts[item_num])) != NULL)
                    {
			if (disp_num == i)
			{
			    disp_info[disp_num].type_info.the_font_struct =
					the_font_struct;
			}

                        gcvalues.font = the_font_struct->fid;
                        XChangeGC(disp_info[i].disp,
                                  disp_info[disp_num].win_gc[i], GCFont,
                                  &gcvalues);
                    }
                    else if  (load_moron_fonts(disp_num, i, &the_font_struct,
                                disp_info[disp_num].cur_font_str) != FALSE)
                    {
                        if (disp_num == i)
                        {
                            disp_info[disp_num].type_info.the_font_struct =
                                                               the_font_struct;
                        }
                        gcvalues.font = the_font_struct->fid;
                        XChangeGC(disp_info[i].disp,
                                  disp_info[disp_num].win_gc[i], GCFont,
                                  &gcvalues);
                    }
		    else
		    {
                        sprintf(tmp_str, "Couldn't allocate font %s on %s%s.\n",
			        fonts[item_num],"the display: ", 
				disp_info[i].disp_name);
			popup_info_dialog(disp_num, tmp_str);
		    }
                }
            }

	    if (disp_info[disp_num].input_mode == TYPING)
	    {
	        erase_type_cursor(disp_num);  /* turn off cursor */
                disp_info[disp_num].type_info.orig_pos.x =
                                     disp_info[disp_num].type_info.cur_pos.x;
                disp_info[disp_num].type_info.orig_pos.y = 
                                     disp_info[disp_num].type_info.cur_pos.y;
                disp_info[disp_num].type_info.prompt_width =
             disp_info[disp_num].type_info.the_font_struct->max_bounds.width;
                disp_info[disp_num].type_info.char_height =
         (disp_info[disp_num].type_info.the_font_struct)->max_bounds.ascent+
         (disp_info[disp_num].type_info.the_font_struct)->max_bounds.descent;
                disp_info[disp_num].type_info.type_history = NULL;
	        draw_type_cursor(disp_num);
	    }
	    break;
        default:
            printf("ERROR: in process_menu_selection, unknown menu_num.\n");
	    break;
    }
}

