/*
 * This file:  wsregion.c
 *
 * This file contains all the functions for manipulating regions.  A region
 * is selected by dragging the "selection box" across it.  Then you can
 * drag the region, or cut, copy, or paste it or something else.
 */

#include "wscrawl.h"

extern int num_disps;
extern XtAppContext appCon;
extern struct disp_info_struct disp_info[MAX_NUM_DISPS];
extern XImage *read_image_from_disk();
unsigned char *get_x_selection_as_text();


/*
 * draw_selection_outlineCB - this function draws the selection outline,
 *                            and then sets itself as a timeout function again
 *                            so that the selection outline will "shimmer".
 */
XtTimerCallbackProc draw_selection_outlineCB(client_data, id)
int *client_data;
XtIntervalId *id;
{
    int disp_num, which, i;
    int new_offset;
    XGCValues gcvalues;

    disp_num = ((int) client_data) >> 8;
    which = ((int) client_data) & 0x00FF;

    disp_info[disp_num].selected_region.timer_id[which] =XtAppAddTimeOut(appCon,
		        500, (XtTimerCallbackProc)draw_selection_outlineCB, 
			(int *) (disp_num << 8 | which));

    if (disp_info[disp_num].selected_region.dash_offset[which] == 0)
	disp_info[disp_num].selected_region.dash_offset[which] = -4;
    else
	disp_info[disp_num].selected_region.dash_offset[which] = 0;

    new_offset = disp_info[disp_num].selected_region.dash_offset[which];

    i = which;
    XDrawRectangle(disp_info[i].disp,
	       XtWindow(disp_info[disp_num].selected_region.selection_area[i]),
	       disp_info[disp_num].selected_region.surround_line_gc[i], 
	       0, 0, disp_info[disp_num].selected_region.width - 1,
	       disp_info[disp_num].selected_region.height - 1);
    gcvalues.dash_offset = new_offset;
    XChangeGC(disp_info[i].disp,
		      disp_info[disp_num].selected_region.surround_line_gc[i],
		      GCDashOffset, &gcvalues);
    XDrawRectangle(disp_info[i].disp,
	       XtWindow(disp_info[disp_num].selected_region.selection_area[i]),
	       disp_info[disp_num].selected_region.surround_line_gc[i], 
	       0, 0, disp_info[disp_num].selected_region.width - 1,
	       disp_info[disp_num].selected_region.height - 1);
    XFlush(disp_info[i].disp);
}


/* 
 * define_selection - this function is called upon the release of the
 *                    mouse button when selecting a region.  This takes
 *                    care of setting the currect selected region, etc.
 */
define_selection(disp_num)
int disp_num;
{
    int i, x, y, width, height, root_x, root_y;

    set_x_y_width_height(disp_num, &x, &y, &width, &height);

    if (width == 0 || height == 0)
    {
        disp_info[disp_num].selected_region.region_has_moved = FALSE;
        make_selection_dependant_menus_stippled(disp_num);
	return;  /* just clicked mouse to set a selection */
    }

    width++;	/* one wider and taller for good effect */
    height++;

    disp_info[disp_num].selected_region.x = x;
    disp_info[disp_num].selected_region.y = y;
    disp_info[disp_num].selected_region.cur_x = x;
    disp_info[disp_num].selected_region.cur_y = y;
    disp_info[disp_num].selected_region.width = width;
    disp_info[disp_num].selected_region.height = height;

    for (i=0; i < num_disps; i++)
        if (disp_info[i].in_session_bool)
            disp_info[disp_num].selected_region.dash_offset[i] = 0;

    for (i=0; i < num_disps; i++)
    {
        if (disp_info[i].in_session_bool)
        {
	    prepare_region(disp_num, i, x, y, width, height);

            XCopyArea(disp_info[i].disp, disp_info[i].backing_store,
	       disp_info[disp_num].selected_region.backing_store[i],
	       disp_info[disp_num].win_gc[i], x, y, width, height, 0, 0);

	    XtMapWidget(disp_info[disp_num].selected_region.selection_shell[i]);
        }
    }

    disp_info[disp_num].selected_region.region_is_selected = TRUE;
    disp_info[disp_num].selected_region.region_has_moved = FALSE;

    make_selection_dependant_menus_unstippled(disp_num);

    set_regions_flickering(disp_num);
}


/* 
 * define_super_selection - this function is called upon the release of the
 *                    mouse button when selecting a SUPER region.  This takes
 *                    care of setting the current selected region, etc.
 */
define_super_selection(disp_num)
int disp_num;
{
    int i, x, y, width, height, root_x, root_y;
    char tmp_name[50];
    XImage *the_image;
    int depth;
    Window dummy;

    set_x_y_width_height(disp_num, &root_x, &root_y, &width, &height);

    XTranslateCoordinates(disp_info[disp_num].disp, 
                          RootWindow(disp_info[disp_num].disp,
			  DefaultScreen(disp_info[disp_num].disp)),
                          disp_info[disp_num].draw_win_id, 
			  root_x, root_y, &x, &y, &dummy);

    if (width == 0 || height == 0)
    {
        disp_info[disp_num].selected_region.region_has_moved = FALSE;
        make_selection_dependant_menus_unstippled(disp_num);
	return;  /* just clicked mouse to set a selection */
    }

    disp_info[disp_num].selected_region.x = x;
    disp_info[disp_num].selected_region.y = y;
    disp_info[disp_num].selected_region.cur_x = x;
    disp_info[disp_num].selected_region.cur_y = y;
    disp_info[disp_num].selected_region.width = width;
    disp_info[disp_num].selected_region.height = height;

    for (i=0; i < num_disps; i++)
        if (disp_info[i].in_session_bool)
            disp_info[disp_num].selected_region.dash_offset[i] = 0;

    sprintf(tmp_name, "/tmp/ws.%d.%d", disp_num, getpid());
    save_image_on_disk(disp_info[disp_num].disp,
		       XDefaultRootWindow(disp_info[disp_num].disp),
		       XDefaultRootWindow(disp_info[disp_num].disp),
		       root_x, root_y, width, height, tmp_name);

    for (i=0; i < num_disps; i++)
    {
        if (disp_info[i].in_session_bool)
        {
	    /*
	    XtTranslateCoords(disp_info[i].draw_area, x, y, &root_x, &root_y);
	    */
	    myXtTranslateCoords(i, disp_info[i].draw_area,x,y,&root_x,&root_y);

	    /* prepare_region(disp_num, i, x, y, width, height);*/
            XMoveResizeWindow(disp_info[i].disp, 
	       XtWindow(disp_info[disp_num].selected_region.selection_shell[i]),
               root_x, root_y, width, height);

	    XResizeWindow(disp_info[i].disp,
	       XtWindow(disp_info[disp_num].selected_region.selection_area[i]),
	       width, height);

	    if (disp_info[disp_num].selected_region.backing_store[i] != NULL)
	        XFreePixmap(disp_info[i].disp, 
			disp_info[disp_num].selected_region.backing_store[i]);

	    disp_info[disp_num].selected_region.backing_store[i] =
	                XCreatePixmap(disp_info[i].disp, 
			              disp_info[i].draw_win_id, width, height,
		                      DefaultDepth(disp_info[i].disp,
			              DefaultScreen(disp_info[i].disp)));
            depth = -1;
            the_image = read_image_from_disk(disp_info[i].disp,
			   disp_info[i].draw_win_id,
                           tmp_name, &width, &height, &depth);

            if ((depth == -1) || (the_image == NULL))  /*it failed*/
            {
                printf("ERROR: Something went wrong with the ");
                printf("Read Image in define_super_selection.\n");
                continue;
            }
            else
            {
                XPutImage(disp_info[i].disp, 
		           disp_info[disp_num].selected_region.backing_store[i],
                           disp_info[disp_num].win_gc[i], the_image,
                           0, 0, 0, 0, width, height);
                XFlush(disp_info[i].disp);
                XDestroyImage(the_image);
            }
 
            XRaiseWindow(disp_info[i].disp,
              XtWindow(disp_info[disp_num].selected_region.selection_shell[i]));

	    XtMapWidget(disp_info[disp_num].selected_region.selection_shell[i]);
        }
    }

    unlink(tmp_name);	  /* remove the file */
    disp_info[disp_num].selected_region.region_is_selected = TRUE;
    disp_info[disp_num].selected_region.region_has_moved = TRUE;
    make_selection_dependant_menus_unstippled(disp_num);

    set_regions_flickering(disp_num);
}


/*
 * select_visible_area - this function selects the "visible" portion of the
 *                       whiteboard on this display.
 */
select_visible_area(disp_num)
int disp_num;
{
    int n, i, x, y, width, height;
    Dimension d_width, d_height;
    Arg args[MAX_ARGS];
    Widget clip_window;
    Window dummy;

    n = 0;
    XtSetArg(args[n], XmNclipWindow, &clip_window); n++;
    XtGetValues(disp_info[disp_num].swindow, args, n);

    n = 0;
    XtSetArg(args[n], XmNwidth, &d_width); n++;
    XtSetArg(args[n], XmNheight, &d_height); n++;
    XtGetValues(clip_window, args, n);
    width = d_width;
    height = d_height;

    XTranslateCoordinates(disp_info[disp_num].disp,
                          XtWindow(clip_window),
                          disp_info[disp_num].draw_win_id, 0,0, &x, &y, &dummy);
    /*
     * now fake like the person was stretching a rubber band here, and just
     * let up the mouse
     */
    boink_tool(disp_num, SELECTION_BUTTON);
    disp_info[disp_num].orig_rubberband_point.x = x;
    disp_info[disp_num].orig_rubberband_point.y = y;
    disp_info[disp_num].last_rubberband_point.x = x + width;
    disp_info[disp_num].last_rubberband_point.y = y + height;
    
    define_selection(disp_num);
}


/*
 * select_all - this function selects the entire whiteboard
 */
select_all(disp_num)
int disp_num;
{
    XWindowAttributes attr_ret;

    XGetWindowAttributes(disp_info[disp_num].disp,
                         disp_info[disp_num].draw_win_id, &attr_ret);

    /*
     * now fake like the person was stretching a rubber band here, and just
     * let up the mouse
     */
    boink_tool(disp_num, SELECTION_BUTTON);
    disp_info[disp_num].orig_rubberband_point.x = 0;
    disp_info[disp_num].orig_rubberband_point.y = 0;
    disp_info[disp_num].last_rubberband_point.x = attr_ret.width;
    disp_info[disp_num].last_rubberband_point.y = attr_ret.height;
    
    define_selection(disp_num);
}


/*
 * handle_selection_keypressCB - this function is a hack to get keybord 
 *                               events to the correct place in a selection.
 */
XtEventHandler handle_selection_keypressCB(w, client_data, our_event, cbool)
Widget w;
caddr_t client_data;
XEvent *our_event;
Boolean *cbool;
{
    int disp_num;

    disp_num = ((int) client_data);
 
    if (our_event->type == EnterNotify)
    {
        XSetInputFocus(disp_info[disp_num].disp, 
		      disp_info[disp_num].draw_win_id,
 		      RevertToPointerRoot, CurrentTime);
    }
    else if (our_event->type == LeaveNotify)
    {
        ;
    }
    else
    {
        printf("ERROR: in handle_selection_keypress, unknown event type.\n");
    }
}


/*
 * selection_area_inputCB - this function handles all selection area input 
 *                     events such as ButtonPress or ButtonRelease.
 */
selection_area_inputCB(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int i, disp_num, which, x, y, width, height;
    XmDrawingAreaCallbackStruct *cb_struct;
    XEvent *the_event;
    XButtonPressedEvent *but_event;

    cb_struct = (XmDrawingAreaCallbackStruct *) call_data;
    the_event = (XEvent *) cb_struct->event;
    disp_num = ((int) client_data) >> 8;
    which = ((int) client_data) & 0x00FF;

    switch(the_event->type)
    {
        case ButtonPress:
	    if (disp_num != which)
		return;			/* only owner can pick it up */
            if (XGrabPointer(disp_info[disp_num].disp,
                     disp_info[disp_num].draw_win_id, False,
                     PointerMotionMask | ButtonReleaseMask,
                     GrabModeAsync, GrabModeAsync, None, None,
                     CurrentTime) != GrabSuccess)
            {
                popup_info_dialog(disp_num,
                    "Couldn't grab the pointer.\nAborting Drag of Region.");
                return;
            }
	    disp_info[disp_num].input_mode = DRAGGING_SELECTION;

            but_event = (XButtonPressedEvent *) the_event;
            disp_info[disp_num].pointer_state = PRESSED;
	    disp_info[disp_num].selected_region.drag_offset.x = but_event->x;
	    disp_info[disp_num].selected_region.drag_offset.y = but_event->y;
            x = disp_info[disp_num].selected_region.cur_x;
            y = disp_info[disp_num].selected_region.cur_y;
            width = disp_info[disp_num].selected_region.width;
            height = disp_info[disp_num].selected_region.height;

            if (disp_info[disp_num].selected_region.region_has_moved == FALSE)
            {
                disp_info[disp_num].selected_region.region_has_moved = TRUE;
        
	        if ((but_event->state & ControlMask) == 0) /* <cntl> copies */
	        {
                    for (i=0; i < num_disps; i++)
                    {
	                if (disp_info[i].in_session_bool)
	                {
	                    XClearArea(disp_info[i].disp, 
				   disp_info[i].draw_win_id,
			           x, y, width, height, False);
		            XFillRectangle(disp_info[i].disp,
			           disp_info[i].backing_store,
			           disp_info[i].backing_store_gc, x, y,
			           width, height);
	                }
	            }
	        }
            }
	    else if ((but_event->state & ControlMask) != 0) /* <cntl> copies */
	    {
                for (i=0; i < num_disps; i++)
                {
	            if (disp_info[i].in_session_bool)
	            {
                        XCopyArea(disp_info[i].disp, 
	                   disp_info[disp_num].selected_region.backing_store[i],
	                   disp_info[i].backing_store,
	                   disp_info[disp_num].win_gc[i], 0, 0, 
			   width, height, x, y);
	                XFlush(disp_info[i].disp);
                    }
                }
	    }
	    break;
        case ButtonRelease:
            disp_info[disp_num].pointer_state = NOT_PRESSED;
	    break;
        case KeyPress:
	    handle_KeyPress_event(the_event, disp_num);
	    break;
        case KeyRelease:
	    break;
	default:
	    printf("Unknown event type in selection_area_inputCB.\n");
	    break;
    }
}


/*
 * selection_pointer_moveCB - this function is called when the pointer moves 
 *                  in the selection window.  This probably means the
 *                  person is dragging the window, but maybe not.
 *
 *                  NOTE: this routine IS NOT called as a callback, but
 *                  instead from my own functions because of a hack for
 *                  smooth and speedy dragging.  I am leaving it defined as
 *                  an XtEventHandler just in case I want to revert.
 */
XtEventHandler selection_pointer_moveCB(w, client_data, our_event, cbool)
Widget w;
caddr_t client_data;
XEvent *our_event;
Boolean *cbool;
{
    int i,  disp_num, which;
    XMotionEvent *the_event;
    int x, y;

    if (our_event->type != MotionNotify)
    {
	printf("ERROR: in selection_pointer_moveCB, invalid event.\n");
	return;	    /* sometimes this function gets the NonMaskable NoExpose */
    }

    the_event = (XMotionEvent *) our_event;
    disp_num = ((int) client_data) >> 8;
    which = ((int) client_data) & 0x00FF;

    if (disp_info[disp_num].pointer_state != PRESSED)
        return;		/* not interested in this event */

    disp_info[disp_num].selected_region.cur_x = 
	      the_event->x - disp_info[disp_num].selected_region.drag_offset.x;
    disp_info[disp_num].selected_region.cur_y =
	      the_event->y - disp_info[disp_num].selected_region.drag_offset.y;

    /*
    XtTranslateCoords(disp_info[disp_num].draw_area,
                    disp_info[disp_num].selected_region.cur_x,
		    disp_info[disp_num].selected_region.cur_y,
		    &x, &y);
    */
    myXtTranslateCoords(disp_num, disp_info[disp_num].draw_area,
                    disp_info[disp_num].selected_region.cur_x,
		    disp_info[disp_num].selected_region.cur_y,
		    &x, &y);

    XMoveWindow(disp_info[disp_num].disp, 
       XtWindow(disp_info[disp_num].selected_region.selection_shell[disp_num]),
       x, y);
    XFlush(disp_info[disp_num].disp);

    for (i=0; i < num_disps; i++)
    {
	if ((disp_info[i].in_session_bool == TRUE) && (i != disp_num))
	{
            XMoveWindow(disp_info[i].disp, 
	       XtWindow(disp_info[disp_num].selected_region.selection_shell[i]),
	       disp_info[disp_num].selected_region.cur_x, 
	       disp_info[disp_num].selected_region.cur_y);
	}
    }
}


/*
 * place_selection_permanently - this function places the selection back onto
 *    				 the screen, and sets all variables accordingly.
 */
place_selection_permanently(disp_num)
int disp_num;
{
    int i, x, y, width, height;
    Window dummy;

    x = disp_info[disp_num].selected_region.cur_x;
    y = disp_info[disp_num].selected_region.cur_y;
    width = disp_info[disp_num].selected_region.width;
    height = disp_info[disp_num].selected_region.height;

    disp_info[disp_num].selected_region.region_is_selected = FALSE;
    make_selection_dependant_menus_stippled(disp_num);

    /*
     * get rid of shimmering line
     */
    for (i = 0; i < num_disps; i++)
	if (disp_info[i].in_session_bool)
            XtRemoveTimeOut(disp_info[disp_num].selected_region.timer_id[i]);

    for (i=0; i < num_disps; i++)
    {
        if (disp_info[i].in_session_bool == TRUE)
        {
             XtUnmapWidget(
		       disp_info[disp_num].selected_region.selection_shell[i]);
             XCopyArea(disp_info[i].disp, 
	          disp_info[disp_num].selected_region.backing_store[i],
	          disp_info[i].draw_win_id,
	          disp_info[disp_num].win_gc[i], 0, 0, width, height, x, y);
             XCopyArea(disp_info[i].disp, 
	          disp_info[disp_num].selected_region.backing_store[i],
	          disp_info[i].backing_store,
	          disp_info[disp_num].win_gc[i], 0, 0, width, height, x, y);
	     XFlush(disp_info[i].disp);
        }
    }
    stop_regions_flickering(disp_num);
}


/*
 * selection_area_exposeCB - this function gets called when an expose event
 *                     occurs in the selection area.
 */
selection_area_exposeCB(w, client_data, call_data)
Widget w;
caddr_t client_data;
caddr_t call_data;
{
    int disp_num, which;
    XmDrawingAreaCallbackStruct *cb_struct;
    XGCValues gcvalues;
    XExposeEvent *the_event;

    cb_struct = (XmDrawingAreaCallbackStruct *) call_data;
    the_event = (XExposeEvent *) cb_struct->event;
    disp_num = ((int) client_data) >> 8;
    which = ((int) client_data) & 0x00FF;

    if (the_event->type != Expose)
	printf("ERROR: in select_area_exposeCB.\n");

    XCopyArea(disp_info[which].disp, 
	 disp_info[disp_num].selected_region.backing_store[which],
         XtWindow(disp_info[disp_num].selected_region.selection_area[which]),
	 disp_info[disp_num].win_gc[which], 0, 0, 
	 disp_info[disp_num].selected_region.width, 
	 disp_info[disp_num].selected_region.height, 
	 0, 0);

    /*
     * now draw the dashed line around the outside
     */
    disp_info[disp_num].selected_region.dash_offset[which] = 0;

    gcvalues.dash_offset = 0;
    XChangeGC(disp_info[which].disp,
	      disp_info[disp_num].selected_region.surround_line_gc[which],
	      GCDashOffset, &gcvalues);
    XDrawRectangle(disp_info[which].disp,
               XtWindow(
	       disp_info[disp_num].selected_region.selection_area[which]),
               disp_info[disp_num].selected_region.surround_line_gc[which], 
               0, 0, disp_info[disp_num].selected_region.width - 1,
               disp_info[disp_num].selected_region.height - 1);
}


/*
 * clear_region - this function is called when "Clear" is selected in the
 *                Edit menu, or backspace is hit.
 */
clear_region(disp_num)
int disp_num;
{
    int i, x, y, width, height;

    if (disp_info[disp_num].selected_region.region_is_selected == FALSE)
    {
        popup_error_dialog(disp_num, 
			   "ERROR: in clear region, region not selected");
    }

    /*
     * The user wants to clear this selected region.  Do it.
     */
    disp_info[disp_num].selected_region.region_is_selected = FALSE;
    make_selection_dependant_menus_stippled(disp_num);
    x = disp_info[disp_num].selected_region.cur_x;
    y = disp_info[disp_num].selected_region.cur_y;
    width = disp_info[disp_num].selected_region.width;
    height = disp_info[disp_num].selected_region.height;

    if (disp_info[disp_num].selected_region.region_has_moved == FALSE)
    {
        /* erase it from the original spot */
        for (i=0; i < num_disps; i++)
        {
            if (disp_info[i].in_session_bool)
            {
                XtUnmapWidget(
                        disp_info[disp_num].selected_region.selection_shell[i]);
                XClearArea(disp_info[i].disp, disp_info[i].draw_win_id,
                                   x, y, width, height, False);
                XFillRectangle(disp_info[i].disp, disp_info[i].backing_store,
                               disp_info[i].backing_store_gc, x, y,
                               width, height);
                XFlush(disp_info[i].disp);
            }
        }
    }
    else                /* it has moved, so just UnMap it */
    {
        for (i=0; i < num_disps; i++)
        {
            if (disp_info[i].in_session_bool)
            {
                XtUnmapWidget(
                     disp_info[disp_num].selected_region.selection_shell[i]);
                XFlush(disp_info[i].disp);
            }
        }
    }

    stop_regions_flickering(disp_num);
}


/*
 * copy_region - this function is called when "Copy" is selected in the
 *               Edit menu, or some similiar event or shortcut occurs. It
 *               copies the currently selected region into the clipboard
 *               of this member of the wscrawl session.
 */
copy_region(disp_num)
int disp_num;
{
    int i, x, y, width, height;

    if (disp_info[disp_num].selected_region.region_is_selected == FALSE)
    {
        popup_error_dialog(disp_num, 
			   "ERROR: in copy region, region not selected");
	return;
    }

    width = disp_info[disp_num].selected_region.width;
    height = disp_info[disp_num].selected_region.height;
    disp_info[disp_num].selected_region.clip_board_width = width;
    disp_info[disp_num].selected_region.clip_board_height = height;

    for (i=0; i < num_disps; i++)
    {
        if (disp_info[i].in_session_bool)
        {
            if (disp_info[disp_num].selected_region.clip_board[i] != NULL)
	    {
	        XFreePixmap(disp_info[i].disp,
		            disp_info[disp_num].selected_region.clip_board[i]);
	    }

            disp_info[disp_num].selected_region.clip_board[i] =
                        XCreatePixmap(disp_info[i].disp,
                                      disp_info[i].draw_win_id, width, height,
                                      DefaultDepth(disp_info[i].disp,
                                      DefaultScreen(disp_info[i].disp)));
	    XCopyArea(disp_info[i].disp, 
	         disp_info[disp_num].selected_region.backing_store[i],
		 disp_info[disp_num].selected_region.clip_board[i],
	         disp_info[disp_num].win_gc[i], 0, 0, width, height, 0, 0);
         }
    }
}


/*
 * cut_region - this function is called when "Cut is selected in the
 *               Edit menu, or some similiar event or shortcut occurs. It
 *               copies the current selected region into the clipboard,
 *               clears the region.
 */
cut_region(disp_num)
int disp_num;
{
    copy_region(disp_num);
    clear_region(disp_num);
}


/*
 * paste_region - this function is called when "Paste" is selected in the
 *               Edit menu, or some similiar event or shortcut occurs. It
 *               pastes the current contents of the clipboard into the
 *               selected region and maps the selected region in the center
 *               of the drawing window.
 */
paste_region(disp_num)
int disp_num;
{
    int i, n, x, y, width, height, vis_x, vis_y, vis_width, vis_height;
    int root_x, root_y;
    Arg args[MAX_ARGS];
    Widget clip_window;
    Window dummy;

    if (disp_info[disp_num].selected_region.region_is_selected == TRUE)
    {
        popup_info_dialog(disp_num, 
	      "Unselect the current region\nthen retry this action.");
	return;
    }

    if (disp_info[disp_num].selected_region.clip_board[disp_num] == NULL)
    {
        popup_info_dialog(disp_num, 
	          "There is nothing in the clipboard to paste.");
	return;
    }

    width = disp_info[disp_num].selected_region.clip_board_width;
    height = disp_info[disp_num].selected_region.clip_board_height;

    /* 
     * find the center of the pasting user's drawing window
     */
    find_center_of_drawing_window(disp_num, width, height, &x, &y);

    disp_info[disp_num].selected_region.x = x;
    disp_info[disp_num].selected_region.y = y;
    disp_info[disp_num].selected_region.cur_x = x;
    disp_info[disp_num].selected_region.cur_y = y;
    disp_info[disp_num].selected_region.width = width;
    disp_info[disp_num].selected_region.height = height;

    for (i=0; i < num_disps; i++)
    {
        if (disp_info[i].in_session_bool)
        {
            prepare_region(disp_num, i, x, y, width, height);
	    XCopyArea(disp_info[i].disp, 
		 disp_info[disp_num].selected_region.clip_board[i],
	         disp_info[disp_num].selected_region.backing_store[i],
	         disp_info[disp_num].win_gc[i], 0, 0, width, height, 0, 0);
            XtMapWidget(disp_info[disp_num].selected_region.selection_shell[i]);
        }
    }

    disp_info[disp_num].selected_region.region_is_selected = TRUE;
    disp_info[disp_num].selected_region.region_has_moved = TRUE;  /* hack */

    make_selection_dependant_menus_unstippled(disp_num);
    set_regions_flickering(disp_num);
}


/*
 * copy_selections_to_new_display - this function copies all the selections
 *                           that currently contain something into their
 *                           backingstore for this new display.  This function
 *                           is called when this display is newly added.
 */
copy_selections_to_new_display(disp_num)
int disp_num;
{
    int i, x, y, width, height, root_x, root_y, src_disp;
    int depth;
    char tmp_name[50];
    XImage *the_image;

    sprintf(tmp_name, "/tmp/ws.%d.%d", disp_num, getpid());

    for (src_disp=0; src_disp < num_disps; src_disp++)
    {
        if ((disp_info[src_disp].in_session_bool == TRUE) &&
            (disp_info[src_disp].selected_region.region_is_selected == TRUE))
        {
	    x = disp_info[src_disp].selected_region.cur_x;
	    y = disp_info[src_disp].selected_region.cur_y;
            width = disp_info[src_disp].selected_region.width;
	    height = disp_info[src_disp].selected_region.height;
            disp_info[src_disp].selected_region.dash_offset[disp_num] = 0;

	    prepare_region(src_disp, disp_num, x, y, width, height);

            save_image_on_disk(disp_info[src_disp].disp,
	           disp_info[src_disp].draw_win_id,
	           disp_info[src_disp].selected_region.backing_store[src_disp],
		   0, 0, width, height, tmp_name);

            depth = -1;
            the_image = read_image_from_disk(disp_info[disp_num].disp,
			   disp_info[disp_num].draw_win_id,
                           tmp_name, &width, &height, &depth);

            if ((depth == -1) || (the_image == NULL))  /*it failed*/
            {
                popup_info_dialog(src_disp, 
		     "Could not copy your clipboard\nto new person.");
            }
            else
            {
                XPutImage(disp_info[disp_num].disp, 
	           disp_info[src_disp].selected_region.backing_store[disp_num],
                   disp_info[src_disp].win_gc[disp_num], the_image,
                   0, 0, 0, 0, width, height);
                XDestroyImage(the_image);
            }

	    XtMapWidget(
		disp_info[src_disp].selected_region.selection_shell[disp_num]);

            disp_info[src_disp].selected_region.timer_id[disp_num] =
		   XtAppAddTimeOut(appCon,
		        500, (XtTimerCallbackProc)draw_selection_outlineCB, 
			(int *) (src_disp << 8 | disp_num));
            unlink(tmp_name);
        }
    }
}


/*
 * copy_clipboards_to_new_display - this function copies all the clipboards
 *                           that currently contain something into their
 *                           clipboard for this new display.  This function
 *                           is called when this display is newly added.
 */
copy_clipboards_to_new_display(disp_num)
int disp_num;
{
    int i, x, y, width, height, root_x, root_y, src_disp;
    int depth;
    char tmp_name[50];
    XImage *the_image;

    sprintf(tmp_name, "/tmp/ws.%d.%d", disp_num, getpid());

    for (src_disp=0; src_disp < num_disps; src_disp++)
    {
        if ((disp_info[src_disp].in_session_bool == TRUE) &&
            (disp_info[src_disp].selected_region.clip_board[src_disp] != NULL))
        {
            width = disp_info[src_disp].selected_region.clip_board_width;
            height = disp_info[src_disp].selected_region.clip_board_height;

            save_image_on_disk(disp_info[src_disp].disp,
		       disp_info[src_disp].draw_win_id,
		       disp_info[src_disp].selected_region.clip_board[src_disp],
		       0, 0, width, height, tmp_name);

            disp_info[src_disp].selected_region.clip_board[disp_num] =
                        XCreatePixmap(disp_info[disp_num].disp,
                                disp_info[disp_num].draw_win_id, width, height,
                                DefaultDepth(disp_info[disp_num].disp,
                                DefaultScreen(disp_info[disp_num].disp)));

            depth = -1;
            the_image = read_image_from_disk(disp_info[disp_num].disp,
			   disp_info[disp_num].draw_win_id,
                           tmp_name, &width, &height, &depth);

            if ((depth == -1) || (the_image == NULL))  /*it failed*/
            {
                popup_info_dialog(src_disp, 
		     "Could not copy your clipboard\nto new person.");
                XFillRectangle(disp_info[disp_num].disp, 
		     disp_info[src_disp].selected_region.clip_board[disp_num],
		     disp_info[disp_num].backing_store_gc, 0, 0, width, height);
            }
            else
            {
                XPutImage(disp_info[disp_num].disp, 
	           disp_info[src_disp].selected_region.clip_board[disp_num],
                   disp_info[src_disp].win_gc[disp_num], the_image,
                   0, 0, 0, 0, width, height);
                XDestroyImage(the_image);
            }

            unlink(tmp_name);
        }
    }
}


/*
 * find_center_of_drawing_window - this function takes the width and height
 *                            of a region, and returns the x and y relative
 *                            to the drawing window of this display that
 *                            will center the region.
 */
find_center_of_drawing_window(disp_num, width, height, x, y)
int disp_num, width, height, *x, *y;
{
    int n, vis_x, vis_y;
    Dimension vis_width, vis_height;
    Arg args[MAX_ARGS];
    Widget clip_window;
    Window dummy;

    n = 0;
    XtSetArg(args[n], XmNclipWindow, &clip_window); n++;
    XtGetValues(disp_info[disp_num].swindow, args, n);

    n = 0;
    XtSetArg(args[n], XmNwidth, &vis_width); n++;
    XtSetArg(args[n], XmNheight, &vis_height); n++;
    XtGetValues(clip_window, args, n);

    XTranslateCoordinates(disp_info[disp_num].disp,
                          XtWindow(clip_window),
                          disp_info[disp_num].draw_win_id, 0,0, &vis_x, &vis_y, 
                          &dummy);

    *x = vis_x + ((int)vis_width/2) - (width/2);/* place in center of window */
    *y = vis_y + ((int)vis_height/2) - (height/2);
}


/*
 * prepare_region - this function prepares the region for an upcoming XCopyArea
 *                  into it.  It resizes it, and moves it to the correct 
 *                  location, and allocates a pixmap for it.  "disp_num"
 *                  is the owner of the region, and "target_disp_num" is the
 *                  display this actual pixmap should be created on.   The
 *                  "x" and "y" are relative to the disp_num's drawing area.
 */
prepare_region(disp_num, target_disp_num, x, y, width, height)
int disp_num, target_disp_num, x, y, width, height;
{
    int i, dest_x, dest_y;

    i = target_disp_num;	/* hack, because this is usually in a loop */

    if (disp_info[disp_num].selected_region.backing_store[i] != NULL)
    {
        XFreePixmap(disp_info[i].disp,
                   disp_info[disp_num].selected_region.backing_store[i]);
    }

    disp_info[disp_num].selected_region.backing_store[i] =
                        XCreatePixmap(disp_info[i].disp,
                                      disp_info[i].draw_win_id, width, height,
                                      DefaultDepth(disp_info[i].disp,
                                      DefaultScreen(disp_info[i].disp)));
    XFillRectangle(disp_info[i].disp, 
		  disp_info[disp_num].selected_region.backing_store[i],
		  disp_info[i].backing_store_gc, 0, 0, width, height);

    if (target_disp_num == disp_num)
    {
	/*
        XtTranslateCoords(disp_info[i].draw_area, x, y, &dest_x, &dest_y);
	*/
        myXtTranslateCoords(i, disp_info[i].draw_area, x, y, &dest_x, &dest_y);
    }
    else
    {
	dest_x = x;
	dest_y = y;
    }

    XMoveResizeWindow(disp_info[i].disp,
               XtWindow(disp_info[disp_num].selected_region.selection_shell[i]),
               dest_x, dest_y, width, height);

    XResizeWindow(disp_info[i].disp,
              XtWindow(disp_info[disp_num].selected_region.selection_area[i]),
               width, height);

    XRaiseWindow(disp_info[i].disp,
              XtWindow(disp_info[disp_num].selected_region.selection_shell[i]));
}


/*
 * set_regions_flickering - this function starts all the edges flickering on
 *                          this display's regions.
 */
set_regions_flickering(disp_num)
int disp_num;
{
    int i;

    for (i = 0; i < num_disps; i++)
        if (disp_info[i].in_session_bool)
	    disp_info[disp_num].selected_region.timer_id[i] = XtAppAddTimeOut(
			appCon, 1, (XtTimerCallbackProc)draw_selection_outlineCB, 
			(int *) (disp_num << 8 | i));
}


/*
 * stop_regions_flickering - this function stops all the edges from flickering 
 *                           on this display's regions.
 */
stop_regions_flickering(disp_num)
int disp_num;
{
    int i;

    for (i = 0; i < num_disps; i++)
        if (disp_info[i].in_session_bool)
	    XtRemoveTimeOut(disp_info[disp_num].selected_region.timer_id[i]);
}


/*
 * read_bitmap_file_into_region - this function reads the indicated
 *                                X-Windows bitmap file into all the 
 *                                selected regions
 */
read_bitmap_file_into_region(disp_num, bitmap_filename)
int disp_num;
char *bitmap_filename;
{
    int x, y, xhr, yhr, i;
    unsigned int width, height;

    Pixmap the_pixmap;

    set_paused_cursors();                     /*this may take a while*/

    if (XReadBitmapFile(disp_info[disp_num].disp, 
               disp_info[disp_num].draw_win_id,
               bitmap_filename, &width, &height, &the_pixmap, &xhr, &yhr) !=
            BitmapSuccess)
    { 
        popup_info_dialog(disp_num, "Not a valid bitmap file.");
        unset_paused_cursors();                    
        return;
    }

    /*
     * it is a valid bitmap file, place it
     */
    find_center_of_drawing_window(disp_num, width, height, &x, &y);
    prepare_region(disp_num, disp_num, x, y, width, height);
                
    XCopyPlane(disp_info[disp_num].disp, the_pixmap, 
		  disp_info[disp_num].selected_region.backing_store[disp_num],
                  disp_info[disp_num].win_gc[disp_num], 0, 0, 
		  width, height, 0, 0, 1);
    XtMapWidget(disp_info[disp_num].selected_region.selection_shell[disp_num]);
    XFlush(disp_info[disp_num].disp);

    disp_info[disp_num].selected_region.x = x;
    disp_info[disp_num].selected_region.y = y;
    disp_info[disp_num].selected_region.cur_x = x;
    disp_info[disp_num].selected_region.cur_y = y;
    disp_info[disp_num].selected_region.width = width;
    disp_info[disp_num].selected_region.height = height;

    for (i=0; i < num_disps; i++)
    {
        if ((disp_info[i].in_session_bool) && (i != disp_num))
        {
            if (XReadBitmapFile(disp_info[i].disp, disp_info[i].draw_win_id,
               bitmap_filename, &width, &height, &the_pixmap, &xhr, &yhr) !=
               BitmapSuccess)
            {   
                popup_error_dialog(disp_num, 
				   "ERROR in read_bitmap_file_into_region.");
	    }

            prepare_region(disp_num, i, x, y, width, height);
                
            XCopyPlane(disp_info[i].disp, the_pixmap, 
		  disp_info[disp_num].selected_region.backing_store[i],
                  disp_info[disp_num].win_gc[i], 0, 0, 
		  width, height, 0, 0, 1);
	    XtMapWidget(disp_info[disp_num].selected_region.selection_shell[i]);
            XFlush(disp_info[i].disp);
        }
    }
    
    disp_info[disp_num].selected_region.region_is_selected = TRUE;
    disp_info[disp_num].selected_region.region_has_moved = TRUE;  /* hack */

    make_selection_dependant_menus_unstippled(disp_num);
    set_regions_flickering(disp_num);
    unset_paused_cursors();                    
}


/*
 * save_region_into_xwd_file - this function saves the current selection
 *                             into the indicated file in xwd format.
 */
save_region_into_xwd_file(disp_num, bitmap_filename)
int disp_num;
char *bitmap_filename;
{
    int width, height;

    set_paused_cursors();                     /*this may take a while*/

    if (disp_info[disp_num].selected_region.region_is_selected == FALSE)
    {
        popup_error_dialog(disp_num, "ERROR: region not selected");
        unset_paused_cursors();                    
        return;
    }

    width = disp_info[disp_num].selected_region.width;
    height = disp_info[disp_num].selected_region.height;

    save_image_on_disk(disp_info[disp_num].disp,
              disp_info[disp_num].draw_win_id,
              disp_info[disp_num].selected_region.backing_store[disp_num],
	      0, 0, width, height, bitmap_filename);

    unset_paused_cursors();                    
}


/*
 * save_region_into_ps_file - this function saves the current selection
 *                            into the indicated file in greyscale postscript 
 *                            format.
 */
save_region_into_ps_file(disp_num, ps_filename)
int disp_num;
char *ps_filename;
{
    XImage *the_image;
    int x, y, width, height;
    float red_val, green_val, blue_val;
    int grey_val;
    XColor the_color;
    int num_pixels_on_this_line;
    FILE *fp;

    set_paused_cursors();                     /*this may take a while*/

    if (disp_info[disp_num].selected_region.region_is_selected == FALSE)
    {
        popup_error_dialog(disp_num, "ERROR: region not selected");
        unset_paused_cursors();                    
        return;
    }

    width = disp_info[disp_num].selected_region.width;
    height = disp_info[disp_num].selected_region.height;

    if ((the_image = XGetImage(disp_info[disp_num].disp,
                disp_info[disp_num].selected_region.backing_store[disp_num], 
		0, 0, width, height, (~0), ZPixmap)) == NULL)
    {
	printf("ERROR: could not get the_image.\n");
    }

    if ((fp = fopen(ps_filename, "w")) == NULL)
    {
        popup_error_dialog(disp_num,
                    "could not open file in save_region_into_ps_file");
        unset_paused_cursors();
        return;
    }

    fprintf(fp, "%%!PostScript file for printing\n");
    fprintf(fp, "%% This file was created by wscrawl 2.0 which is by ");
    fprintf(fp, "Brian Wilson\n");
    fprintf(fp, "\n");
    fprintf(fp, "gsave 10 dict begin\n");
    fprintf(fp, "\n");
    fprintf(fp, "/PaperWidth      8.5 4.2 mul def %%  8.5 inchx300dpi/72dpi\n");
    fprintf(fp, "/PaperHeight    11 4.2 mul def   %%  11 inch\n");
    fprintf(fp, "/ImagePaperWidth        %d 4.2 mul def %% size of image", 
	    width);
    fprintf(fp, "on the page\n");
    fprintf(fp, "/ImagePaperHeight       %d 4.2 mul def %%\n", height);
    fprintf(fp, "/ImageWidth             %3d def %% the width in pixels\n",
	    width);
    fprintf(fp, "/ImageHeight            %3d def %% the height in pixels\n", 
	    height);
    fprintf(fp, "\n");
    fprintf(fp, "/XPosition PaperWidth  ImagePaperWidth  sub 2 div def %% ");
    fprintf(fp, "center the image\n");
    fprintf(fp, "/YPosition PaperHeight ImagePaperHeight sub 2 div def\n");
    fprintf(fp, "\n");
    fprintf(fp, "/BitsPerPixel 8 def   %% 8 bits per pixel\n");
    fprintf(fp, "\n");
    fprintf(fp, "/PicStr ImageWidth ImageHeight gt {ImageWidth}{ImageHeight} ");
    fprintf(fp, "ifelse string def\n");
    fprintf(fp, "\n");
    fprintf(fp, "[ImagePaperWidth 0 0 ImagePaperHeight neg XPosition ");
    fprintf(fp, "YPosition] setmatrix\n");
    fprintf(fp, "ImageWidth ImageHeight BitsPerPixel\n");
    fprintf(fp, "[ImageWidth 0 0 ImageHeight neg 0 0]\n");
    fprintf(fp, "{currentfile PicStr readhexstring pop} image\n");

    /*
     * Now write out the greyscale postscript image.  The formula for 
     * converting from color to greyscale is grey_val = .299r + .587g + .114b
     * which I got from the PGM manpage.  The grey_vals range from 0 to 255.
     */
    num_pixels_on_this_line = 0;
    for (y=0; y < height; y++)
    {
	for (x=0; x < width; x++)
	{
	    the_color.pixel = XGetPixel(the_image, x, y);
	    XQueryColor(disp_info[disp_num].disp, DefaultColormap(
			disp_info[disp_num].disp, DefaultScreen(
			disp_info[disp_num].disp)), &the_color);

	    red_val = (float) ((((long) the_color.red) * 255) / ((long) 65535));
	    green_val=(float)((((long)the_color.green) * 255) / ((long) 65535));
	    blue_val =(float) ((((long)the_color.blue) * 255) / ((long) 65535));

            grey_val = (int) ((0.299 * red_val) + (0.587 * green_val) +
                              (0.114 * blue_val));
            /*
	     * the format is in hexidecimal, with two characters representing
	     * one pixel, so...
	     */
	    if (grey_val < 0xf)		
	        fprintf(fp, "0");    /* it is only one digit, so prepend a 0 */
	    fprintf(fp, "%x", grey_val);

	    num_pixels_on_this_line++;
	    if (num_pixels_on_this_line > 35)
	    {
		fprintf(fp, "\n");
		num_pixels_on_this_line = 0;
	    }
	}
    }

    fprintf(fp, "\n");
    fprintf(fp, "grestore end showpage\n");
    fprintf(fp, "\n");
    fclose(fp);

    unset_paused_cursors();                    
}


/*
 * save_region_into_bitmap_file - this function saves the current selection
 *                                into the indicated file in X-Windows bitmap 
 *                                format.
 */
save_region_into_bitmap_file(disp_num, bitmap_filename)
int disp_num;
char *bitmap_filename;
{
    int n;
    Arg args[MAX_ARGS];
    XImage *the_image;
    int x_cord, y_cord, cur_left_byte, cur_right_byte, width, height;
    int num_on_this_line, done, i;
    Pixel bg;
    char *bitmap_name;
    FILE *bitmap_file;

    set_paused_cursors();                     /*this may take a while*/

    if (disp_info[disp_num].selected_region.region_is_selected == FALSE)
    {
        popup_error_dialog(disp_num, "ERROR: region not selected");
        unset_paused_cursors();                    
        return;
    }

    width = disp_info[disp_num].selected_region.width;
    height = disp_info[disp_num].selected_region.height;
    width = (width/8) * 8;        /* even multiple of 8 for bitmap */
    
    /* 
     * find the bitmap name in the file name 
     */
    for (i = strlen(bitmap_filename); i > 0; i--)
    {
	if (bitmap_filename[i] == '/')
	{
            bitmap_name = &(bitmap_filename[i+1]);
	    break;
	}
    }

    if (i == 0)
        bitmap_name = bitmap_filename;

    if (width == 0)
    {
        popup_info_dialog(disp_num, "region is too small, try again");
        unset_paused_cursors();                    
        return;
    }

    if ((the_image = XGetImage(disp_info[disp_num].disp,
                disp_info[disp_num].selected_region.backing_store[disp_num],
                0, 0, width, height, (~0), ZPixmap)) == NULL)
    {
         popup_error_dialog(disp_num, "ERROR: XGetImage failed.\n");
         return;
    }
    else if ((bitmap_file = fopen(bitmap_filename, "w")) == NULL)
    {
         /*fopen failed*/
         XDestroyImage(the_image);
         popup_info_dialog(disp_num, 
	    "Could not open bitmap file\nfor writing: bitmap not saved.\n");
         return;
    }

    n = 0;
    XtSetArg(args[n], XmNbackground, &bg); n++;
    XtGetValues(disp_info[disp_num].draw_area, args, n);

    fprintf(bitmap_file, "#define %s_width %d\n", bitmap_name, width);
    fprintf(bitmap_file, "#define %s_height %d\n", bitmap_name, height);
    fprintf(bitmap_file, "static char %s_bits[] = {\n", bitmap_name);
    fprintf(bitmap_file, "   ");
    /*
     * the following blows the data out to the file on disk
     */
    num_on_this_line = 0;
    for (y_cord=0, done=FALSE; done==FALSE; y_cord++)
    {
        for (x_cord=0; x_cord < width; x_cord+=8)
        {
            cur_right_byte = 0;
            cur_left_byte = 0;

            for (i=0; i<4; i++)
            {
                if (XGetPixel(the_image, x_cord+i, y_cord) != bg)
                {
                    cur_right_byte |= 1<<i;
                }
            }
            for (i=0; i<4; i++)
            {
                if (XGetPixel(the_image, x_cord+i+4, y_cord) != bg)
                {
                    cur_left_byte |= 1<<i;
                }
            }
            fprintf(bitmap_file, "0x%x%x", cur_left_byte, cur_right_byte);
            num_on_this_line++;

            if ((y_cord == height-1) && (x_cord == (width-8)))
            {
                fprintf(bitmap_file, "};\n");
                done = TRUE;
                break;
            }
            else if (num_on_this_line == 12)
            {
                fprintf(bitmap_file, ",\n");
                fprintf(bitmap_file, "   ");
                num_on_this_line = 0;
            }
            else
                fprintf(bitmap_file, ", ");
        }
    }
    fclose(bitmap_file);
    XDestroyImage(the_image);

    unset_paused_cursors();                    
}


/*
 * read_xwd_file_into_region - this function reads the indicated
 *                             XWD file (as produced from the xwd program)
 *                             into this display's selected regions
 */
read_xwd_file_into_region(disp_num, xwd_filename)
int disp_num;
char *xwd_filename;
{
    XImage *the_image;
    int width, height, depth, x, y, xhr, yhr, i;
    char tmp_str[200];

    set_paused_cursors();                     /*this may take a while*/

    if ((the_image = read_image_from_disk(disp_info[disp_num].disp,
                           disp_info[disp_num].draw_win_id,
                           xwd_filename, &width, &height, &depth)) == NULL)
    {
        popup_info_dialog(disp_num, "Not a valid xwd file.");
        unset_paused_cursors();                    
        return;
    }

    /*
     * it is a valid xwd file, place it
     */
    find_center_of_drawing_window(disp_num, width, height, &x, &y);
    prepare_region(disp_num, disp_num, x, y, width, height);
                
    if (DefaultDepth(disp_info[disp_num].disp,
                     DefaultScreen(disp_info[disp_num].disp)) != depth)
    {
        sprintf(tmp_str, 
               "Image not displayed on display %s.\nInconsistent Depths.",
               disp_info[disp_num].disp_name);
        popup_info_dialog(disp_num, tmp_str);
        XDestroyImage(the_image);       /*free the image memory*/
    }
    else
    {
        XPutImage(disp_info[disp_num].disp, 
              disp_info[disp_num].selected_region.backing_store[disp_num],
              disp_info[disp_num].win_gc[disp_num], the_image,
              0, 0, 0, 0, width, height);

        XDestroyImage(the_image);
        XtMapWidget(
             disp_info[disp_num].selected_region.selection_shell[disp_num]);
        XFlush(disp_info[disp_num].disp);
    }

    disp_info[disp_num].selected_region.x = x;
    disp_info[disp_num].selected_region.y = y;
    disp_info[disp_num].selected_region.cur_x = x;
    disp_info[disp_num].selected_region.cur_y = y;
    disp_info[disp_num].selected_region.width = width;
    disp_info[disp_num].selected_region.height = height;

    for (i=0; i < num_disps; i++)
    {
        if ((disp_info[i].in_session_bool) && (i != disp_num))
        {
            prepare_region(disp_num, i, x, y, width, height);
	    read_image_into_region(disp_num, i, xwd_filename);

            XtMapWidget(disp_info[disp_num].selected_region.selection_shell[i]);
            XFlush(disp_info[i].disp);
        }
    }
    
    disp_info[disp_num].selected_region.region_is_selected = TRUE;
    disp_info[disp_num].selected_region.region_has_moved = TRUE;  /* hack */

    make_selection_dependant_menus_unstippled(disp_num);
    set_regions_flickering(disp_num);
    unset_paused_cursors();                    
}


/*
 * get_token - this function takes the current line and current position
 *             and finds the first non-whitespace after that point, and
 *             copies up to the next whitespace into the "token_str".  If
 *             there is no "token" on this line, it reads the next line
 *             from the file and keeps going.
 */
get_token(cur_line, cur_pos, fp, token_str)
char *cur_line;
int *cur_pos;
FILE **fp;
char *token_str;
{
    int i, token_count;

    /*
     * find the first non-whitespace
     */
    for (i=0; ; i++)
    {
        if ((cur_line[*cur_pos + i] == '#') ||
	    (cur_line[*cur_pos + i] == '\n'))
	{
            fgets(cur_line, 100, *fp);       /* get a new line */
            i = -1;
            *cur_pos = 0;
	}
        else if (!my_isspace(cur_line[*cur_pos + i]))
	{
	    break;
	}
    }

    /*
     * copy until find the first whitespace
     */
    token_count = 0;
    for (*cur_pos += i; ; (*cur_pos)++)
    {
        if (my_isspace(cur_line[*cur_pos]))
	{
	    token_str[token_count] = '\0';
	    return;
	}
	else
	{
            token_str[token_count] = cur_line[*cur_pos];
	    token_count++;
	}
    }
}


/*
 * read_image_into_region - this function reads the image from the
 *                          disk into the region.  It assumes
 *                          that everything is prepared for this 
 *                          transfer, such as the destination region
 *                          is large enough.
 */
read_image_into_region(disp_num, i, name)
int disp_num, i;
char *name;
{
    XImage *the_image;
    char tmp_str[50];
    int width, height, depth;

    if ((the_image = read_image_from_disk(disp_info[i].disp,
                           disp_info[i].draw_win_id,
                           name, &width, &height, &depth)) == NULL)
    {
         popup_error_dialog(disp_num, 
			   "ERROR in read_image_into_region.");
         return;
    }
        
    /*
     * it is a valid xwd file, place it
     */
    if (DefaultDepth(disp_info[i].disp,
                     DefaultScreen(disp_info[i].disp)) != depth)
    {
         sprintf(tmp_str, 
                   "Image not displayed on display %s.\nInconsistent Depths.",
                    disp_info[i].disp_name);
         popup_info_dialog(disp_num, tmp_str);
         XDestroyImage(the_image);       /*free the image memory*/
    }
    else
    {
         XPutImage(disp_info[i].disp, 
                   disp_info[disp_num].selected_region.backing_store[i],
                   disp_info[disp_num].win_gc[i], the_image,
                   0, 0, 0, 0, width, height);
         XDestroyImage(the_image);
     }
}


/*
 * read_ppm_file_into_region - this function reads the indicated
 *                             ppm file (Portable PixMap)
 *                             into this display's selected regions
 */
read_ppm_file_into_region(disp_num, ppm_filename)
int disp_num;
char *ppm_filename;
{
    FILE *fp;
    XImage *the_image, *example_image;
    int loc_x, loc_y, width, height, depth, x, y, xhr, yhr, i, j;
    int red_val, green_val, blue_val;
    char cur_line[102];	/* no ppm line is longer than 70 characters */
    char token_str[50]; 
    char *data;
    int cur_pos, max_color;
    char tmp_str[200];
    char tmp_name[50];
    XWindowAttributes win_attr;
    XColor the_color;

    set_paused_cursors();                     /*this may take a while*/

    if ((fp = fopen(ppm_filename, "r")) == NULL)
    {
        popup_error_dialog(disp_num,
                    "could not open file in read_ppm_file_into_region");
        unset_paused_cursors();
        return;
    }

    fgets(cur_line, 100, fp);       /*get the line*/
    cur_pos = 0;
    get_token(cur_line, &cur_pos, &fp, token_str);

    if (strcmp(token_str, "P3") != 0)
    {
        popup_info_dialog(disp_num,
                    "This is not a PPM file.\nPPM files start with P3.");
        unset_paused_cursors();
        return;
    }

    XGetWindowAttributes(disp_info[disp_num].disp, 
			 disp_info[disp_num].draw_win_id, &win_attr);
    depth = win_attr.depth;

    get_token(cur_line, &cur_pos, &fp, token_str);
    sscanf(token_str, "%d", &width);
    get_token(cur_line, &cur_pos, &fp, token_str);
    sscanf(token_str, "%d", &height);
    get_token(cur_line, &cur_pos, &fp, token_str);
    sscanf(token_str, "%d", &max_color);

    find_center_of_drawing_window(disp_num, width, height, &x, &y);
    prepare_region(disp_num, disp_num, x, y, width, height);

    if ((the_image = XGetImage(disp_info[disp_num].disp,
                disp_info[disp_num].selected_region.backing_store[disp_num], 
		0, 0, width, height, (~0), ZPixmap)) == NULL)
    {
	printf("ERROR: could not get the_image.\n");
    }

    for (y=0; y < height; y++)
    {
        for (x=0; x < width; x++)
        {
            get_token(cur_line, &cur_pos, &fp, token_str);
            sscanf(token_str, "%d", &red_val);
            get_token(cur_line, &cur_pos, &fp, token_str);
            sscanf(token_str, "%d", &green_val);
            get_token(cur_line, &cur_pos, &fp, token_str);
            sscanf(token_str, "%d", &blue_val);
            
	    /*
	     * now that we have the R, G, and B values for this pixel, 
	     * scale them to 16 bit for X-Windows, and then allocate that
	     * color and place it in the image.
	     */
	    the_color.flags = DoRed | DoGreen | DoBlue;
	    the_color.red = (unsigned short) (((long) 65535 * (long) red_val) / 
			           ((long) max_color));
	    the_color.green=(unsigned short) (((long) 65535 * (long)green_val) /
			           ((long) max_color));
	    the_color.blue =(unsigned short) (((long) 65535 * (long)blue_val) /
			           ((long) max_color));
	    if (!XAllocColor(disp_info[disp_num].disp, DefaultColormap(
			disp_info[disp_num].disp, DefaultScreen(
			disp_info[disp_num].disp)), &the_color))
	    {
		printf("ERROR: XAllocColor in read_ppm_file_into_region.\n");
	    }
	    XPutPixel(the_image, x, y, the_color.pixel);
        }
    }

    find_center_of_drawing_window(disp_num, width, height, &x, &y);
    printf("x=%d, y=%d, width=%d, height=%d.\n", x, y, width, height);
    disp_info[disp_num].selected_region.x = x;
    disp_info[disp_num].selected_region.y = y;
    disp_info[disp_num].selected_region.cur_x = x;
    disp_info[disp_num].selected_region.cur_y = y;
    disp_info[disp_num].selected_region.width = width;
    disp_info[disp_num].selected_region.height = height;

    prepare_region(disp_num, disp_num, x, y, width, height);
                       
    XPutImage(disp_info[disp_num].disp, 
              disp_info[disp_num].selected_region.backing_store[disp_num],
              disp_info[disp_num].win_gc[disp_num], the_image,
              0, 0, 0, 0, width, height);
    XtMapWidget(disp_info[disp_num].selected_region.selection_shell[disp_num]);
    XFlush(disp_info[disp_num].disp);

    sprintf(tmp_name, "/tmp/ws.%d.%d", disp_num, getpid());
    save_image_on_disk(disp_info[disp_num].disp,
              disp_info[disp_num].draw_win_id,
              disp_info[disp_num].selected_region.backing_store[disp_num],
	      0, 0, width, height, tmp_name);

    for (i=0; i < num_disps; i++)
    {
        if ((disp_info[i].in_session_bool) && (i != disp_num))
        {
            prepare_region(disp_num, i, x, y, width, height);
            read_image_into_region(disp_num, i, tmp_name);
            XtMapWidget(disp_info[disp_num].selected_region.selection_shell[i]);
            XFlush(disp_info[i].disp);
        }
    }
    unlink(tmp_name);

    disp_info[disp_num].selected_region.region_is_selected = TRUE;
    disp_info[disp_num].selected_region.region_has_moved = TRUE;  /* hack */

    make_selection_dependant_menus_unstippled(disp_num);
    set_regions_flickering(disp_num);
    unset_paused_cursors();                    
}


/*
 * save_region_into_ppm_file - this function saves the current selection
 *                             into the indicated file in ppm format.
 */
save_region_into_ppm_file(disp_num, ppm_filename)
int disp_num;
char *ppm_filename;
{
    XImage *the_image;
    XColor the_color;
    int width, height, max_color;
    FILE *fp;
    int x, y, num_pixels_on_this_line;
    int red_val, green_val, blue_val;

    set_paused_cursors();                     /*this may take a while*/

    if (disp_info[disp_num].selected_region.region_is_selected == FALSE)
    {
        popup_error_dialog(disp_num, "ERROR: region not selected");
        unset_paused_cursors();                    
        return;
    }

    width = disp_info[disp_num].selected_region.width;
    height = disp_info[disp_num].selected_region.height;

    if ((the_image = XGetImage(disp_info[disp_num].disp,
                disp_info[disp_num].selected_region.backing_store[disp_num], 
		0, 0, width, height, (~0), ZPixmap)) == NULL)
    {
	printf("ERROR: could not get the_image.\n");
    }

    if ((fp = fopen(ppm_filename, "w")) == NULL)
    {
        popup_error_dialog(disp_num,
                    "could not open file in save_region_into_ppm_file");
        unset_paused_cursors();
        return;
    }

    fprintf(fp, "P3\n");
    fprintf(fp, "# This ppm file was generated by wscrawl 2.0\n");
    fprintf(fp, "%d %d\n", width, height);
    max_color = 255;
    fprintf(fp, "%d\n", max_color);      /* max value */

    num_pixels_on_this_line = 0;
    for (y=0; y < height; y++)
    {
	for (x=0; x < width; x++)
	{
	    the_color.pixel = XGetPixel(the_image, x, y);
	    XQueryColor(disp_info[disp_num].disp, DefaultColormap(
			disp_info[disp_num].disp, DefaultScreen(
			disp_info[disp_num].disp)), &the_color);
	    the_color.flags = DoRed | DoGreen | DoBlue;

	    red_val = (int) ((((long) the_color.red) * 255) / ((long) 65535));
	    green_val = (int)((((long)the_color.green) * 255) / ((long) 65535));
	    blue_val = (int)((((long)the_color.blue) * 255) / ((long) 65535));

	    fprintf(fp, "%d %d %d  ", red_val, green_val, blue_val);

	    num_pixels_on_this_line++;
	    if (num_pixels_on_this_line > 4)
	    {
		fprintf(fp, "\n");
	        num_pixels_on_this_line = 0;
	    }
	}
    }

    fclose(fp);
    unset_paused_cursors();                    
}


/*
 * read_text_file_into_region - this function reads the indicated
 *                             text file into this display's selected regions
 */
read_text_file_into_region(disp_num, text_filename)
int disp_num;
char *text_filename;
{
    XImage *the_image;
    int width, height, depth, x, y, xhr, yhr, i;
    char tmp_str[200];
    int c, length, top;
    char current_line[512];
    FILE *fp;

    set_paused_cursors();                     /*this may take a while*/

    if ((fp = fopen(text_filename, "r")) == NULL)
    {
	popup_error_dialog(disp_num, 
		    "could not open file in read_text_file_into_region");
        unset_paused_cursors();                    
        return;
    }

    disp_info[disp_num].type_info.char_height =
         (disp_info[disp_num].type_info.the_font_struct)->max_bounds.ascent +
         (disp_info[disp_num].type_info.the_font_struct)->max_bounds.descent;
    get_dimensions_of_text_file(disp_num, fp, &width, &height);
    fclose(fp);
    find_center_of_drawing_window(disp_num, width, height, &x, &y);
                
    disp_info[disp_num].selected_region.x = x;
    disp_info[disp_num].selected_region.y = y;
    disp_info[disp_num].selected_region.cur_x = x;
    disp_info[disp_num].selected_region.cur_y = y;
    disp_info[disp_num].selected_region.width = width;
    disp_info[disp_num].selected_region.height = height;

    for (i=0; i < num_disps; i++)
    {
        if (disp_info[i].in_session_bool)
        {
            prepare_region(disp_num, i, x, y, width, height);
            fp = fopen(text_filename, "r");
            top = disp_info[disp_num].type_info.char_height;

	    while ((c = getc(fp)) != EOF)
	    {
                ungetc(c, fp);
                fgets(current_line, 510, fp);       /*get the line*/
                for (length=0; current_line[length] != '\n'; length++)
                    ;

                XDrawString(disp_info[i].disp, 
		     disp_info[disp_num].selected_region.backing_store[i],
                     disp_info[disp_num].win_gc[i],
                     0, top, current_line, length);

                top += disp_info[disp_num].type_info.char_height;
            }
	    fclose(fp);

            XtMapWidget(disp_info[disp_num].selected_region.selection_shell[i]);
            XFlush(disp_info[i].disp);
        }
    }
    
    disp_info[disp_num].selected_region.region_is_selected = TRUE;
    disp_info[disp_num].selected_region.region_has_moved = TRUE;  /* hack */

    make_selection_dependant_menus_unstippled(disp_num);
    set_regions_flickering(disp_num);
    unset_paused_cursors();                    
}


/*
 * paste_x_selection_region - this function is called when "Paste X Selection" 
 *               is selected in the Edit menu.  If not in typing mode, it
 *               pastes the current contents of the X-Windows Selection
 *               into the selected region and maps the selected region 
 *               in the center of the drawing window. If in typing mode,
 *		 it pastes the current X-Selection into the window.
 */
paste_x_selection_region(disp_num)
int disp_num;
{
    int i, n, x, y, width, height;
    Arg args[MAX_ARGS];
    Pixmap the_pixmap;
    unsigned char *the_text;

    if (disp_info[disp_num].selected_region.region_is_selected == TRUE)
    {
        popup_info_dialog(disp_num, 
	      "Unselect the current region\nthen retry this action.");
	return;
    }

    if (disp_info[disp_num].input_mode == TYPING)
    {
        the_text = get_x_selection_as_text(disp_num, disp_info[disp_num].disp, 
	                        disp_info[disp_num].draw_win_id);
	for (i=0; the_text[i] != '\0'; i++)
	{
	    draw_char_on_screen(disp_num, (char *) &(the_text[i]));
	}

	free(the_text);
    }
    else
    {
        get_x_selection_as_pixmap(disp_num, disp_info[disp_num].disp, 
	                      disp_info[disp_num].draw_win_id,
			      &width, &height);
    }
}


/*
 * get_x_selection_as_text - this function gets the X-Windows X-Selection
 *			     and returns it as a text string.  This should
 *		             be an Xlib call, but the X Consortium dropped
 *		             the ball in a big way.  What total bozos.
 */
unsigned char *get_x_selection_as_text(disp_num, disp, win_id)
int disp_num;
Display *disp;
Window win_id;
{
    unsigned int property_data[4];
    Atom the_property, the_type;
    Atom the_selection, the_target;
    Atom actual_type_ret;
    int actual_format_ret;
    unsigned long nitems_ret, bytes_after_ret;
    unsigned char *the_text;
    Window root_ret;
    unsigned int x_ret, y_ret, width_ret, height_ret, bw_ret, depth_ret;
    char tmp_name[256];
    static struct timezone dummy;
    static struct timeval start_time, cur_time;
    static char msg[256];
    XWindowAttributes attr_ret;
    XEvent the_event;

    /* 
     * prepare the win_id window to be the communication medium between
     * us and the server (who has the selection).  We want the text,
     * and we are attaching a property to the window called
     * "WS_SELECTION" to be the rendezvous point.
     */
    property_data[0] = XInternAtom(disp, "STRING", False);
    property_data[1] = XInternAtom(disp, "STRING_WS_SELECTION", False);

    the_property = XInternAtom(disp, "WS_SELECTION", False);
    the_type = XInternAtom(disp, "ATOM_PAIR", False);

    XChangeProperty(disp, win_id, the_property, the_type, 32, PropModeReplace, 
		    (unsigned char *) property_data, 2);
    
    /*
     * Add PropertyChanges to the events we are looking for on this win_id.
     */
    XGetWindowAttributes(disp, win_id, &attr_ret);
    XSelectInput(disp, win_id, PropertyChangeMask | attr_ret.your_event_mask);

    /*
     * Now tell the server to send us the selection by attaching it to 
     * the rendezvous window in the format specified.
     */
    the_selection = XInternAtom(disp, "PRIMARY", False);
    the_target = XInternAtom(disp, "MULTIPLE", False);

    XConvertSelection(disp, the_selection, the_target, the_property, win_id, 
		      CurrentTime);
    /*
     * now wait for the selection to arrive from the server, timeout in 5 
     * seconds with a failure dialog.
     */
    gettimeofday(&start_time, &dummy);
    while(XCheckWindowEvent(disp,win_id,PropertyChangeMask,&the_event) == False)
    {
	gettimeofday(&cur_time, &dummy);
	if (cur_time.tv_sec > start_time.tv_sec + 5)	/* 5 seconds elapsed */
	{
	    sprintf(msg, "%s\n%s\n%s", 
	        "There seems to be no text X-Selection.  If",
		"you are trying to paste graphics, first select",
		"some other tool than text input tool.");
	    popup_info_dialog(disp_num, msg);
            XSelectInput(disp, win_id, attr_ret.your_event_mask);
	    return(NULL);
	}
    }

    /*
     * return the event mask to the way we found it
     */
    XSelectInput(disp, win_id, attr_ret.your_event_mask);

    /*
     * Now get the selection that has arrived at the window
     */
    if (XGetWindowProperty(disp, win_id, property_data[1], 0, 1024, True, 
	       AnyPropertyType, &actual_type_ret, &actual_format_ret,
	       &nitems_ret, &bytes_after_ret, &the_text) != Success)
    {
	printf("ERROR: XGetWindowProperty for the string failed.\n");
    }
    
    if (bytes_after_ret != 0)
    {
        popup_info_dialog(disp_num, 
	    "Could not paste entire X-Selection\nbecause it was too large.");
    }

    return(the_text);
}

/*
 * get_x_selection_as_pixmap - this function returns the X-Windows selection
 *                             as a pixmap for this display.  This (of course)
 *                             should be an Xlib call, but they are hoseheads.
 */
get_x_selection_as_pixmap(disp_num, disp, win_id, width, height)
int disp_num;
Display *disp;
Window win_id;
int *width, *height;
{
    unsigned int property_data[4];
    Atom the_property, the_type;
    Atom the_selection, the_target;
    Atom actual_type_ret;
    int actual_format_ret;
    unsigned long nitems_ret, bytes_after_ret;
    unsigned char *the_pixmap, *the_colormap;
    Window root_ret;
    unsigned int width_ret, height_ret, bw_ret, depth_ret;
    int x_ret, y_ret;
    char tmp_name[256];
    static struct timezone dummy;
    static struct timeval start_time, cur_time;
    static char msg[256];
    XWindowAttributes attr_ret;
    XEvent the_event;

    /* 
     * prepare the win_id window to be the communication medium between
     * us and the server (who has the selection).  We want the pixmap and
     * the colormap, and we are attaching a property to the window called
     * "WS_SELECTION" to be the rendezvous point.
     */
    property_data[0] = XInternAtom(disp, "PIXMAP", False);
    property_data[1] = XInternAtom(disp, "PIXMAP_WS_SELECTION", False);
    property_data[2] = XInternAtom(disp, "COLORMAP", False);
    property_data[3] = XInternAtom(disp, "COLORMAP_WS_SELECTION", False);

    the_property = XInternAtom(disp, "WS_SELECTION", False);
    the_type = XInternAtom(disp, "ATOM_PAIR", False);

    XChangeProperty(disp, win_id, the_property, the_type, 32, PropModeReplace, 
		    (unsigned char *) property_data, 4);
    
    /*
     * Add PropertyChanges to the events we are looking for on this win_id.
     */
    XGetWindowAttributes(disp, win_id, &attr_ret);
    XSelectInput(disp, win_id, PropertyChangeMask | attr_ret.your_event_mask);

    /*
     * Now tell the server to send us the selection by attaching it to 
     * the rendezvous window in the format specified.
     */
    the_selection = XInternAtom(disp, "CLIPBOARD", False);
    the_target = XInternAtom(disp, "MULTIPLE", False);

    XConvertSelection(disp,the_selection, the_target, the_property, win_id, 
		      CurrentTime);
    /*
     * now wait for the selection to arrive from the server, timeout in 5 
     * seconds with a failure dialog.
     */
    gettimeofday(&start_time, &dummy);
    while(XCheckWindowEvent(disp,win_id,PropertyChangeMask,&the_event) == False)
    {
	gettimeofday(&cur_time, &dummy);
	if (cur_time.tv_sec > start_time.tv_sec + 5)	/* 5 seconds elapsed */
	{
	    sprintf(msg, "%s\n%s\n%s", 
	        "There seems to be no graphic X-Selection.",
		"If you are trying to paste text, first",
		"select the text input tool.");
	    popup_info_dialog(disp_num, msg);
            XSelectInput(disp, win_id, attr_ret.your_event_mask);
	    return;
	}
    }

    /*
     * return the event mask to the way we found it
     */
    XSelectInput(disp, win_id, attr_ret.your_event_mask);

    /*
     * Now get the selection that has arrived at the window
     */
    if (XGetWindowProperty(disp, win_id, property_data[1], 0, 1, True, 
	       AnyPropertyType, &actual_type_ret, &actual_format_ret,
	       &nitems_ret, &bytes_after_ret, &the_pixmap) != Success)
    {
	printf("ERROR: XGetWindowProperty for the pixmap failed.\n");
    }

    if (XGetWindowProperty(disp, win_id, property_data[3], 0, 1, True, 
	       AnyPropertyType, &actual_type_ret, &actual_format_ret,
	       &nitems_ret, &bytes_after_ret, &the_colormap) != Success)
    {
	printf("ERROR: XGetWindowProperty for the colormap failed.\n");
    }

    XGetGeometry(disp, *the_pixmap, &root_ret, &x_ret, &y_ret, &width_ret, 
		 &height_ret, &bw_ret, &depth_ret);

    *width = (int) width_ret;
    *height = (int) height_ret;
    sprintf(tmp_name, "/tmp/ws.tmp.%d", getpid());

    /*
     * assume the standard colormap for our purposes
     */
    save_image_on_disk(disp, win_id, *the_pixmap, 0, 0, *width, *height, 
		       tmp_name);
    read_xwd_file_into_region(disp_num, tmp_name);
}

