/*
 * This file:  wsxlib.c
 *
 * This file contains most of the really hairy xlib stuff for wscrawl
 */

#include "wscrawl.h"

extern int num_disps;
extern struct disp_info_struct disp_info[MAX_NUM_DISPS];


/*
 * adjust_graphics_contexts - this function re-allocates all new graphics
 *                            contexts for all the displays involved in this
 *                            session, assuming the one who has just joined
 *                            is the display index passed in.
 */
adjust_graphics_contexts(disp_num)
int disp_num;
{
    int i, n;
    Arg args[MAX_ARGS];
    XGCValues gcvalues;
    XFontStruct *the_font_struct;
    unsigned long nothing_mask = 0;
    unsigned long alter_these_mask = GCLineWidth | GCForeground | GCCapStyle |
				     GCBackground | GCFont;
    unsigned long gc_copy_mask = GCSubwindowMode | GCLineWidth | 
				     GCForeground | GCCapStyle | GCFont |
				     GCBackground | GCGraphicsExposures;
    XColor scrn_def_ret, exact_def_ret;
    XGCValues cursor_gc_values;
    XColor the_color;

    /*
     * The following the new "user" N graphics contexts, where N
     * is the number of open displays.  This "user" wants his particular
     * color on every display, and since gc's are attached to displays,
     * we need one with this color loaded on every display.  These gc's
     * are then used for pen width, pen caps, etc.
     *
     * Then all the OTHER members of the session must get a GC BACK to
     * this guy.
     */

    /*
     * first assign a GC for this guy to every other member of the session's
     * screen
     */
    for (i=0; i < num_disps; i++)
    { 
	if (disp_info[i].in_session_bool)  /*only if this session is wanted*/
	{
            /*
	     * create the gc from scratch
	     */
            if (XAllocNamedColor(disp_info[i].disp,
				 XDefaultColormapOfScreen(
				 XDefaultScreenOfDisplay(disp_info[i].disp)),
			         disp_info[disp_num].cur_color_str, 
			         &scrn_def_ret, &exact_def_ret))
             {
                 ;  /*great!  it worked!*/
	     }
	     else if (!XAllocNamedColor(disp_info[i].disp, 
			     DefaultColormap(disp_info[i].disp, 
			     DefaultScreen(disp_info[i].disp)), 
		             "black", &scrn_def_ret, &exact_def_ret))
	     {
	          printf("The color %s is not supported on the display %s", 
			 disp_info[disp_num].cur_color_str,
			 disp_info[i].disp_name);
                  printf("Using black.\n");
	     }

             if ((the_font_struct = XLoadQueryFont(disp_info[i].disp,
                              disp_info[disp_num].cur_font_str)) != NULL)
             {
		  if (disp_num == i)
		  {
                      disp_info[disp_num].type_info.the_font_struct = 
							       the_font_struct;
		  }
                  gcvalues.font = the_font_struct->fid;
             }
             else if  (load_moron_fonts(disp_num, i, &the_font_struct,
				disp_info[disp_num].cur_font_str) != FALSE)
	     {
		  if (disp_num == i)
		  {
                      disp_info[disp_num].type_info.the_font_struct = 
							       the_font_struct;
		  }
                  gcvalues.font = the_font_struct->fid;
	     }
	     else
             {
                  printf("Couldn't allocate font %s.\n", 
	                  disp_info[disp_num].cur_font_str);
             }

	     if (disp_num == i)
                 disp_info[disp_num].current_pixel = scrn_def_ret.pixel;
             gcvalues.foreground = scrn_def_ret.pixel;
             gcvalues.background = WhitePixelOfScreen(DefaultScreenOfDisplay(
				     disp_info[i].disp));
        
             gcvalues.line_width = 8;
	     gcvalues.cap_style = CapRound;
            
             disp_info[disp_num].win_gc[i] = XCreateGC(disp_info[i].disp, 
		             disp_info[i].draw_win_id, alter_these_mask, 
			     &gcvalues);
	}
    }

    /*
     * now go through and give each OTHER member of the session a GC for
     * this guy
     */
    for (i=0; i < num_disps; i++)
    { 
	if ((disp_info[i].in_session_bool) && (i != disp_num))
	{
	    /* 
	     * find out what R, G, and B values were being used, then allocate
	     * that color on the new dude's display
	     */
            the_color.pixel = disp_info[i].current_pixel;
	    XQueryColor(disp_info[i].disp, DefaultColormap(disp_info[i].disp, 
		        DefaultScreen(disp_info[i].disp)), &the_color);
            /*
	     * create the gc from scratch
	     */
            if (XAllocColor(disp_info[disp_num].disp,
			    DefaultColormap(disp_info[disp_num].disp, 
			    DefaultScreen(disp_info[disp_num].disp)), 
			    &the_color))
             {
		 scrn_def_ret.pixel = the_color.pixel;  /*great!  it worked!*/
	     }
	     else if (!XAllocNamedColor(disp_info[disp_num].disp, 
			     DefaultColormap(disp_info[disp_num].disp, 
			     DefaultScreen(disp_info[disp_num].disp)), 
		             "black", &scrn_def_ret, &exact_def_ret))
	     {
	          printf("The color %s is not supported on the display %s", 
			 disp_info[disp_num].cur_color_str,
			 disp_info[disp_num].disp_name);
                  printf("Using black.\n");
	     }

             if ((the_font_struct = XLoadQueryFont(disp_info[disp_num].disp,
                              disp_info[i].cur_font_str)) != NULL)
             {
                  gcvalues.font = the_font_struct->fid;
             }
             else if  (load_moron_fonts(i, disp_num, &the_font_struct,
				disp_info[i].cur_font_str) != FALSE)
	     {
                  gcvalues.font = the_font_struct->fid;
	     }
             else
             {
                  printf("Couldn't allocate font %s.\n", 
	                  disp_info[i].cur_font_str);
             }
            
             gcvalues.foreground = scrn_def_ret.pixel;
             gcvalues.background = WhitePixelOfScreen(DefaultScreenOfDisplay(
				     disp_info[disp_num].disp));
        
	     switch (disp_info[i].input_mode)
	     {
	         case PAINTING:
		     if (disp_info[i].paintbrush.paint_mode == NORMAL)
		     {
                         gcvalues.line_width = 
			    disp_info[i].paintbrush.brush_width;
		     }
		     else
		     {
                         gcvalues.line_width = 0;
		     }
		     break;
		 case PENCILING:
                     gcvalues.line_width = 0;
		     break;
	         case DRAWING_LINES:
                     gcvalues.line_width = 
			       disp_info[i].line_dialog.line_width;
		     break;
		 default:
                     gcvalues.line_width = 8;
		     break;
	     }
	     gcvalues.cap_style = CapRound;
            
             disp_info[i].win_gc[disp_num] = XCreateGC(disp_info[disp_num].disp,
		             disp_info[disp_num].draw_win_id, alter_these_mask, 
			     &gcvalues);
	}
    }
}

